% clev_em: test the E-M procedure with the cantilever psf
%

% Simulation options
bDebug 					= 1;
bPause 					= 0;
bPrint 					= 0;
bCheckStoppingCriteria 	= 1;
bStopAfterEachIter		= 0;

% Experimental parameters
numEmIters 				= 20;
alphaFac 				= 1;
sigmaFac 				= 1.6;
mseThresFac 			= 1/3;

par = psf_param3( 1, 3e4, 10 );
stMesh = psf_mesh( par );
numPts = stMesh.numXPts * stMesh.numYPts * stMesh.numZPts;

%%% XXX: assume for now that we only have a 2-d image
if ( stMesh.numZPts ~= 1 ), error( 'not using a 2-d image' ), end;

% Generate the output image
out = gen_img( par, stMesh, 'random', sigmaFac, [], 10 );
figure( 1 ), imagesc( out.f ), colormap( gray), colorbar;
title( 'Original image' );

% Globals
%global H alpha sigma fftSize hty HpH

% Compute the variance in the image domain
h = real( get_h_op( par, stMesh, 1 ) );
%H = im2BlockCirc( h );

mseThres = mseThresFac * mean( out.f(:).^2 );
disp( sprintf( 'ms of f = %g, mseThres = %g', mean( out.f(:).^2 ), mseThres ) );

sigma1 = out.sigma1;

% Compute the upper bound for alpha

% Old way: inefficient
%vecLambdaMax = eigs( H*H', 10, 'la' );
%lambdaMax = max( vecLambdaMax(1) );

% New way: use FFTs
fftSize = pad2_fft_size( [stMesh.numXPts, stMesh.numYPts, stMesh.numZPts], ...
	0 );
H = fftn( h, fftSize );
H2 = H .* conj( H );
lambdaMax = max( max( H2 ) );
alphaUpperBound = sigma1/sqrt( lambdaMax ); 
alpha = alphaFac * alphaUpperBound;
r = alpha/sigma1;
disp( sprintf( 'r = %g, r^2 = %g, log r = %g', r, r^2, log(r) ) );

% E-M algorithm
vecEStepMse = [];
vecMStepMse = [];

% Initial conditions

fML = real( f_deconv( stMesh, out.h, out.gNoisy ) );
results = ebayesthresh( fML(:), 'laplace', [], 0, [], 1, 'median', 0 );
fMLClean = reshape( results.muhat, [stMesh.numXPts, stMesh.numYPts] );

thetaHat_0 = fML;						% Use the ML estimator
%thetaHat_0 = zeros( size( out.f ) );	% Use the all-zero image
%thetaHat_0 = fMLClean;					% Use the cleaned ML estimate
figure( 2 ), imagesc( thetaHat_0 ), colormap( gray), colorbar;
title( 'thetaHat_0' );

thetaHatCurr = thetaHat_0;

Y = fftn( real(out.gNoisy), fftSize );
hty = real( ifftn( H' .* Y, fftSize ) ); 
HpH = conj( H ) .* H;

for iterInd = 1:numEmIters
	szIterMsg = sprintf( 'iterInd=%g', iterInd );
	disp( szIterMsg );

	%%% E-step %%%
	zHatCurr = estep( fftSize, hty, HpH, r, thetaHatCurr );

	eStepMse = mean( (out.f(:) - zHatCurr(:)).^2 );
	disp( sprintf( 'E-step mse: %g', eStepMse ) );
	vecEStepMse = [vecEStepMse; eStepMse];

	if bDebug %%% DEBUG %%%
		%create_int_plot( 2, strcat( 'zHatCurr:', szIterMsg ), [], ...
		%	out.X, out.Y, zHatCurr, 1, 1 );
		figure( 3 ), imagesc( zHatCurr ), colormap( gray ), colorbar;
		title( strcat( 'zHatCurr:', szIterMsg ) );
		szFile = sprintf( 'estep_%d.ps', iterInd );
		if bPrint, print( '-f3', '-depsc2', szFile ), end;
	end 
	%pause;

	%%% M-step %%%
	try
		vecZHatCurr = reshape( zHatCurr, [numPts, 1] );
		vecThetaHatNext = mstep( alpha, vecZHatCurr, 1 );
		thetaHatNext = reshape( vecThetaHatNext, ...
			[stMesh.numXPts, stMesh.numYPts] );
	catch
		disp( sprintf( 'Caught error from mstep at iterInd=%g', iterInd ) );
		disp( lasterr );

		% Option 1: break
		break;

		% Option 2: skip the m-step and continue the e-step
		%disp( 'Skipping m-step and continuing the e-step...' );
		%thetaHatCurr = zHatCurr;
		%continue;
	end

	mStepMse = mean( (out.f(:) - vecThetaHatNext(:)).^2 );
	disp( sprintf( 'M-step mse: %g', mStepMse ) );
	vecMStepMse = [vecMStepMse; mStepMse];
	
	% DEBUG
	%numNeg = sum( vecThetaHatNext < -eps );
	%if ( numNeg > 0 ), warning( 'negative values from M-step' ), end;

	if bDebug %%% DEBUG %%%
		%create_int_plot( 3, strcat( 'thetaHatNext:', szIterMsg ), [], ...
		%	out.X, out.Y, thetaHatNext, 1, 1 );
		figure( 4 ), imagesc( thetaHatNext ), colormap( gray ), colorbar;
		title( strcat( 'thetaHatNext:', szIterMsg ) );

		if ( ( iterInd == 1 ) | ( (iterInd > 1) & bPause ) | ...
			 ( bStopAfterEachIter ) )
			disp( 'Pausing--press any key to continue...' );
			pause;
		end

		szFile = sprintf( 'mstep_%d.ps', iterInd );
		if bPrint, print( '-f4', '-depsc2', szFile ), end;
	end 

	% Update the variables
	thetaHatCurr = thetaHatNext;

	% Check for stopping criteria
	if bCheckStoppingCriteria
		if ( mStepMse < mseThres ), break, end;
	end
end

fEm = thetaHatCurr;

%%% Now plot the images and compare them
if ~bDebug
	create_int_plot( 1, 'f', [], out.X, out.Y, real(out.f), 1, 1 );
	create_int_plot( 2, 'fEm', [], out.X, out.Y, fEm, 1, 1 );
	%create_int_plot( 3, 'g', [], out.X, out.Y, real(out.g), 1 );
	%create_int_plot( 4, 'gNoisy', [], out.X, out.Y, real(out.gNoisy), 1 );
end

finalMse = mean( (out.f(:) - fEm(:)).^2 );
disp( sprintf( '   *** Final MSE: %g ***', finalMse ) );
