% em_restore: Implements the EM algorithm by Figueiredo and Nowak. 
%
% Usage:
%    sRecon = em_restore( hForwardOp, imgNoisy, noiseSigma, thetaTrue,
%       szMSelector, numEmIters, numMaxGradientSearchFailures,
%       szInitialSelector, sOptions )
%
%    sRecon - structure containing the reconstructed image
%    hForwardOp - forward operator in convolution form
%    imgNoisy - noise image
%    noiseSigma - noise variance
%    thetaTrue - true value of theta
%    szMSelector - string to describe method to use for M step: 'eb', 'map',
%       'map_prior', 'eb_fixed', 'map_fixed', 'id'
%    numEmIters - # of EM iterations
%    numMaxGradientSearchFailures - # of max gradient search failures before
%       we give up gradient search altogether
%    szInitialSelector - string to describe the initial point to use in the
%       algorithm: 'deconv', 'zero', 'checker', 'true', 'custom', 'obs'
%    sOptions - structure containing the elements:
%       imgInitial - initial image to use when szInitialSelector is set to
%          'custom'
%       bSaveHistory - boolean flag to save the history of the estimates
%       atHistory - numEmIters x 2 matrix containing the values of (a,t) to
%          use in the M step
%

function sRecon = em_restore( hForwardOp, imgNoisy, noiseSigma, thetaTrue, ...
	szMSelector, numEmIters, numMaxGradientSearchFailures, ...
	szInitialSelector, sOptions )

sRestoreOptions.GRADIENT_SEARCH_FAILURES_MAX = numMaxGradientSearchFailures;
sRestoreOptions.EM_SKIP_INITIAL_GRID_THRES = 5;
sRestoreOptions.EM_PAUSE_NUMSECS = 0;
sRestoreOptions.NUM_EM_ITERS = numEmIters;
sRestoreOptions.B_PLOT = 0;

try
	% Try to see if these have been specified in the sOptions structure.
	% If not, then use defaults
	sRestoreOptions.EB_FIXED_AVAL = sOptions.aVal;
	sRestoreOptions.EB_FIXED_WVAL = sOptions.wVal;
	sRestoreOptions.EB_FIXED_TVAL = sOptions.tVal;
	sRestoreOptions.EB_FIXED_SPARSE_IND = sOptions.szSparseInd;
	sRestoreOptions.EB_FIXED_USE_HISTORY = sOptions.bUseHistory;
    sRestoreOptions.STOPPING_COND = sOptions.szStoppingCond;
    sRestoreOptions.STOPPING_COND_TOL = sOptions.stoppingCondTol;    
    sRestoreOptions.THETASTAR_FIXED = sOptions.thetaStar;
    sRestoreOptions.SOFT_THRES_T = sOptions.softThresT;
catch
	
    warning( 'missing some fields in sOption' );
    disp( 'press a key to continue' ), pause;
    warning( 'using default values...' );

    % Didn't work, so use default values
	sRestoreOptions.EB_FIXED_AVAL = 1;
	sRestoreOptions.EB_FIXED_WVAL = 0.01; 
	sRestoreOptions.EB_FIXED_TVAL = -1;
	sRestoreOptions.EB_FIXED_SPARSE_IND = 'w';
	sRestoreOptions.EB_FIXED_USE_HISTORY = 0;
    sRestoreOptions.STOPPING_COND = 'fixed';
    sRestoreOptions.STOPPING_COND_TOL = 1e-2;
    sRestoreOptions.THETASTAR_FIXED = 1;
    sRestoreOptions.SOFT_THRES_T = 1;
end

sRestoreOptions.numGradientSearchFailures = 0;

% Wtf? Convert thetaTrue to double. It seems that sometimes we get a logical
% array!
thetaTrue = double( thetaTrue );

thetaHist = {};

try
	bSaveHistory = sOptions.bSaveHistory;
catch
	bSaveHistory = 0;
end

sClock = clock;

% No need for log file
%----------------------
% szLogFilename = sprintf( 'em_restore_%02d%02d%02d.log', sClock(3), ...
% 	sClock(4), sClock(5) );
%disp( sprintf( 'Using log file: %s', szLogFilename ) );
% Start diary
%diary off, delete( szLogFilename ), diary( szLogFilename );

n = length( imgNoisy(:) );
   
imgNoisyDft = fftn( imgNoisy );

hForwardOpDft = fftn( hForwardOp );
HEigMax = max( abs( hForwardOpDft(:) ) );
HEigMin = min( abs( hForwardOpDft(:) ) );
disp( sprintf( 'H max eig: %g, min eig: %g, cond no: %g', HEigMax, HEigMin, ...
	HEigMax / HEigMin ) );

Hty = real( ifftn( conj(hForwardOpDft) .* imgNoisyDft ) );

gamH_gam = conj( hForwardOpDft ) .* hForwardOpDft;
gam_gamH = hForwardOpDft .* conj( hForwardOpDft );

lambdaMax = max( gam_gamH(:) );
alphaUpperBound = noiseSigma / sqrt( lambdaMax );

disp( sprintf( 'lambdaMax: %g, noiseSigma: %g, alphaUpperBound: %g', ...
	lambdaMax, noiseSigma, alphaUpperBound ) );

switch ( lower(szInitialSelector) )
case 'deconv'
	deconvSoln = pseudoinv_eqdim( hForwardOp, imgNoisy );
	thetaHat0 = deconvSoln;
case 'zero'
	thetaHat0 = zeros( size( imgNoisy ) );
case 'checker'
	thetaHat0 = create_checker( size( imgNoisy ), 0.5, 1 );
case 'custom'
	thetaHat0 = sOptions.imgInitial;
case 'true'
	thetaHat0 = thetaTrue;
case 'obs'
	thetaHat0 = imgNoisy;
otherwise
	error( 'invalid szInitialSelector' );
end

alpha = alphaUpperBound;
sRestoreOptions.alpha = alpha;

thetaHatPrev = thetaHat0;

% History of hyperparameter estimates
tHist = [];
aHist = [];
wHist = [];
thetaStarHist = [];

% Metrics for convergence of iterations
thetaMaxDiffHist = []; % l_inf norm of difference of 2 consecutive estimates
thetaSumDiffHist = []; % l_1 norm of difference ...

% Metrics for accuracy of reconstruction. Note that depending on the method
% used, some of these might not be computed.
jointProbHist = [];         % Joint probability p(y,theta)

% l_1 norm of difference
errorNorm1Hist = [ norm( thetaHat0(:) - thetaTrue(:), 1 ) ];
% l_2 norm ...
errorNorm2Hist = [ norm( thetaHat0(:) - thetaTrue(:), 2 ) ];

logPcond = logp_ytheta_cond_gauss( imgNoisy, thetaHat0, hForwardOpDft, ...
	noiseSigma );
cost = -2*noiseSigma^2 * logPcond + ...
    2*sRestoreOptions.SOFT_THRES_T*norm(thetaHat0(:),1);
costFuncHist = [cost];

%close all;

i = 1; % Loop counter

while 1
    
    switch ( lower(sRestoreOptions.STOPPING_COND) )
        case 'fixed'
            % Perform a fixed number of iterations
            if ( i > sRestoreOptions.NUM_EM_ITERS ), break, end
        case 'conv'
            % Find the max value in the last 50 iterations, and if it is
            % less than a certain tolerance, quit.
            histLen = length( thetaMaxDiffHist );
            if ( histLen >= 51 )
                maxChange = max( thetaMaxDiffHist( max(1,histLen-50): histLen ) );
                if ( maxChange < sRestoreOptions.STOPPING_COND_TOL )
					break;
				end
            end
                                
            if ( i > sRestoreOptions.NUM_EM_ITERS )
                warning( 'stopping cond=conv, exceeded NUM_EM_ITERS' );
                break;
            end
        otherwise
            error( 'unrecognized stopping condition' );
    end
    
	% E step
	zHatNext = em_restore_e( alpha, noiseSigma, Hty, gamH_gam, thetaHatPrev );

	if ( sRestoreOptions.EB_FIXED_USE_HISTORY == 1 )
		try
			sRestoreOptions.EB_FIXED_AVAL = sOptions.atHistory( i, 1 );
			sRestoreOptions.EB_FIXED_TVAL = sOptions.atHistory( i, 2 );
		catch
			error( 'user must supply AVAL and TVAL' );
		end
	end

	thetaSnr = mean( abs(thetaHatPrev(:)) ) / alpha^2;
	if ( thetaSnr > eps )
		thresHash = min( [1/thetaSnr, sqrt(2*log(n))*alpha] );
	else
		thresHash = sqrt(2*log(n))*alpha;
	end
	thresHash = max( [0, thresHash] );
	%disp( sprintf( 'thetaSnr = %g, thresHash = %g', thetaSnr, thresHash ) );

	% M step
    sRestoreOptions.nEmIter = i;
    [thetaHatNext, estParams] = em_restore_m( zHatNext, lower(szMSelector), ...
        sRestoreOptions );

    disp( sprintf( '\tEM iter=%g\n', i ) ); % End of EM iteration
    
	% Compute error metrics

    errorNorm1Hist = [ errorNorm1Hist, ...
		norm( thetaHatNext(:) - thetaTrue(:), 1 ) ];
    errorNorm2Hist = [ errorNorm2Hist, ...
		norm( thetaHatNext(:) - thetaTrue(:), 2 ) ];
    
	logPcond = logp_ytheta_cond_gauss( imgNoisy, thetaHatNext, hForwardOpDft, noiseSigma );
    bComputeLogPtheta = 1;
    
    switch ( estParams.szMethod )
        case { 'eb', 'eb_fixed', 'eb_better', 'map', 'map_prior' }            
            logPtheta = logp_hyper_john_unif( thetaHatNext, estParams.aRange );
        case { 'binaryd', 'binaryd_map' }
            logPtheta = logp_hyper_binaryd_unif( thetaHatNext, ...
                estParams.thetaStarRange );     
		case 'eb_fix_w'
			logPtheta = logp_hyper_laplace_unif( thetaHatNext, ...
				estParams.aRange );
        otherwise
            bComputeLogPtheta = 0; 
    end
    
    if bComputeLogPtheta
        jointProbHist = [ jointProbHist, logPcond + logPtheta ];
    end
    
    % Compute the cost function
    
    bComputeCostFunc = 1;
    
    switch ( estParams.szMethod )
        case 'soft_fixed'
            cost = -2*noiseSigma^2 * logPcond + ...
                2*sRestoreOptions.SOFT_THRES_T*norm(thetaHatNext(:),1);
        otherwise
            bComputeCostFunc = 0;
    end
            
    if bComputeCostFunc
        costFuncHist = [costFuncHist, cost];
    end
    
    if 0 % DEBUG
        if ( i == 1 ), save 'zIter1' zHatNext sRestoreOptions -v6, end
        error( 'XXX: Saving zHatNext after first iter, Aborting' );
    end
    
    if ( ~isempty( estParams ) )
        switch ( estParams.szMethod )
            case { 'eb', 'eb_fixed', 'map', 'map_prior' }
                % estParams should contain elements a, t, w 
                try
                    tHist = [tHist; estParams.t];
                    aHist = [aHist; estParams.a];
                    wHist = [wHist; estParams.w];
                catch
                    error( 'estParams missing elements a, w, t' );                                       
                end
                
            case 'eb_better'
                % estParams should contain the elements: x(1) -> a, x(2) -> t
                try
                    aHist = [aHist; estParams.x(1)];
                    tHist = [tHist; estParams.x(2)];
                    wHist = [wHist; convert_tsigma2w( estParams.x(1), ...
                        estParams.x(2), alpha, 'median' )];
                catch
                    error( 'estParams missing elements x(1), x(2)' );
                end
                 
            case 'eb_fix_w'
                % estParams should contain the elements: x -> a, w
                try
                    aHist = [aHist; estParams.x];
                    wHist = [wHist; estParams.w];
                    tHist = [tHist; convert_wsigma2t( estParams.x, ...
                        estParams.w, alpha, 'median' )];
                catch
                    error( 'estParams missing elements x, w' );
                end
                
            case { 'binaryd_wfree', 'binaryd', 'binaryd_map' }
                % estParams should contain the elements: thetaStar, w
                try
                    thetaStarHist = [thetaStarHist; estParams.thetaStar];
                    wHist = [wHist; estParams.w];
                catch
                    error( 'estParams missing elements thetaStar, w' );
                end
                
            otherwise
%                 warning( sprintf( ...
%                     'History handling: cannot handle estParams.szMethod=%s', ...
%                     estParams.szMethod ) );
                    
        end			
	end

	if ( sRestoreOptions.B_PLOT == 1 )	

		% Unimplemented

    end

	if ( sRestoreOptions.EM_PAUSE_NUMSECS > 0 )
		disp( sprintf( 'Pausing for %g seconds...', ...
			sRestoreOptions.EM_PAUSE_NUMSECS ) );
		pause( sRestoreOptions.EM_PAUSE_NUMSECS );
	end

	% Save history?
	if ( bSaveHistory == 1 ), thetaHist{i} = thetaHatNext; end;

    % Store the difference in the estimates between iterations
    thetaHatDiff = thetaHatNext(:) - thetaHatPrev(:);
    % The l_infinity norm of the difference
    thetaMaxDiffHist = [thetaMaxDiffHist; max( abs( thetaHatDiff ) )];
    % The l_1 norm of the difference
    thetaSumDiffHist = [thetaSumDiffHist; sum( abs( thetaHatDiff ) )];

	% Update thetaHat
	thetaHatPrev = thetaHatNext;

    % Increment loop counter
    i = i + 1;
    
end % while

% Turn off diary
%diary off;

% Compute the actual joint probability of y and theta
jointProbReal = -1; % Initialize
logPcond = logp_ytheta_cond_gauss( imgNoisy, thetaTrue, hForwardOpDft, noiseSigma );
bComputeLogPtheta = 1;
 
switch ( estParams.szMethod )
    case {'eb', 'eb_fixed', 'eb_better', 'map', 'map_prior' }
        logPtheta = logp_hyper_john_unif( thetaTrue, estParams.aRange );        
    case { 'binaryd', 'binaryd_map' }
        logPtheta = logp_hyper_binaryd_unif( thetaHatNext, ...
            estParams.thetaStarRange );
	case 'eb_fix_w'
		logPtheta = logp_hyper_laplace_unif( thetaHatNext, ...
			estParams.aRange );
    otherwise
        bComputeLogPtheta = 0;
end

if bComputeLogPtheta
    jointProbReal = logPcond + logPtheta;
else
    jointProbReal = -1; % Invalid
end

reconElems = { 'last', 'thetaHatPrev'; ...
    'init', 'thetaHat0'; ...
    't', 'tHist'; ...
    'a', 'aHist'; ...
    'w', 'wHist'; ...
    'thetaStar', 'thetaStarHist'; ...
    'thetaMaxDiff', 'thetaMaxDiffHist'; ...
    'thetaSumDiff', 'thetaSumDiffHist'; ...
    'errorNorm1', 'errorNorm1Hist'; ...
    'errorNorm2', 'errorNorm2Hist'; ...
    'jointProb', 'jointProbHist'; ...
    'jointProbReal', 'jointProbReal'; ...
    'cost', 'costFuncHist' };

for i = 1:size( reconElems, 1 )
    szCmd = sprintf( 'sRecon.%s = %s;', char( reconElems(i,1) ), ...       
        char( reconElems(i,2) ) );
    try
        eval( szCmd );
    catch
        warning( sprintf( 'Could not execute: %s', szCmd ) );
    end
end

if ( bSaveHistory == 1 ), sRecon.thetaHist = thetaHist; end

% em_restore_e: Performs the E step in the EM algorithm

% ...................................................................
function zHat = em_restore_e( alpha, sigma, Hty, gamH_gam, thetaHat )
% ###################################################################

thetaHatDft = fftn( thetaHat );
residual = Hty - real( ifftn( gamH_gam .* thetaHatDft ) );
zHat = thetaHat + (alpha/sigma)^2*residual;

% em_restore_m: Performs the M step in the EM algorithm. See switch 
% statement in function body for valid options.

% ..............................................................................
function [thetaHat, estParams] = em_restore_m( zHat, szMethod, sRestoreOptions )
% ##############################################################################

persistent estParamsPrev

estParams = [];

switch ( lower( szMethod ) )
    case 'eb'
        %
        % XXX: Depreciated. Use eb_better option instead.
        %
        error( 'DEPRECIATED: use eb_better instead' );

        % Default value of options structure
        sLocOptions = struct( 'bSkipInitialGridSearch', 0, 'nDebugFlag', 0 );

        if ( sRestoreOptions.numGradientSearchFailures < ...
                sRestoreOptions.GRADIENT_SEARCH_FAILURES_MAX )

            if ( sRestoreOptions.nEmIter >= ...
                    sRestoreOptions.EM_SKIP_INITIAL_GRID_THRES )

                if ( isempty( estParamsPrev ) )
                    error( 'estParamsPrev should not be empty' );
                end
                sLocOptions = struct( 'bSkipInitialGridSearch', 1, ...
                    'alpha0', [estParamsPrev.a, estParamsPrev.t]', ...
                    'nDebugFlag', 0 );
            end

            [estParams, est] = embayfunc( zHat(:), 'laplace', [], ...
                sRestoreOptions.alpha, 'median', sLocOptions );
            if ( estParams.exitFlag <= 0 )
                sRestoreOptions.numGradientSearchFailures = ...
                    sRestoreOptions.numGradientSearchFailures + 1;
            end

        else
            [estParams, est] = embayfunc( zHat(:), 'laplace', [], ...
                sRestoreOptions.alpha, 'median_no_grad', sLocOptions );
        end

    case 'eb_fixed'
        if ( strcmp( lower(sRestoreOptions.EB_FIXED_SPARSE_IND), 't' ) == 1 )
            disp( sprintf( 'eb_fixed using a=%g, t=%g', ...
                sRestoreOptions.EB_FIXED_AVAL, sRestoreOptions.EB_FIXED_TVAL ) );
            est = laplace_posterior_med( zHat(:), ...
                sRestoreOptions.EB_FIXED_AVAL, sRestoreOptions.EB_FIXED_TVAL, ...
                sRestoreOptions.alpha );
        else
            % EB_FIXED_SPARSE_IND == w
            thresMed = convert_wsigma2t( sRestoreOptions.EB_FIXED_AVAL, ...
                sRestoreOptions.EB_FIXED_WVAL, sRestoreOptions.alpha, ...
                'median' );
            disp( sprintf( 'eb_fixed using a=%g, w=%g -> t=%g', ...
                sRestoreOptions.EB_FIXED_AVAL, sRestoreOptions.EB_FIXED_WVAL, ...
                thresMed ) );
            est = laplace_posterior_med( zHat(:), sRestoreOptions.EB_FIXED_AVAL, ...
                thresMed, sRestoreOptions.alpha );
        end
          
    case 'eb_better'
        %
        % Replaces eb. This is a cleaner interface, as it separates the 
        % maximization of the marginal ML separately from the thresholding
        % step. 
        %
        estParams = john_optim2( zHat(:), sRestoreOptions.alpha, 0 );
        est = laplace_posterior_med( zHat(:), estParams.x(1), estParams.x(2), ...
            sRestoreOptions.alpha );

    case 'eb_fix_w'
        %
        % EB with w FIXED
        %
        estParams = john_optim1a( zHat(:), sRestoreOptions.alpha, ...
            sRestoreOptions.EB_FIXED_WVAL, 0 );
        t = convert_wsigma2t( estParams.x, estParams.w, sRestoreOptions.alpha, ...
            'median' );
        % DEBUG
        disp( sprintf( 'eb_fix_w: x=%g, w=%g, t=%g', estParams.x, ...
            estParams.w, t ) );
        est = laplace_posterior_med( zHat(:), estParams.x, estParams.w, ...
            sRestoreOptions.alpha, 'w' );
        
    case 'map'
        %
        % MAP without bias correction prior. Produces results with high
        % bias.
        %
        if ( sRestoreOptions.numGradientSearchFailures < ...
                sRestoreOptions.GRADIENT_SEARCH_FAILURES_MAX )

            [estParams, est] = mapfunc( zHat(:), 'laplace', [], ...
                sRestoreOptions.alpha, 'map', 0 );
            if ( estParams.exitFlag <= 0 )
                sRestoreOptions.numGradientSearchFailures = ...
                    sRestoreOptions.numGradientSearchFailures + 1;
            end

        else
            [estParams, est] = mapfunc( zHat(:), 'laplace', [], ...
                sRestoreOptions.alpha, 'map_no_grad', 0 );
        end

    case 'map_prior'

        % Latest version of map_2d
        [aHat, tHat, wHat, muHat] = map_2d_ver4( zHat(:), sRestoreOptions.alpha );

        estParams = struct( 'a', aHat, 't', tHat, 'w', wHat, 'sdev', ...
            sRestoreOptions.alpha );
        est = muHat(:);

        banner( 'map_prior estimates' );
        disp( sprintf( 'aHat: %g, tHat: %g, wHat: %g', estParams.a, ...
            estParams.t, estParams.w ) );

    case 'map_fixed'
        est = laplace_map_thres0( zHat(:), sRestoreOptions.EB_FIXED_AVAL, ...
            sRestoreOptions.EB_FIXED_WVAL, sRestoreOptions.alpha );

    case 'id'
        est = zHat(:);

    case 'soft_fixed'
        %
        % Use a fixed soft threshold
        %
        est = soft_thres( zHat(:), sRestoreOptions.SOFT_THRES_T );

    case 'binaryd_wfree'
        %
        % Binary delta with thetaStar fixed, so only have to estimate
        % tuning parameter w
        %
        estParams = binaryd_optim1w( zHat(:), sRestoreOptions.alpha, ...
            sRestoreOptions.THETASTAR_FIXED, 1, 1 );
        est = binaryd_thres( zHat(:), estParams.thetaStar, estParams.threshold );     

    case 'binaryd'
        %
        % Binary delta with w, thetaStar free
        %
        estParams = binaryd_optim2( zHat(:), sRestoreOptions.alpha, 1, 1 );
        est = binaryd_thres( zHat(:), estParams.thetaStar, estParams.threshold );
        
    case 'binaryd_map'
        %
        % Binary delta under MAP
        %
        estParams = binaryd_map_optim2( zHat(:), sRestoreOptions.alpha, 1, 1 );
        est = binaryd_thres( zHat(:), estParams.thetaStar, estParams.threshold );
        
    otherwise
        error( 'unimplemented' );
end

estParams.szMethod = lower( szMethod );
if ( strcmp( estParams.szMethod, 'eb' ) )
    estParamsPrev = estParams;
end

thetaHat = reshape( est, size(zHat) );


