% em_restore_m: Performs the M step in the EM algorithm. See switch 
% statement in function body for valid options.
%##########################################################################

function [mData, hParamEst] = em_restore_m( zHat, szMethod, hParamEstPrev )

global sBoundingBox bCloseBoundingBox sRestoreOptions

UPPER_BOUND_SCALE_MAX = 1e3;
LOWER_BOUND_SCALE_MIN = 1e-6;

% Initialize with the previous hyper-parameter estimates
hParamEst = hParamEstPrev; 

n = length( zHat(:) );
empSdev = std( zHat(:) );
tUniv = sqrt( 2*log(n) );
alpha = sRestoreOptions.alpha;

szMethodLower = lower( szMethod );

% Only initialize this once
if isempty( sBoundingBox )
	switch ( szMethodLower )
		case { 'eb_better_nov05', 'eb_complete', 'eb_fix_w_nov05' }
			sBoundingBox = struct( 'loBoundScale', [1e-3, 1e-3, 1e-3], ...
				'hiBoundScale', [3, tUniv, 1], ...
				'loBoundScaleMin', 1e-6, ...
				'hiBoundScaleMax', 1e3, ...
				'loBoundInit', [1/empSdev, alpha, empSdev], ...
				'hiBoundInit', [1/empSdev, alpha, empSdev] ...
			);
		case { 'exp1', 'exp2' }
			sBoundingBox = struct( 'loBoundScale', [1e-3, 1e-2], ...
				'hiBoundScale', [1e3, 1e2], ...
				'loBoundScaleMin', 1e-6, ...
				'hiBoundScaleMax', 1e6, ...
				'loBoundInit', [1, alpha], ...
				'hiBoundInit', [1, alpha] ...
			);
		otherwise
			% Dummy
			sBoundingBox = struct( 'loBoundScale', 1, ...
				'hiBoundScale', 1, ...
				'loBoundInit', -1, ...
				'hiBoundInit', -1 ...
			);
	end

	% Initialize bCloseBoundingBox whenever we initialize sBoundingBox
	if ( length( sBoundingBox.loBoundScale ) ~= length( sBoundingBox.hiBoundScale ) )
		error( 'unequal lengths of bounding box vectors' );
	end
	bCloseBoundingBox = zeros( length(sBoundingBox.loBoundScale), 2 );

end
 
paramLowerBound = sBoundingBox.loBoundScale .* sBoundingBox.loBoundInit;
paramUpperBound = sBoundingBox.hiBoundScale .* sBoundingBox.hiBoundInit;

pLen = length( paramLowerBound(:) );
tolMatrix = [ 1e-4*ones(1,pLen); ...
              1e-4*( paramUpperBound(:)' - paramLowerBound(:)' ); ...
              0.1*paramLowerBound(:)' ];
tol = min( tolMatrix );

% tol is only used in methods that have a search in the M-step
switch ( szMethodLower )
    case { 'eb_better_nov05', 'eb_complete', 'eb_fix_w_nov05', 'exp1', 'exp2' }
        disp( 'tol' ), disp( tol );
end
        
switch ( szMethodLower )
    case 'eb_fixed' 
        % EBD but with a fixed thresholding rule
        if ( strcmp( lower(sRestoreOptions.EB_FIXED_SPARSE_IND), 't' ) == 1 )
            % Use (a,t) to determine threshold function
            disp( sprintf( 'eb_fixed using a=%g, t=%g', ...
                sRestoreOptions.EB_FIXED_AVAL, sRestoreOptions.EB_FIXED_TVAL ) );
            est = laplace_posterior_med( zHat(:), ...
                sRestoreOptions.EB_FIXED_AVAL, ...
				sRestoreOptions.EB_FIXED_TVAL, ...
                alpha );
        else
            % Use (a,w) to determine threshold function
            thresMed = convert_wsigma2t( sRestoreOptions.EB_FIXED_AVAL, ...
                sRestoreOptions.EB_FIXED_WVAL, alpha, ...
                'median' );
            disp( sprintf( 'eb_fixed using a=%g, w=%g -> t=%g', ...
                sRestoreOptions.EB_FIXED_AVAL, sRestoreOptions.EB_FIXED_WVAL, ...
                thresMed ) );
            est = laplace_posterior_med( zHat(:), ...
				sRestoreOptions.EB_FIXED_AVAL, ...
                thresMed, ...
				alpha );
        end
          
    case 'eb_better'        
        % Replaces eb. This is a cleaner interface, as it separates the
        % maximization of the marginal ML separately from the thresholding
        % step.
        
        hParamEst = john_optim2( zHat(:), alpha, 0 );                
        est = laplace_posterior_med( zHat(:), hParamEst.x(1), hParamEst.x(2), ...
            alpha );
        
    case 'eb_better_nov05'
        
        maxIters = 3e2;
        debugFlag = 0;
        
        % 22-jan-06: in order to speed up the algorithm, do not perform the
        % estimation of the hyperparameters in each iteration.
        
        if ( sRestoreOptions.nHparamRestInterval <= 0 ) | ...
           ( ( sRestoreOptions.nHparamRestInterval > 0 ) & ...
             ( mod( sRestoreOptions.nEmIter, sRestoreOptions.nHparamRestInterval ) == 1 ) )

            sOptim2 = probjs_wrapper( 'y', zHat(:), ...
                'ylength', n, ...
                'lowerbound', paramLowerBound(1:2), ...
                'upperbound', paramUpperBound(1:2), ...
                'tol', tol(1:2), ...
                'maxiters', maxIters, ...
                'sigma', alpha, ...
                'debugflag', debugFlag ...
                );

            % Re-adjust bounding box for hyperparameters if necessary
            sBoundingBox = change_scale( sBoundingBox, bCloseBoundingBox, ...
                sOptim2.voptim, 1e3*eps, paramLowerBound(1:2), paramUpperBound(1:2) );

            hParamEst = sOptim2;
            hParamEst.aRange = [paramLowerBound(1), paramUpperBound(1)];
            hParamEst.tRange = [paramLowerBound(2), paramUpperBound(2)];
		else
			% Re-use previous hyperparameter estimates
		end

        est = laplace_posterior_med( zHat(:), hParamEst.voptim(1), hParamEst.voptim(2), alpha );


    case 'eb_complete'
        maxIters = 2e2;
        debugFlag = 0;

        sOptim3 = probjs_wrapper( 'y', zHat(:), ...
            'ylength', n, ...
            'lowerbound', paramLowerBound, ...
            'upperbound', paramUpperBound, ...
            'tol', tol, ...
            'maxiters', maxIters, ...
            'debugflag', debugFlag ...
            );    

		sBoundingBox = change_scale( sBoundingBox, bCloseBoundingBox, ...
			sOptim3.voptim, 1e3*eps, paramLowerBound, paramUpperBound );

        est = laplace_posterior_med( zHat(:), ...
			sOptim3.voptim(1), sOptim3.voptim(2), sOptim3.voptim(3) );

        hParamEst = sOptim3;
		hParamEst.hyperParamRange = [ paramLowerBound(1), paramUpperBound(1), ...
			paramLowerBound(2), paramUpperBound(2), ...
			paramLowerBound(3), paramUpperBound(3) ];

    case 'eb_fix_w'
        % EB with w FIXED. We estimate a via marginal ML, however.
        hParamEst = john_optim1a( zHat(:), alpha, ...
            sRestoreOptions.EB_FIXED_WVAL, 0 );
        t = convert_wsigma2t( hParamEst.x, hParamEst.w, alpha, ...
            'median' );
        % DEBUG
%         disp( sprintf( 'eb_fix_w: x=%g, w=%g, t=%g', hParamEst.x, ...
%             hParamEst.w, t ) );
        % In the general case, the posterior median function is obtained
        % via laplace_posterior_med.
        est = laplace_posterior_med( zHat(:), hParamEst.x, hParamEst.w, ...
            alpha, 'w' );

    case 'eb_fix_w_nov05'
        maxIters = 2e2;
        debugFlag = 0;

		if 0 % DEBUG
			numTmpPts = 200;
			aVecTmp = logspace( log10(paramLowerBound(1)), ...
				log10(paramUpperBound(1)*10), numTmpPts );
			valTmp = zeros( size( aVecTmp ) );
			for indTmp = 1:numTmpPts
				s = struct( 'a', aVecTmp(indTmp), ...
					't', -1, ...
					'w', sRestoreOptions.EB_FIXED_WVAL, ...
					'sigma', alpha );
				valTmp( indTmp ) = logliksparseJS( s, zHat(:) );
			end
			figure(1), semilogx( aVecTmp, valTmp, 'b' ), xlabel( 'a' );
			[valTmpMax, valTmpMaxInd] = max( valTmp );
			disp( sprintf( 'Max at a=%g, value=%g', ...
				aVecTmp(valTmpMaxInd(1)), valTmpMax(1) ) );
			disp( 'Pausing...' ), pause;
		end

        sOptim1 = probjs_wrapper( 'y', zHat(:), ...
            'ylength', n, ...
            'lowerbound', paramLowerBound(1), ...
            'upperbound', paramUpperBound(1), ...
            'tol', tol(1), ...
            'maxiters', maxIters, ...
            'w', sRestoreOptions.EB_FIXED_WVAL, ...
            'sigma', alpha, ...
            'debugflag', debugFlag ...
            );

		sBoundingBox = change_scale( sBoundingBox, bCloseBoundingBox, ...
			sOptim1.voptim, 1e3*eps, paramLowerBound(1), paramUpperBound(1) );
       
        if sOptim1.optimstatus == 4            
            % In case of error, make this a normal Landweber iteration
            est = zHat(:);
        else
            % Only apply the posterior median function if we succeeded in
            % obtaining an estimate of a
            est = laplace_posterior_med( zHat(:), sOptim1.voptim(1), ...
                sRestoreOptions.EB_FIXED_WVAL, alpha, 'w' );
        end
                
        hParamEst = sOptim1;
        hParamEst.w = sRestoreOptions.EB_FIXED_WVAL;
        hParamEst.aRange = [paramLowerBound(1), paramUpperBound(1)];

    case 'map1_fixed'
        % MAP1 with fixed thresholding rule
        est = laplace_map_thres0( zHat(:), sRestoreOptions.EB_FIXED_AVAL, ...
            sRestoreOptions.EB_FIXED_WVAL, alpha );
        
	case 'map1'

		% Note that we use the GIVEN alpha here, and are NOT estimating it
		% using the MAP principle

		thres0 = hParamEst.a * alpha^2;

		% SANITY on rMap1
		if ( abs( hParamEst.w ) < eps )
			error( 'map1: |wHat| < eps' );
		else
			rMap1 = (1/hParamEst.w-1);
		end

		if ( hParamEst.w <= 0.5 ), thres = thres0 + alpha*sqrt( 2*log(rMap1) );
		else, thres = thres0;
		end

		hParamEst.thres = thres;

		% SANITY on thres (lower bound)
		if ( thres < 0 ), error( 'Negative thres is invalid' ), end;

		% SANITY on thres (upper bound)
		if ( thres > tUniv*alpha )
			% Print warning only once
			if ( sRestoreOptions.numThresWarning == 0 )
				warning( 'Pos thres exceeds univ. threshold: check numThresWarning counter' )
			end
			sRestoreOptions.numThresWarning = sRestoreOptions.numThresWarning + 1;
			% Limit threshold to the universal threshold
			if sRestoreOptions.bUpperBoundThres, thres = tUniv*alpha; end;
		end;
        
% 		est1 = zHat - thres0;
% 		est2 = zHat + thres0;
% 		est = est1 .* ( zHat > thres ) + est2 .* ( zHat < -thres );
	
        % In map1, do not combine the estimate of \tilde{\theta} and I
        % together. Instead, keep them separate.
        
        estI = ( abs( zHat ) >= thres );
        estThetaTilde = ( abs( zHat ) >= thres0 ) .* ( zHat - sign(zHat)*thres0 ) .* estI;
        
    case 'map2_fixed'
        
        gSpecial = sRestoreOptions.gSpecial;
        rMap2 = gSpecial/(sRestoreOptions.EB_FIXED_AVAL/2) * ...
            (1/sRestoreOptions.EB_FIXED_WVAL-1);

        thres0 = sRestoreOptions.EB_FIXED_AVAL * alpha^2;
        
		thres = -1; % Initialize
		if ( rMap2 < 0 )
			error( sprintf( 'rMap2=%g < 0?', rMap2 ) );
		elseif ( ( rMap2 >= 0 ) & ( rMap2 < 1 ) )
			thres = thres0;
		elseif ( rMap2 >= 1 )
			thres = thres0 + alpha*sqrt( 2*log( rMap2 ) );
        end
        
        est1 = zHat - thres0;
        est2 = zHat + thres0;
        est = est1 .* ( zHat > thres ) + est2 .* ( zHat < -thres );
        
	case 'map2'

		% Note comment in map1, which also applies here.

		thres0 = hParamEst.a * alpha^2;

		if ( sRestoreOptions.gSpecial < 0 ), gSpecial = hParamEst.a/2;
		else, gSpecial = sRestoreOptions.gSpecial;
		end

		% SANITY on rMap2
		if ( abs( hParamEst.a ) < eps ) 
			error( 'map2: |aHat| < eps' );
		elseif ( abs( hParamEst.w ) < eps )
			error( 'map2: |wHat| < eps' );
		else
			rMap2 = gSpecial/(hParamEst.a/2) * (1/hParamEst.w-1);
		end

		thres = -1; % Initialize
		if ( rMap2 < 0 )
			error( sprintf( 'rMap2=%g < 0?', rMap2 ) );
		elseif ( ( rMap2 >= 0 ) & ( rMap2 < 1 ) )
			thres = thres0;
		elseif ( rMap2 >= 1 )
			thres = thres0 + alpha*sqrt( 2*log( rMap2 ) );
		end

		hParamEst.thres = thres;
		
		% SANITY on thres (lower bound)
		if ( thres < 0 ), error( 'Negative thres is invalid' ), end;

		% SANITY on thres (upper bound)
		if ( thres > tUniv*alpha )
			% Print warning only once
			if ( sRestoreOptions.numThresWarning == 0 )
				warning( 'Pos thres exceeds univ. threshold: check numThresWarning counter' )
			end
			sRestoreOptions.numThresWarning = sRestoreOptions.numThresWarning + 1;
			% Limit threshold to the universal threshold
			if sRestoreOptions.bUpperBoundThres, thres = tUniv*alpha; end;
		end;

		est1 = zHat - thres0;
		est2 = zHat + thres0;
		est = est1 .* ( zHat > thres ) + est2 .* ( zHat < -thres );
			 
    case 'map_prior'
        % MAP with bias correction prior. Use with latest version of
        % map_2d. XXX: this is numerically unstable and should be avoided!
        
        [aHat, tHat, wHat, muHat, aRange, tRange] = map_2d_ver4( zHat(:), ...
			alpha );

        hParamEst = struct( 'a', aHat, 't', tHat, 'w', wHat, 'sdev', ...
            alpha, 'aRange', aRange, 'tRange', tRange );
        est = muHat(:);

        banner( 'map_prior estimates' );
        disp( sprintf( 'aHat: %g, tHat: %g, wHat: %g', hParamEst.a, ...
            hParamEst.t, hParamEst.w ) );

    case 'id'
        % Use the identity operator in the M step. This produces the
        % standard Landweber iterations.
        est = zHat(:);

    case 'idpos'
        % Use the identity operator, but project onto the positive orthant.
        % This effectively zeros out the negative components of zHat.
        est = zHat .* (zHat >= 0);
        est = est(:);
        
    case 'soft_fixed'
        % Use a fixed soft threshold. These iterations will produce the MAP
        % estimate under a Laplacian prior on theta.
        est = soft_thres( zHat(:), sRestoreOptions.SOFT_THRES_T );

    case 'daub1'
        est = soft_thres( zHat(:), 0.5*sRestoreOptions.lambda/sRestoreOptions.daub1NormConst );       

    case 'daub_hard_soft'
        softThres = 0.5*sRestoreOptions.lambda/sRestoreOptions.daub1NormConst;
        frac = sRestoreOptions.hard_soft_frac;
        if ( ( frac < 0 ) | ( frac > 1 ) ), error( 'illegal hard_soft_frac' ), end;
        est = hard_soft_thres( zHat(:), softThres, softThres*frac );
        
    case 'daub2'
        est = soft_thres( zHat(:), 0.5*sRestoreOptions.lambda/sRestoreOptions.daub2NormConst );
               
    case 'binaryd_wfree'
        % Binary delta with thetaStar fixed, so only have to estimate
        % tuning parameter w. The parameter thetaStar is the position of
        % the impulse that represents the non-zero values of theta.
        hParamEst = binaryd_optim1w( zHat(:), alpha, ...
            sRestoreOptions.THETASTAR_FIXED, 1, 1 );
        est = binaryd_thres( zHat(:), hParamEst.thetaStar, hParamEst.threshold );     

    case 'binaryd'
        % Binary delta with w AND thetaStar free
        hParamEst = binaryd_optim2( zHat(:), alpha, 1, 1 );
        est = binaryd_thres( zHat(:), hParamEst.thetaStar, hParamEst.threshold );
        
    case 'binaryd_map'
        % Binary delta under MAP
        hParamEst = binaryd_map_optim2( zHat(:), alpha, 1, 1 );
        est = binaryd_thres( zHat(:), hParamEst.thetaStar, hParamEst.threshold );

    case 'exp1'
        
        maxIters = 2e2;
        debugFlag = 0;

        sOptim1 = probexp_wrapper( 'y', zHat(:), ...
            'ylength', n, ...
            'lowerbound', paramLowerBound(1), ...
            'upperbound', paramUpperBound(1), ...
            'tol', tol(1), ...
            'maxiters', maxIters, ...
            'sigma', alpha, ...
            'debugflag', debugFlag ...
            );

        sBoundingBox = change_scale( sBoundingBox, bCloseBoundingBox, ...
			sOptim1.voptim, 1e3*eps, paramLowerBound, paramUpperBound );

        est = exp_thres( zHat(:), sOptim1.voptim(1), alpha );

		hParamEst = sOptim1;
		hParamEst.vLowerBound = paramLowerBound(1);
		hParamEst.vUpperBound = paramUpperBound(1);
            
    case 'exp2'
        
        maxIters = 2e2;
        debugFlag = 0;
        
        sOptim2 = probexp_wrapper( 'y', zHat(:), ...
            'ylength', n, ...
            'lowerbound', paramLowerBound(1:2), ...
            'upperbound', paramUpperBound(1:2), ...
            'tol', tol(1:2), ...
            'maxiters', maxIters, ...          
            'debugflag', debugFlag ...
            );
        
		sBoundingBox = change_scale( sBoundingBox, bCloseBoundingBox, ...
			sOptim2.voptim, 1e3*eps, paramLowerBound, paramUpperBound );

        est = exp_thres( zHat(:), sOptim2.voptim(1), sOptim2.voptim(2) );
        
		hParamEst = sOptim2;
		hParamEst.vLowerBound = [paramLowerBound(1:2)];
		hParamEst.vUpperBound = [paramUpperBound(1:2)];

    otherwise
        error( 'unimplemented M-step method' );
end

% Produce the output mData
switch ( szMethodLower )
    case 'map1'
        mData = [ estI(:), estThetaTilde(:) ];
    otherwise
        mData = est(:);
end



