% // ORIGINAL COMMENT // Performs the M step in the EM algorithm. See
%    switch statement in function body for valid options.
% // 2008-jul-16 // Modify so that it can be more generally used. In
%    particular, we remove the global variables and comment out any methods
%    that depend on them. In particular, we rely on everything contained in
%    sOptions. Do not do any field checking in this structure. If a field
%    is needed that's missing, Matlab is just going to crap out.
%##########################################################################

function [mData, hParamEst] = em_restore_m2( zHat, szMethod, hParamEstPrev, ...
    sOptions )

% Initialize with the previous hyper-parameter estimates
hParamEst = hParamEstPrev; 

n = length( zHat(:) );
empSdev = std( zHat(:) );
tUniv = sqrt( 2*log(n) );
alpha = sOptions.alpha;

szMethodLower = lower( szMethod );
        
switch ( szMethodLower )
    
    case 'map1_fixed'
        % MAP1 with fixed thresholding rule
        est = laplace_map_thres0( zHat(:), sOptions.map1_fixed.a, ...
            sOptions.map1_fixed.w, alpha );
        
	case 'map1'

		% Note that we use the GIVEN alpha here, and are NOT estimating it
		% using the MAP principle

		thres0 = hParamEst.a * alpha^2;

		% SANITY on rMap1
		if ( abs( hParamEst.w ) < eps )
			error( 'map1: |wHat| < eps' );
		else
			rMap1 = (1/hParamEst.w-1);
		end

		if ( hParamEst.w <= 0.5 ), thres = thres0 + alpha*sqrt( 2*log(rMap1) );
		else, thres = thres0;
		end

		% SANITY on thres (lower bound)
		if ( thres < 0 ), error( 'Negative thres is invalid' ), end;

		% SANITY on thres (upper bound)
		if ( thres > tUniv*alpha )
            fprintf( 1, 'thres=%g > tUniv*alpha=%g\n', thres, tUniv*alpha );
			sOptions.map1.numThresWarning = sOptions.map1.numThresWarning + 1;
			% Limit threshold to the universal threshold
			if sOptions.map1.bUpperBoundThres, thres = tUniv*alpha; end;
        end
        
% 		est1 = zHat - thres0;
% 		est2 = zHat + thres0;
% 		est = est1 .* ( zHat > thres ) + est2 .* ( zHat < -thres );
	
        % In map1, do not combine the estimate of \tilde{\theta} and I
        % together. Instead, keep them separate.
        
        if ( sOptions.map1.bDebug )
            fprintf( 1, 'em_restore_m2 [%g]: map1: thres0=%g, thres=%g\n', ...
                sOptions.map1.nIter, thres0, thres );
        end
            
        if ( thres0 > thres )
            warning( 'thres0 > thres: setting thres0 <- thres' );
            thres0 = thres;
        end
        
        hParamEst.thres = thres;
                
        estI = ( abs( zHat ) >= thres );
        estThetaTilde = ( abs( zHat ) >= thres0 ) .* ( zHat - sign(zHat)*thres0 ) .* estI;
        
    case 'map2_fixed'
        
        gSpecial = sOptions.map2_fixed.gSpecial;
        rMap2 = gSpecial/(sOptions.map2_fixed.a/2) * ...
            (1/sOptions.map2_fixed.w-1);

        thres0 = sOptions.map2_fixed.a * alpha^2;
        
		thres = -1; % Initialize
		if ( rMap2 < 0 )
			error( sprintf( 'rMap2=%g < 0?', rMap2 ) );
		elseif ( ( rMap2 >= 0 ) & ( rMap2 < 1 ) )
			thres = thres0;
		elseif ( rMap2 >= 1 )
			thres = thres0 + alpha*sqrt( 2*log( rMap2 ) );
        end
        
        est1 = zHat - thres0;
        est2 = zHat + thres0;
        est = est1 .* ( zHat > thres ) + est2 .* ( zHat < -thres );
        
	case 'map2'

		% Note comment in map1, which also applies here.

		thres0 = hParamEst.a * alpha^2;

		if ( sOptions.map2.gSpecial < 0 ), gSpecial = hParamEst.a/2;
		else, gSpecial = sOptions.map2.gSpecial;
		end

		% SANITY on rMap2
		if ( abs( hParamEst.a ) < eps ) 
			error( 'map2: |aHat| < eps' );
		elseif ( abs( hParamEst.w ) < eps )
			error( 'map2: |wHat| < eps' );
		else
			rMap2 = gSpecial/(hParamEst.a/2) * (1/hParamEst.w-1);
		end

		thres = -1; % Initialize
        
        if ( true ) % XXX
            fprintf( 1, '\t/// rMap2=%g\n', rMap2 );
        end
        
		if ( rMap2 < 0 )
			error( sprintf( 'rMap2=%g < 0?', rMap2 ) );
		elseif ( ( rMap2 >= 0 ) & ( rMap2 < 1 ) )
			thres = thres0;
		elseif ( rMap2 >= 1 )
			thres = thres0 + alpha*sqrt( 2*log( rMap2 ) );
        end
		
		% SANITY on thres (lower bound)
		if ( thres < 0 ), error( 'Negative thres is invalid' ), end;

		% SANITY on thres (upper bound)
        if ( thres > tUniv*alpha )
            fprintf( 1, 'thres=%g > tUniv*alpha=%g\n', thres, tUniv*alpha );
            sOptions.map2.numThresWarning = sOptions.map2.numThresWarning + 1;
            % Limit threshold to the universal threshold
            if sOptions.map2.bUpperBoundThres, thres = tUniv*alpha; end;
        end

        if ( sOptions.map2.bDebug )
            fprintf( 1, 'em_restore_m2 [%g]: map2: thres0=%g, thres=%g\n', ...
                sOptions.map2.nIter, thres0, thres );
        end

        if ( thres0 > thres )
            warning( 'thres0 > thres: setting thres0 <- thres' );
            thres0 = thres;
        end
        
        hParamEst.thres = thres;
        
		est1 = zHat - thres0;
		est2 = zHat + thres0;
		est = est1 .* ( zHat > thres ) + est2 .* ( zHat < -thres );
			 
    case 'id'
        % Use the identity operator in the M step. This produces the
        % standard Landweber iterations.
        est = zHat(:);

    case 'idpos'
        % Use the identity operator, but project onto the positive orthant.
        % This effectively zeros out the negative components of zHat.
        est = zHat .* (zHat >= 0);
        est = est(:);
        
    case 'soft_fixed'
        % Use a fixed soft threshold. These iterations will produce the MAP
        % estimate under a Laplacian prior on theta.
        est = soft_thres( zHat(:), sOptions.soft_fixed.thres );

    case 'daub1'
        est = soft_thres( zHat(:), 0.5*sOptions.daub1.lambda/sOptions.daub1.normConst );       

    case 'daub_hard_soft'
        softThres = 0.5*sOptions.daub_hard_soft.lambda/sOptions.daub1.normConst;
        frac = sOptions.daub_hard_soft.frac;
        if ( ( frac < 0 ) | ( frac > 1 ) ), error( 'illegal hard_soft_frac' ), end;
        est = hard_soft_thres( zHat(:), softThres, softThres*frac );
        
    case 'daub2'
        est = soft_thres( zHat(:), 0.5*sOptions.daub2.lambda/sOptions.daub2.normConst );
               
    otherwise
        error( 'unimplemented M-step method' );
end

% Produce the output mData
switch ( szMethodLower )
    case 'map1'
        mData = [ estI(:), estThetaTilde(:) ];
    otherwise
        mData = est(:);
end



