% empbayfunc: Our home-cooked version of ebayesthresh
%
% Usage:
%    [estParams, thetaHat] = embayfunc( vecY, szPrior, valA, valSdev, ...
%       szThresRule, sOptions )
%

function [estParams, thetaHat] = embayfunc( vecY, szPrior, valA, valSdev, ...
	szThresRule, sOptions )

% Globals
global y tmp aGiven sdevGiven sMapInfo nThresFunc thresUniv

global paramA paramW paramSdev

global THRESFUNC_LAPLACE_MEDIAN THRESFUNC_LAPLACE_MAP

% Constants

PRIOR_LAPLACE = 1;

THRES_MEDIAN = 1;
THRES_MAP0 = 2;
THRES_MAP1 = 3;
THRES_MAP2 = 4;

THRESFUNC_LAPLACE_MEDIAN = 1;
THRESFUNC_LAPLACE_MAP = 2;

y = vecY(:);
ySigma = std( y );
tmp = zeros( size( y ) );
n = max( size( y ) );
aGiven = valA;
sdevGiven = valSdev;

switch ( lower( szPrior ) )
case 'laplace'
	disp( 'using Laplace prior' );
	prior = PRIOR_LAPLACE;

otherwise
	error( 'unsupported prior' );
end

% Initially, we had this only for mapfunc, but it seems that the gradient 
% search *DOES* also fail for the EBayes method. So borrow the handling
% routines from mapfunc.
bGradientSearch = 1; 

switch ( lower( szThresRule ) )
case 'median'
	disp( 'using median thresholding rule' );
	thres = THRES_MEDIAN;
	if ( prior == PRIOR_LAPLACE ), nThresFunc = THRESFUNC_LAPLACE_MEDIAN; end;

case 'median_no_grad'
	disp( 'using median thresholding rule with no gradient search' );
	thres = THRES_MEDIAN;
	if ( prior == PRIOR_LAPLACE ), nThresFunc = THRESFUNC_LAPLACE_MEDIAN; end;
	bGradientSearch = 0;

case 'map0'
	disp( 'using MAP0 thresholding rule' );
	thres = THRES_MAP0;
case 'map1'
	disp( 'using MAP1 thresholding rule' );
	thres = THRES_MAP1;
case 'map2'
	disp( 'using MAP2 thresholding rule' );
	thres = THRES_MAP2;

otherwise
	error( 'unsupported thresholding rule' );
end

if ( prior == PRIOR_LAPLACE )
	if ( thres == THRES_MEDIAN )
		nThresFunc = THRESFUNC_LAPLACE_MEDIAN;
	else
		nThresFunc = THRESFUNC_LAPLACE_MAP;
	end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Step 1: find ML estimates of w (optionally a and sdev also)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

banner( 'Stage 1: finding ML estimates' );

thresUniv = sqrt( 2*log(n) );
disp( sprintf( 'thresUniv = %g', thresUniv ) );

tic; % Start timing

% Initialize
paramT = -1;
paramA = -1;
paramW = -1;
paramSdev = -1;

exitFlag = 1;

ALPHA_DELTA_MULT = 0.5;
ALPHA_PERCENT_ACCURACY = 1e-2;
ALPHA1_GRID_SIZE = 10;
ALPHA2_GRID_SIZE = 10;
ALPHA3_GRID_SIZE = 10;

if ( ~isempty( sdevGiven ) ) 

	%--- CASE 1: sigma is given
	%##########################

	paramSdev = sdevGiven;
	
	options = optimset( 'Display', 'final', 'MaxFunEvals', 5e3, ...
		'MaxIter', 1e3 );
	%options = optimset( 'Display', 'final' );

	if ( isempty( aGiven ) )
		% CASE 1a: a isn't given, so we have to do a 2-d search for [a,t]
		% Define: alpha = [a, t], i.e. find a_ML and t_ML
		alphaLBound = [1e-6/paramSdev, 1e-6*paramSdev];
		alphaUBound = [5/paramSdev, thresUniv*paramSdev];

		%alpha0 = [0.5/paramSdev, paramSdev];

		if ( bGradientSearch == 1 )

			if ( sOptions.bSkipInitialGridSearch == 1 )
				alpha0 = sOptions.alpha0;
			else

				% Do an initial grid search to locate alpha0 that we use
				% to initialize fmincon

				alpha0 = pick_2d_initial( @neg_laplace_log_lik_at, ...
					[alphaLBound(1),alphaUBound(1)], ...
					[alphaLBound(2),alphaUBound(2)], ...
					(alphaUBound(1)-alphaLBound(1))/ALPHA1_GRID_SIZE, ...
					(alphaUBound(2)-alphaLBound(2))/ALPHA1_GRID_SIZE, ...
					'none', 0 );

				w0 = convert_tsigma2w( alpha0(1), alpha0(2), paramSdev, ...
					'median' );
				f0 = neg_laplace_log_lik_at( alpha0 );
				disp( sprintf( 'alpha0 = [%g,%g], f0 = %g', alpha0(1), ...
					alpha0(2), f0 ) );

				if ( (alpha0(1) == alphaLBound(1)) | ...
				 	 (alpha0(1) == alphaUBound(1)) | ...
					 (alpha0(2) == alphaLBound(2)) | ...
					 (alpha0(2) == alphaUBound(2)) ) 
					warning( 'alpha0 is on boundary' );
				end

			end

			[alphaML, fFinal, exitFlag] = fmincon( @neg_laplace_log_lik_at, ...
				alpha0, [], [], [], [], alphaLBound, alphaUBound, [], options );

			if ( exitFlag <= 0 ) 

				% fmincon was called, but failed to converge. So call 
				% pick_2d_initial again to find the ML parameters.

				alpha1Delta = (alphaUBound(1)-alphaLBound(1))/ALPHA1_GRID_SIZE
				alpha2Delta = (alphaUBound(2)-alphaLBound(2))/ALPHA1_GRID_SIZE

				alpha1Lower = ...
					max( [alphaLBound(1), ...
						alpha0(1)-ALPHA_DELTA_MULT*alpha1Delta] );
				alpha1Upper = ...
					min( [alphaUBound(1), ...
						alpha0(1)+ALPHA_DELTA_MULT*alpha1Delta] );

				alpha2Lower = ...
					max( [alphaLBound(2), ...
						alpha0(2)-ALPHA_DELTA_MULT*alpha2Delta] );
				alpha2Upper = ...
					min( [alphaUBound(2), ...
						alpha0(2)+ALPHA_DELTA_MULT*alpha2Delta] );

				alphaML = pick_2d_initial( @neg_laplace_log_lik_at, ...
					[alpha1Lower, alpha1Upper], [alpha2Lower, alpha2Upper], ...
					(alpha1Upper - alpha1Lower)/ALPHA1_GRID_SIZE, ...
					(alpha2Upper - alpha2Lower)/ALPHA2_GRID_SIZE, 'none', 0 );

				if ( (alphaML(1) == alpha1Lower) | ...
					 (alphaML(1) == alpha1Upper) | ...
					 (alphaML(2) == alpha2Lower) | ... 
					 (alphaML(2) == alpha2Upper) )
					warning( 'alphaML is on boundary' );
				end

			end
			
		else

			% Skip fmincon altogether.

			alphaML = pick_2d_initial( @neg_laplace_log_lik_at, ...
				[alphaLBound(1),alphaUBound(1)], ...
				[alphaLBound(2),alphaUBound(2)], ...
				(alphaUBound(1)-alphaLBound(1))/ALPHA1_GRID_SIZE, ...
				(alphaUBound(2)-alphaLBound(2))/ALPHA1_GRID_SIZE, ...
				'none', 1 );

			if ( (alphaML(1) == alphaLBound(1)) | ...
			     (alphaML(1) == alphaUBound(1)) | ...
		         (alphaML(2) == alphaLBound(2)) | ... 
			     (alphaML(2) == alphaUBound(2)) )
				warning( 'alphaML is on boundary' );
			end

		end

		fFinal = neg_laplace_log_lik_at( alphaML );

		disp( sprintf( 'a_ML = %g, t_ML = %g -> w = %g', ...
			alphaML(1), alphaML(2), ...
			convert_tsigma2w(alphaML(1),alphaML(2),paramSdev,'median') ) );
		disp( sprintf( 'fFinal = %g', fFinal ) );

		paramA = alphaML(1);
		paramT = alphaML(2);
		paramW = convert_tsigma2w( alphaML(1), alphaML(2), paramSdev, ...
			'median' );

	elseif ( ~isempty( aGiven ) ) 
	
		if ( sOptions.bComputeThres == 1 ) 
			% CASE 1b: a is given, so only have to do a 1-d search for t
			% Define: alpha = [t], i.e. find t_ML

			alphaLBound = 1e-6*paramSdev;
			alphaUBound = thresUniv*paramSdev;
			alpha0 = paramSdev;

			[alphaML, fFinal, exitFlag] = fmincon( @neg_laplace_log_lik_t, ...
				alpha0, [], [], [], [], alphaLBound, alphaUBound, [], options );

			if ( exitFlag <= 0 ), error( 'fmincon experienced problems' ), end;

			disp( sprintf( 'aGiven = %g, t_ML = %g -> w = %g', ...
				aGiven, alphaML(1), ...
				convert_tsigma2w(aGiven,alphaML(1),paramSdev,'median') ) );
			disp( sprintf( 'fFinal = %g', fFinal ) );

			paramA = aGiven;
			paramT = alphaML(1);
			paramW = convert_tsigma2w( aGiven, alphaML(1), paramSdev, ...
				'median' );
		else
			paramA = aGiven;
			paramW = sOptions.w;
			paramT = convert_wsigma2t( paramA, paramW, paramSdev, 'median' );
		end
	end

else

	%--- CASE 2: neither a, t, sigma are given
	%#########################################

	% Assume that we want to find the ML estimates of a, t, sigma
	% alpha = [a, t, sigma]
	alphaLBound = [ 1e-6/ySigma, 1e-6*ySigma, 1e-6*ySigma ];
	alphaUBound = [ 5/ySigma, thresUniv*ySigma, ySigma ];

	%alpha0 = [ 1/ySigma, ySigma, 0.5*ySigma ];

	warning off; % Turn off log of zero warnings

	alpha0 = pick_3d_initial( @neg_laplace_log_lik_atsigma, ...
		[alphaLBound(1),alphaUBound(1)], ...
		[alphaLBound(2),alphaUBound(2)], ...
		[alphaLBound(3),alphaUBound(3)], ...
		(alphaUBound(1)-alphaLBound(1))/ALPHA1_GRID_SIZE, ...
		(alphaUBound(2)-alphaLBound(2))/ALPHA2_GRID_SIZE, ...
		(alphaUBound(3)-alphaLBound(3))/ALPHA3_GRID_SIZE, 'none', 0 );

	warning on;

	if ( alpha0(1) == alphaLBound(1) | alpha0(1) == alphaUBound(1) | ...
		 alpha0(2) == alphaLBound(2) | alpha0(2) == alphaUBound(2) | ...
		 alpha0(3) == alphaLBound(3) | alpha0(3) == alphaUBound(3) )
		warning( 'alpha0 is on boundary' );
	end

	% Evaluate the cost function at alpha0
	w0 = convert_tsigma2w( alpha0(1), alpha0(2), alpha0(3), 'median' );
	f0 = neg_laplace_log_lik_atsigma( alpha0 );
	disp( sprintf( 'alpha0 = [%g,%g,%g]', alpha0(1), alpha0(2), ...
		alpha0(3) ) );
	disp( sprintf( 'w0 = %g, f0 = %g', w0, f0 ) );
	if ( ~isfinite( w0 ) | ~isfinite( f0 ) )
		error( 'w0 and/or f0 is not finite' );
	end

	options = optimset( 'Display', 'final', 'MaxFunEvals', 1e4, ...
		'MaxIter', 1e4 );

	% Turn off divide-by-zero warnings
	warning( 'off', 'MATLAB:divideByZero' );

	[alphaML, fFinal, exitFlag] = fmincon( @neg_laplace_log_lik_atsigma, ...
		alpha0, [], [], [], [], alphaLBound, alphaUBound, [], options );

	% Turn back on
	warning( 'on', 'MATLAB:divideByZero' );

	if ( exitFlag <= 0 ), error( 'fmincon did not return ok' ), end;

	disp( sprintf( 'a_ML = %g, t_ML = %g -> w = %g, sdev_ML = %g', ...
		alphaML(1), alphaML(2), ...
		convert_tsigma2w(alphaML(1),alphaML(2),alphaML(3),'median'), ...
		alphaML(3) ) );
	disp( sprintf( 'fFinal = %g', fFinal ) );

	paramA = alphaML(1);
	paramT = alphaML(2);
	paramW = convert_tsigma2w( alphaML(1), alphaML(2), alphaML(3), 'median' );
	paramSdev = alphaML(3);

end

toc; % Stop timing

estParams = struct( 'a', paramA, 't', paramT, 'w', paramW, 'sdev', paramSdev );

% Piggy back the estParams structure to return the status of fmincon
estParams.exitFlag = exitFlag;

if ( sOptions.nDebugFlag == 1 )
	disp( 'Stopping after step 1' );
	thetaHat = [];
	return;
end

% From this point onwards, use variable names param{A,T,W,Sdev}. Note that
% paramT refers to the threshold using the median thresholding rule.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Step 2: apply the appropriate threshold function
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

banner( 'Stage 2: applying threshold function' );

disp( sprintf( 'paramA: %g, paramT: %g, paramW: %g, paramSdev: %g', ...
	paramA, paramT, paramW, paramSdev ) );

% Print some statistics about the vector y
yMin = min( y(:) );
yMax = max( y(:) );
yAbsMin = min( abs( y(:) ) );
yAbsMax = max( abs( y(:) ) );

disp( sprintf( 'yMin: %g, ymax: %g, yAbsMin: %g, yAbsMax: %g', ...
	yMin, yMax, yAbsMin, yAbsMax ) );

if ( prior == PRIOR_LAPLACE )

	switch ( thres )
	case THRES_MEDIAN
		thetaHat = laplace_posterior_med( y, paramA, paramT, paramSdev );
	
	case THRES_MAP0
		thetaHat = laplace_map_thres0( y, paramA, paramW, paramSdev );

	case THRES_MAP1
		thetaHat = laplace_map_thres1( y, paramA, paramT, paramW, paramSdev );

	case THRES_MAP2
		thetaHat = laplace_map_thres2( y, paramA, paramT, paramW, paramSdev );
	end

end

%%% Functions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% neg_laplace_log_lik_t: use paramSdev, paramA
function f = neg_laplace_log_lik_t( params )

global paramSdev paramA

t = params(1);
f = neg_laplace_log_lik_at( [paramA, t, paramSdev] );


