% Estimates lambda from samples. Assumes that the noise sigma is given.
% Basically a stripped down version of exp_optim2.
% Usage: sOptim = exp_optim1( varargin )
% Keys: 'y', 'lambda0LowerBound', 'lambda0UpperBound', 'sigma'

function sOptim = exp_optim1( varargin )

% Initialize
y = [];
lambda0 = -1;
lambda0LowerBound = -1;
lambda0UpperBound = -1;
noiseSigma = -1;
B_PLOT = 0;

% Process arguments

numArgs = length( varargin ) / 2;

for i = 1:numArgs
    szKey = varargin{2*i-1};
    switch lower( szKey )
        case 'y', y = varargin{2*i};
        case 'lambda0lowerbound', lambda0LowerBound = varargin{2*i};
		case 'lambda0upperbound', lambda0UpperBound = varargin{2*i};
        case 'sigma', noiseSigma = varargin{2*i};
        case 'bplot', B_PLOT = varargin{2*i};
        otherwise
            error( 'invalid szKey' );
    end
end

% Check variables
if ( noiseSigma < eps ), error( 'invalid noiseSigma' ), end;

if ( lambda0LowerBound < 0 | lambda0UpperBound < 0 | ...
	 lambda0LowerBound >= lambda0UpperBound )
	error( 'invalid bounds on lambda0' );
end

y = y(:);
n = length( y );

% Use a grid search to determine the best lambda0
lenLambda0Grid = 100;
lambda0Grid = logspace( log10(lambda0LowerBound), log10(lambda0UpperBound), lenLambda0Grid );
initValue = zeros( size( lambda0Grid ) );

disp( sprintf( 'lambda0 interval: [%g,%g]', lambda0LowerBound, lambda0UpperBound ) );

warning off;
for i = 1:lenLambda0Grid
	initValue(i) = minusLogf( lambda0Grid(i) );
end
warning on;

if B_PLOT
    figure( 1 ), clf, semilogx( lambda0Grid, initValue, 'bx' );
    figure( 2 ), clf, hist( y, 50 );
    pause;
end

[gridMin, gridMinInd] = min( initValue );

% Check if initValue is finite
if ~isfinite( initValue(gridMinInd(1)) )
    error( 'not finite' );
end

lambda0 = lambda0Grid( gridMinInd(1) );
disp( sprintf( '>USING lambda0=%g', lambda0 ) );

    function [f,g] = minusLogf( lambda )
        
        tmp = sum( log( normcdf( y/noiseSigma - noiseSigma*lambda ) ) - lambda*y );
        tmp = tmp + n*log(lambda) + n/2*noiseSigma^2*lambda^2;
        f = -tmp;
        
        if nargout > 1 % Compute the gradient
            tmp = sum( -noiseSigma * 1./normcdf( y/noiseSigma - noiseSigma*lambda ) .* ...
                normpdf( y/noiseSigma - noiseSigma*lambda, 0, 1 ) - y );
            tmp = tmp + n/lambda + n*noiseSigma^2*lambda;
            dLambda = -tmp;

            g = dLambda;
        end
        
    end

%lambdaLowerBound = eps;
%lambdaUpperBound = Inf;
lambdaLowerBound = lambda0LowerBound;
lambdaUpperBound = lambda0UpperBound;

% Call fmincon
options = optimset( 'GradObj', 'on', 'Display', 'final', 'MaxIter', 1e4, 'TolX', 1e-6 );

try
    warning off;
    [lambda, logfVal, exitFlag] = fmincon( @minusLogf, lambda0, [], [], [], [], ...
        lambdaLowerBound, lambdaUpperBound, [], options );
    warning on;
catch
    disp( 'caught error' );
    error( lasterr );
end

% Defines the structure of sOptim
sOptimElems = { ...
    'lambda0', 'lambda0'; ...
    'lambdaLowerBound', 'lambdaLowerBound'; ...
    'lambdaUpperBound', 'lambdaUpperBound'; ...
    'lambda', 'lambda'; ...
    'logfVal', 'logfVal'; ...
    'exitFlag', 'exitFlag' ...
    };

for elemInd = 1:size( sOptimElems, 1 )
    szCmd = sprintf( 'sOptim.%s = %s;', ...
        char( sOptimElems(elemInd,1) ), ...       
        char( sOptimElems(elemInd,2) ) );
    try
        eval( szCmd );
    catch
        warning( sprintf( 'Could not execute: %s', szCmd ) );
    end
end

end

