% Estimates lambda and noiseSigma from samples
% Usage: sOptim = exp_optim2( varargin )
% Keys: 'y', 'v0LowerBound', 'v0UpperBound'

function sOptim = exp_optim2( varargin )

% Initialize
y = [];
% v0 = [lambda0, noiseSigma0]
v0 = [-1, -1];
v0LowerBound = zeros( size(v0) );
v0UpperBound = zeros( size(v0) );
B_PLOT = 0;

% Process arguments

numArgs = length( varargin ) / 2;

for i = 1:numArgs
    szKey = varargin{2*i-1};
    switch lower( szKey )
        case 'y', y = varargin{2*i};
		case 'v0lowerbound', v0LowerBound = varargin{2*i};
		case 'v0upperbound', v0UpperBound = varargin{2*i};
        case 'bplot', B_PLOT = varargin{2*i};
        otherwise
            error( 'invalid szKey' );
    end
end

% Check variables
if ( sum(v0LowerBound < 0) > 0 | sum(v0UpperBound < 0) > 0 | ...
	 sum(v0LowerBound >= v0UpperBound) > 0 )
	error( 'invalid bounds for v0' );
end

y = y(:);
n = length( y );

% Use a grid search to determine the best v0
lenLambda0Grid = 40;
lenSigma0Grid = 40;
lambda0Grid = logspace( log10(v0LowerBound(1)), log10(v0UpperBound(1)), ...
    lenLambda0Grid );
sigma0Grid = logspace( log10(v0LowerBound(2)), log10(v0UpperBound(2)), ...
    lenSigma0Grid );
initValue = zeros( lenSigma0Grid, lenLambda0Grid );

disp( sprintf( 'v0 interval: [%.4g,%.4g] x [%.4g,%.4g]', ...
    v0LowerBound(1), v0UpperBound(1), v0LowerBound(2), v0UpperBound(2) ) );

warning off;
for j = 1:lenSigma0Grid
	for i = 1:lenLambda0Grid
		initValue( j, i ) = minusLogf( [lambda0Grid(i),sigma0Grid(j)] );
	end
end
warning on;

if B_PLOT
    figure(1), clf, contour( sigma0Grid, lambda0Grid, initValue );
    figure(2), clf, hist( y, 50 );
    pause;
end

% Use the same trick as in gridsearch_at.m to find the max of a matrix
[colMin, colMinInd] = min( initValue, [], 1 );
[valMin, lambda0MinInd] = min( colMin );
sigma0MinInd = colMinInd( lambda0MinInd );

% Check if initValue is finite
if ~isfinite( initValue( sigma0MinInd, lambda0MinInd ) )
    error( 'initValue not finite' );
end

v0 = [ lambda0Grid(lambda0MinInd), sigma0Grid(sigma0MinInd) ];
disp( sprintf( '>USING v0 = [%g,%g], initValue=%g', v0(1), v0(2), ...
    initValue( sigma0MinInd, lambda0MinInd ) ) );

    function [f,g] = minusLogf( v )
        
        lambda = v(1);
        noiseSigma = v(2);
        
        tmp = sum( log( normcdf( y/noiseSigma - noiseSigma*lambda ) ) - lambda*y );
        tmp = tmp + n*log(lambda) + n/2*noiseSigma^2*lambda^2;
        f = -tmp;
        
        % DEBUG
        disp( sprintf( 'f=%g', f ) );
        
        if nargout > 1 % Compute the gradient
            tmp = sum( -noiseSigma * 1./normcdf( y/noiseSigma - noiseSigma*lambda ) .* ...
                normpdf( y/noiseSigma - noiseSigma*lambda, 0, 1 ) - y );
            tmp = tmp + n/lambda + n*noiseSigma^2*lambda;
            dLambda = -tmp;

            tmp = sum( -( y/noiseSigma^2 + lambda ) .* 1./normcdf( y/noiseSigma - noiseSigma*lambda ) .* ...
                normpdf( y/noiseSigma - noiseSigma*lambda, 0, 1 ) );
            tmp = tmp + n*lambda^2*noiseSigma;
            dSigma = -tmp;
            
            g = [dLambda, dSigma];
            
            % DEBUG
            disp( sprintf( 'g=[%g,%g]', dLambda, dSigma ) );
        end
        
    end

%vLowerBound = eps*[1, 1];
%vUpperBound = [Inf, Inf];
vLowerBound = v0LowerBound;
vUpperBound = v0UpperBound;

% Call fmincon
options = optimset( 'GradObj', 'on', 'Display', 'final', 'MaxIter', 1e4, 'TolX', 1e-6 );

try
    warning off;
    [v, logfVal, exitFlag] = fmincon( @minusLogf, v0, [], [], [], [], vLowerBound, vUpperBound, [], options );
    warning on;
catch
    disp( 'caught error' );
    error( lasterr );
end


% Defines the structure of sOptim
sOptimElems = { ...
    'v0', 'v0'; ...
    'vLowerBound', 'vLowerBound'; ...
    'vUpperBound', 'vUpperBound'; ...
    'v', 'v'; ...
    'logfVal', 'logfVal'; ...
    'exitFlag', 'exitFlag' ...
    };

for elemInd = 1:size( sOptimElems, 1 )
    szCmd = sprintf( 'sOptim.%s = %s;', ...
        char( sOptimElems(elemInd,1) ), ...       
        char( sOptimElems(elemInd,2) ) );
    try
        eval( szCmd );
    catch
        warning( sprintf( 'Could not execute: %s', szCmd ) );
    end
end

end
