% john_optim2: Find the optimum values of (a,t). Assume the noise variance
% is given.
% Usage: sOptim = john_optim2( y, sigma, bMakePlot )

function sOptim = john_optim2( y, sigma, bMakePlot )

% Initialize
sOptim = [];

sdev = std( y(:) );
n = length( y(:) );
tUniv = sqrt( 2*log(n) )*sigma;
aRange = [1e-6/sdev, 3/sdev];
tRange = [1e-6, tUniv];
% aRange = [0.5, 4];
% tRange = [1e-2, tUniv];

disp(	[sprintf( 'sigma=%g, std(y)=%g, tUniv=%g\n', sigma, sdev, tUniv ), ...
		 sprintf( 'aRange: [%g,%g]\n', aRange(1), aRange(2) ), ...
		 sprintf( 'tRange: [%g,%g]', tRange(1), tRange(2) )] ...
);

% DEBUG
%error( 'stop' );

% Study plots of log likelihood

aNumPts = 15;
tNumPts = 20;

aVec = linspace( aRange(1), aRange(2), aNumPts );
tVec = linspace( tRange(1), tRange(2), tNumPts );

s = struct( 'a', 0, 't', 0, 'sigma', sigma, 'bVerbose', 0 );

if ( ( nargin == 3 ) & ( bMakePlot == 1 ) )

	[A,T] = meshgrid( aVec, tVec );
	Phi = zeros( size( A ) );
	PhiDa = zeros( size( Phi ) );
	PhiDt = zeros( size( Phi ) );
	PhiHessDef = zeros( size( Phi ) );

	for i1 = 1:size( Phi, 1 )
		for i2 = 1:size( Phi, 2 )
			%disp( sprintf( 'i1=%g, i2=%g', i1, i2 ) );

			s.bComputeGradOnly = 0;
			s.a = A(i1,i2);
			s.t = T(i1,i2);
			s.w = -1;

			Phi( i1, i2 ) = logliksparseJS( s, y(:) );

			%-- Matlab: XXX - DEPRECIATED
			%[phiGrad, phiHess] = phi_deriv( y(:), A(i1,i2), T(i1,i2), ...
			%	sigma, 'grad' );

			%-- Call mex function
			[phiGrad, phiHess] = logliksparseJS_deriv( s, y(:) );

			gradNorm = sqrt( phiGrad(1)^2 + phiGrad(2)^2 );
			PhiDa( i1, i2 ) = phiGrad(1) / gradNorm;
			PhiDt( i1, i2 ) = phiGrad(2) / gradNorm;

			PhiHessDef( i1, i2 ) = det_sym2_def( phiHess );
		end
	end

	figure( 1 ), surfc( A, T, Phi );
	xlabel( 'a' ), ylabel( 't' );

	figure( 2 ), contour( A, T, Phi ), hold on;
	[iRow, iCol] = find( PhiHessDef > 0 );
	AA = A( iRow, iCol );
	TT = T( iRow, iCol );
	plot( AA(:), TT(:), 'k+' );
	[iRow, iCol] = find( PhiHessDef < 0 );
	AA = A( iRow, iCol );
	TT = T( iRow, iCol );
	plot( AA(:), TT(:), 'ro' );
	xlabel( 'a' ), ylabel( 't' ); 

	figure( 3 ), quiver( A, T, PhiDa, PhiDt );
	xlabel( 'a' ), ylabel( 't' );
    return; % XXX
end

% Implement the gradient projection method

% First: try to find an estimate for the Lipschitz condition number L
numRandPoints = 30;
gridPoint = [];
gridValue = [];
gridGrad = [];
aDelta = aRange(2) - aRange(1);
tDelta = tRange(2) - tRange(1);

s.bComputeGradOnly = 1; % Only compute the gradient for the part below

for i = 1:numRandPoints
	a = aRange(1) + aDelta * rand;
	t = tRange(1) + tDelta * rand;
	gridPoint = [gridPoint; a, t];

	s.a = a;
	s.t = t;
	s.w = -1;

	[phiGrad1, phiHess1] = logliksparseJS_deriv( s, y(:) );
	gridGrad = [gridGrad; phiGrad1(:)'];

	val = logliksparseJS( s, y(:) );
	gridValue = [gridValue; val];
end

LVal = [];
for i = 2:numRandPoints
	for j = 1:(i-1)
		lval = norm( gridGrad(i,:) - gridGrad(j,:), 2 ) / ...
			norm( gridPoint(i,:) - gridPoint(j,:), 2 );
		LVal = [LVal; lval];
	end
end

L = max( LVal );
alpha = 1/L;

if ( L < eps )
	disp( '   gridPoint:' ), disp( gridPoint' );
	disp( '   gridGrad:' ), disp( gridGrad' );
	disp( '   gridValue:' ), disp( gridValue' );
	disp( '   LVal:' ), disp( LVal' );
	error( '   L < eps!' );
end

disp( sprintf( '\tEstimate of Lipschitz constant: L=%g, alpha=%g', L, alpha ) );

% Secondly: pick a starting point
[gridMax, indGridMax] = max( gridValue );
xInit = gridPoint( indGridMax, : );
s.a = xInit(1);
s.t = xInit(2);
s.w = -1;
xValInit = logliksparseJS( s, y(:) );

xPrev = xInit(:);

disp( sprintf( '\tStarting point xInit: [%g,%g], xValInit=%g', ...
	xInit(1), xInit(2), xValInit ) );

% Thirdly: iterate
numIter = 0;

% Sets stopping condition
X_DIFF_TOL = 1e-4 * ([aRange(2),tRange(2)] - [aRange(1),tRange(1)]);
X_MAX_ITER = 500;

% Sanity: make sure that tol isn't unrealistically small
if ( X_DIFF_TOL(1) < 1e-3 ), X_DIFF_TOL(1) = 1e-3, end;
if ( X_DIFF_TOL(2) < 1e-6 ), X_DIFF_TOL(2) = 1e-6, end;

xOptim = []; % Initialize
nHessDef = 0;
nStop = 0;

s.alphaMax = 1/L;
s.maxNumIter = 500;
s.bVerbose = 1;

% In this for loop, refrain from computing the value of the cost function
% and only do this when looking for the next point in lsearch_loglik().

% Define status constants
N_STATUS_INIT = 0;
N_STATUS_SMALLER_DIFF = 1;
N_STATUS_EXCEED_ITER = 2;
N_STATUS_NO_ALPHAOPT = 3;
sszStatusDesc = { 'init', 'diff < tol', 'exceed iter', 'no alphaopt' };

while ( 1 )

	nStop = N_STATUS_INIT;

	s.a = xPrev(1);
	s.t = xPrev(2);
	s.bComputeGradOnly = 0;
	[phiGrad, phiHess] = logliksparseJS_deriv( s, y(:) );
	nHessDef = det_sym2_def( phiHess );
	disp( sprintf( '\tnHessDef=%g', nHessDef ) );

	% Do a line search for the best alpha
    s.xPrev = xPrev(:);
    s.phiGrad = phiGrad(:);
    s.aRange = aRange(:);
    s.tRange = tRange(:);
    
	alphaOpt = logliksparseJS_lsearch( s, y(:) );

	if ( abs( alphaOpt ) < eps )

		warning( 'cannot find alphaOpt > 0' );
		nStop = N_STATUS_NO_ALPHAOPT;

	else

		xNextRaw = xPrev + alphaOpt * phiGrad;

		% Project xNextRaw onto the constrained set (box)
		xNext = proj_2d_box( xNextRaw, aRange, tRange );

		% Check for stopping condition
		xDiff = xNext - xPrev;
		if ( ( abs(xDiff(1)) < X_DIFF_TOL(1) ) & ...
		 	 ( abs(xDiff(2)) < X_DIFF_TOL(2) ) )
			nStop = N_STATUS_SMALLER_DIFF;
		end

		numIter = numIter + 1;
		if ( numIter >= X_MAX_ITER ), nStop = N_STATUS_EXCEED_ITER; end;

        % DEBUG
        disp( sprintf( '\tloop %g, nStop=%s, |xDiff(1)|=%g, |xDiff(2)|=%g', ...
            numIter, char(sszStatusDesc(1+nStop)), abs(xDiff(1)), abs(xDiff(2)) ) );
	end
	


	if ( nStop == N_STATUS_INIT )
		% Update prev variables and continue
		xPrev = xNext;
	else
		% Set result variables. For these 2 conditions, we are definitely
		% breaking out of the loop
		if ( ( nStop == N_STATUS_SMALLER_DIFF ) | ...
			 ( nStop == N_STATUS_EXCEED_ITER ) )
			xOptim = xNext;
		elseif ( nStop == N_STATUS_NO_ALPHAOPT )
			xOptim = xPrev;
		end

		% Investigate the other conditions.
		s.a = xOptim(1);
		s.t = xOptim(2);
		s.bComputeGradOnly = 0;
		[phiGrad, phiHess] = logliksparseJS_deriv( s, y(:) );
		nHessDef = det_sym2_def( phiHess );

		% Non-negative definite Hessian
		if ( nHessDef >= 0 )

			warning( 'nHessDef is non-negative!' );

			if ( ( nStop == N_STATUS_EXCEED_ITER ) | ...
				 ( nStop == N_STATUS_NO_ALPHAOPT) )
				% Done with iterations
				disp( 'john_optim2 finished' );
			else
				% Otherwise, continue iterations
				xPrev = xNext;
				continue;
			end

		end

		break;
	end

end

disp( 	[sprintf( '\t\tSummary:\n' ), ...
	  	 sprintf( '\t\titers=%g, xOptim=[%g,%g], w=%g\n', ...
	  		numIter, xOptim(1), xOptim(2), ...
			convert_tsigma2w( xOptim(1), xOptim(2), sigma, 'median' ) ), ...
		 sprintf( '\t\tnHessDef=%g, nStop=%s', nHessDef, ...
		 	char(sszStatusDesc(1+nStop)) )] ...
	);

sOptim = struct( 'x', xOptim, 'numIter', numIter, ...
	'grad', phiGrad, 'hess', phiHess, 'nHessDef', nHessDef, 'nStop', nStop, ...
    'aRange', aRange, 'tRange', tRange );

% Utility functions

% Determines the definiteness of a 2x2 symmetric matrix A
function nDef = det_sym2_def( A )
a = A(1,1);
d = A(2,2);
b = A(1,2);
tmp1 = a + d;
tmp2 = sqrt( (a-d)^2 + 4*b^2 );
if ( (tmp1 + tmp2 > 0) & (tmp1 - tmp2 > 0) )
	nDef = 1; 		% Positive definite
elseif ( (tmp1 + tmp2 < 0) & (tmp1 - tmp2 < 0) )
	nDef = -1;		% Negative definite
else
	nDef = 0;		% Indeterminate
end

% Projects a 2d vector into a constrained box
function xProj = proj_2d_box( x, x1Box, x2Box )

xProj = zeros( 2, 1 );

if ( x(1) > x1Box(2) )
	xProj(1) = x1Box(2);
elseif ( x(1) < x1Box(1) )
	xProj(1) = x1Box(1);
else
	xProj(1) = x(1);
end

if ( x(2) > x2Box(2) )
	xProj(2) = x2Box(2);
elseif ( x(2) < x2Box(1) )
	xProj(2) = x2Box(1);
else
	xProj(2) = x(2);
end


