% mapfunc: Version of thresholding that applies the MAP principle to estimate
% the nuisance parameters and to do the thresholding.
%
% Usage:
%    [estParams, thetaHat] = mapfunc( vecY, szPrior, valA, valSdev, ...
%       szThresRule, nDebugFlag )
%

function [estParams, thetaHat] = mapfunc( vecY, szPrior, valA, valSdev, ...
	szThresRule, nDebugFlag )

% Globals
global y tmp aGiven sdevGiven sMapInfo nThresFunc thresUniv

global paramA paramW paramSdev

global THRESFUNC_LAPLACE_MEDIAN THRESFUNC_LAPLACE_MAP

% Constants

PRIOR_LAPLACE = 1;

THRES_MEDIAN = 1;
THRES_MAP = 2;

THRESFUNC_LAPLACE_MEDIAN = 1;
THRESFUNC_LAPLACE_MAP = 2;

y = vecY(:);
ySigma = std( y );
tmp = zeros( size( y ) );
n = max( size( y ) );
aGiven = valA;
sdevGiven = valSdev;

switch ( lower( szPrior ) )
case 'laplace'
	disp( 'using Laplace prior' );
	prior = PRIOR_LAPLACE;

otherwise
	error( 'unsupported prior' );
end

bGradientSearch = 1; % Default value

switch ( lower( szThresRule ) )

case 'map'
	disp( 'using MAP thresholding rule' );
	thres = THRES_MAP;
	if ( prior == PRIOR_LAPLACE ), nThresFunc = THRESFUNC_LAPLACE_MAP; end;

case 'map_no_grad'
	disp( 'using MAP thresholding rule with no gradient search' );
	thres = THRES_MAP;
	if ( prior == PRIOR_LAPLACE ), nThresFunc = THRESFUNC_LAPLACE_MAP; end;
	bGradientSearch = 0;

otherwise
	error( 'unsupported thresholding rule' );
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Step 1: find MAP estimates of w (optionally a and sdev also)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

banner( 'Stage 1: finding MAP estimates' );

thresUniv = sqrt( 2*log(n) );
disp( sprintf( 'thresUniv = %g', thresUniv ) );

% Initialize
paramT = -1;
paramA = -1;
paramW = -1;
paramSdev = -1;

exitFlag = 1; 

tic; % Start timing

if ( ~isempty( sdevGiven ) ) 

	%--- CASE 1: sigma is given
	%##########################

	paramSdev = sdevGiven;
	
	options = optimset( 'Display', 'final', 'MaxFunEvals', 5e3, ...
		'MaxIter', 1e3 );

	if ( isempty( aGiven ) )
		% CASE 1a: a isn't given, so have to do a 2-d search for [a,w]
		% Define: beta = [a, w], i.e. find a_ML and w_ML
		betaLBound = [1e-6/paramSdev, 1e-6];
		betaUBound = [5/paramSdev, 0.5];

		warning off; % Turn off log of zero and divide by zero warnings

		beta0 = pick_2d_initial( @neg_laplace_log_lik_map_aw, ...
			[betaLBound(1),betaUBound(1)], ...
			[betaLBound(2),betaUBound(2)], ...
			(betaUBound(1)-betaLBound(1))/20, ...
			(betaUBound(2)-betaLBound(2))/20, 'none', 0 );

		f0 = neg_laplace_log_lik_map_aw( beta0 );
		disp( sprintf( 'f0 = %g, beta0 = [%g, %g]', f0, ...
			beta0(1), beta0(2) ) );


		if ( bGradientSearch == 1 )
			% Do a gradient search
			[betaMAP, fFinal, exitFlag] = ...
				fmincon( @neg_laplace_log_lik_map_aw, beta0, ...
					[], [], [], [], betaLBound, betaUBound, [], options );
		end

		% If bGradientSearch == 0 OR fmincon didn't successfully 
		% terminate, do another pick_2d_initial to determine the MAP
		% parameters

		if ( ( bGradientSearch == 0 ) | ...
			 ( ( bGradientSearch == 1 ) & ( exitFlag <= 0 ) ) )

			BETA_DELTA_MULT = 3;
			BETA_PERCENT_ACCURACY = 1e-2;
			BETA1_GRID_SIZE = 20;
			BETA2_GRID_SIZE = 20;

			disp( 'Doing another 2d search for MAP parameters' );
			
			beta1Delta = (betaUBound(1)-betaLBound(1))/20;
			beta2Delta = (betaUBound(2)-betaLBound(2))/20;

			beta1Lower = max( [betaLBound(1), ...
				beta0(1)-BETA_DELTA_MULT*beta1Delta] );
			beta1Upper = 2*beta0(1) - beta1Lower;

			beta2Lower = max( [betaLBound(2), ...
				beta0(2)-BETA_DELTA_MULT*beta2Delta] );
			beta2Upper = 2*beta0(2) - beta2Lower;

			betaMAP = pick_2d_initial( @neg_laplace_log_lik_map_aw, ...
				[beta1Lower, beta1Upper], [beta2Lower, beta2Upper], ...
				(beta1Upper - beta1Lower)/BETA1_GRID_SIZE, ...
				(beta2Upper - beta2Lower)/BETA2_GRID_SIZE, 'none', 0 );

			fFinal = neg_laplace_log_lik_map_aw( betaMAP );

		end

		warning on;

		disp( sprintf( 'a_MAP = %g, w_MAP = %g', betaMAP(1), betaMAP(2) ) );
		paramT = convert_wsigma2t( betaMAP(1), betaMAP(2), paramSdev, 'map' );
		disp( sprintf( 'thres = %g, fFinal = %g', paramT, fFinal ) );
		paramA = betaMAP(1);
		paramW = betaMAP(2);

	else
		% CASE 1b: a is given
		% Define: beta = [w]
		error( 'CASE 1b: unimplemented' );
	end

else

	%--- CASE 2: neither a, w, sigma are given
	%#########################################

	% Assume that we want to find the ML estimates of a, w, sigma
	% beta = [a, w, sigma]
	betaLBound = [ 1e-3/ySigma, 1e-5, 1e-6*ySigma ];
	betaUBound = [ 5/ySigma, 0.5, ySigma ];

	warning off;

	beta0 = pick_3d_initial( @neg_laplace_log_lik_map_awsigma, ...
		[betaLBound(1),betaUBound(1)], ...
		[betaLBound(2),betaUBound(2)], ...
		[betaLBound(3),betaUBound(3)], ...
		(betaUBound(1)-betaLBound(1))/20, ...
		(betaUBound(2)-betaLBound(2))/20, ...
		(betaUBound(3)-betaLBound(3))/20, 'none', 0 );

	% Evaluate the cost function at beta0
	f0 = neg_laplace_log_lik_map_awsigma( beta0 );
	disp( sprintf( 'f0 = %g, beta0 = [%g, %g, %g]', f0, ...
		beta0(1), beta0(2), beta0(3) ) );

	options = optimset( 'Display', 'final', 'MaxFunEvals', 1e3, ...
		'MaxIter', 1e3 );

	exitFlag = 1; % Initialize

	if ( bGradientSearch == 1 )
		% Do a gradient search
		[betaMAP, fFinal, exitFlag] = ...
			fmincon( @neg_laplace_log_lik_map_awsigma, beta0, ...
				[], [], [], [], betaLBound, betaUBound, [], options );
	end

	% If bGradientSearch == 0 OR fmincon didn't successfully 
	% terminate, do another pick_3d_initial to determine the MAP
	% parameters

	if ( ( bGradientSearch == 0 ) | ...
		 ( ( bGradientSearch == 1 ) & ( exitFlag <= 0 ) ) )

		disp( 'Doing another 3d search for MAP parameters' );
			
		beta1Delta = (betaUBound(1)-betaLBound(1))/20;
		beta2Delta = (betaUBound(2)-betaLBound(2))/20;
		beta3Delta = (betaUBound(3)-betaLBound(3))/20;
		disp( sprintf( 'beta{1,2,3}Delta = %g, %g, %g', ...
			beta1Delta, beta2Delta, beta3Delta ) );

		% DEBUG
		%disp( 'DEBUG: press any key to continue' );
		%pause;

		betaMAP = pick_3d_initial( @neg_laplace_log_lik_map_awsigma, ...
			[0.5*beta0(1),1.5*beta0(1)], ...
			[0.5*beta0(2),1.5*beta0(2)], ...
			[0.5*beta0(3),1.5*beta0(3)], ...
			beta0(1) * 0.1, beta0(2) * 0.1, beta0(3) * 0.1, 'none', 0 );
		fFinal = neg_laplace_log_lik_map_awsigma( betaMAP );

	end

	warning on;

	disp( sprintf( 'a_MAP = %g, w_MAP = %g, sigma_MAP = %g', ...
		betaMAP(1), betaMAP(2), betaMAP(3) ) );
	paramT = convert_wsigma2t( betaMAP(1), betaMAP(2), betaMAP(3), 'map' );
	disp( sprintf( 'thres = %g, fFinal = %g', paramT, fFinal ) );

	paramA = betaMAP(1);
	paramW = betaMAP(2);
	paramSdev = betaMAP(3);
end

toc; % Stop timing

estParams = struct( 'a', paramA, 't', paramT, 'w', paramW, 'sdev', paramSdev );

% Piggy back the estParams structure to return the status of fmincon
estParams.exitFlag = exitFlag;

if ( nDebugFlag == 1 )
	disp( 'Stopping after step 1' );
	thetaHat = [];
	return;
end

% From this point onwards, use variable names param{A,T,W,Sdev}. The value
% of the threshold here is obtained from the MAP thresholding rule.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Step 2: apply the appropriate threshold function
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

banner( 'Stage 2: applying threshold function' );

disp( sprintf( 'paramA: %g, paramW: %g, paramSdev: %g', ...
	paramA, paramW, paramSdev ) );

if ( prior == PRIOR_LAPLACE )

	switch ( thres )

	case THRES_MAP
		thetaHat = laplace_map_thres0( y, paramA, paramW, paramSdev );

	otherwise
		error( 'unsupported thresholding rule' );
		
	end	

end


