% min_log_lik: minimizes the EB log likelihood
%-------------------------------------------------------------------------------
% Uses a simple grid-like method to find the maximizer of the log likelihood
%
% Usage:
%    [w, a] = min_log_lik( aRange, vecX )
%
function [w,a] = min_log_lik( aRange, vecX )

global bVerbose bDebug bStrictJohnstone

bVerbose = 1;
bDebug = 0;
bStrictJohnstone = 1;

numAPts = 50;
numWPts = 1e3;
numMaxIters = 3;
tMax = sqrt( 2 * log( length(vecX) ) );

disp( sprintf( 'tMax = %g', tMax ) );

% First iteration
aInterLength = aRange(2) - aRange(1);
vecA = linspace( aRange(1), aRange(2), numAPts );
[w, a] = min_grid( tMax, numAPts, numWPts, vecA, vecX )

% Check boundary
if ( ( abs( a-aRange(1) ) < 1e4*eps ) | ...
	 ( abs( a-aRange(2) ) < 1e4*eps ) | ...
	 ( abs( w ) < 1e4*eps ) )
	error( 'optimal values close to boundary edge' );
end

% Keep looping
for iterInd = 2:numMaxIters
	aInterLength = aInterLength / 4;
	aInterLower = max( [ aRange(1), a - aInterLength ] );
	aInterUpper = min( [ aRange(2), a + aInterLength ] );
	if bVerbose, disp( sprintf( 'iter %g: a interval = [%g,%g]', iterInd, ...
		aInterLower, aInterUpper ) );
	end

	bVerbose = 0; % Turn off verbose flag
	vecA = linspace( aInterLower, aInterUpper, numAPts );
	[wNew, aNew] = min_grid( tMax, numAPts, numWPts, vecA, vecX )

	% Check boundary
	if ( ( abs( aNew-aInterLower ) < 1e4*eps ) | ...
		 ( abs( aNew-aInterUpper ) < 1e4*eps ) | ...
		 ( abs( wNew ) < 1e4*eps ) )
		error( sprintf( '%d: optimal values close to boundary edge', ...
			iterInd ) );
	end

	% Stopping condition
	if ( abs( wNew - w ) < 0.05*w & abs( aNew - a ) < 0.05*a )
		w = wNew; a = aNew;
		break;
	end

	% Update the variables
	w = wNew;
	a = aNew;

end

function [wMin, aMin] = min_grid( tMax, numAPts, numWPts, vecA, vecX )

global bVerbose bDebug bStrictJohnstone

bHaveMin = 0;
LLmax = 0;
vecLL = zeros( 1, numWPts );

for indA = 1:numAPts
	wThres = wfromt( tMax, 'laplace', vecA(indA) );
	if bStrictJohnstone
		vecW = linspace( wThres, 1, numWPts );
		if bDebug
			disp( sprintf( 'a = %g, wThres = %g', vecA(indA), wThres ) )
		end
	else
		vecW = linspace( 1e-3, 1, numWPts );
	end
	for indW = 1:numWPts
		vecLL( indW ) = eb_log_lik( vecW(indW), vecA(indA), vecX );
	end
	
	[val, ind] = max( vecLL );
	if ( ( bHaveMin & (val > LLmax) ) | (~bHaveMin) )
		LLmax = val;
		aMin = vecA( indA );
		wMin = vecW( ind );
		bHaveMin = 1;
	end
end


