% pdb_sim_start: Starts a pdb simulation
%
% sSim = pdb_sim_start( szFilename, szRePlotType, snr )
%
% szFilename: name of PDB file
% szRePlotType: atom type to match in PDB file
% snr: SNR of output

function sSim = pdb_sim_start( szFilename, szRePlotType, snr )

bRotate = 0;
szRotateAxis = 'x';
rotateDeg = 45; % Degrees to rotate in CCW direction 

% Read in molecule definition from PDB file. The matrix spinPosNm will
% contain the locations of the selected atoms in dimensions of nm.
[A, AType, H, HType, res] = pdb_read_file( szFilename );
spinPosNm = pdb_match( A, AType, H, HType, szRePlotType );

% If applicable, rotate molecule in order to make it fit into the grid better
if bRotate 
	disp( 'Rotating spin positions' );
	spinPosNm = rotate_3d( spinPos, szRotateAxis, rotateDeg * (pi/180) );
end

% Find the bounding box size, i.e., the smallest box volume that completely
% encloses the molecule
molBoundBoxSize = [ max(spinPosNm(:,1))-min(spinPosNm(:,1)), ...
    max(spinPosNm(:,2))-min(spinPosNm(:,2)), ...
    max(spinPosNm(:,3))-min(spinPosNm(:,3)) ...
    ];
disp( 'molBoundBoxSize:' ), disp( molBoundBoxSize );

% Find the smallest distances between atoms in the x, y, z directions and
% the smallest Euclidean distance between atoms
molSmallestCoordDist = [ inf, inf, inf ];
molSmallestDist = inf;
numAtoms = size( spinPosNm, 1 );

for i = 1:numAtoms
    for j = (i+1):numAtoms
        coordDist = abs( spinPosNm( i, : ) - spinPosNm( j, : ) );
        for k = 1:3
            if coordDist( k ) < molSmallestCoordDist( k )
                molSmallestCoordDist( k ) = coordDist( k ); % Swap
            end
        end
        l2Dist = sqrt( sum( coordDist.^2 ) );
        if l2Dist < molSmallestDist
            molSmallestDist = l2Dist; % Swap
        end
    end
end

disp( 'molSmallestCoordDist:' ), disp( molSmallestCoordDist );
disp( 'molSmallestDist:' ), disp( molSmallestDist );  

% Obtain the parameters of the MRFM psf
BresVal = 1e4;
dVal = 3;
sPsf = psf_param4( 'Bres', BresVal, ...
    'd', dVal, ...    
    'bUseSmallerR0', 1, ...
    'bUseSmallerB0', 0, ...
    'bComputeOptR0', 1, ...
    'bComputeOptBext', 1 );

% Plot resonant surface
sPsfResSurf = bmag_lev_surf2( sPsf.Bext, sPsf.Bres^2, sPsf.m, sPsf.R0 );
disp( 'Pausing...' ), pause;

% Figure out the support of the MRFM psf with these parameters
sPsfResSurf.x = sPsfResSurf.r .* sin( sPsfResSurf.phi ) .* cos( sPsfResSurf.theta );
sPsfResSurf.y = sPsfResSurf.r .* sin( sPsfResSurf.phi ) .* sin( sPsfResSurf.theta );
sPsfResSurf.z = sPsfResSurf.r .* cos( sPsfResSurf.phi ) - sPsf.R0; % normalized z
psfSupp = [ min( sPsfResSurf.x(:) ), max( sPsfResSurf.x(:) ); ...
    min( sPsfResSurf.y(:) ), max( sPsfResSurf.y(:) ); ...
    min( sPsfResSurf.z(:) ), max( sPsfResSurf.z(:) ) ...
    ]; % in nm
disp( 'psfSupp:' ), disp( psfSupp );

% Construct mesh. An optimal mesh configuration takes into account the size
% of the molecule and the size of the psf.

% The bounding box for the mesh
padFactor = [ 0.7, 0.7, 1 ];
meshBoundBox = [ min(spinPosNm(:,1)) - padFactor(1)*abs(psfSupp(1,2)-psfSupp(1,1)), ...
    max(spinPosNm(:,1)) + padFactor(1)*abs(psfSupp(1,2)-psfSupp(1,1)); ...
    min(spinPosNm(:,2)) - padFactor(2)*abs(psfSupp(2,2)-psfSupp(2,1)), ...
    max(spinPosNm(:,2)) + padFactor(2)*abs(psfSupp(2,2)-psfSupp(2,1)); ...
    min(spinPosNm(:,3)), ...
    max(spinPosNm(:,3)) + padFactor(3)*abs(psfSupp(3,2)-psfSupp(3,1)) ];
disp( 'meshBoundBox:' ), disp( meshBoundBox );

vecNumPts = [ 64, 64, 64 ];
vecMeshBoundBoxDelta = [ abs(meshBoundBox(1,2)-meshBoundBox(1,1)), ...
	abs(meshBoundBox(2,2)-meshBoundBox(2,1)), ...
	abs(meshBoundBox(3,2)-meshBoundBox(3,1)) ] ./ vecNumPts;

% Adjust the mesh delta so that they are all equal. This involves taking
% the max of the deltas.
vecMeshBoundBoxDelta = max( vecMeshBoundBoxDelta ) * ones( 1, 3 );

sMesh = struct( 'lowerLeftCoord', [1.25*psfSupp(1,1), 1.1*psfSupp(2,1), sPsf.R0], ... % [nm]
    'originCoord', zeros( 1, 3 ), ... % [nm]
    'numPts', vecNumPts, ...
    'delta', vecMeshBoundBoxDelta, ... % [nm]
    'szPosition', 'lowerleft' ...
    );
disp( 'sMesh' ), disp( sMesh );

% Plot the MRFM psf going through each slice of the psf
psf = get_h_op( sPsf, sMesh );
% Normalize psf so that its max value is 1
psfMax = max( psf(:) );
psf = psf / psfMax;
disp( 'Plotting psf' ), plotSlices( 2, psf, sMesh.numPts(3), 0, 2 );

% Discretize. Put the lower left-hand corner at meshBoundBox(:,1). This
% ensures that there is enough padding for the psf.
spinPosDis = discretizeSpinPos( meshBoundBox(:,1), sMesh.delta, spinPosNm );
%disp( 'spinPosDis:' ), disp( spinPosDis );

% Plot the discrete locations of the spins
figure( 1 ), clf; 
plot3( spinPosDis(:,1), spinPosDis(:,2), spinPosDis(:,3), 'ko' );
grid on, axis equal tight;
title( 'Spin locations - discretized' );

% Create image
xtrue = create_orig_img( sMesh, spinPosDis );
disp( 'Plotting xtrue' ), plotSlices( 3, xtrue, sMesh.numPts(3), 1, 2 );

% Convolve with psf

psfShiftAmount = ceil( [ -(1/2)*(psfSupp(1,2)-psfSupp(1,1)), -(1/2)*(psfSupp(2,2)-psfSupp(2,1)), 0 ] ./ sMesh.delta );
disp( 'psfShiftAmount:' ), disp( psfShiftAmount );

mask = logical(ones(sMesh.numPts));
Gfft = Gblur( mask, 'psf', psf, 'type', 'conv,fft', 'nDim', ndims(xtrue) );

y0 = Gfft * xtrue;
y0 = y0 .* ( y0 > eps );
y0Display = circshift( y0, psfShiftAmount );
disp( 'Plotting y0Display' ), plotSlices( 4, y0Display, sMesh.numPts(3), 0, 2 );

% Create noisy version
y0Ms = 1/prod( sMesh.numPts ) * sum( y0(:).^2 );
noiseSigma = sqrt( y0Ms/snr );
disp( sprintf( 'snr = %g, noiseSigma = %g', snr, noiseSigma ) );
yi = y0 + noiseSigma * randn( size( y0 ) );
yiDisplay = circshift( yi, psfShiftAmount );
disp( 'Plotting yiDisplay' ), plotSlices( 5, yiDisplay, sMesh.numPts(3), 0, 2 );

% Defines the structure of sSim
sSimElems = { 'szFilename', 'szFilename'; ...
    'szRePlotType', 'szRePlotType'; ...
    'snr', 'snr'; ...
    'sPsf', 'sPsf'; ...
    'spinPosDis', 'spinPosDis'; ...
    'sMesh', 'sMesh'; ...
    'xtrue', 'xtrue'; 'y0', 'y0'; 'yi', 'yi'; 'psf', 'psf'; ...
    'psfShiftAmount', 'psfShiftAmount'; ...
    'noiseSigma', 'noiseSigma' ...
    };

for elemInd = 1:size( sSimElems, 1 )
    szCmd = sprintf( 'sSim.%s = %s;', ...
        char( sSimElems(elemInd,1) ), char( sSimElems(elemInd,2) ) );
    try
        eval( szCmd );
    catch
        warning( sprintf( 'Could not execute: %s', szCmd ) );
    end
end

% Pilfered from gen_img.m
function spinPosDis = discretizeSpinPos( vecLowerLeftCoord, vecDelta, spinPosNm )
%##########################################################################

c = vecLowerLeftCoord;
spinPosNmSize = size( spinPosNm );
numAtoms = spinPosNmSize( 1 );
thres = 1e2*eps;

if ( (spinPosNmSize(1) < 1) | (spinPosNmSize(2) ~= 3) )
    error( 'Invalid dimensions for spinPosNm' );
end

spinPosDis = spinPosNm - repmat( c(:)', numAtoms, 1 ); % Intermediate value

for i = 1:3
    if ( ~isempty( find( spinPosDis(:,i) < 0 ) ) )        
        warning( sprintf( 'Mesh lower left corner coord %g is too large', i ) );        
        disp( 'spinPosDis:' ), disp( spinPosDis );
        disp( 'c:' ), disp( c );
        error( 'Aborting' );
    else
        spinPosDis( :, i ) = 1 + floor( spinPosDis( :, i ) ./ vecDelta(i) );
    end
end

% Yet another function pilfered from gen_img.m
function f = create_orig_img( sMesh, spinPosDis )
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% the original image
vecMeshSize = sMesh.numPts;
f = zeros( vecMeshSize );
spinPosDisSize = size( spinPosDis );

if ( (spinPosDisSize(1) < 1) | (spinPosDisSize(2) ~= 3) )
	error( 'Invalid dimensions for spinPosDis' );
end

for locInd = 1:spinPosDisSize(1)
	vecLoc = spinPosDis( locInd, : );
    % Sanity: check that vecLoc is a valid location vector in the mesh
    for coordInd = 1:3
        if ( vecLoc( coordInd ) < 1 ) | ( vecLoc( coordInd ) > vecMeshSize( coordInd ) )
            error( sprintf( 'Spin pos %g, bad coord index %g', locInd, coordInd ) );
        end
    end
    % Passes sanity
	f( vecLoc(1), vecLoc(2), vecLoc(3) ) = 1;
end


