% pdb_sim_start2d: Starts a pdb simulation but in 2 dimensions
%
% sSim = pdb_sim_start2d( szFilename, szRePlotType, snr )
%
% szFilename: name of PDB file
% szRePlotType: atom type to match in PDB file
% snr: SNR of output

function sSim = pdb_sim_start2d( szFilename, szRePlotType, snr )

global bMol3d

% Read in molecule definition from PDB file. The matrix spinPosNm will
% contain the locations of the selected atoms in dimensions of nm.
[A, AType, H, HType, res] = pdb_read_file( szFilename );
spinPosNm = pdb_match( A, AType, H, HType, szRePlotType );

% At this point, figure out if the molecule is only 3d or 2d. If it's 3d,
% then form a 2d projection.
bMol3d = 0; % Initialize

spinPosNm3Unique = unique( spinPosNm(:,3) );

if length( spinPosNm3Unique ) == 1, disp( 'Molecule is 2d' );
else
    disp( 'Molecule is 3d, must project later on' );
    bMol3d = 1;
    % TODO. For now, quit with an error
    error( 'Unimplemented' );
end
    
% Find the bounding box size, i.e., the smallest area that completely
% encloses the molecule
molBoundBoxSize = [ max(spinPosNm(:,1))-min(spinPosNm(:,1)), ...
    max(spinPosNm(:,2))-min(spinPosNm(:,2)) ];

disp( 'molBoundBoxSize:' ), disp( molBoundBoxSize );

% Find the smallest distances between atoms in the x, y directions and
% the smallest Euclidean distance between atoms
molSmallestCoordDist = [ inf, inf ];
molSmallestDist = inf;
numAtoms = size( spinPosNm, 1 );

for i = 1:numAtoms
    for j = (i+1):numAtoms
        coordDist = abs( spinPosNm( i, : ) - spinPosNm( j, : ) );
        for k = 1:2
            if coordDist( k ) < molSmallestCoordDist( k )
                molSmallestCoordDist( k ) = coordDist( k ); % Swap
            end
        end
        l2Dist = sqrt( sum( coordDist.^2 ) );
        if l2Dist < molSmallestDist
            molSmallestDist = l2Dist; % Swap
        end
    end
end

disp( 'molSmallestCoordDist:' ), disp( molSmallestCoordDist );
disp( 'molSmallestDist:' ), disp( molSmallestDist );  

% Obtain the parameters of the MRFM psf
BresVal = 3e4;
dVal = 2;
sPsf = psf_param4( 'Bres', BresVal, 'd', dVal, ...
    'bComputeOptR0', 1, 'bComputeOptBext', 1, ...
    'bUseSmallerR0', 1, 'bUseSmallerB0', 1 );

% Plot resonant surface
sPsfResSurf = bmag_lev_surf2( sPsf.Bext, sPsf.Bres^2, sPsf.m, sPsf.R0 );

% Figure out the support of the MRFM psf with these parameters
sPsfResSurf.x = sPsfResSurf.r .* sin( sPsfResSurf.phi ) .* cos( sPsfResSurf.theta );
sPsfResSurf.y = sPsfResSurf.r .* sin( sPsfResSurf.phi ) .* sin( sPsfResSurf.theta );
sPsfResSurf.z = sPsfResSurf.r .* cos( sPsfResSurf.phi ) - sPsf.R0; % normalized z
psfSupp = [ min( sPsfResSurf.x(:) ), max( sPsfResSurf.x(:) ); ...
    min( sPsfResSurf.y(:) ), max( sPsfResSurf.y(:) ); ...
    min( sPsfResSurf.z(:) ), max( sPsfResSurf.z(:) ) ...
    ]; % in nm
disp( 'psfSupp:' ), disp( psfSupp );

% Construct mesh. An optimal mesh configuration takes into account the size
% of the molecule and the size of the psf.

% The bounding box for the mesh
padFactor = [ 0.6, 0.6, 1 ];
meshBoundBox = [ min(spinPosNm(:,1)) - padFactor(1)*abs(psfSupp(1,2)-psfSupp(1,1)), ...
    max(spinPosNm(:,1)) + padFactor(1)*abs(psfSupp(1,2)-psfSupp(1,1)); ...
    min(spinPosNm(:,2)) - padFactor(2)*abs(psfSupp(2,2)-psfSupp(2,1)), ...
    max(spinPosNm(:,2)) + padFactor(2)*abs(psfSupp(2,2)-psfSupp(2,1)); ...
    min(spinPosNm(:,3)), ...
    max(spinPosNm(:,3)) + padFactor(3)*abs(psfSupp(3,2)-psfSupp(3,1)) ];
disp( 'meshBoundBox:' ), disp( meshBoundBox );

vecNumPts = [ 64, 64, 1 ];

% Note that the z-delta value is 0
vecMeshBoundBoxDelta = [ abs(meshBoundBox(1,2)-meshBoundBox(1,1)), ...
	abs(meshBoundBox(2,2)-meshBoundBox(2,1)), 0 ] ./ vecNumPts;

% Adjust the mesh delta so that they are all equal. This involves taking
% the max of the deltas.
vecMeshBoundBoxDelta(1:2) = max( vecMeshBoundBoxDelta(1:2) ) * ones( 1, 2 );

sMesh = struct( 'lowerLeftCoord', [psfSupp(1,1), psfSupp(2,1), sPsf.R0], ... % [nm]
    'originCoord', zeros( 1, 3 ), ... % [nm]
    'numPts', vecNumPts, ...
    'delta', vecMeshBoundBoxDelta, ... % [nm]
    'szPosition', 'lowerleft' ...
    );
disp( 'sMesh' ), disp( sMesh );

% Plot the MRFM psf going through each slice of the psf
psf = get_h_op( sPsf, sMesh );
% Normalize psf so that its max value is 1
psfMax = max( psf(:) );
psf = psf / psfMax;
disp( 'Plotting psf' );
figure( 1 ), clf, im( psf ), colorbar;
axis equal tight;

% Discretize. Put the lower left-hand corner at meshBoundBox(:,1). This
% ensures that there is enough padding for the psf.
[spinPosDis, spinVal] = discretizeSpinPos2d( meshBoundBox(:,1), ...
    [sMesh.delta(1), sMesh.delta(2), 0], spinPosNm );
disp( 'spinPosDis:' ), disp( spinPosDis );

% Plot the discrete locations of the spins
figure( 2 ), clf, plot( spinPosDis(:,1), spinPosDis(:,2), 'ko' );
grid on, axis equal tight;
title( 'Spin locations - discretized' );

% Create image
xtrue = create_orig_img( sMesh, spinPosDis, spinVal );
disp( 'Plotting xtrue' );
figure( 3 ), clf, im( xtrue );
colormap gray, colorbar;
axis equal tight;

% Convolve with psf

psfShiftAmount = round( [ -(1/2)*(psfSupp(1,2)-psfSupp(1,1)), -(1/2)*(psfSupp(2,2)-psfSupp(2,1)) ] ./ sMesh.delta(1:2) );
disp( 'psfShiftAmount:' ), disp( psfShiftAmount );

mask = logical(ones(sMesh.numPts));
Gfft = Gblur( mask, 'psf', psf, 'type', 'conv,fft', 'nDim', ndims(xtrue) );

y0 = Gfft * xtrue;
y0 = y0 .* ( y0 > eps );
y0Display = circshift( y0, psfShiftAmount );
disp( 'Plotting y0Display' );
figure( 4 ), clf, im( y0Display ), colorbar;
axis equal tight;

% Create noisy version
y0Ms = 1/prod( sMesh.numPts ) * sum( y0(:).^2 );
noiseSigma = sqrt( y0Ms/snr );
disp( sprintf( 'snr = %g, noiseSigma = %g', snr, noiseSigma ) );
yi = y0 + noiseSigma * randn( size( y0 ) );
yiDisplay = circshift( yi, psfShiftAmount );
disp( 'Plotting yiDisplay' );
figure( 5 ), clf, im( yiDisplay ), colorbar;
axis equal tight;

% Defines the structure of sSim
sSimElems = { 'szFilename', 'szFilename'; ...
    'szRePlotType', 'szRePlotType'; ...
    'snr', 'snr'; ...
    'sPsf', 'sPsf'; ...
    'spinPosDis', 'spinPosDis'; ...
    'sMesh', 'sMesh'; ...
    'xtrue', 'xtrue'; 'y0', 'y0'; 'yi', 'yi'; 'psf', 'psf'; ...
    'psfShiftAmount', 'psfShiftAmount'; ...
    'noiseSigma', 'noiseSigma' ...
    };

for elemInd = 1:size( sSimElems, 1 )
    szCmd = sprintf( 'sSim.%s = %s;', ...
        char( sSimElems(elemInd,1) ), char( sSimElems(elemInd,2) ) );
    try
        eval( szCmd );
    catch
        warning( sprintf( 'Could not execute: %s', szCmd ) );
    end
end

% Pilfered from gen_img.m
% Here, if we are projection from a 3d molecule, we must assign the
% "values" at the positions depending on the distance.

function [spinPosDis, spinVal] = discretizeSpinPos2d( vecLowerLeftCoord, ...
    vecDelta, spinPosNm )
%##########################################################################

% Initialize
c = vecLowerLeftCoord;
thres = 1e2*eps;
pos3HalfTime = 1;

spinPosNmSize = size( spinPosNm );
numAtoms = spinPosNmSize( 1 );

spinPosDis = zeros( numAtoms, 2 );
spinVal = ones( numAtoms, 1 );

% Intermediate value
spinPosNmOffset = spinPosNm - repmat( c(:)', numAtoms, 1 ); 

% Check to see that no spinPosDis values are negative

for i = 1:2
    if ( ~isempty( find( spinPosNmOffset(:,i) < 0 ) ) )        
        warning( sprintf( 'Mesh lower left corner coord %g is too large', i ) );        
        disp( 'spinPosNmOffset:' ), disp( spinPosNmOffset );
        disp( 'c:' ), disp( c );
        error( 'Aborting' );
    else
        spinPosDis( :, i ) = 1 + floor( spinPosNmOffset( :, i ) ./ vecDelta(i) );
    end
end

pos3Max = max( spinPosNmOffset( :, 3 ) );
pos3Min = min( spinPosNmOffset( :, 3 ) );
pos3Delta = pos3Max - pos3Min;

if pos3Delta > thres
    for atomInd = 1:numAtoms
        dNorm = ( spinPosNmOffset( atomInd, 3 ) - pos3Min ) / pos3Delta;
        spinVal( atomInd ) = exp( -dNorm/pos3HalfTime );
    end
end

% Yet another function pilfered from gen_img.m
function f = create_orig_img( sMesh, spinPosDis, spinVal )
%#########################################################

% the original image
vecMeshSize = sMesh.numPts;
f = zeros( vecMeshSize );
f = squeeze( f );
spinPosDisSize = size( spinPosDis );

for locInd = 1:spinPosDisSize(1)
	vecLoc = spinPosDis( locInd, : );
    % Sanity: check that vecLoc is a valid location vector in the mesh
    for coordInd = 1:2
        if ( vecLoc( coordInd ) < 1 ) | ( vecLoc( coordInd ) > vecMeshSize( coordInd ) )
            error( sprintf( 'Spin pos %g, bad coord index %g', locInd, coordInd ) );
        end
    end
    % Passes sanity
	f( vecLoc(1), vecLoc(2) ) = spinVal( locInd );
end


