% pick_2d_initial: Picks an initial point in 2d for a minimization routine.
% Basically lay a grid and pick the point with min value to start from. If
% this point is on the boundary, return with error. Assume that we are only
% interested in strictly positive quantities.
%
% Usage:
%    x0 = pick_2d_initial( f, [x1min,x1max], [x2min,x2max], x1Res, x2Res,
%       szOptions, numRecursions )
%

function x0 = pick_2d_initial( f, x1Range, x2Range, x1Res, x2Res, szOptions, ...
	numRecursions )

DEFAULT_X1_PTS = 20;
DEFAULT_X2_PTS = 20;

if ( x1Res <= 0 )
	if ( x1Range(1) ~= x1Range(2) )
		error( 'x1Res <= 0 BUT x1Range is non-singular' );
	end
	x1Low = 0.5*x1Range(1)
	x1High = 1.5*x1Range(1);
	x1Res = (x1High - x1Low) / DEFAULT_X1_PTS;
else
	x1Low = x1Range(1);
	x1High = x1Range(2);
end

if ( x2Res <= 0 )
	if ( x2Range(1) ~= x2Range(2) )
		error( 'x2Res <= 0 BUT x2Range is non-singular' );
	end
	x2Low = 0.5*x2Range(1);
	x2High = 1.5*x2Range(1);
	x2Res = (x2High - x2Low) / DEFAULT_X2_PTS;
else
	x2Low = x2Range(1);
	x2High = x2Range(2);
end

numX1Pts = ceil( (x1High - x1Low)/x1Res );
x1 = linspace( x1Low, x1High, numX1Pts );

numX2Pts = ceil( (x2High - x2Low)/x2Res );
x2 = linspace( x2Low, x2High, numX2Pts );

disp( sprintf( '         [%g,%g] (X2: %3d pts)', x2Low, x2High, numX2Pts ) );
disp( sprintf( '   Grid: XXX     [%g,%g] (X1: %3d pts)', x1Low, x1High, ...
	numX1Pts ) );

% rows of X1 are copies of x1; columns of X2 are copies of x2
[X1,X2] = ndgrid( x1, x2 );

% DEBUG
%size( X1 )
%size( X2 )

fVal = zeros( numX1Pts, numX2Pts );

numPts = numX1Pts*numX2Pts;

disp( 'Finding grid values' );
for i1 = 1:numX1Pts
	for i2 = 1:numX2Pts
		fVal(i1,i2) = feval( f, [X1(i1,i2), X2(i1,i2)] );
	end
	numPtsDone = i1*numX2Pts;
	disp( sprintf( '%5.2f %% done', 100*numPtsDone/numPts ) );
end

if ( strcmp( lower(szOptions), 'adjust' ) == 1 )

	% Adjust the grid in the x1 and x2 directions.

	w1 = 1/2;
	w2 = 1/2;

	% Adjust x1-dir
	disp( 'Adjusting x1-dir' );
	for i2 = 2:(numX2Pts-1)
	for i1 = 2:(numX1Pts-1)
		if ( (fVal(i1-1,i2) < fVal(i1,i2)) & (fVal(i1-1,i2) < fVal(i1+1,i2)) )
			X1(i1,i2) = w1*X1(i1,i2) + (1-w1)*X1(i1-1,i2);
		elseif ( (fVal(i1+1,i2) < fVal(i1,i2)) & ...
				 (fVal(i1+1,i2) < fVal(i1-1,i2)) )
			X1(i1,i2) = w1*X1(i1,i2) + (1-w1)*X1(i1+1,i2);
		end
		fVal(i1,i2) = feval( f, [X1(i1,i2), X2(i1,i2)] );
	end
	end

	% Adjust x2-dir
	disp( 'Adjusting x2-dir' );
	for i1 = 2:(numX1Pts-1)
	for i2 = 2:(numX2Pts-1)
		if ( (fVal(i1,i2-1) < fVal(i1,i2)) & (fVal(i1,i2-1) < fVal(i1,i2+1)) )
			X2(i1,i2) = w2*X2(i1,i2) + (1-w2)*X2(i1,i2-1);
		elseif ( (fVal(i1,i2+1) < fVal(i1,i2)) & ...
				 (fVal(i1,i2+1) < fVal(i1,i2-1)) )
			X2(i1,i2) = w2*X2(i1,i2) + (1-w2)*X2(i1,i2+1);
		end
		fVal(i1,i2) = feval( f, [X1(i1,i2), X2(i1,i2)] );
	end
	end

end

% Now find the min in fVal
fValMin = min( fVal(:) );
[i,j] = find( fVal == fValMin );
x0 = [ X1(i,j), X2(i,j) ]';

if ( numRecursions > 0 )

	% Call pick_2d_initial another time. Decide where to place the grid.
	X1_MULT_RES = 0.5;
	X2_MULT_RES = 0.5;

	if ( x0(1) == x1Low ), x1Min = 0.5*x1Low;
	else 
		x1Min = max( [ x1Low, x0(1) - X1_MULT_RES*x1Res ] );
	end

	if ( x0(1) == x1High ), x1Max = 1.5*x1High;
	else
		x1Max = min( [ x1High, x0(1) + X1_MULT_RES*x1Res ] );
	end

	if ( x0(2) == x2Low ), x2Min = 0.5*x2Low;
	else
		x2Min = max( [ x2Low, x0(2) - X2_MULT_RES*x2Res ] );
	end

	if ( x0(2) == x2High ), x2Max = 1.5*x2High;
	else
		x2Max = min( [ x2High, x0(2) + X2_MULT_RES*x2Res ] );
	end

	x1NewRes = ( x1Max - x1Min ) / DEFAULT_X1_PTS;
	x2NewRes = ( x2Max - x2Min ) / DEFAULT_X2_PTS;

	x0 = pick_2d_initial( f, [x1Min,x1Max], [x2Min,x2Max], x1NewRes, ...
		x2NewRes, 'none', numRecursions - 1 );

end


