% pick_3d_initial: Picks an initial point in 3d for a minimization routine.
% Basically lay a grid and pick the point with min value to start from. If
% this point is on the boundary, return with error. An extension of
% pick_2d_initial.
%
% Usage:
%    x0 = pick_3d_initial( f, [x1min,x1max], [x2min,x2max], [x3min,x3max],
%       x1Res, x2Res, x3Res, szOptions, numRecursions )
%

function x0 = pick_3d_initial( f, x1Range, x2Range, x3Range, x1Res, x2Res, ...
	x3Res, szOptions, numRecursions )

DEFAULT_X_PTS = [ 20, 20, 10 ];

numPts = zeros( 3, 1 );
numPts(1) = ceil( (x1Range(2) - x1Range(1))/x1Res );
numPts(2) = ceil( (x2Range(2) - x2Range(1))/x2Res );
numPts(3) = ceil( (x3Range(2) - x3Range(1))/x3Res );

% Limit the number of points

for i = 1:3
	if ( numPts(i) > DEFAULT_X_PTS(i) )
		disp( sprintf( 'numPts(%d) = %g too large!', i, numPts(i) ) );
		numPts(i) = DEFAULT_X_PTS(i)
	end
end

numX1Pts = numPts(1);
numX2Pts = numPts(2);
numX3Pts = numPts(3);

x1 = linspace( x1Range(1), x1Range(2), numX1Pts );
x2 = linspace( x2Range(1), x2Range(2), numX2Pts );
x3 = linspace( x3Range(1), x3Range(2), numX3Pts );

[X1,X2,X3] = ndgrid( x1, x2, x3 );

fVal = zeros( numX1Pts, numX2Pts, numX3Pts );

numPts = numX1Pts * numX2Pts * numX3Pts;

disp( 'Finding grid values' );
for i1 = 1:numX1Pts
	for i2 = 1:numX2Pts
		for i3 = 1:numX3Pts
			fVal(i1,i2,i3) = feval( f, ...
				[X1(i1,i2,i3), X2(i1,i2,i3), X3(i1,i2,i3)] );
		end
	end
	numPtsDone = i1 * numX2Pts * numX3Pts;
	disp( sprintf( '%5.2f %% done', 100*numPtsDone / numPts ) );
end

if ( strcmp( lower(szOptions), 'adjust' ) == 1 )

	w1 = 1/2;
	w2 = 1/2;
	w3 = 1/2;

	% Adjust x1-dir
	disp( 'Adjusting x1-dir' );
	for i2 = 2:(numX2Pts-1)
	for i3 = 2:(numX3Pts-1)
	for i1 = 2:(numX1Pts-1)
		if ( (fVal(i1-1,i2,i3) < fVal(i1,i2,i3)) & ...
			 (fVal(i1-1,i2,i3) < fVal(i1+1,i2,i3)) )
			X1(i1,i2,i3) = w1*X1(i1,i2,i3) + (1-w1)*X1(i1-1,i2,i3);
		elseif ( (fVal(i1+1,i2,i3) < fVal(i1,i2,i3)) & ...
			     (fVal(i1+1,i2,i3) < fVal(i1-1,i2,i3)) )
			X1(i1,i2,i3) = w1*X1(i1,i2,i3) + (1-w1)*X1(i1+1,i2,i3);
		end
		fVal(i1,i2,i3) = feval( f, ...
			[X1(i1,i2,i3), X2(i1,i2,i3), X3(i1,i2,i3)] );
	end
	end
	end

	% Adjust x2-dir
	disp( 'Adjusting x2-dir' );
	for i1 = 2:(numX1Pts-1)
	for i3 = 2:(numX3Pts - 1)
	for i2 = 2:(numX2Pts-1)
		if ( (fVal(i1,i2-1,i3) < fVal(i1,i2,i3)) & ...
			 (fVal(i1,i2-1,i3) < fVal(i1,i2+1,i3)) )
			X2(i1,i2,i3) = w2*X2(i1,i2,i3) + (1-w2)*X2(i1,i2-1,i3);
		elseif ( (fVal(i1,i2+1,i3) < fVal(i1,i2,i3)) & ...
			     (fVal(i1,i2+1,i3) < fVal(i1,i2-1,i3)) )
			X2(i1,i2,i3) = w2*X2(i1,i2,i3) + (1-w2)*X2(i1,i2+1,i3);
		end
		fVal(i1,i2,i3) = feval( f, ...
			[X1(i1,i2,i3), X2(i1,i2,i3), X3(i1,i2,i3)] );
	end
	end
	end

	% Adjust x3-dir
	disp( 'Adjusting x3-dir' );
	for i1 = 2:(numX1Pts-1)
	for i2 = 2:(numX2Pts-1)
	for i3 = 2:(numX3Pts - 1)
		if ( (fVal(i1,i2,i3-1) < fVal(i1,i2,i3)) & ...
			 (fVal(i1,i2,i3-1) < fVal(i1,i2,i3+1)) )
			X3(i1,i2,i3) = w3*X3(i1,i2,i3) + (1-w3)*X3(i1,i2,i3-1);
		elseif ( (fVal(i1,i2,i3+1) < fVal(i1,i2,i3)) & ...
			     (fVal(i1,i2,i3+1) < fVal(i1,i2,i3-1)) )
			X3(i1,i2,i3) = w3*X3(i1,i2,i3) + (1-w3)*X3(i1,i2,i3-1);
		end
		fVal(i1,i2,i3) = feval( f, ...
			[X1(i1,i2,i3), X2(i1,i2,i3), X3(i1,i2,i3)] );
	end
	end
	end

end

% Now find the min in fVal
fValMin = min( fVal(:) );
fValLinInd = find( fVal == fValMin );
[i,j,k] = ind2sub( size(fVal), fValLinInd );
x0 = [ X1(i,j,k), X2(i,j,k), X3(i,j,k) ]';

if ( numRecursions > 0 )

	% Call pick_3d_initial another time. Decide where to place the grid.
	X_MULT_RES = [ 0.5, 0.5, 0.5 ];
	xRes = [ x1Res, x2Res, x3Res ];

	xOldLow = [ x1Range(1), x2Range(1), x3Range(1) ];
	xOldHigh = [ x1Range(2), x2Range(2), x3Range(2) ];

	xNewLow = zeros( size( xOldLow ) );
	xNewHigh = zeros( size( xNewLow ) );

	for i = 1:3
		if ( x0(i) == xOldLow(i) ), xNewLow(i) = 0.5 * xOldLow(i);
		else
			xNewLow(i) = max( [ xOldLow(i), x0(i) - X_MULT_RES(i)*xRes(i) ] );
		end

		if ( x0(i) == xOldHigh(i) ), xNewHigh(i) = 1.5 * xOldHigh(i);
		else
			xNewHigh(i) = min( [ xOldHigh(i), x0(i) + X_MULT_RES(i)*xRes(i) ] );
		end
	end

	xNewRes = [ (xNewHigh(1)-xNewLow(1))/DEFAULT_X_PTS(1), ...
		(xNewHigh(2)-xNewLow(2))/DEFAULT_X_PTS(2), ...
		(xNewHigh(3)-xNewLow(3))/DEFAULT_X_PTS(3) ];

	x0 = pick_3d_initial( f, [xNewLow(1),xNewHigh(1)], ...
		[xNewLow(2),xNewHigh(2)], [xNewLow(3),xNewHigh(3)], ...
		xNewRes(1), xNewRes(2), xNewRes(3), 'none', numRecursions - 1 );

end
