% recon_cmp: compare some reconstruction techniques for MRFM
%-------------------------------------------------------------------------------
% Compare some recon methods for the white paper to John Sidles to motivate
% discussion re MRFM image reconstruction.
%

%%% Parameters
numSpinCentres = 10;
sigmaFac = 1.6;

par = psf_param3( 1, 3e4, 10 );
stMesh = psf_mesh( par );
out = gen_img( par, stMesh, 'random', sigmaFac, [], numSpinCentres, 0 );
L = prod( size( out.f ) );

disp( '*** Running reconstruction algorithms ***' );

% ML
fML = real( f_deconv( stMesh, out.h, out.gNoisy ) );
mseML = (1/L)*sum( (fML(:) - out.f(:)).^2 );
disp( sprintf( 'mseML: %.5e', mseML ) );

% Constrained least squares (CLS) with Laplacian penalty function
vecGamma = [linspace( 0, 1, 20 ), logspace( 0, 20, 100 )];
for i = 1:length( vecGamma )
	gamma = vecGamma(i);
	fTest = real( f_conls( stMesh, out.h, out.gNoisy, gamma ) );
	mseTest = (1/L)*sum( (fTest(:) - out.f(:)).^2 );
	if ( i == 1 )
		fCLS = fTest;
		mseCLS = mseTest;
		bestGamma = gamma;
	else % i > 1, so compare with best estimate so far
		if ( mseTest < mseCLS ) % found a better estimate
			fCLS = fTest;
			mseCLS = mseTest;
			bestGamma = gamma;
		end
	end
end
disp( sprintf( 'mseCLS: %.5e, bestGamma: %g', mseCLS, bestGamma ) );
 
% Empirical Bayes (EB). Define the following operators:
%    Teb - apply the EB method to "clean" up the image
%    Hp  - apply the pseudo-inverse operator

%%% (Teb o Hp):

sOptions = struct( 'bSkipInitialGridSearch', 0, 'nDebugFlag', 0 );

h = real( get_h_op( par, stMesh ) );
facVar = sum( sum( h.^2 ) ); % Compute the variance in the image domain
[estParams, ebOut.muhat] = embayfunc( fML(:), 'laplace', [], out.sigma1/sqrt(facVar), ...
	'median', sOptions );
fEB1 = real( reshape( ebOut.muhat, [stMesh.numPts(1), stMesh.numPts(2)] ) );
mseEB1 = (1/L)*sum( (fEB1(:) - out.f(:)).^2 );
disp( sprintf( 'mseEB1: %.5e', mseEB1 ) );

%%% (Hp o Teb):

[estParams, ebOut.muhat] = embayfunc( real( out.gNoisy(:) ), 'laplace', [], out.sigma1, ...
	'median', sOptions );
gClean = reshape( ebOut.muhat, [stMesh.numPts(1), stMesh.numPts(2)] );
fEB2 = real( f_deconv( stMesh, out.h, gClean ) );
mseEB2 = (1/L)*sum( (fEB2(:) - out.f(:)).^2 );
disp( sprintf( 'mseEB2: %.5e', mseEB2 ) );

%%% (Teb o Hp o Teb):

[estParams, ebOut.muhat] =embayfunc( fEB2(:), 'laplace', [], [], 'median', ...
    struct( 'bSkipInitialGridSearch', 0, 'nDebugFlag', 0 ) );
fEB3 = real( reshape( ebOut.muhat, [stMesh.numPts(1), stMesh.numPts(2)] ) );
mseEB3 = (1/L)*sum( (fEB3(:) - out.f(:)).^2 );
disp( sprintf( 'mseEB3: %.5e', mseEB3 ) );

% Optionally, make plots and save to hard disk
if 1
	close all;
	figure(1), create_int_plot( 1, out.X, out.Y, out.f, [], 'atom locations (f)', 1, 1 );
	figure(2), create_int_plot( 2, out.X, out.Y, real(out.g), [], 'noiseless obs (g)', 1 );
	figure(3), ...
		create_int_plot( 3, out.X, out.Y, real(out.gNoisy), [], 'noisy obs (gNoisy)', 1 );

	figure(4), create_int_plot( 4, out.X, out.Y, real(fML), [], 'deconvolved soln', 1, 1 );
	figure(5), create_int_plot( 5, out.X, out.Y, real(fCLS), [], 'const. LS soln', 1, 1 );

	figure(6), create_int_plot( 6, out.X, out.Y, fEB1, [], 'EB1', 1, 1 );
	figure(7), create_int_plot( 7, out.X, out.Y, fEB2, [], 'EB2', 1, 1 );
	figure(8), create_int_plot( 8, out.X, out.Y, fEB3, [], 'EB3', 1, 1 );

% 	print -f1 -depsc2 f.eps
% 	print -f2 -depsc2 g.eps
% 	print -f3 -depsc2 gTilde.eps
% 	print -f4 -depsc2 fML.eps
% 	print -f5 -depsc2 fCLS.eps
% 	print -f6 -depsc2 fEB1.eps
% 	print -f7 -depsc2 fEB2.eps
% 	print -f8 -depsc2 fEB3.eps
end

% save recon_cmp; 
