% recon_htest: Tests reconstruction with forward operator
%

% There is already a diary command in em_restore
diary off;
if cdir_file_exists( 'recon_htest.log' ), delete 'recon_htest.log', end;
diary 'recon_htest.log';

tic;

if 0

    % ........
    % 1-d case
    % ########

    n = 1e4;
    sigma = 1;

    % Laplacian prior
    a = 1;
    w = 0.05;
    [fClean, fNoisy] = gen_laplace_data( n, a, w, sigma );

    % Johnstone's data
    k = floor( 0.01*n );
    mu = 5;
    [fClean, fNoisy] = gen_john_data( n, k, mu, sigma );

    % Must randomly permute fClean
    fClean = fClean( randperm(n) );

    h = zeros( size(fClean) );

    % FIR
    h(1) = 1;
    h(2) = -0.5;
    h(3) = 0.25;

    % Parabolic blur
    %fK = (3)^3;
    %fa = 10;
    %supp = 1:2:fa;
    %h(supp) = 6*fK/(fa^3) * supp .* ( fa - supp );

    % Gaussian blur
    h(1:11) = normpdf( linspace(-2,2,11), 0, 1 );
    h(1:11) = h(1:11) / max( h(1:11) );

else

    % ........
    % 2-d case
    % ########

    sigmaFac = 1;
    numSpinCentres = 10;

    % Random 2-d data from cantilever
    if 0
        par = psf_param3( 1, 3e4, 10 );
        mesh = psf_mesh( par, [], [100 100 1] );
        out = gen_img( par, mesh, 'random', sigmaFac, [], numSpinCentres, 0 );
        fClean = out.f;
        h = out.h;
        sigma = out.sigma1;
        fh = out.g;
        fhNoisy = out.gNoisy;
        % DEBUG
        %return;
    end

    % Random 2-d data using Gaussian blur
    if 1

        if ~isvar( 'xtrue' )
            % Generate our data
            f = rand( 64, 64 ) > 0.995;
            h = fspecial( 'gaussian', size(f), 0.3 );
            fClean = f;
            %fh = imfilter( f, h );
            fh = real( ifftn( fftn(h) .* fftn(fClean) ) );
            sigma = 0.1;
            fhNoisy = fh + randn( size(fh) ) * sigma;
        else
            % Re-use existing data
            fClean = xtrue;
            %h = zeros( size( fClean ) );
            %h( 1:size(psf,1), 1:size(psf,2) ) = psf;
            %fh = real( ifftn( fftn(h) .* fftn(fClean) ) );
            fh = y0;
            sigma = estd;
            %fhNoisy = fh + randn( size(fh) ) * sigma;
            fhNoisy = yi;

            if 1 % Sanity
                disp( 'SANITY: figuring out h' );
                disp( 'size of psf: ' ), sizePsf = size( psf )

                psfSize = max( sizePsf );
                bPsfMatch = 0;

                for i = 1:( size(fClean,2) - size(psf,2) + 1 )
                    h = zeros( size( fClean ) );
                    h( i:(i+size(psf,1)-1), i:(i+size(psf,2)-1) ) = psf;
                    fhTest = real( ifftn( fftn(h) .* fftn(fClean) ) );
                    normDiff = norm( fh - fhTest, 2 )
                    if ( abs( normDiff ) < eps )
                        disp( 'Matched psf, i:' ), disp( i ), bPsfMatch = 1;
                        break;
                    end
                end

            end % Sanity

            if 0 % Plots
                figure( 1 ), imagesc( fClean ), title( 'fClean' );
                figure( 2 ), imagesc( fhTest ), title( 'fhTest' );
                figure( 3 ), imagesc( fh ), title( 'fh' );
                figure( 4 ), imagesc( fhNoisy ), title( 'fhNoisy' );
                disp( 'Pausing...press key to continue' ), pause;
                %close all;
            end

            if ( bPsfMatch == 0 ), error( 'No match for psf!' ), end;

            if 1 % XXX Debug
                return;
            end
            
        end

    end % Random 2d data using Gaussian blur

end % 2d test case


%........
% OPTIONS
%########

NUM_EM_ITERS = 300;
GRADIENT_SEARCH_FAILURES_MAX = NUM_EM_ITERS;
SZ_RESTORE_METHOD = 'eb_better';
SZ_INITIAL_SELECTOR = 'deconv';

%.......................
% Start restoration code
%#######################

switch ( lower( SZ_RESTORE_METHOD ) )

    case 'eb13'
        sRecon = eb13( h, fhNoisy, sigma );

    case 'none'
        % Do nothing

    otherwise

        %tmp = nnz( fClean(:) ) / length( fClean(:) );
        tmp = 1;
        l2b_n = 0;
        rx_delta = 1e-2;
        
        sOptions = struct( 'aVal', sqrt(2)/2, ...
            'wVal', tmp, ...
            'tVal', -1, ...
            'szSparseInd', 'w', ...
            'bUseHistory', 0, ...
            'bSaveHistory', 0, ...
            'szStoppingCond', 'fixed', 'stoppingCondTol', 1e-2, ...
            'thetaStar', 1, 'softThresT', (1/sqrt(3))*(2^l2b_n)*rx_delta );

        disp( sprintf( 'softThresT=%g', sOptions.softThresT ) );

        sRecon = em_restore( h, fhNoisy, sigma, fClean, ...
            lower(SZ_RESTORE_METHOD), ...
            NUM_EM_ITERS, ...
            GRADIENT_SEARCH_FAILURES_MAX, ...
            lower(SZ_INITIAL_SELECTOR), ...
            sOptions );

end

if ( ( ndims( fClean ) == 2 ) & ( min( size(fClean) ) == 1 ) )

	% If we are dealing with 1-d data, plot the following
	%    1. clean signal before the forward operator
	%    2.  "   "   "   after   "   "   "
	%    3. noisy signal after the forward operator
	%    4. reconstructed signal

	figure( 2 ), clf, hold on, plot( fClean, 'bx' ), plot( fh, 'ro' );
	plot( fhNoisy, 'g+' );

	switch ( lower( SZ_RESTORE_METHOD ) )
	case 'eb'
		plot( sRecon.last, 'k+' );
		legend( 'fClean', 'fh', 'fhNoisy', 'EB M-step', 0 );

	case 'eb13'
		legend( 'fClean', 'fh', 'fhNoisy', 0 );
		figure( 3 ), clf, hold on, plot( sRecon.eb1, 'r+' );
		plot( sRecon.eb2, 'gx' ), plot( sRecon.eb3, 'bo' );	
		title( 'EB1-3 methods' );
		legend( 'eb1', 'eb2', 'eb3', 0 );

	case 'map'
		plot( sRecon.last, 'k+' );
		legend( 'fClean', 'fh', 'fhNoisy', 'MAP M-step', 0 );

	case 'map_prior'
		plot( sRecon.last, 'k+' );
		legend( 'fClean', 'fh', 'fhNoisy', 'MAP+prior M-step', 0 );

	otherwise
		legend( 'fClean', 'fh', 'fhNoisy', 0 );
		fDeconv = pseudoinv_eqdim( h, fhNoisy );
		figure( 3 ), clf, plot( fDeconv, 'k+' ), title( 'fDeconv' );

		%sOptions = struct( 'bSkipInitialGridSearch', 0, 'nDebugFlag', 0 );
		%[estParams, thetaHat] = embayfunc( fhNoisy, 'laplace', [], sigma, ...
		%	'median', sOptions );
		%figure( 3 ), clf, plot( thetaHat, 'rv' ), title( 'thetaHat' );
		%fDeconv2 = pseudoinv_eqdim( h, thetaHat );
		%figure( 4 ), clf, plot( fDeconv2, 'k+' ), title( 'fDeconv2' );
	end

elseif ( ndims( fClean ) == 2 )

	% We are dealing with genuine 2-d data. Recall that vectors in Matlab
	% have #dim = 2.

	%create_int_plot( 1, out.X, out.Y, fClean, [], 'fClean', 1, 1 );
	%create_int_plot( 2, out.X, out.Y, fhNoisy, [], 'fhNoisy', 1, 1 );
    if 1       
        figure( 1 ), imagesc( fClean' ), colormap gray, cbar;
        title( 'image' );
        figure( 2 ), imagesc( fhNoisy' ), colormap gray, cbar;
        title( 'noisy blurred obs' ); 
        figure( 3 ), imagesc( sRecon.last' ), colormap gray, cbar;
        title( 'reconstruction' );
    end

	switch ( lower( SZ_RESTORE_METHOD ) )

	case 'eb13'
		eb1DiffNorm1 = sum( abs( fClean(:) - sRecon.eb1(:) ) );
		eb2DiffNorm1 = sum( abs( fClean(:) - sRecon.eb2(:) ) );
		eb3DiffNorm1 = sum( abs( fClean(:) - sRecon.eb3(:) ) );

		disp( [ ...
            sprintf( 'eb1 norm-1 error: %g\n', eb1DiffNorm1 ), ...
			sprintf( 'eb2 norm-1 error: %g\n', eb2DiffNorm1 ), ...
			sprintf( 'eb3 norm-1 error: %g\n', eb3DiffNorm1 ), ...
        ] );

	end

end

diary off;

toc;
