% rocmult: multiple-file version of roc
%-------------------------------------------------------------------------------
% Function to present the output generated by the mrfmsim C program. Derived
% from roc.m. Currently, its abilities are:
% - loading and plotting the ROC curves
% - finding the power (P_D) corresponding to a specified false alarm (P_F)
%
% Note that you have to make sure the variable values are correct.
%
% History:
% Oct 2003 mlting - Initial coding
% Nov 2004 mlting - Updated to multiple-file version
%
% Usage:
%    [aveVar, rocPow] = rocmult( nRocMethod, numFiles, numSims, szPlotType, 
%       bFindPower )
%
% Outputs:
%    aveVar - the average variance at each ROC data point
%    rocPow - an N x 2 matrix, where each row corresponds to (PD, sigma) for
%             the specified PF
% Inputs:
%    nRocMethod - 1: use read_lrt_vals, 2: use read_lrt_vals2
%    numFiles   - # output files
%    numSims    - # runs to divide the output into
%    szPlotTYpe - 'sym'|'asym'|'rwsym'|'rwasym'
%    bFindPower - fill in rocPow, if set to false, then rocPow is all zeros
%
function [aveVar, rocPow] = rocmult( nRocMethod, numFiles, numSims, ...
	szPlotType, bFindPower )

% Parameters for nRocMethod 
lrt_root 	= 'lrt';
en_u_root 	= 'energy_u';
en_f_root 	= 'energy_f';
mf_root 	= 'match';
map_root 	= 'map';
rt_root 	= 'frank';
lrt_rw_root	= 'lrt_rw';
amp_root	= 'amp';
en_fm_root 	= 'energy_fm';

global ascii_opt dat_ext t0_ext t1_ext

ascii_opt 	= '-ascii';
dat_ext 	= '.dat';
t0_ext 		= '.ts0';
t1_ext 		= '.ts1';

numThresPts = 1e3;
bVerbose	= 1;

% ROC parameters 
alpha		= 0.1;
numRocPts	= 50;

% Plotting constants. NB. all the colours are black, and that is because
% the plots will be printed on black and white.
markMf 		= 'k.';
markLrtRw 	= 'k-.';
markLrt 	= 'kd';
markMap		= 'go';
markRt 		= 'k-';
markEnf 	= 'ks';
markEnfm	= 'k-';
markAmp		= 'k--';
markEnu		= 'kv';


% Boolean masks for the plot types. They are as follows:
% RT-LRT, Enu, Enf, MF, MAP, RT (??), RW-LRT, Amp, Hybrid
bPlotMaskSym	= [ 1 1 1 1 0 0 0 1 1 ];
bPlotMaskAsym	= bPlotMaskSym;
%bPlotMaskRwSym	= [ 1 1 1 1 0 0 1 1 0 ];
bPlotMaskRwSym = [ 0 1 1 1 0 0 1 0 0 ];
%bPlotMaskRwAsym	= [ 0 1 1 1 0 0 1 1 0 ];
bPlotMaskRwAsym = [0 1 1 1 0 0 1 0 0 ];

switch ( lower( szPlotType ) )
case 'sym'
	bMask = bPlotMaskSym;
case 'asym'
	bMask = bPlotMaskAsym;
case 'rwsym'
	bMask = bPlotMaskRwSym;
case 'rwasym'
	bMask = bPlotMaskRwAsym;
otherwise
	error( 'unrecognized szPlotType' );
end

%%% Load and plot ROCs %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

bCreateRoc = 1; % Initialize

switch ( nRocMethod )
case 1	% Use read_lrt_vals

	if ( numSims ~= numFiles )
		error( 'unexpected params for nRocMethod = 1' );
	end

	disp( 'Calling read_lrt_vals...' );
	if ( bMask(1) == 1 ), MLrt = read_lrt_vals( lrt_root, numFiles ); end
	if ( bMask(2) == 1 ), MEnu = read_lrt_vals( en_u_root, numFiles ); end
	if ( bMask(3) == 1 ), MEnf = read_lrt_vals( en_f_root, numFiles ); end
	if ( bMask(4) == 1 ), MMf = read_lrt_vals( mf_root, numFiles ); end
	if ( bMask(5) == 1 ), MMap = read_lrt_vals( map_root, numFiles ); end
	if ( bMask(6) == 1 ), MRt = read_lrt_vals( rt_root, numFiles ); end
	if ( bMask(7) == 1 ), MLrtRw = read_lrt_vals( lrt_rw_root, numFiles ); end
	if ( bMask(8) == 1 ), MAmp = read_lrt_vals( amp_root, numFiles ); end
	if ( bMask(9) == 1 ), MEnfm = read_lrt_vals( en_fm_root, numFiles ); end

case 2 	% Use read_lrt_vals2

	disp( 'Calling read_lrt_vals2...' );
	if ( bMask(1) == 1 )
		MLrt = read_lrt_vals2( lrt_root, numFiles, numSims, numThresPts ); 
	end
	if ( bMask(2) == 1 )
		MEnu = read_lrt_vals2( en_u_root, numFiles, numSims, numThresPts ); 
	end
	if ( bMask(3) == 1 )
		MEnf = read_lrt_vals2( en_f_root, numFiles, numSims, numThresPts ); 
	end
	if ( bMask(4) == 1 )
		MMf = read_lrt_vals2( mf_root, numFiles, numSims, numThresPts ); 
	end
	if ( bMask(5) == 1 )
		MMap = read_lrt_vals2( map_root, numFiles, numSims, numThresPts ); 
	end
	if ( bMask(6) == 1 )
		MRt = read_lrt_vals2( rt_root, numFiles, numSims, numThresPts ); 
	end
	if ( bMask(7) == 1 )
		MLrtRw = read_lrt_vals2( lrt_rw_root, numFiles, numSims, numThresPts ); 
	end
	if ( bMask(8) == 1 )
		MAmp = read_lrt_vals2( amp_root, numFiles, numSims, numThresPts ); 
	end
	if ( bMask(9) == 1 )
		MEnfm = read_lrt_vals2( en_fm_root, numFiles, numSims, numThresPts ); 
	end

otherwise
	bCreateRoc = 0; % Don't create any ROC plots
end

aveVar = zeros( size( bMask ) );

if ( bCreateRoc == 1 )
	close all, figure;
	vecAlpha = [linspace( 0, 0.1, numRocPts/4 ), ...
		linspace( 0.1 + 0.1*(4/numRocPts), 1, floor(3/4*numRocPts) )];

	% We have to resort to this trick in order to get the legend correctly
	% done and still use the errorbar function
	switch ( lower( szPlotType ) )
	case { 'sym', 'asym' }
		% Create dummy plots 
		plot( 0, 0, markMf ), hold on;
		plot( 0, 0, markLrt ), ...
		plot( 0, 0, markEnf ), ...
		plot( 0, 0, markEnfm ), ...
		plot( 0, 0, markAmp ), ...
		plot( 0, 0, markEnu );
		% Create the legend
		legend( 'Matched filter bound', 'RT-LRT', 'Filtered energy',...
			'Hybrid', 'Amplitude', 'Unfiltered energy', 0 );
		% Then erase the dummy plots
		plot( 0, 0, whiteOut( markMf ) ), ...
		plot( 0, 0, whiteOut( markLrt ) ), ...
		plot( 0, 0, whiteOut( markEnf ) ), ...
		plot( 0, 0, whiteOut( markEnfm ) ), ...
		plot( 0, 0, whiteOut( markAmp ) ), ...
		plot( 0, 0, whiteOut( markEnu ) );

	case 'rwsym'
		% Create dummy plots
		plot( 0, 0, markMf ), hold on;
		plot( 0, 0, markLrtRw ), ...
%		plot( 0, 0, markLrt ), ...
		plot( 0, 0, markEnf ), ...
%		plot( 0, 0, markAmp ), ...
		plot( 0, 0, markEnu );
		% Create the legend
%		legend( 'Matched filter bound', 'RW-LRT', 'RT-LRT', ...
%			'Filtered energy', 'Amplitude', 'Unfiltered energy', 0 );
		legend( 'Matched filter bound', 'RW-LRT', 'Filtered energy', ...
			'Unfiltered energy', 0 );
		% Erase the dummy plots
		plot( 0, 0, whiteOut( markMf ) ), ...
		plot( 0, 0, whiteOut( markLrtRw ) ), ...
%		plot( 0, 0, whiteOut( markLrt ) ), ...
		plot( 0, 0, whiteOut( markEnf ) ), ...
%		plot( 0, 0, whiteOut( markAmp ) ), ...
		plot( 0, 0, whiteOut( markEnu ) );

	case 'rwasym'
		% Create dummy plots
		plot( 0, 0, markMf ), hold on;
		plot( 0, 0, markLrtRw ), ...
		plot( 0, 0, markEnf ), ...
%		plot( 0, 0, markAmp ), ...
		plot( 0, 0, markEnu );
		% Create the legend
%		legend( 'Matched filter bound', 'RW-LRT', 'Filtered energy', ...
%			'Amplitude', 'Unfiltered energy', 0 );
		legend( 'Matched filter bound', 'RW-LRT', 'Filtered energy', ...
			'Unfiltered energy', 0 );
		% Erase the dummy plots
		plot( 0, 0, whiteOut( markMf ) ), ...
		plot( 0, 0, whiteOut( markLrtRw ) ), ...
		plot( 0, 0, whiteOut( markEnf ) ), ...
%		plot( 0, 0, whiteOut( markAmp ) ), ...
		plot( 0, 0, whiteOut( markEnu ) );

	end

	% We will always plot the MF detector
	if ( bMask(4) ~= 1 ), error( 'not plotting MF detector???' ), end

	[vecBeta, vecBetaStd] = roc_get_plot( MMf, numSims, vecAlpha );
	errorbar( vecAlpha, vecBeta, vecBetaStd, markMf );
	aveVar(4) = mean( vecBetaStd.^2 );
	if ( bVerbose ), disp( sprintf( 'mf av var=%g', aveVar(4) ) );
	end

	%%% XXX: Debug
	%bMask = zeros( 1, 9 );

	% LRT-RW
	if ( bMask(7) == 1 )
		[vecBeta, vecBetaStd] = roc_get_plot( MLrtRw, numSims, vecAlpha );
		errorbar( vecAlpha, vecBeta, vecBetaStd, markLrtRw );
		aveVar(7) = mean( vecBetaStd.^2 );
		if ( bVerbose ), disp( sprintf( 'lrtrw av var=%g', aveVar(7) ) );
		end
	end

	% LRT-RT
	if ( bMask(1) == 1 )
		[vecBeta, vecBetaStd] = roc_get_plot( MLrt, numSims, vecAlpha );
		errorbar( vecAlpha, vecBeta, vecBetaStd, markLrt );
		aveVar(1) = mean( vecBetaStd.^2 );
		if ( bVerbose ), disp( sprintf( 'lrt av var=%g', aveVar(1) ) );
		end
	end

	% MAP
	if ( bMask(5) == 1 )
		[vecBeta, vecBetaStd] = roc_get_plot( MMap, numSims, vecAlpha );
		errorbar( vecAlpha, vecBeta, vecBetaStd, markMap );
		aveVar(5) = mean( vecBetaStd.^2 );
		if ( bVerbose ), disp( sprintf( 'map av var=%g', aveVar(5) ) );
		end
	end

	% RT
	if ( bMask(6) == 1 )
		[vecBeta, vecBetaStd] = roc_get_plot( MRt, numSims, vecAlpha );
		errorbar( vecAlpha, vecBeta, vecBetaStd, markRt );
		aveVar(6) = mean( vecBetaStd.^2 );
		if ( bVerbose ), disp( sprintf( 'rt av var=%g', aveVar(6) ) );
		end
	end

	% Filtered energy
	if ( bMask(3) == 1 )
		[vecBeta, vecBetaStd] = roc_get_plot( MEnf, numSims, vecAlpha );
		errorbar( vecAlpha, vecBeta, vecBetaStd, markEnf );
		aveVar(3) = mean( vecBetaStd.^2 );
		if ( bVerbose ), disp( sprintf( 'enf av var=%g', aveVar(3) ) );
		end
	end

	% Hybrid
	if ( bMask(9) == 1 )
		[vecBeta, vecBetaStd] = roc_get_plot( MEnfm, numSims, vecAlpha );
		errorbar( vecAlpha, vecBeta, vecBetaStd, markEnfm );
		aveVar(9) = mean( vecBetaStd.^2 );
		if ( bVerbose ), disp( sprintf( 'hybrid av var=%g', aveVar(9) ) );
		end
	end

	% Amplitude
	if ( bMask(8) == 1 )
		[vecBeta, vecBetaStd] = roc_get_plot( MAmp, numSims, vecAlpha );
		errorbar( vecAlpha, vecBeta, vecBetaStd, markAmp );
		aveVar(8) = mean( vecBetaStd.^2 );
		if ( bVerbose ), disp( sprintf( 'amp av var=%g', aveVar(8) ) );
		end
	end

	% Unfiltered energy
	if ( bMask(2) == 1 )
		[vecBeta, vecBetaStd] = roc_get_plot( MEnu, numSims, vecAlpha );
		errorbar( vecAlpha, vecBeta, vecBetaStd, markEnu );
		aveVar(2) = mean( vecBetaStd.^2 );
		if ( bVerbose ), disp( sprintf( 'enu av var=%g', aveVar(2) ) );
		end
	end

	axis tight;

	xlabel( 'Probability of False Alarm (P_F)' );
	ylabel( 'Probability of Detection (P_D)' );
end

%%% Figure out the PD for a given PFA %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

rocPow = zeros( length( bMask ), 2 );

if ( bFindPower == 1 )
	disp( sprintf( 'alpha = %g', alpha ) );
	if ( bMask(1) == 1 )
		[rocPow(1,1),rocPow(1,2)] = roc_interp_mult( MLrt, numSims, alpha );
		disp( 'lrt beta = ' ), disp( rocPow( 1,: ) );
	end
	if ( bMask(2) == 1 )
		[rocPow(2,1),rocPow(2,2)] = roc_interp_mult( MEnu, numSims, alpha );
		disp( 'enu beta = ' ), disp( rocPow( 2,: ) );
	end
	if ( bMask(3) == 1 )
		[rocPow(3,1),rocPow(3,2)] = roc_interp_mult( MEnf, numSims, alpha );
		disp( 'enf beta = ' ), disp( rocPow( 3,: ) );
	end
	if ( bMask(4) == 1 )
		[rocPow(4,1),rocPow(4,2)] = roc_interp_mult( MMf, numSims, alpha );
		disp( 'mf beta = ' ), disp( rocPow( 4,: ) );
	end
	if ( bMask(5) == 1 )
		[rocPow(5,1),rocPow(5,2)] = roc_interp_mult( MMap, numSims, alpha );
		disp( 'map beta = ' ), disp( rocPow( 5,: ) );
	end
	if ( bMask(6) == 1 )
		[rocPow(6,1),rocPow(6,2)] = roc_interp_mult( MRt, numSims, alpha );
		disp( 'rt beta = ' ), disp( rocPow( 6,: ) );
	end
	if ( bMask(7) == 1 )
		[rocPow(7,1),rocPow(7,2)] = roc_interp_mult( MLrtRw, numSims, alpha );
		disp( 'lrtrw beta = ' ), disp( rocPow( 7,: ) );
	end
	if ( bMask(8) == 1 )
		[rocPow(8,1),rocPow(8,2)] = roc_interp_mult( MAmp, numSims, alpha );
		disp( 'amp beta = ' ), disp( rocPow( 8,: ) );
	end
	if ( bMask(9) == 1 )
		[rocPow(9,1),rocPow(9,2)] = roc_interp_mult( MEnfm, numSims, alpha );
		disp( 'enfm beta = ' ), disp( rocPow( 9,: ) );
	end
end

%%% 
%%% FUNCTIONS
%%%

%...............................................................................
function szOut = whiteOut( sIn )
%###############################################################################
% Assume that the plot background is white. If not, then change to the
% appropriate colour.
szOut = strcat( 'w', sIn(2:end) );

%...............................................................................
function M = read_lrt_vals( szRoot, numFiles )
%###############################################################################
global ascii_opt dat_ext t0_ext t1_ext

% Read the first file
szFirstFile = sprintf( '%s%d%s', szRoot, 1, dat_ext );
vecTmp = load( ascii_opt, szFirstFile );
Msize = max( size( vecTmp ) );

% So now that we have the number of records in each data file, create a
% matrix of size numFiles x Msize x 2 (=M). Then, M( i, :, : ) will index the
% data from the i-th data file.
M = zeros( numFiles, Msize, 2 );
M( 1, :, : ) = vecTmp;

% Go through and read in the rest of the files
for fileInd = 2:numFiles
	szNthFile = sprintf( '%s%d%s', szRoot, fileInd, dat_ext );
	vecTmp = load( ascii_opt, szNthFile );
	M( fileInd, :, : ) = vecTmp;
end

% 
% Given the test statistics under H0 and H1, generate a ROC curve. Basically
% a port of the similarly named function in mrfmsim.c
%...............................................................................
function [vecPf, vecPd] = thres_and_save( numThresPts, vecTs0, vecTs1 )
%###############################################################################

n = length( vecTs0 );
if ( length( vecTs1 ) ~= n )
	error( 'thres_and_save: incompatible vectors' );
end

thresMin = min( [min(vecTs0), min(vecTs1)] );
thresMax = max( [max(vecTs0), max(vecTs1)] );
thresEps = 1.0e-2 * ( thresMax - thresMin );
thresMin = thresMin - thresEps;
thresMax = thresMax + thresEps;
thresDiff = thresMax - thresMin;

vecPf = []; % Initialize
vecPd = [];

for thresInd = 1:numThresPts
	thres = thresMin + (thresInd - 1)/(numThresPts - 1) * thresDiff;
	numH1Correct = sum( vecTs1 > thres );
	numH0Correct = sum( vecTs0 < thres );
	Pd = numH1Correct / n;
	Pf = 1 - numH0Correct / n;
	vecPf = [vecPf; Pf];
	vecPd = [vecPd; Pd];	
end

%
% A more general version of read_lrt_vals. Instead of using the #sims as the
% #files, we allow #sims to be arbitrary. This means that we have to read in
% the H0 and H1 test statistics, divide them into #sims blocks, and generate
% a ROC curve for each block.
%...............................................................................
function M = read_lrt_vals2( szRoot, numFiles, numSims, numThresPts )
%###############################################################################
global ascii_opt dat_ext t0_ext t1_ext

vecTs0 = [];
vecTs1 = [];

% Read in all the Ts0 and Ts1 files
for fileInd = 1:numFiles
	szTs0File = sprintf( '%s%d%s', szRoot, fileInd, t0_ext );
	vecTmp = load( ascii_opt, szTs0File );
	vecTs0 = [vecTs0; vecTmp(:)];

	szTs1File = sprintf( '%s%d%s', szRoot, fileInd, t1_ext );
	vecTmp = load( ascii_opt, szTs1File );
	vecTs1 = [vecTs1; vecTmp(:)];
end

n = length( vecTs0 );
if ( length( vecTs1 ) ~= n )
	error( 'read_lrt_vals2: length of two Ts0 and Ts1 vectors unequal' );
end

% After this, chop up the test statistics into numSims cases. Fo each "case",
% call thres_and_save to create an individual ROC curve.
simBoundary = round( linspace( 1, n+1, numSims+1 ) );

% The matrix M will be of dimension numSims x numThresPts x 2
M = zeros( numSims, numThresPts, 2 );

for simInd = 1:numSims
	simBoundaryStart = simBoundary( simInd );
	simBoundaryEnd = simBoundary( simInd + 1 ) - 1;
	[vecPf, vecPd] = thres_and_save( numThresPts, ...
		vecTs0( simBoundaryStart:simBoundaryEnd ), ...
		vecTs1( simBoundaryStart:simBoundaryEnd ) );
	M( simInd, :, 1 ) = vecPf;
	M( simInd, :, 2 ) = vecPd;
end

%...............................................................................
function [beta, betaStd] = roc_interp_mult( M, numSims, alpha )
%###############################################################################

vecBeta = zeros( 1, numSims );

for fileInd = 1:numSims
	P = roc_interp( M( fileInd, :, 1 ), M( fileInd, :, 2 ) );
	vecBeta( fileInd ) = get_beta_value( alpha, P );
end

beta = mean( vecBeta );
betaStd = std( vecBeta );

%
% Given a 3-d matrix that contains several ROC curves, we plot the average
% ROC curve with error bars.
%...............................................................................
function [vecBeta, vecBetaStd] = roc_get_plot( M, numSims, vecAlpha )
%###############################################################################

lenVecAlpha = length( vecAlpha );
vecBeta = zeros( lenVecAlpha, 1 );
vecBetaStd = zeros( lenVecAlpha, 1 );

for alphaInd = 1:lenVecAlpha
	% Value of alpha in this iteration
	alpha = vecAlpha( alphaInd );

	% Contains the vector of different betas that correspond to this value
	% of alpha
	vecBetaTmp = zeros( numSims, 1 );

	for setInd = 1:numSims
		P = roc_interp( squeeze( M( setInd, :, 1 ) ), ...
			squeeze( M( setInd, :, 2 ) ) );
		vecBetaTmp( setInd ) = get_beta_value( alpha, P );
	end

	vecBeta( alphaInd ) = mean( vecBetaTmp );
	vecBetaStd( alphaInd ) = std( vecBetaTmp );
end


