
% SBL
%
% Given a point spread function, the observation t, estimate the 
% initial weight vector w according to
% t = Phi * w + n
%
% Note that some code from emgauss_restore has been re-used. Hence the
% appearance of variable names with theta in them.
%

function sRestore = sbl_restore2( varargin )

% Default restore options

sRestoreOptions = struct( 'B_SBL_PLOT_SANITY', 0, ...
	'Phi', [], ...
    'STOPPING_COND_LAST_FAC', 1, ...
    'GAMMA0_SELECTOR', 'ones', ...
    'GAMMA0', [], ...
	'LAST_MUHAT', [], ...
    'B_UPDATE_NOISESIGMA', 0 ...
    );
    
% Process input arguments

lenVarargin = length( varargin );
numArgsIn = floor( lenVarargin/2 );

if 2*numArgsIn ~= lenVarargin, error( 'even ip args expected' ), end

for argInInd = 1:numArgsIn
    
    szKey = lower( varargin{2*argInInd-1} );    
    switch szKey
        case 'forward_op', hForwardOp = varargin{2*argInInd};
        case 'noisy_output', t = varargin{2*argInInd};
        case 'noise_sigma', noiseSigma = varargin{2*argInInd};
        case 'theta_true', wTrue = varargin{2*argInInd}; 
                 
        case 'stopping_cond', sRestoreOptions.STOPPING_COND = varargin{2*argInInd};
        case 'stopping_cond_tol', sRestoreOptions.STOPPING_COND_TOL = varargin{2*argInInd};
        case 'stopping_cond_win', sRestoreOptions.STOPPING_COND_WIN = varargin{2*argInInd};
        case 'stopping_cond_last_fac', sRestoreOptions.STOPPING_COND_LAST_FAC = varargin{2*argInInd};
        case 'num_em_iters', sRestoreOptions.NUM_EM_ITERS = varargin{2*argInInd};    
		case 'phi', sRestoreOptions.Phi = varargin{2*argInInd};
        case 'gamma0_selector', sRestoreOptions.GAMMA0_SELECTOR = varargin{2*argInInd};
        case 'gamma0', sRestoreOptions.GAMMA0 = varargin{2*argInInd};
		case 'last_muhat', sRestoreOptions.LAST_MUHAT = varargin{2*argInInd};
            
        otherwise
            error( sprintf( 'unrecognized option %s', szKey ) );
    end
end

% Note that we expect to be given the zero-padded psf
N = size( hForwardOp, 1 ); 
if ( N ~= size( hForwardOp, 2 ) ), error( 'h is not a square matrix' ), end;
NN = N^2; % The matrix phi will be of size NN x NN

% Check to see if noiseSigma has been initialized. If not, do it ourselves
if ~exist( 'noiseSigma', 'var' )
    disp( 'Initializing noiseSigma with image std. dev.' );
    noiseSigma = std( t(:) );
    disp( sprintf( 'Setting noiseSigma = %g', noiseSigma ) );
    sRestoreOptions.B_UPDATE_NOISESIGMA = 1;
end

% Check validity of arguments
if noiseSigma < 0, error( 'noiseSigma < 0' ), end;
switch ( sRestoreOptions.STOPPING_COND )
    case 'conv_win'
        if sRestoreOptions.STOPPING_COND_WIN < 0 | sRestoreOptions.STOPPING_COND_TOL < 0
            error( 'cond_win and cond_tol must be non-neg' );
        end
    case 'conv_last'
        if sRestoreOptions.STOPPING_COND_LAST_FAC < 0 | sRestoreOptions.STOPPING_COND_TOL < 0
            error( 'cond_last_fac and cond_tol must be non-neg' );
        end
end

% Allocate space for system matrix. Bad idea: Matlab cannot build matrices
% whose N^2 value exceeds the max representable integer. Use sparse
% matrices instead. 
%Phi = zeros( N, N );

bCreatedPhi = 0;

% Build the system matrix. Note that the operations below are valid when we
% assume that the matrix can be represented by a straightforward
% convolution kernel. That is, it can be diagonalized using DFTs.

if isempty( sRestoreOptions.Phi ), disp( 'Creating Phi' )
    if ( ndims( hForwardOp ) ~= 2 ), error( 'psf must be 2d' ), end;
    sRestoreOptions.Phi = create_smatrix_from_psf2( hForwardOp, 1 );
	bCreatedPhi = 1;
else
	disp( 'Using supplied Phi matrix' );
end

if sRestoreOptions.B_SBL_PLOT_SANITY
    tNoiseless = sRestoreOptions.Phi * wTrue;
    tNoiselessImg = reshape( tNoiseless, N, N );
    imagesc( tNoiselessImg ), colorbar;
end

switch sRestoreOptions.GAMMA0_SELECTOR
    case 'ones'
        vecGamma0 = ones( NN, 1 );
        
    case 'custom'
        vecGamma0 = sRestoreOptions.GAMMA0;
        
    otherwise
        error( 'invalid GAMMA0_SELECTOR' );
end

errorNorm1Hist = [];
errorNorm2Hist = [];
costFuncHist = [];
thetaMaxDiffHist = [];
thetaSumDiffHist = [];

% Initial values of gamma
vecGammaPrev = vecGamma0;

if ~isempty( sRestoreOptions.LAST_MUHAT )
	% Not really necessary--the only thing is that the stats for
	% thetaMaxDiff and thetaSumDiff are preserved between simulation
	% runs.
	muCompHatPrevPrev = sRestoreOptions.LAST_MUHAT(:);
else
	muCompHatPrevPrev = zeros( size( vecGammaPrev ) );
end

emCounter = 0; % Loop counter
bStopIterations = 0;
stoppingCondWin = sRestoreOptions.STOPPING_COND_WIN;

tCpuTimeStart = cputime;

while 1, % Iterate until convergence
     
    Gamma = diag( vecGammaPrev );
    % Print out the minimum and maximum gamma value
    gammaMin = min( vecGammaPrev );
    gammaMax = max( vecGammaPrev );
    disp( sprintf( 'gamma: min=%g, max=%g', gammaMin, gammaMax ) );
    
    % Select the non-zero subset of the gamma vector
    gammaNzIndSet = sort( find( vecGammaPrev > 1e2*eps ) );
    disp( sprintf( 'non-zero subset size: %g/%g', length(gammaNzIndSet), NN ) );
    vecGammaNz = vecGammaPrev( gammaNzIndSet );
    GammaNz = diag( vecGammaNz );
    PhiNz = sRestoreOptions.Phi( :, gammaNzIndSet );
    
    tmp = PhiNz * GammaNz;
     
    Sigma_t = noiseSigma^2 * eye( NN ) + tmp * PhiNz';    
    Log_Det_Eigval_Sigma_t = sum( log ( eig( Sigma_t ) ) );
    InvSigma_t = inv( Sigma_t );
        
    Sigma_w = GammaNz - tmp' * InvSigma_t * tmp;
    Diag_Sigma_w = diag( Sigma_w );    
    muHatPrev = (1/noiseSigma^2) * Sigma_w * PhiNz' * t;

    % DEBUG
    if 0
        figure( 1 ), semilogy( Diag_Sigma_w, 'bx' );
        figure( 2 ), semilogy( Eigval_Sigma_t, 'ro' );
        disp( 'Press key to continue...' ), pause;
    end   
                      
    %vecGammaNext = Diag_Sigma_w + muHatPrev.^2;
    % Use an alternative update which is NOT guaranteed to decrease the
    % cost in that it's not an EM iteration
    size( muHatPrev )
    size( Diag_Sigma_w )
    size( vecGammaNz )
    vecGammaNext = (muHatPrev.^2) ./ (1 - Diag_Sigma_w ./ vecGammaNz );
    
    noiseVarNext = (1/NN) * ( norm( t - PhiNz * muHatPrev, 2 ) + ...
        noiseSigma*noiseSigma * sum( 1 - Diag_Sigma_w ./ vecGammaNz ) );
    
    % Increase iteration counter
    emCounter = emCounter + 1;
       
    % Form muCompHatPrev, which is the COMPLETE muHatPrev. Likewise do the
    % same for vecCompGammaNext.
    muCompHatPrev = zeros( NN, 1 );
    muCompHatPrev( gammaNzIndSet ) = muHatPrev;
    vecCompGammaNext = zeros( NN, 1 );    
    vecCompGammaNext( gammaNzIndSet ) = vecGammaNext;
    
    % Compute the l_1 and l_2 error norm
    errorNorm1Hist = [ errorNorm1Hist, norm( muCompHatPrev(:) - wTrue(:), 1 ) ];
    errorNorm2Hist = [ errorNorm2Hist, norm( muCompHatPrev(:) - wTrue(:), 2 ) ];   
    
    % Compute the cost function that EM is trying to maximize, which is
    % L = log|Sigma_t| + t^T Sigma_t^{-1} t
    
    % N.B. it seems that Sigma_t is very close to being singular, so the
    % |Sigma_t| ~ 0, which will result in L being -inf we we try to compute
    % everything. Later note: it is stupid to use det(). Must explicitly
    % compute the eigenvalues and take sum of the log.
    costFuncHist = [ costFuncHist; Log_Det_Eigval_Sigma_t + t' * InvSigma_t * t ];
    
    if ( length( costFuncHist ) >= 2 ) & ( costFuncHist(end) > costFuncHist(end-1) ) 
        warning( 'Iterations increased cost function!' );
        bStopIterations = 1;
    end
    
    % Store the difference in the estimates between iterations. The
    % stopping condition is based on how "small" the difference between
    % consecutive reconstructions are.    
    thetaHatDiff = muCompHatPrev(:) - muCompHatPrevPrev(:);
    % The l_infinity norm of the difference
    thetaMaxDiffHist = [thetaMaxDiffHist; max( abs( thetaHatDiff ) )];
    % The l_1 norm of the difference
    thetaSumDiffHist = [thetaSumDiffHist; sum( abs( thetaHatDiff ) )];
     
    disp( sprintf( 'loop %g: cost = %g, l_1(error) = %g, l_inf(diff) = %g', ...
        emCounter, costFuncHist(end), errorNorm1Hist(end), thetaMaxDiffHist(end) ) );
    
    % Check stopping condition
    
    switch ( sRestoreOptions.STOPPING_COND )
        case 'fixed'
            % Perform a fixed number of iterations
            if ( emCounter >= sRestoreOptions.NUM_EM_ITERS )
                bStopIterations = 1;
                warning( 'stopping cond=fixed, exceeded NUM_EM_ITERS' );
            end
            
        case 'conv_win'
            % Find the max value in the last STOPPING_COND_WINDOW iterations,
            % and if it is less than a certain tolerance, quit.
            histLen = length( thetaMaxDiffHist );
            if ( histLen >= stoppingCondWin )
                maxChange = max( ...
                    thetaMaxDiffHist( max(1,histLen-stoppingCondWin): histLen )  ...
                    );
                if ( maxChange < sRestoreOptions.STOPPING_COND_TOL )
                    bStopIterations = 1;
				else
    				disp( sprintf( 'stop cond: maxChange=%g, tol=%g', ...
						maxChange, sRestoreOptions.STOPPING_COND_TOL ) );
                end
            end

            % Never exceed NUM_EM_ITERS
            if ( emCounter > sRestoreOptions.NUM_EM_ITERS )
                warning( 'stopping cond=conv_win, but exceeded NUM_EM_ITERS' );
                bStopIterations = 1;
            end
            
        case 'conv_last'
            % Find the last time T that we saw a value exceed
            % STOPPING_COND_TOL. We don't want to see another value exceed
            % STOPPING_COND_TOL for the next STOPPING_COND_LAST_FAC*T samples. 
			% More rigorous stopping condition than conv_win.
            indLastViolation = ...
				find( thetaMaxDiffHist > sRestoreOptions.STOPPING_COND_TOL, ...
					1, 'last' );
            if isempty( indLastViolation )
                indLastViolation = -1;
            end

			stopSpacing = emCounter - indLastViolation + 1;
            stopThres = sRestoreOptions.STOPPING_COND_LAST_FAC * indLastViolation;
            if ( stopSpacing > stopThres )
                bStopIterations = 1;
            else
                disp( sprintf( 'stop cond: spacing=%g, thres=%g', ...
                    stopSpacing, stopThres ) );
            end
 
            % Never exceed NUM_EM_ITERS
            if ( emCounter > sRestoreOptions.NUM_EM_ITERS )
                warning( 'stopping cond=conv_last, but exceeded NUM_EM_ITERS' );
                bStopIterations = 1;
            end

        otherwise
            error( 'unrecognized stopping condition' );
    end
    
    % Update previous variables with next
    vecGammaPrev = vecCompGammaNext;
    muCompHatPrevPrev = muCompHatPrev;
    
    % Check to see if we should update noiseSigma
    if sRestoreOptions.B_UPDATE_NOISESIGMA
        noiseSigma = sqrt( noiseVarNext );
        disp( sprintf( 'updating noiseSigma = %g', noiseSigma ) );
    end
    
    if bStopIterations, break, end
    
end

tCpuTimeStop = cputime;

% Defines the structure of sRestore
sRestoreElems = { 'last', 'muCompHatPrevPrev'; ...
    'initGam', 'vecGamma0'; ...
    'lastGam', 'vecGammaPrev'; ...
    'thetaMaxDiff', 'thetaMaxDiffHist'; ...
    'thetaSumDiff', 'thetaSumDiffHist'; ...
    'errorNorm1', 'errorNorm1Hist'; ...
    'errorNorm2', 'errorNorm2Hist'; ...
    'cost', 'costFuncHist'; ...
    'cputime', '[tCpuTimeStart, tCpuTimeStop]'; ...
    'noiseSigma', 'noiseSigma' ...
    };

if bCreatedPhi

	% If we had to create phi, return the system matrix in sRestore. Else,
	% the caller already has phi, so there's no need to return it in
	% sRestore.

    sRestoreElems = vertcat( sRestoreElems, { 'phi', 'sRestoreOptions.Phi' } );
    
else
    sRestoreElems = vertcat( sRestoreElems, { 'phi', '[]' } );
end

for elemInd = 1:size( sRestoreElems, 1 )
    szCmd = sprintf( 'sRestore.%s = %s;', ...
        char( sRestoreElems(elemInd,1) ), ...       
        char( sRestoreElems(elemInd,2) ) );
    try
        eval( szCmd );
    catch
        warning( sprintf( 'Could not execute: %s', szCmd ) );
    end
end
