
%--------------------------------------------------------------------------
% Create framework for testing image reconstruction based on different 
% image types and different point spread functions. Parts copied from
% Prof. Fessler's restore_sparse.m file.
%--------------------------------------------------------------------------

SZ_IMAGE_TYPE = 'hex';
SZ_PSF_TYPE = 'mrfm';

B_PSF_SANITY = 1;

N_GEN_PLOTS = 2; % Set to 1 or 2 for plots. Otherwise, set to 0 for no plots.

B_GEN_SAME_NOISE = 0;

if ~exist( 'xtrue', 'var' ), disp 'Generating image'
    
    % Reset state so that we get the same "random" image
    rand( 'state', 0 );
    
    switch ( SZ_IMAGE_TYPE )
        
        % Nonzero value distributed as one-sided exponential density
        case 'exp'
            
            if strcmp( SZ_PSF_TYPE, 'gaussian' ) ~= 1
                error( 'exp only works with the gaussian blur function for now' );
            end
            
            rand( 'state', -1 );
            propImage = struct( 'N', 32, 'w', 5e-2, 'lambda', .1, 'emptySpace', 3 );
%             rand( 'state', 2005 );
%             propImage = struct( 'N', 32, 'w', 5e-2, 'lambda', 3, 'emptySpace', 3 );
            xtrue0Size = propImage.N - 2*propImage.emptySpace;

            xtrue0 = reshape( gen_exp_data( xtrue0Size^2, propImage.lambda, propImage.w ), ...
				xtrue0Size, xtrue0Size );
            xtrue = zeros( propImage.N, propImage.N );
            indStart = propImage.emptySpace + 1;
            xtrue( indStart:(indStart+size(xtrue0,1)-1), indStart:(indStart+size(xtrue0,2)-1) ) = xtrue0;

        % Uniform valued image
        case 'uniform'
                        
            if strcmp( SZ_PSF_TYPE, 'gaussian' ) | strcmp( SZ_PSF_TYPE, 'gaussian2' )
                % Tuned for Gaussian blur    
                propImage = struct( 'N', 32, 'uniformLevel', 0.98, 'ampl', 1, ...
                    'emptySpace', 3 );
            else                            
                % Tuned for MRFM blur           
                % Large picture
%                 propImage = struct( 'N', 128, 'uniformLevel', 0.997, 'ampl', 1, ...
%                     'emptySpace', 22 );
                % Smaller picture
                rand( 'state', 5 );
                propImage = struct( 'N', 64, 'uniformLevel', 0.985, 'ampl', 1, ...
                    'emptySpace', 16 );
            end
            
            xtrue0 = propImage.ampl * ...
                (rand( propImage.N-2*propImage.emptySpace, ...
                    propImage.N-2*propImage.emptySpace ) > propImage.uniformLevel);
            xtrue = zeros( propImage.N, propImage.N );
            indStart = propImage.emptySpace + 1;
            xtrue( indStart:(indStart+size(xtrue0,1)-1), ...
                   indStart:(indStart+size(xtrue0,2)-1) ) = xtrue0;
            disp( sprintf( 'Number of non-zero values: %g', ...
                sum( xtrue(:) ) / propImage.ampl ) );

        % Image where the non-zero values come from a uniform distribution
        case 'uniform2'
            if strcmp( SZ_PSF_TYPE, 'gaussian' ) ~= 1
                error( 'uniform2 only works with gaussian blur function for now' );
            end
            
            propImage = struct( 'N', 32, 'uniformLevel', 0.98', ...
                'uMin', 5, 'uMax', 50, 'emptySpace', 3 );
            xtrue0Size = propImage.N - 2*propImage.emptySpace;
            rand( 'state', 2 );
            xtrue0 = (propImage.uMin + (propImage.uMax-propImage.uMin)*rand( xtrue0Size, xtrue0Size ) ) .* ...
                (rand( xtrue0Size, xtrue0Size ) > propImage.uniformLevel );
            xtrue = zeros( propImage.N, propImage.N );
            indStart = propImage.emptySpace + 1;
            xtrue( indStart:(indStart+size(xtrue0,1)-1), ...
                   indStart:(indStart+size(xtrue0,2)-1) ) = xtrue0;
            
        % Image that is distributed according to the J&S prior
        case 'js_prior'
            
            if strcmp( SZ_PSF_TYPE, 'gaussian' )
                % Tuned for the Gaussian blur
                rand( 1996 );
                %propImage = struct( 'N', 32, 'a', .1, 'w', 5e-2, ...
                %    'emptySpace', 3 );                
                propImage = struct( 'N', 32, 'a', 1, 'w', 4e-2, ...
                    'emptySpace', 3 );
                %propImage = struct( 'N', 32, 'a', 1e-2, 'w', 1, ...
                %    'emptySpace', 3 );
            else
                % Tuned for the MRFM blur
                propImage = struct( 'N', 128, 'a', sqrt(2), 'w', 7.09e-3, ...
                    'emptySpace', 22 );
            end
            
            Ntilde = propImage.N - 2*propImage.emptySpace;
            [jsClean, jsNoisy] = gen_laplace_data( Ntilde * Ntilde, ...
                propImage.a, propImage.w, 1 ); % set sigma=1, don't use jsNoisy
            jsCleanMat = reshape( jsClean(:), Ntilde, Ntilde );
            xtrue = zeros( propImage.N, propImage.N );
            indStart = propImage.emptySpace + 1;
            xtrue( indStart:(indStart+size(jsCleanMat,1)-1), ...
                   indStart:(indStart+size(jsCleanMat,2)-1) ) = jsCleanMat;

        case 'hex'
            
            if strcmp( SZ_PSF_TYPE, 'gaussian' )
                % Tuned for the Gaussian blur
                propImage = struct( 'N', 32, 'ampl', 1, 'emptySpace', 6, ...
                    'numHex', 1, 'hexRadius', 4 );
            else
                % Tuned for the MRFM blur
                propImage = struct( 'N', 64, 'ampl', 1, 'emptySpace', 20, ...
                    'numHex', 3, 'hexRadius', 6 );
            end
            
            xtrue = zeros( propImage.N, propImage.N );           
            centre = zeros( propImage.numHex, 2 );

            % Generate centres randomly            
            % Initialize seed                       
%             rand( 'state', 653265 ); 
%             for hexInd = 1:propImage.numHex            
%                 centre( hexInd, : ) = propImage.emptySpace * ones(2,1) + ...
%                     round( rand(2,1) * ( propImage.N - 2*propImage.emptySpace ) );
%                 disp( sprintf( 'hex center: %g, %g', centre(1), centre(2) ) );
%             end

            % Generate centres deterministically
            centre( 1, : ) = [ 20; 20 ];
            centre( 2, : ) = [ 32; 42 ];
            centre( 3, : ) = [ 42; 32 ];
            
            for hexInd = 1:propImage.numHex
                hexCoords = round( gen_hex_coords( centre( hexInd, : ), propImage.hexRadius ) );
                for vertexInd = 1:6
                    xtrue( hexCoords(vertexInd,1), hexCoords(vertexInd,2) ) = ...
                        xtrue( hexCoords(vertexInd,1), hexCoords(vertexInd,2) ) + propImage.ampl;
                end
            end

        case 'uniform3'
            nnz=10;
            x=[ones(nnz,1);zeros(32*32-nnz,1)];
            p=randperm(32*32);
            x=x(p);
            x=reshape(x,32,32);
            xtrue = x;
            
        % For ICIP06 - binary valued image
        case 'icip06_binary'

            propImage = struct( 'N', 32, 'ampl', 1, 'nnz', 36, 'emptySpace', 3 );
            dim0 = propImage.N - 2*propImage.emptySpace;
            xtrue0 = [propImage.ampl*ones( propImage.nnz, 1 ); zeros( dim0^2 - propImage.nnz, 1 )];
            xtrue0 = xtrue0( randperm( dim0^2 ) );
            xtrue0 = reshape( xtrue0, dim0, dim0 );
            xtrue = zeros( propImage.N, propImage.N );
            indStart = propImage.emptySpace + 1;
            xtrue( indStart:(indStart+dim0-1), indStart:(indStart+dim0-1) ) = xtrue0;

        % Extension of icip06_binary to include negative values as well               
        case 'trinary'
            
            propImage = struct( 'N', 32, 'ampl', 1e2, 'nnzPos', 10, 'nnzNeg', 10, 'emptySpace', 3 );
            dim0 = propImage.N - 2*propImage.emptySpace;
            xtrue0 = [ 1*propImage.ampl*ones( propImage.nnzPos, 1 ); ...
                -1*propImage.ampl*ones( propImage.nnzNeg, 1 ); ...
                zeros( dim0^2 - propImage.nnzPos - propImage.nnzNeg, 1 ) ];
            xtrue0 = xtrue0( randperm( dim0^2 ) );
            xtrue0 = reshape( xtrue0, dim0, dim0 );
            xtrue = zeros( propImage.N, propImage.N );
            indStart = propImage.emptySpace + 1;
            xtrue( indStart:(indStart+dim0-1), indStart:(indStart+dim0-1) ) = xtrue0;
                        
        % For ICIP06 - uniform dist centred around 0
        case 'icip06_udca0'
            
            propImage = struct( 'N', 32, 'u1', -1, 'u2', 1, 'nnz', 12, 'emptySpace', 3 );            
            ud = (propImage.u2-propImage.u1) * rand( propImage.nnz, 1 ) + propImage.u1;
            dim0 = propImage.N - 2*propImage.emptySpace;
            xtrue0 = [ud; zeros( dim0^2 - propImage.nnz, 1 )];
            xtrue0 = xtrue0( randperm( dim0^2 ) );
            xtrue0 = reshape( xtrue0, dim0, dim0 );
            xtrue = zeros( propImage.N, propImage.N );
            indStart = propImage.emptySpace + 1;
            xtrue( indStart:(indStart+dim0-1), indStart:(indStart+dim0-1) ) = xtrue0;            
            
        otherwise
            error( 'unrecognized SZ_IMAGE_TYPE' );
            
    end

	[imgNx imgNy] = size( xtrue );
    
    if N_GEN_PLOTS == 1
        im clf, pl = 130;
        im( pl+1, xtrue, 'xtrue' ), colorbar
    end
    
end

if ~exist( 'psf', 'var' ), disp 'Generating psf of blur function'

    switch ( SZ_PSF_TYPE )

        % Gaussian blur function, from restore_sparse.m
        case 'gaussian'
            propPsf = struct( 'fwhm', 3, 'nk_half', 5 );
            psf = gaussian_kernel( propPsf.fwhm, propPsf.nk_half );
            psf = psf * psf';

        % MRFM blur function
        case 'mrfm'

            % Because we need more variables in constructing the MRFM psf,
            % use a cell array. The first index is the 'par' struct, and
            % the second is the 'mesh' struct.

            propPsf{1} = psf_param4( 'Bres', 1e4, ...
                'd', 6, ...
                'nVerboseLevel', 1, ...
                'bUseSmallerR0', 1, ...
                'bUseSmallerB0', 0, ...
                'bComputeOptR0', 0, ...
                'R0', 4, ...
                'bComputeOptBext', 1 );
            propPsf{1}.ySpan = propPsf{1}.xSpan;
            propPsf{1}.vecNumPts = [imgNx, imgNy, 1]; % Only consider 2d arrays
            propPsf{2} = psf_mesh( propPsf{1}, [], [propImage.N, propImage.N, 1] );
            psf = get_h_op( propPsf{1}, propPsf{2}  );
 
            % Normalize psf so that its max value is 1
            psfMax = max( psf(:) );
            psf = psf / psfMax;

            % Compute the SUPPORT of the psf in the x and y plane
            psfNnz = abs( psf ) > eps;
            [nnzI, nnzJ] = find( abs( psf ) > eps );
            psfSuppX = [ min(nnzI), max(nnzI)];
            psfSuppY = [ min(nnzJ), max(nnzJ)];
            disp( sprintf( 'Support of psf in x-y plane: [%g,%g] x [%g,%g]', ...
                psfSuppX(1), psfSuppX(2), psfSuppY(1), psfSuppY(2) ) );
            propPsf{3} = [ psfSuppX; psfSuppY ];

        % Another Gaussian blur function
        case 'gaussian2'
            tmp=[-5:1:5]';
            psf=exp(-((tmp*ones(size(tmp))').^2+(ones(size(tmp))*tmp').^2)/2/2);

        otherwise
            error( 'unrecognized SZ_PSF_TYPE' );

    end

    [psfNx psfNy] = size( psf );
    disp( sprintf( 'size of psf is %g x %g', psfNx, psfNy ) );

end

bPsfMatch = 0;
iMatch = 0;

mask = logical(ones(imgNx, imgNy));
Gfft = Gblur( mask, 'psf', psf, 'type', 'conv,fft', 'nDim', 2 );
Gconv = Gblur( mask, 'psf', psf, 'type', 'conv,same', 'nDim', 2 );

nConvMethod = 1; % 1 => use DFTs, 2 => use convolution

if B_PSF_SANITY, disp( 'Doing sanity check on psf' )

    disp( 'Size of psf: ' ), sizePsf = size( psf )
    psfSize = max( sizePsf );

    switch nConvMethod
        case 1
            y0Tmp = Gfft * xtrue;
            % Figure out the position where the psf matrix has to be
            for i = 1:( size(xtrue,2) - size(psf,2) + 1 )
                psfZeroPadded = zeros( size( xtrue ) );
                psfZeroPadded( i:(i+size(psf,1)-1), i:(i+size(psf,2)-1) ) = psf;
                y0Test = real( ifftn( fftn(psfZeroPadded) .* fftn(xtrue) ) );
                normDiff = norm( y0Tmp - y0Test, 2 );
                if ( abs( normDiff ) < eps )
                    disp( 'Matched psf, i:' ), disp( i );
                    iMatch = i;
                    bPsfMatch = 1;
                    break;
                end
            end
        case 2
            psfZeroPadded = zeros( size( xtrue ) );
            psfZeroPadded( 1:size(psf,1), 1:size(psf,2) ) = psf;
            bPsfMatch = 1; % XXX
        otherwise
            error( 'illegal nConvMethod' );
    end
    
end 

if ~bPsfMatch, error( 'No match for psf!' ), end

if ~exist( 'y0', 'var' ), disp 'Generating y0'
    switch nConvMethod
        case 1, y0 = Gfft * xtrue;
        case 2, y0 = Gconv * xtrue;
        otherwise, error( 'unrecognized nConvMethod' );
    end
end


% SNR = 10;
% SNRdb = 10*log10(SNR);

if ~exist( 'SNRdb', 'var' ), SNRdb = 4;
else, disp( 'Using supplied SNRdb value' );
end

SNR = 10^(SNRdb/10);

disp( sprintf( 'SNR=%g, SNRdb=%g', SNR, SNRdb ) );

if ~exist( 'yi', 'var' ), disp 'Generating noisy data yi'
    
    % KEEP FOR POSTERITY
    
%     if strcmp( SZ_PSF_TYPE, 'gaussian' )
%         % Tuned for Gaussian blur
%         switch ( SZ_IMAGE_TYPE )
%             case 'uniform'
%                 propNoise = struct( 'sigma', 1e-2 );
%                 %propNoise = struct( 'sigma', 0.1 );
%             case 'exp'
%                 propNoise = struct( 'sigma', .05 * max(y0(:)) );
%                 %propNoise = struct( 'sigma', 1e-2 );
%             case 'uniform2'
%                 propNoise = struct( 'sigma', 0.16 );
%             case 'js_prior'
%                 %propNoise = struct( 'sigma', 1e-2 );
%                 %propNoise = struct( 'sigma', 0.1 );
%                 %propNoise = struct( 'sigma', 2 );
%                 propNoise = struct( 'sigma', .05 * max(y0(:)) );
%             otherwise
%                 error( 'propNoise.sigma = ???' );
%         end
% 
%     else
%         % Tuned for MRFM blur
%         propNoise = struct( 'sigma', 5e-2*max(y0(:)) );
%     end

    % Compute n^{-1} * \sum_i y0_i^2
    numImageVoxels = numel( y0 );
    y0Vec = y0(:);
    Esignal2 = (1/numImageVoxels) * sum( y0Vec.^2 );
    propNoise = struct( 'snr', SNR, 'sigma', sqrt( Esignal2/SNR ) );
    
    if B_GEN_SAME_NOISE
        randn( 'state', 0 ); % Reset state so we get the same "random" noise
    else
        randn( 'state', sum(100*clock) );
    end
    
    yi = y0 + propNoise.sigma * randn(size(y0));

    % If we implement convolution via DFTs, do a circular shift to 
    % correctly display y0 and yi. 
    
    if nConvMethod == 1        
        y0Display = img_correct_cc( SZ_PSF_TYPE, propPsf, y0 );
        yiDisplay = img_correct_cc( SZ_PSF_TYPE, propPsf, yi );
        if N_GEN_PLOTS == 1
            disp( 'Correcting for circular shift...' );
            im( pl+2, y0Display, 'y0' ), colorbar
            im( pl+3, yiDisplay, 'yi' ), colorbar
        end
    else
        if N_GEN_PLOTS == 1
            disp( 'Not correcting for circular shift' );
            im( pl+2, y0, 'y0' ), colorbar
            im( pl+3, yi, 'yi' ), colorbar
        end
    end
end

if N_GEN_PLOTS == 2

    figure( 1 ), im( xtrue ), title( 'xtrue' ), colorbar;
    colormap gray;
    figure( 2 ), im( y0Display ), title( 'y0' ), colorbar;
    figure( 3 ), im( yiDisplay ), title( 'yi' ), colorbar;

end

        
