% binary_tst: test various approach to recovering binary images
%-------------------------------------------------------------------------------
% Note that this should not be used to compare the different methods. However,
% it can be used to test the merit of each method. The reason for this is that
% we haven't ensured that the methods are optimizing the same function.
%
% See also test_bsolv.
%
% binary_tst( n, sigma [, sDisplayMethod] )
%    n              - dimension
%    sigma          - noise std deviation
%    sDisplayMethod - string that describes how to display the results. Valid
%                     options are 'print', 'stem'.
%
function binary_tst( n, sigma, sDisplayMethod )

% Globals
global pi0 pi1

% Parameters
numCases = 2;
maxIter = 500;
pi0 = 0.9;
pi1 = 1 - pi0;

if ( nargin < 2 ), error( 'missing arguments' ), end;

numErrs = zeros( numCases, 1 );

for i = 1:numCases

	% Generate a random A and x0
	preA = randn( n, n );
	A = preA' * preA;
	Ainv = inv( A );

	preX0 = rand( n, 1 );
	x0 = ( preX0 > pi0 );
	x0Noisy = x0 + sigma*Ainv*randn( n, 1 );


	x0Hat1 = test_bnb_min( n, sigma, maxIter, A, x0Noisy );
	%x0Hat2 = test_subopt_binary( n, sigma, pi0, pi1, A, Ainv, x0Noisy );
	x0Hat3 = test_grad_method( n, sigma, maxIter, A, x0Noisy );

	if ( nargin == 3 )
		switch ( lower( sDisplayMethod ) )
		case 'print'
			disp( 'x0:' );
			disp( x0' );
			disp( 'x0Noisy:' );
			disp( x0Noisy' );

		case 'stem'
			figure( 1 ), stem( x0 ), title( 'x0' );
			figure( 2 ), stem( x0Noisy ), title( 'x0Noisy' );

		otherwise
			% Do nothing
		end
	end

	disp( 'x0Hat1:' ), disp( x0Hat1 );
	numErrs1( i ) = sum( abs( x0 - x0Hat1 ) );
	disp( sprintf( 'numErrs1 = %4d/%4d bits', numErrs1(i), n ) );

	disp( 'x0Hat3:' ), disp( x0Hat3 );
	numErrs3( i ) = sum( abs( x0 - x0Hat3 ) );
	disp( sprintf( 'numErrs3 = %4d/%4d bits', numErrs3(i), n ) );
end

% Tests bnb_min
%...............................................................................
function x0Hat = test_bnb_min( n, sigma, maxIter, A, x0Noisy )
%...............................................................................

%[newA, newX0Noisy, T] = bnb_reorder2( A, x0Noisy );
[newA, newX0Noisy, T] = bnb_reorder( A, x0Noisy );
sol = bnb_min( newA, newX0Noisy, maxIter, 0 );
x0Hat = T' * sol;

% Tests the (suboptimal) algorithm where we treat each dimension as a binary
% detection problem.
%...............................................................................
function x0Hat = test_subopt_binary( n, sigma, pi0, pi1, A, Ainv, x0Noisy )
%...............................................................................

x0Hat = zeros( n, 1 );
Rprime = sigma^2*Ainv*Ainv';

for i = 1:n
	z = x0Noisy( i );
	thres = 1/2 + Rprime( i, i ) * log( pi0/pi1 );
	if ( z > thres ), x0Hat(i) = 1; else, x0Hat(i) = 0; end;
end

% Tests the discrete gradient ("steepest descent") method.
%...............................................................................
function x0Hat = test_grad_method( n, sigma, maxIter, A, x0Noisy )
%...............................................................................

matSigma = A'*A;

% Formulate the initial guess
xInitial = zeros( n, 1 );

for i = 1:n
	z = x0Noisy( i );
	if ( z > 0.5 )
		xInitial( i ) = 1;
	else
		xInitial( i ) = 0;
	end
end

xOld = xInitial; % Initialize
xNew = xInitial;

for i = 1:maxIter
	possibleDeltaX = 1 - 2*xOld;
	delta = zeros( n, 1 );
	gradObjFunc = 2*(xOld - x0Noisy)'*matSigma;

	for j = 1:n
		if ( possibleDeltaX(j) > 0 ) 	% equals to +1
			if ( gradObjFunc(j) < 0 )
				delta( j ) = 1;
			end
		else							% equals to -1
			if ( gradObjFunc(j) > 0 )
				delta( j ) = -1;
			end
		end
	end

	if ( max( abs(delta) ) < 0.5 )		% delta is equal to all zeros
		disp( 'test_grad_method: delta=0' );
		break;
	end

	xNew = xOld + delta;
	xOld = xNew;
end

x0Hat = xNew;
