% bmag_lev_surf: plots the level surface of Bmag(x,y,z) for z > 0
%-------------------------------------------------------------------------------
% Plots the level surface of Bmag(x,y,z) for z > 0.
%
% [theta, phi, r] = bmag_lev_surf( Bext, c, m, [, bRetSpherCoords] 
%                      [, maxPhiValue] [, figNum] )
%
%    Bext, m - parameters of the experiment
%    c       - value of Bmag^2 for which we want to plot the level surface. 
%              Note that Bmag^2 >= Bext^2
%    bRetSpherCoords - boolean flag on whether to return the spherical coords
%                      computed. We shall implicitly assume that the user
%                      wants the plot suppressed
%    maxPhiValue     - max phi value to use in radians
%
function [theta, phi, r] = bmag_lev_surf( Bext, c, m, bRetSpherCoords, ...
	ArgMaxPhiValue, ArgFigNum )

% Constants
numThetaPts = 90; % These define our (theta,phi) mesh
numPhiPts = 45;

maxThetaValue = 2*pi; % The default (theta,phi) space to use
maxPhiValue = pi/2;
figNum = 1;

% Initialize
theta = [];
phi = [];
r = [];

if ( nargin >= 5 )
	if ( ArgMaxPhiValue <= 0 ), error( 'ArgMaxPhiValue <= 0' ), end;
	if ( ArgMaxPhiValue > pi/2 ), error(  'ArgMaxPhiValue > pi/2' ), end;
	maxPhiValue = ArgMaxPhiValue;
end

if ( nargin == 6 ), figNum = ArgFigNum; end;

vecTheta = linspace( 0, maxThetaValue, numThetaPts );
vecPhi = linspace( 0, maxPhiValue, numPhiPts );
[meshTheta, meshPhi] = meshgrid( vecTheta, vecPhi );
meshR = zeros( length(vecPhi), length(vecTheta) );

disp( sprintf( '(theta,phi) space: [0, %.3e] x [0, %.3e]', ...
	maxThetaValue, maxPhiValue ) );

a = (Bext^2 - c);
b = 2*m*Bext*(2 - 3*(sin(meshPhi)).^2);
c = 9*m^2* (1 - (sin(meshTheta)).^4) .* (sin(meshPhi)).^4 + ...
	3*m^2* (3*(sin(meshTheta)).^2 - 4) .* (sin(meshPhi)).^2 + ...
	4*m^2;

meshThetaSize = size( meshTheta );
meshPhiSize = size( meshPhi );
meshRSize = size( meshR );

if 0 % DEBUG: Check that the array sizes are compatible
disp( sprintf( 'meshTheta has size %5d x %5d', meshThetaSize(1), ...
	meshThetaSize(2) ) );
disp( sprintf( 'meshPhi   has size %5d x %5d', meshPhiSize(1), ...
	meshPhiSize(2) ) );
disp( sprintf( 'meshR     has size %5d x %5d', meshRSize(1), meshRSize(2) ) );
end

for i = 1:meshRSize( 1 )
	for j = 1:meshRSize( 2 )
		quadSol1 = ( -b(i,j) + sqrt( b(i,j)^2 - 4*a*c(i,j) ) )/( 2*a );
		quadSol2 = ( -b(i,j) - sqrt( b(i,j)^2 - 4*a*c(i,j) ) )/( 2*a );
		rootPos = quad_pos_sol( quadSol1, quadSol2 );
		r = rootPos^(1/3);
		if ( r < 0 ), error( 'negative r' ), end;
		meshR( i, j ) = r;
	end
end

if ( ( nargin >= 4 ) & ( bRetSpherCoords == 1 ) )
	theta = meshTheta;
	phi = meshPhi;
	r = meshR;
	return; % Skip plot
end

% Convert to Cartesian coords for plotting
meshX = meshR .* sin( meshPhi ) .* cos( meshTheta );
meshY = meshR .* sin( meshPhi ) .* sin( meshTheta );
meshZ = meshR .* cos( meshPhi );

% Create surface plot
create_surf_plot( figNum, 'B_{mag} level surface for z >= 0', [], ...
	meshX, meshY, meshZ, 1, 'z' );

% Compute the largest value of z
meshZMax = max( max( meshZ ) );
disp( sprintf( 'max z = %.6e', meshZMax ) );
