% bstr_conj_grad: binary string version of the conjugate gradient method
%-------------------------------------------------------------------------------
%
% {{{ History }}}
% 13-sep-04: changed function handle calls so that feval is used
%
% x = bstr_conj_grad( maxIter, x0, nVerboseLevel, fnPhi, fnPhiRowGrad )
%    maxIter            - maximum number of iterations
%    x0                 - initial point where the algorithm starts
%    nVerboseLevel      - verbose level: 0 for silent, and higher values for
%                         more output
%    fn{Phi,PhiRowGrad} - function handles to the cost function and the row
%                         gradient of the cost function
%
% Output
%    x - output at the termination of the algorithm
%
function x = bstr_conj_grad( maxIter, x0, nVerboseLevel, fnPhi, fnPhiRowGrad )

% Constants
nSlackIter = 0;
FTOL = 1e-8;
MAX_ITER = maxIter + nSlackIter;

xInitial = x0;
n = length( x0 );

if ( nVerboseLevel >= 2 )
	disp( sprintf( 'xInitial: weight=%.3e', feval(fnPhi,xInitial) ) ); 
	disp( xInitial' );
end

% The g, h vectors in the conjugate gradient method. The 'Curr' suffix denotes
% '_i' and the 'Prev' suffix '_{i-1}'.
vecGCurr = zeros( 1, n );
vecGPrev = zeros( 1, n ); 
vecHCurr = zeros( 1, n );
vecHPrev = zeros( 1, n );

xOld = xInitial; 			% Initialize
xNew = xInitial;
%vecGCurr = rand( 1, n ); 	% g0 <- random vector
% DEBUG
%isa( fnPhiRowGrad, 'function_handle' )
vecGCurr = feval( fnPhiRowGrad, xInitial );
vecHCurr = vecGCurr;		% h0 <- g0

for iterInd = 1:MAX_ITER
	possibleDeltaX = 1 - 2*xOld;
	delta = steer_dprod( vecHCurr, possibleDeltaX, 'positive' );
	xNew = xOld + delta';

	phiOld = feval( fnPhi, xOld );
	phiNew = feval( fnPhi, xNew );
	if ( nVerboseLevel >= 1 )
		disp( sprintf( 'phiOld = %.5e, phiNew = %.5e', phiOld, phiNew ) );
	end

	% Stopping condition
	if ( iterInd > nSlackIter )
		if ( 2*abs( phiOld - phiNew ) < FTOL*( abs(phiOld) + abs(phiNew) ) )
			break;
		end
	end

	% Update the g and h vectors
	vecGPrev = vecGCurr;
	vecGCurr = -1*feval( fnPhiRowGrad, xNew );
	%gamma = dot( vecGCurr - vecGPrev, vecGCurr ) / dot( vecGPrev, vecGPrev );
	gamma = dot( vecGCurr, vecGCurr ) / dot( vecGPrev, vecGPrev );
	vecHPrev = vecHCurr;
	vecHCurr = vecGCurr + gamma*vecHPrev;

	xOld = xNew; % replace old with new

	if ( nVerboseLevel >= 2 )
		disp( sprintf( 'xNew: weight=%.3e', feval(fnPhi,xNew) ) ); 
		disp( xNew' );
	end
end

if ( nVerboseLevel >= 1 )
	disp( sprintf( 'interInd=%6d', iterInd ) );
end

x = xOld;

