% bstr_steepest_desc: binary string version of steepest descent
%-------------------------------------------------------------------------------
%
% {{{ History }}}
% 13-sep-04: changed the function handle calls so that they use feval. For
% some reason, Matlab 6.5 needs to have this done.
%
% Implements the binary string version of steepest descent. The initial x0 from
% which to start is supplied. There are two variants we have experimented with.
% They are:
%    a. 'greedy'
%       Try to go in the direction of -grad. This is achieved by computing
%       the gradient of the cont. fn. at the current point and trying to 
%       compute a vector v that: (i) when added to the current point will 
%       result in another valid point; (ii) the dot product <v,grad> is 
%       minimal (i.e., we want <v,grad> to be as negative as possible).
% 
%    b. 'notgreedy'
%       Instead of being "greedy" as in (a), restrict v so that it only has
%       one non-zero component. The thinking is that maybe we'll do better
%       by not being as greedy.
% 
% x = bstr_steepest_desc( maxIter, x0, nVerboseLevel, sType, fnPhi,
%    fnPhiRowGrad )
%    maxIter            - maximum number of iterations
%    x0                 - initial point where the algorithm starts
%    nVerboseLevel      - verbose level: 0 for silent, and higher values for
%                         more output
%    fn{Phi,PhiRowGrad} - function handles to the cost function and the row
%                         gradient of the cost function
%
% Output
%    x - output at the termination of the algorithm
%
function x = bstr_steepest_desc( maxIter, x0, nVerboseLevel, sType, ...
	fnPhi, fnPhiRowGrad )
%...............................................................................

% Constants
FTOL = 1e-7;
LARGE_NUMBER = 1e10;
nSlackIter = max( [4, .05*maxIter] );
MAX_ITER = maxIter + nSlackIter;

xInitial = x0;
[n, m] = size( x0 );
if ( ( n < 2 ) | ( m ~= 1 ) ), error( 'invalid x0' ), end;

if ( nVerboseLevel >= 2 )
	disp( sprintf( 'xInitial: weight=%.3e', feval(fnPhi,xInitial) ) ); 
	disp( xInitial' );
end

xOld = xInitial; % Initialize
xNew = xInitial;
xBestPreSlack = xInitial;

switch ( lower( sType ) )
case 'greedy'
	nType = 0; % method (a)
case 'notgreedy'
	nType = 1; % method (b)
otherwise
	error( 'invalid sType in bstr_steepest_desc' );
end

% DEBUG
%functions( fnPhi )
%functions( fnPhiRowGrad )

for iterInd = 1:MAX_ITER

	possibleDeltaX = 1 - 2*xOld;
	delta = zeros( n, 1 );

	% gradient row vector at current point
	rowGrad = -1*feval( fnPhiRowGrad, xOld );  

	if ( nVerboseLevel >= 1 )
		disp( sprintf( 'iterInd: %4d', iterInd ) );
		if ( nVerboseLevel >= 2 ) 
			disp( 'rowGrad:' ), disp( rowGrad );
		end
	end

	delta = steer_dprod( rowGrad, possibleDeltaX, 'positive' );

	% cannot move to another valid point: stop
	if ( max( abs(delta) ) < 0.5 )
		if ( nVerboseLevel >= 1 ), disp( 'bstr_steepest_desc: delta=0' ), end;
		break; 
	end

	if ( nType == 0 )	
		wBest = feval( fnPhi, xOld + delta' );
		xNew = xOld + delta';
	else 
		%%% nType == 1

		ind = find( abs(delta) > 0.5 );
		indLen = length( ind );

		% Initialize: make the 1st index the best
		delta1Cand = zeros( n, 1 );
		delta1Cand( ind(1) ) = delta( ind(1) );
		indBest = 1;
		wBest = feval( fnPhi, xOld + delta1Cand );

		for j = 2:indLen
			deltaCand = zeros( n, 1 );
			deltaCand( ind(j) ) = delta( ind(j) );
			xNewCand = xOld + deltaCand;
			wCand = feval( fnPhi, xNewCand );
			if ( wCand < wBest )
				indBest = j;
				wBest = wCand;
			end
		end

		deltaCand = zeros( n, 1 );
		deltaCand( ind(indBest) ) = delta( ind(indBest) );
		xNew = xOld + deltaCand;
	end

	phiOld = feval( fnPhi, xOld );
	phiNew = feval( fnPhi, xNew );
	if ( nVerboseLevel >= 1 )
		disp( sprintf( 'phiOld = %.5e, phiNew = %.5e', phiOld, phiNew ) );
	end

	% Stopping condition
	if ( iterInd > nSlackIter )
		%if ( 2*abs( phiOld - phiNew ) < FTOL*( abs(phiOld) + abs(phiNew) ) )
		if ( phiNew >= phiOld )
			break;
		end
	else
		% iterInd <= nSlackIter
		if ( phiNew < feval( fnPhi, xBestPreSlack ) )
			xBestPreSlack = xNew;
		end
	end

	if ( iterInd == nSlackIter )
		% Replace xOld with the best value seen so far
		xOld = xBestPreSlack;
	else
		xOld = xNew; % replace old with new
	end

	if ( nVerboseLevel >= 2 )
		disp( sprintf( 'xNew: weight=%.3e', feval(fnPhi,xNew) ) ); 
		disp( xNew' );
	end
end

if ( nVerboseLevel >= 1 )
	disp( sprintf( 'interInd=%6d', iterInd ) );
end

x = xOld;


