% create_int_plot: create an intensity plot using pcolor()
%-------------------------------------------------------------------------------
% 
% create_int_plot( nFig, meshX, meshY, meshP, vecRescale, szTitle,
%    [, bSetAxisImage] [, bSetColourmapGray] [, sRescaleType] )
%    nFig          - figure number
%    szTitle       - title
%    vecRescale    - if nonzero, is a 2-element vector that contains the
%                    coefficients by which to scale the range
%    mesh{X,Y,P}   - the x, y, and output values
%    bSetAxisImage - boolean flag to run "axis image" on the intensity plot.
%                    By default, this is 0 (false).
%    bSetColourmapGray - boolean flag to set the colour map to gray instead
%                        of jet (default)
%    sRescaleType  - string description of format of vecRescale. There are
%                    two options: 'ratio', or 'abs'. By default, it is 
%                    equal to 'ratio'. Case-insensitive.
%
function create_int_plot( nFig, meshX, meshY, meshP, vecRescale, szTitle, ...
	bSetAxisImage, bSetColourmapGray, sRescaleType )

% Check parameters
if ( nFig < 1 ), error( 'invalid nFig value' ), end;

Pmin = min( min( meshP ) );
Pmax = max( max( meshP ) );

disp( sprintf( 'Figure %3d: Pmin = %15.3e, Pmax = %15.3e', nFig, Pmin, ...
	Pmax ) );

% Default value
sRescaleConvention = 'ratio';

if ( (nargin == 9) )
	sRescaleConvention = sRescaleType;
end

switch ( lower( sRescaleConvention ) )
case 'ratio'
	if ( min( size( vecRescale ) ) == 1 & max( size( vecRescale ) ) == 2 )
		if ( vecRescale(1) < 0 | vecRescale(1) > 1 | ...
		 	 vecRescale(2) < 0 | vecRescale(2) > 1 | ...
		 	 vecRescale(1) + vecRescale(2) > 1 ) 
			error( 'invalid vecRescale values' );
		end
		colourVals = [ Pmin + vecRescale(1)*(Pmax - Pmin), ...
			Pmax - vecRescale(2)*(Pmax - Pmin) ];
	else
		colourVals = [ Pmin, Pmax ];
	end

case 'abs'
	colourVals = vecRescale;

otherwise
	error( 'invalid rescale type' );
end

disp( sprintf( 'colourVals: [%g,%g]', colourVals(1), colourVals(2) ) );

meshXSize = size( meshX );
meshYSize = size( meshY );
meshPSize = size( meshP );

if ( (meshXSize(1) < 1) | (meshXSize(2) < 1) )
	meshX = [1:meshPSize(2)];
end
if ( (meshYSize(1) < 1) | (meshYSize(2) < 1) )
	meshY = [1:meshPSize(1)];
end

figure( nFig ), hold off, clf, hold on;

% Create the main plot

subplot( 121 ), pcolor( meshX, meshY, meshP ), hold on;
colormap( jet  ), shading flat, caxis( colourVals );
if ( (nargin >= 7) & (bSetAxisImage == 1) ), axis image, end;
if ( (nargin >= 8) & (bSetColourmapGray == 1) ), colormap( gray ), end;
xlabel( 'x [nm]' ), ylabel( 'y [nm]' ), title( szTitle );
hP = gca;

% Create the colour bar. Bug fixed: it should depend on colourVals
Pmin = min( colourVals(:) );
Pmax = max( colourVals(:) );
vecP = linspace( Pmin, Pmax, 100 );
[meshPintX, meshPintY] = meshgrid( [1:3], vecP );
meshPint = repmat( vecP', 1, 3 );

% DEBUG
if 0 
	meshPintXSize = size( meshPintX );
	meshPintYSize = size( meshPintY );
	meshPintSize = size( meshPint );
	disp( sprintf( 'meshPintX size: %5d x %5d', meshPintXSize(1), ...
		meshPintXSize(2) ) );
	disp( sprintf( 'meshPintY size: %5d x %5d', meshPintYSize(1), ...
		meshPintYSize(2) ) );
	disp( sprintf( ' meshPint size: %5d x %5d', meshPintSize(1), ...
		meshPintSize(2) ) );
end

if ( Pmax > Pmin )
	subplot( 122 ), pcolor( meshPintX, meshPintY, meshPint ), hold on;
	colormap( jet ), shading flat, caxis( colourVals );
	if ( (nargin >= 8) & (bSetColourmapGray == 1) ), colormap( gray ), end;
else
	subplot( 122 ), title( 'INVALID' );
end

hPint = gca;
set( hPint, 'XTick', [] );
set( hPint, 'YAxisLocation', 'right' );

% Adjust axes
PWidth = 0.6;
PintWidth = 0.05;
spaceWidth = 0.03;

PPos = get( hP, 'Position' );
PintPos = get( hPint, 'Position' );
set( hP, 'Position', [PPos(1), PPos(2), PWidth, PPos(4)] );
set( hPint, 'Position', ...
	[PPos(1) + PWidth + spaceWidth, PintPos(2), PintWidth, PintPos(4)] );
