% create_surf_plot: create a surface plot using surfc()
%-------------------------------------------------------------------------------
% 
% create_surf_plot( nFig, sTitle, vecRescale, meshX, meshY, meshP
%    [, bSetAxisImage] [, sPTitle] )
%
%    nFig          - figure number
%    sTitle        - title
%    vecRescale    - if nonzero, is a 2-element vector that contains the
%                    coefficients by which to scale the range
%    mesh{X,Y,P}   - the x, y, and output values
%    bSetAxisImage - boolean flag as to whether to run 'axis image' on the
%                    plot. It might not be advisable to do so for some
%                    plots.
%    sPTitle       - title for the P-axis
%
function create_surf_plot( nFig, sTitle, vecRescale, meshX, meshY, meshP, ...
	bSetAxisImage, sPTitle )

% Check parameters
if ( nFig < 1 ), error( 'invalid nFig value' ), end;

Pmin = min( min( meshP ) );
Pmax = max( max( meshP ) );

disp( sprintf( 'Figure %3d: Pmin = %15.3e, Pmax = %15.3e', nFig, Pmin, ...
	Pmax ) );

if ( min( size( vecRescale ) ) == 1 & max( size( vecRescale ) ) == 2 )
	if ( vecRescale(1) < 0 | vecRescale(1) > 1 | ...
		 vecRescale(2) < 0 | vecRescale(2) > 1 | ...
		 vecRescale(1) + vecRescale(2) > 1 ) 
		error( 'invalid vecRescale values' );
	end
	colourVals = [ Pmin + vecRescale(1)*(Pmax - Pmin), ...
		Pmax - vecRescale(2)*(Pmax - Pmin) ];
else
	colourVals = [ Pmin, Pmax ];
end

figure( nFig ), hold off, clf;

subplot( 121 ), surf( meshX, meshY, meshP );
colormap( jet ), shading interp, caxis( colourVals );
xlabel( 'x [nm]' ), ylabel( 'y [nm]' ), title( sTitle );
if ( nargin >= 7 & bSetAxisImage == 1 ), axis image, end;
if ( nargin == 8 & length( sPTitle ) > 0 ), zlabel( sPTitle ), end;
hP = gca;

Pmin = min( min( meshP ) );
Pmax = max( max( meshP ) );
vecP = linspace( Pmin, Pmax, 100 );
[meshPintX, meshPintY] = meshgrid( [1:3], vecP );
meshPint = repmat( vecP', 1, 3 );
if 0 % Debug
	meshPintXSize = size( meshPintX );
	meshPintYSize = size( meshPintY );
	meshPintSize = size( meshPint );
	disp( sprintf( 'meshPintX size: %5d x %5d', meshPintXSize(1), ...
		meshPintXSize(2) ) );
	disp( sprintf( 'meshPintY size: %5d x %5d', meshPintYSize(1), ...
		meshPintYSize(2) ) );
	disp( sprintf( ' meshPint size: %5d x %5d', meshPintSize(1), ...
		meshPintSize(2) ) );
end
subplot( 122 ), pcolor( meshPintX, meshPintY, meshPint );
colormap( jet ), shading interp, caxis( colourVals );
set( gca, 'XTick', [] );
set( gca, 'YAxisLocation', 'right' );
hPint = gca;

% Adjust axes
PWidth = 0.6;
PintWidth = 0.05;
spaceWidth = 0.05;

PPos = get( hP, 'Position' );
PintPos = get( hPint, 'Position' );
set( hP, 'Position', [PPos(1), PPos(2), PWidth, PPos(4)] );
set( hPint, 'Position', ...
	[PPos(1) + PWidth + spaceWidth, PintPos(2), PintWidth, PintPos(4)] );
