% gen_img: generates an image using the vertical cantilever PSF
%-------------------------------------------------------------------------------
% out = gen_img( par, mesh, sImgType, sigmaFac, spinPos, numRandLoc 
%    [, bPlot] )
%    par        - parameter structure
%    mesh     - mesh structure
%    sImgType   - valid options are:
%                    'fixed'  -> positions provided in spinPos
%                    'pdb'    -> positions returned from pdb_cl_img
%                    'random' -> generate the positions randomly
%    sigmaFac   - factor for determining sigma (variance of WGN)
%    spinPos    - 3-d matrix with each row containing the Cartesian coords of
%                 an electron. Note that the coords must be discrete.
%    numRandLoc - number of random spins
%    bPlot      - boolean flag that, if set to 1, will create plots of f,g,h
%                 if numZPts = 1
%
% Outputs
%    out.{X, Y}       - coords of f, g
%    out.{f, h, g}    - f: input, h: operator, g: output
%    out.{gNoisy}     - g + AWGN
%    out.sigma        - value of sigma
% 
function out = gen_img( par, mesh, sImgType, sigmaFac, spinPos, ...
	numRandLoc, bPlot )

% Constants
xRangeFrac = 1/3;
yRangeFrac = 1/3;
zRangeFrac = 1/3;

CONV_METHOD = 'dft';

psf_mesh_chk( mesh );

numXPts = mesh.numPts(1);
numYPts = mesh.numPts(2);
numZPts = mesh.numPts(3);

% Generate the original image
switch ( lower(sImgType) )
case 'fixed'
	f = create_orig_img( mesh, spinPos );

case 'pdb'
	spinPosSize = size( spinPos );
	if ( (spinPosSize(1) < 1) | (spinPosSize(2) ~= 3) )
		error( 'invalid dimensions for spinPos' );
	end

	% Proceed as in fixed, but we have to do two things here:
	%    1. Translate the positions 
	%    2. Convert units to discrete spacing

	c = mesh.vecLowerLeftCoords(:);
	spinPos = spinPos - repmat( [c(1:2)', min(spinPos(:,3))], ...
			size( spinPos, 1 ), 1 );
	thres = 100*eps;

	if ( ~isempty( find( spinPos(:,1) < -thres ) ) )
		error( 'mesh lower left corner x-coord is too large' );
	else
		spinPos( :, 1 ) = 1 + floor( abs(spinPos( :, 1 ) ./ mesh.deltaX) );
	end

	if ( ~isempty( find( spinPos(:,2) < -thres ) ) )
		error( 'mesh lower left corner y-coord is too large' );
	else
		spinPos( :, 2 ) = 1 + floor( abs(spinPos( :, 2 ) ./ mesh.deltaY) );
	end

	if ( ~isempty( find( spinPos(:,3) < -thres ) ) )
		error( 'mesh lower left corner z-coord is too large' );
	else
		spinPos( :, 3 ) = 1 + floor( abs(spinPos( :, 3 ) ./ mesh.deltaZ) );
	end

	f = create_orig_img( mesh, spinPos );

case 'random'
	if ( numRandLoc < 1 ), error( 'invalid arguments' ), end;

	f = zeros( numXPts, numYPts, numZPts );

	xRange = numXPts * xRangeFrac;
	yRange = numYPts * yRangeFrac;
	zRange = numZPts * zRangeFrac;

	for locInd = 1:numRandLoc
		locX = max( [1, ceil( rand * xRange )] );
		locY = max( [1, ceil( rand * yRange )] );
		locZ = max( [1, ceil( rand * zRange )] );
		f( locX, locY, locZ ) = 1;
		disp( sprintf( 'loc %3d: %5d %5d %5d', locInd, locX, locY, locZ ) );
	end

otherwise
	error( 'unrecognized sImgType' );
end

% Generate the output 

%%%
% The straightforward way is to just use the definition of g, e.g. g = f*h.
% However, this will involve nested for loops six levels deep, and will
% probably take a long time.
% 
% A better idea is to use DFTs. This assume that the convolution operator can 
% be well approximated by the circular convolution operator though. However,
% if we zero-pad h and f, circular convolution will yield linear convolution.
% Convolution has the effect of smearing the original image, however, and so
% g will be larger than f. Because we don't want to consider images of
% different sizes, make sure that the mesh is large enough to accomodate the
% expansion. This can be summarized into the rule that the mesh must be at
% least twice as large as h and f.
%%%

close all;

if 0 % Try to be "smart" in obtaining the h operator

	% Save the h operator to disk so that it might be re-used. If the disk 
	% version is different, then only will we call get_h_op(). We overwrite 
	% the disk version with the newer operator. This only saves time if we use 
	% the same parameters and mesh, but change the molecule under study.

	shDiskFile = 'h_op.mat';
	hDisk = read_h_op( par, mesh, shDiskFile );

	if ( min( size( hDisk ) ) == 0 )
		disp( 'h operator file did not check out...calling get_h_op' );
 
		tic; % Start timing
		h = get_h_op( par, mesh  );
		tElapsed = toc; % Stop timing
		disp( sprintf( 'gen_h_op took %g secs', tElapsed ) );

		disp( 'saving h operator to disk' );
		save_h_op( par, mesh, h, shDiskFile );
	else
		h = hDisk;
		clear hDisk; % don't need the disk version anymore
		disp( 'using saved h operator' );
	end
else
	% No need to get fancy, just call get_h_op. With psf_vert in get_h_op
	% converted to mex_psf_vert, this doesn't take as long as it used to.
	h = get_h_op( par, mesh );
    % DEBUG
    disp( 'size h:' ), disp( size( h ) );
end

mask = logical( ones( numXPts, numYPts, numZPts ) );
Gfft = Gblur( mask, 'psf', h, 'type', 'conv,fft' );
Gconv = Gblur( mask, 'psf', h, 'type', 'conv,same' );

switch CONV_METHOD
    case 'dft'
        
        % Old method of implementing convolution using DFTs
        %
%         fftSize = [numXPts, numYPts, numZPts*2];
%         F = fftn( f, fftSize );
%         H = fftn( h, fftSize );
%         G = H .* F;
%         gLarge = real( ifftn( G ) );
%         g = gLarge( 1:numXPts, 1:numYPts, 1:numZPts );
        
        g = Gfft * f;
        
    case 'convn'
        
        g = Gconv * f;
        
    otherwise
        
        error( 'Unrecognized CONV_METHOD' );
                
end

% Do this b/c sometimes we get negative values in the output, which we
% isn't permissible as F is positive valued
g = g .* ( g >= 0 );

% Add AWGN noise to g in an ad-hoc fashion. To determine an appropriate sigma,
% use the variance of g.
gSdev = std( g(:) );
sdev = sigmaFac * gSdev;
disp( sprintf( 'gSdev = %.4e, sdev = %.4e', gSdev, sdev ) );
gNoisy = g + sdev * randn( numXPts, numYPts, numZPts );

% DEBUG
%return;

vecX = [];
for i = 1:numXPts
	[xC, yC, zC] = ind_cart_d2c( mesh, i, 1, 1 );
	vecX(i) = xC;
end

vecY = [];
for j = 1:numYPts
	[xC, yC, zC] = ind_cart_d2c( mesh, 1, j, 1 );
	vecY(j) = yC;
end

vecZ = [];
for k = 1:numZPts
	[xC, yC, zC] = ind_cart_d2c( mesh, 1, 1, k );
	vecZ(k) = zC;
end

X = repmat( vecX(:), [1, numYPts, numZPts] );
Y = repmat( vecY(:)', [numXPts, 1, numZPts] );
Z = repmat( vecZ(:)', [numYPts, 1, numXPts] );
Z = shiftdim( Z, 2 );

% Create plots
if ( (nargin == 7) & (bPlot == 1) & (numZPts == 1) )
	create_int_plot( 1, 'f (input)', [], X, Y, f, 1, 1 );
	create_int_plot( 2, 'h (operator)', [], X, Y, h, 1 );
	create_int_plot( 3, 'g (clean output)', [], X, Y, g, 1 );
	create_int_plot( 3, 'gNoisy (noisy output)', [], X, Y, gNoisy, 1 );
end

% Create the output structure
out = struct( 'X', X, 'Y', Y, 'Z', Z, 'sigma1', sdev, 'spinPos', spinPos, ...
	'f', f, 'h', h, 'g', g, 'gNoisy', gNoisy );

%...............................................................................
function f = create_orig_img( mesh, spinPos )
%...............................................................................

% the original image
f = zeros( numXPts, numYPts, numZPts );

vecMeshSize = [ numXPts, numYPts, numZPts ];
spinPosSize = size( spinPos );

if ( (spinPosSize(1) < 1) | (spinPosSize(2) ~= 3) )
	error( 'invalid dimensions for spinPos' );
end


for locInd = 1:spinPosSize(1)
	% Check the spin location coords
	vecLoc = [ spinPos(locInd,1), spinPos(locInd,2), spinPos(locInd,3) ];
	for i = 1:3
		if ( ( spinPos(locInd,i) < 1 ) | ...
			 ( spinPos(locInd,i) > vecMeshSize(i)/2 ) )

			% If in the x- or y-dir, not meeting the conditions is an error
			if ( (i == 1) | (i == 2) )
				error( sprintf( 'illegal location for spin %3d', locInd ) );

			elseif ( i == 3 )
				% Otherwise, if in the z-dir, there's an error if the location
				% exceeds the size. However, if it exceeds half the size and
				% the mesh size in the z-dir is > 1, just generate a warning
				if ( spinPos(locInd,3) > vecMeshSize(3) )
					error( sprintf( 'spin %3d: z-pos %g is too large', ...
						locInd, spinPos( locInd, 3 ) ) );
				elseif ( vecMeshSize(3) > 1 )
					warning( sprintf( 'spin %3d: z-pos %g is > 1/2 mesh', ...
						locInd, spinPos( locInd, 3 ) ) );
				end

			end

		end
	end

	% If it's ok, add it to f
	f( vecLoc(1), vecLoc(2), vecLoc(3) ) = 1;
	disp( sprintf( 'loc %3d: %5d %5d %5d', locInd, vecLoc(1), ...
		vecLoc(2), vecLoc(3) ) );
end

