% ind_cart_d2c: translate cartesian coords from discrete to continuous
%-------------------------------------------------------------------------------
% Assume that we enumerate the points in the order of x, y, z.
%
% [xC, yC, zC] = ind_cart_d2c( sMesh, xD, yD, zD )
%    sMesh        - mesh structure 
%    xD, yD, zD    - discrete coords
%
% Outputs
%    xC, yC, zC    - continuous coords
%
function [xC, yC, zC] = ind_cart_d2c( sMesh, xD, yD, zD )

vD = [xD, yD, zD]; 	% discrete (input)
vC = [0, 0, 0];		% continuous (output)

vecNumPts = sMesh.numPts;
vecDelta = sMesh.delta;

switch ( lower( sMesh.szPosition ) )

case 'origin'

%%%
% If the number of points is even, centre the mesh so that it straddles the
% origin's x-value. Otherwise, if there is only one point, centre the mesh 
% on the origin's x-value.
%
% Do the same for the y and z coords. Note that the image should have
%        1 <= xD <= numXPts, 1 <= yD <= numYPts, 1 <= zD <= numZPts,
% but we might fall outside this range when computing the operator matrix A.
%%%

	for vInd = 1:3
		if ( mod( vecNumPts(vInd), 2 ) == 0 )
			vC(vInd) = sMesh.originCoord( vInd ) + ...
				( 2*vD(vInd) - 1 - vecNumPts(vInd) )*vecDelta(vInd)/2;
		elseif ( vecNumPts(vInd) == 1 )
			vC(vInd) = sMesh.originCoord( vInd ) + ...
				vD(vInd)*vecDelta(vInd);
		else
			error( sprintf( 'invalid value for numPts(%d)', vInd ) );
		end
	end

case 'lowerleft'

% Things are simpler here
	for vInd = 1:3
		vC(vInd) = sMesh.lowerLeftCoord(vInd) + vecDelta(vInd) * vD(vInd);
	end

otherwise
	error( 'unknown mesh position value' );
end

xC = vC( 1 );
yC = vC( 2 );
zC = vC( 3 );
