% obj_hinv_energy: objective function for computing energy of HInv
%-------------------------------------------------------------------------------
% If one assume AWGN for noise, then to achieve a good measure of performance
% for the inverse problem, one criterion is to minimize the energy of the
% inverse filter. That is, the objective function to minimize is given by
% Psi = \sum_k 1/|H[k]|^2, where H[k] is the DFT of the PSF.
%
% We see to minimize Psi in terms of the variables Bext, Bres, R0, xPk. If
% there is a zero-valued H[k] term, then Psi = inf.
%
% energy = obj_hinv_energy( x )
%
function energy = obj_hinv_energy( x )

persistent h H HNorm HPhase HNormFix HInv matHInvEnergy

% Constants
NORM_THRES = 100*eps;

% Elements of x
ParBext = x(1);
ParBres = x(2);
ParR0 = x(3);

% Defaults: refer to psf_param3
ParM = 1700;
Parm = 4*pi/3*ParM*ParR0^3;
Pard = 3;								% This should be 10 nm
Parz0 = Pard + ParR0;

optXFac = -( sqrt(35) - sqrt(19) )/4; 	% optimal x-pos factor for max G
ParxOpt = optXFac*Parz0;
ParxSpan = abs( 1.5*ParxOpt );

optGFac = 2.7428; 						% max factor for G in terms of m,z0
ParGMax = optGFac * Parm / Parz0^4;

deltaB0 = 20;							% This should really be 50 Gauss
ParxPk = deltaB0/ParGMax;

par = struct( ... 
	'Bext', ParBext, ... % Simulation parameters
	'Bres', ParBres, ...
	'R0', ParR0, ...
	'M', ParM, ...	
	'm', Parm, ...	
	'xPk', ParxPk, ...	
	'xOpt', ParxOpt, ...
	'z0', Parz0, ...
	'd', Pard, ...
	'GMax', ParGMax, ...
	'xSpan', ParxSpan, ...	% Specify the grid for the x-y plane
	'ySpan', ParxSpan );

stMesh = psf_mesh( par );

h = get_h_op( par, stMesh );
H = fftn( h );
HSize = size( H );

warning( 'off', 'MATLAB:divideByZero' );
HInv = ones( HSize ) ./ H;
warning( 'on', 'MATLAB:divideByZero' );

matHInvEnergy = abs( HInv ).^2;
HInvEnergy = sum( sum( sum( matHInvEnergy ) ) );

if ( HInvEnergy < 0 ), error( 'negative energy??' ), end;
energy = HInvEnergy;
