% pdb_cl_img: create the cantilever intensity plot of a PDB molecule
%-------------------------------------------------------------------------------
% Takes ATOM and HETATM data and plots atoms of a specific type. The type of
% interest is specified by a regular expression (RE). 
%
% {{{ History }}}
%
% 30-jul-04: modified to work with psf_param3. Also, changed lengths to nm.
%
% spinPos = pdb_cl_img( A, AType, H, HType, sRePlotType [, bSuppressImg]
%    A, AType, H, HType - output of pdb_read_file
%    sRePlotType        - a RE specifying atoms of interest to plot
%	 bSuppressImg       - boolean flag to suppress the generation of the
%                         cantilever image. By default, it is set to 0
%                         (false)
%    
% Output
%    spinPos            - matrix containing the spin positions
%
function spinPos = pdb_cl_img( A, AType, H, HType, sRePlotType, bSuppressImg )

% Load parameter values from another file
%par = psf_param3( 0 );
par = psf_param3( 1, 3e4, 3 );

xMeshSize = par.xSpan * par.xInc;
yMeshSize = par.ySpan * par.yInc;
disp( sprintf( 'Mesh size: x-dir = %.3e, y-dir = %.3e', xMeshSize, ...
	yMeshSize ) );

spinPos = pdb_match( A, AType, H, HType, sRePlotType );

spinPosSize = size( spinPos );
numSpins = spinPosSize( 1 );
vecOnes = ones( numSpins, 1 );

if 0 % Rotate the spin locations
	disp( '   >>> Rotating spin locations <<<' );
	beep;
	spinPos = rotate_3d( spinPos, 'x', pi/2 );
end

% Centre the position of the spins around (0,0,z0).
spinPosExp = mean( spinPos );
spinPos( :, 1 ) = spinPos( :, 1 ) - spinPosExp(1)*vecOnes;
spinPos( :, 2 ) = spinPos( :, 2 ) - spinPosExp(2)*vecOnes;
spinPos( :, 3 ) = spinPos( :, 3 ) + (par.z0 - spinPosExp(3))*vecOnes;

%-------------------------------------------------------------------------------
% Figure 1: Plot the spin locations
%-------------------------------------------------------------------------------
close all, figure( 1 );
plot3( spinPos( :,1 ), spinPos( :,2 ), spinPos( :,3 ), 'bo' );
grid on, ...
	xlabel( 'x [nm]' ), ylabel( 'y [nm]' ), zlabel( 'z [nm]' ), ...
	axis image 
	title( 'Spin locations of interest' );
print( '-f1', '-djpeg', 'pics/pdb_spin_loc.jpg' );

if ( (nargin == 6) & (bSuppressImg == 1) ), return, end;

% Construct the cantilever (intensity) image: at each spin location, create a 
% PSF response.

% First, we have to determine how big our mesh is in order to capture 
% everything. Compute this from the spinPos array. 
spinsXMin = min( spinPos( :, 1 ) );
spinsXMax = max( spinPos( :, 1 ) );
spinsYMin = min( spinPos( :, 2 ) );
spinsYMax = max( spinPos( :, 2 ) );

% Figure out how large out map has to be in x-y in order to cover the spin
% locations of interest and the additional area due to the impulse response
% at each spin location.
mapXMin = spinsXMin - par.xSpan;
mapXMax = spinsXMax + par.xSpan;
mapYMin = spinsYMin - par.ySpan;
mapYMax = spinsYMax + par.ySpan;
disp( 'Map dimensions:' );
disp( sprintf( '%12s %12s %12s %12s', 'xMin', 'xMax', 'yMin', 'yMax' ) );
disp( sprintf( '%12.3e %12.3e %12.3e %12.3e', mapXMin, mapXMax, mapYMin, ...
	mapYMax ) );

% Round the map into multiples of XSpan and YSpan
numXSpans = 1 + ceil( (mapXMax - mapXMin)/par.xSpan );
numYSpans = 1 + ceil( (mapYMax - mapYMin)/par.ySpan );

% Create a mesh for a single PSF
vecSpinX = par.xSpan*[-1:par.xInc:1];
vecSpinY = par.ySpan*[-1:par.yInc:1];
vecSpinXLen = length( vecSpinX );
vecSpinYLen = length( vecSpinY );
spinXMid = floor( vecSpinXLen / 2 );
spinYMid = floor( vecSpinYLen / 2 );

MESH_SPIN_ALL_ZEROS = zeros( vecSpinYLen, vecSpinXLen ); % Constants
MESH_SPIN_ALL_ONES = ones( vecSpinYLen, vecSpinXLen ); 

%Bmag = MESH_SPIN_ALL_ZEROS;
%G = MESH_SPIN_ALL_ZEROS;
%s = MESH_SPIN_ALL_ZEROS;
P = MESH_SPIN_ALL_ZEROS;

[meshSpinX, meshSpinY] = meshgrid( vecSpinX, vecSpinY );

% Create a mesh for the molecule
xSpaceFac = ceil( numXSpans / 2 );
ySpaceFac = ceil( numYSpans / 2 );
disp( 'Mesh of molecule:' );
disp( sprintf( '%12s %12s %12s %12s', 'xSpaceFac', 'xInc', 'ySpaceFac', ...
	'yInc' ) );
disp( sprintf( '%12.3e %12.3e %12.3e %12.3e', xSpaceFac, par.xInc, ...
	ySpaceFac, par.yInc ) ); 

vecMolX = par.xSpan*[-xSpaceFac:par.xInc:xSpaceFac];
vecMolY = par.ySpan*[-ySpaceFac:par.yInc:ySpaceFac];
vecMolXLen = length( vecMolX );
vecMolYLen = length( vecMolY );

MESH_MOL_ALL_ZEROS = zeros( vecMolYLen, vecMolXLen ); % Constant
meshMolP = MESH_MOL_ALL_ZEROS;
disp( sprintf( 'size of meshMolP: %6d x %6d', vecMolYLen, vecMolXLen ) );

% Create a mesh for the molecule
[meshMolX, meshMolY] = meshgrid( vecMolX, vecMolY );

spinPosDiscrete = []; % Contains the spin locations, but with the x and y
					  % coords discretized

% Discretize the x and y coords of the spins, but leave the z coord alone.
% Additionally, shift the x and y coords since arrays start at 1.
for spinInd = 1:numSpins
	xDis = round( spinPos( spinInd, 1 ) / xMeshSize ) + ...
		floor( vecMolXLen / 2 );
	yDis = round( spinPos( spinInd, 2 ) / yMeshSize ) + ...
		floor( vecMolYLen / 2 );
	spinPosDiscrete = [ spinPosDiscrete; [xDis, yDis, spinPos( spinInd, 3 )] ];
end

disp( 'Computing image...' );
tic; % Start timing

for spinInd = 1:numSpins % Superimpose the spin energies
	xSpinDis = spinPosDiscrete( spinInd, 1 );
	ySpinDis = spinPosDiscrete( spinInd, 2 );
	zSpin = spinPosDiscrete( spinInd, 3 );

	% Translation in the x-y plane 
	xDelta = xSpinDis - spinXMid;
	yDelta = ySpinDis - spinYMid;

	disp( sprintf( 'spin %4d: xSpinDis = %6d, ySpinDis = %6d, z = %12.3e', ...
		spinInd, xSpinDis, ySpinDis, zSpin ) );

	if 0 % Non-MEX fashion
		for j = 1:vecSpinYLen
			for k = 1:vecSpinXLen
				% First, generate the output due to a spin with a certain z 
				% value; use {Bmag,G,s}Tmp as temporary vars.
				[nStatus, BmagTmp, GTmp, sTmp, P(j,k)] = ...
					psf_vert( meshSpinX(j,k), meshSpinY(j,k), zSpin, ...
						par.Bext, par.Bres, par.m, par.xPk, 0 );
			end
		end
	else
		% Use MEX function
		disp( '   NOTE: using mex_psf_vert' );
		meshSpinZ = zSpin*MESH_SPIN_ALL_ONES;
		[nErrs, Bmag,G, s, P] = ...
			mex_psf_vert( meshSpinX, meshSpinY, meshSpinZ, par.Bext, ...
				par.Bres, par.m, par.xPk, 0 );
		disp( sprintf( 'nErrs = %g', nErrs ) );
	end

	% Then, add it to the molecule
	for k = 1:vecSpinXLen
		for j = 1:vecSpinYLen
			meshMolP( j + yDelta, k + xDelta ) = ...
				meshMolP( j + yDelta, k + xDelta ) + P( j, k );
		end
	end
end

tElapsed = toc; % Stop timing
disp( sprintf( 'Elapsed time: %g secs', tElapsed ) );

%-------------------------------------------------------------------------------
% Figure 2: Plot the cantilever image
%-------------------------------------------------------------------------------
sTitle = sprintf( 'Cantilever image of %d spins, z_0 = %d nm', ...
	numSpins, par.z0 );

[llRow, llCol, urRow, urCol] = get_int_bbox( meshMolP, 1e-3 );

% Unfortunately, we can't just use the bounding box computed above. We want to
% have some 0-space so that it seems like we're not cutting off anything. 

colRangeMin = llCol - floor( 0.2*( llCol - 1 ) ); 
colRangeMax = urCol + floor( 0.2*( vecMolXLen - urCol ) );
colRange = [colRangeMin:colRangeMax];

rowRangeMin = urRow - floor( 0.2*( urRow - 1 ) );
rowRangeMax = llRow + floor( 0.2*( vecMolYLen - llRow ) );
rowRange = [rowRangeMin:rowRangeMax];

disp( 'Plotting image...' );
tic; % Start timing

create_int_plot( 2, sTitle, [], ...
	meshMolX( rowRange, colRange ), meshMolY( rowRange, colRange ), ...
	meshMolP( rowRange, colRange ), 1 );
print( '-f2', '-djpeg', 'pics/pdb_cl_img.jpg' );

tElapsed = toc; % Stop timing
disp( sprintf( 'Elapsed time: %g secs', tElapsed ) );


