% pdb_match: extract positions of interest from ATOM and HETATM records
%-------------------------------------------------------------------------------
% spinPos = pdb_match( A, AType, H, HType, sRePlotType )
%    A, AType, H, HType - output of pdb_read_file
%    sRePlotType        - a RE specifying atoms of interest
%
% Output
%    spinPos            - a matrix of positions where each row specifies the 
%                         Cartesian coords of record types that matched
%                         sRePlotType. Units are in nm.
%
% NB. regexp is not available on all Matlab version
%
function spinPos = pdb_match( A, AType, H, HType, sRePlotType )

spinPos = []; % Contains the "spin" positions

if ( max( size( AType ) ) >= 1 )
	ATypeCellArr = my_strtrim( cellstr( AType ) );

	[nStatus, ATypeMatch] = safe_regexp( char(ATypeCellArr), sRePlotType );
	if ( nStatus < 0 ), error( 'no regexp' ), end;

	for atomInd = 1:length( ATypeMatch )
		if ( ~isempty( ATypeMatch{atomInd} ) & ( ATypeMatch{atomInd} >= 1 ) )
			spinPos = [ spinPos; A( atomInd, : ) ];
		end
	end
end

if ( max( size( HType ) ) >= 1 )
	HTypeCellArr = my_strtrim( cellstr( HType ) );

	[nStatus, HTypeMatch] = safe_regexp( char(HTypeCellArr), sRePlotType );
	if ( nStatus < 0 ), error( 'no regexp' ), end;

	for hetatmInd = 1:length( HTypeMatch )
		if ( ~isempty( HTypeMatch{hetatmInd} ) & ( HTypeMatch{hetatmInd} >= 1 ) )
			spinPos = [ spinPos; H( hetatmInd, : ) ];
		end
	end
end

% Cartesian coords are listed in Angstroms (1e-10 m). Convert to nm.
spinPos = spinPos * 1e-1;
spinPosSize = size( spinPos );
disp( sprintf( '%5d atoms matched', spinPosSize(1) ) );

% Find the smallest and largest spin distance
[smallestSpinDist, largestSpinDist] = find_dist_minmax( spinPos );
disp( sprintf( 'smallest/largest spin dist: %.3e, %.3e nm', ...
	smallestSpinDist, largestSpinDist ) );


%...............................................................................
function [nStatus, match] = safe_regexp( sa, re )
%...............................................................................
nStatus = 0;
bHomebrew = 0;

numStrs = size( sa, 1 );
match = cell( numStrs, 1 ); 

try
    for i = 1:numStrs
        res = regexp( sa( i, : ), re );
        if ( ~isempty( res ) ), match{i} = res(1); end;
    end
catch
	warning( sprintf( 'regexp produced error: %s', lasterr ) );
	% Have to implement our version of regexp
	disp( 'Running home-brew version of regexp' );
	disp( 'NOTE: do not use any RE expressions, as these will not work' );
	bHomebrew = 1;
end

if ( bHomebrew == 1 )		
	for i = 1:numStrs
		res = findstr( sa( i, : ), re );
		if ( ~isempty( res ) ), match{i} = res(1); end;
	end 
end

