% pdb_read_file: obtain ATOM and HETATM records from a PDB file
%-------------------------------------------------------------------------------
% Read a PDB file and return the coordinates of ATOMs and HETATMs. Consult
% the PDB guide for more info at http://www.rcsb.org.
%
% [A, AType, H, HType, res] = pdb_read_file( szFilename )
%    szFilename - file name of PDB file
% 
% Outputs:
%    A     - matrix of ATOM entries, with each row being the (x,y,z) coords for
%            an atom
%    AType - a column vector with the corresponding type
%    H     - matrix of HETATM entries,  "   "
%    HType -   "   "
%    res   - resolution of image in Angstroms
%
function [A, AType, H, HType, res] = pdb_read_file( szFilename )
global PDB_IGNORE PDB_REMARK2 PDB_ATOM PDB_HETATM

% Initialize
PDB_IGNORE = -1; 
PDB_REMARK2 = 0; 
PDB_ATOM = 1; 
PDB_HETATM = 2;

A = []; 
AType = []; 
H = []; 
HType = [];
res = -1;

fid = fopen( szFilename, 'r' );
if ( fid == -1 ), error( 'could not open file' ), end;

while ( 1 )
	% Read new line from file and discard NL character
	sLine = fgetl( fid );
	if ( ~ischar( sLine ) ), break, end;
	[nStatus, stParseInfo] = pdb_parse_line( sLine );
	switch ( nStatus )
	case PDB_REMARK2
		if ( ~isstruct( stParseInfo ) | ~isfield( stParseInfo, 'resolution' ) )
			error( '???' );
		end
		res = stParseInfo.resolution;

	case PDB_ATOM
		if ( ~isstruct( stParseInfo ) | ...
			 ~isfield( stParseInfo, 'vCoords' ) | ...
			 ~isfield( stParseInfo, 'atomType' ) ), error( '???' ), end;
		A = [A; stParseInfo.vCoords'];
		AType = [AType; stParseInfo.atomType];
	
	case PDB_HETATM	
		if ( ~isstruct( stParseInfo ) | ...
			 ~isfield( stParseInfo, 'vCoords' ) | ...
			 ~isfield( stParseInfo, 'atomType' ) ), error( '???' ), end;
		H = [H; stParseInfo.vCoords'];
		HType = [HType; stParseInfo.atomType];

	case PDB_IGNORE
		continue;

	otherwise
		error( 'pdb_parse_line returned illegal value' );
	end
end
fclose( fid );

Adims = size( A );
Hdims = size( H );
disp( sprintf( 'Read %5d ATOMs and %5d HETATMs', Adims(1), Hdims(1) ) );

if ( res == -1 )
	warning( 'Did not obtain any resolution info' );
end

%-------------------------------------------------------------------------------
function [nStatus, stParseInfo] = pdb_parse_line( sLine )
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
global PDB_IGNORE PDB_REMARK2 PDB_ATOM PDB_HETATM

%disp( sprintf( 'Parsing: %s', sLine ) );

% Ranges for the field values of interest in REMARK records
remarkNumRange = 10;
angstromStrRange = [29:38];
resolutionRange = [23:27];

% Ranges for the field values of interest in ATOM and HETATM records
recordRange = [1:6];
serialRange = [7:11];
atomRange = [13:16];
xRange = [31:38];
yRange = [39:46];
zRange = [47:54];

if ( length( sLine ) >= recordRange(end) )
	recordType = upper( deblank( sLine( recordRange ) ) );
else
	warning( sprintf( 'unrecognized PDB line: %s', sLine ) );
	nStatus = PDB_IGNORE;
	stParseInfo = [];
	return;
end

switch recordType
case 'REMARK'
	remarkNum = str2num( sLine( remarkNumRange ) );
	if ( remarkNum == 2 )
		sAngstrom = sLine( angstromStrRange );
		if ( strcmp( sAngstrom, 'ANGSTROMS.' ) == 1 )
			resolution = str2double( sLine( resolutionRange ) );
			if ( isfinite( resolution ) )
				stParseInfo = struct( 'resolution', resolution );
				nStatus = PDB_REMARK2;
				disp( sprintf( 'Resolution %g Angstroms', ...
					stParseInfo.resolution ) );
			else
				error( 'illegal value in parsing REMARK2 line' );
			end
			return;
		end
	end
		 
case 'ATOM'
	serial = str2num( sLine( serialRange ) );
	x = str2double( sLine( xRange ) );
	y = str2double( sLine( yRange ) );
	z = str2double( sLine( zRange ) );
	if ( isfinite( x ) & isfinite( y ) & isfinite( z ) )
		stParseInfo = struct( 'vCoords', [x; y; z], ...
			'atomType', sLine( atomRange ) );
		nStatus = PDB_ATOM;
		%disp( sprintf( 'Read ATOM %5d of type %s', serial, ...
		%	stParseInfo.atomType ) );
	else
		error( 'illegal coord values in parsing ATOM line' );
	end
	return;

case 'HETATM'
	serial = str2num( sLine( serialRange ) );
	x = str2double( sLine( xRange ) );
	y = str2double( sLine( yRange ) );
	z = str2double( sLine( zRange ) );
	if ( isfinite( x ) & isfinite( y ) & isfinite( z ) )
		stParseInfo = struct( 'vCoords', [x; y; z], ...
			'atomType', sLine( atomRange ) );
		nStatus = PDB_HETATM;
		%disp( sprintf( 'Read HETATM %5d of type %s', serial, ...
		%	stParseInfo.atomType ) );
	else
		error( 'illegal coord values in parsing HETATM line' );
	end
	return;
end

nStatus = PDB_IGNORE;
stParseInfo = [];
