% create_rand_Prw: create a random Prw transition matrix
%-------------------------------------------------------------------------------
% Prw = create_rand_Prw( M, sPrwOuterConstructionMethod,
%	sPrwInnerConstructionMethod )
%    - M = number of positive levels. Note that the total number of levels is 
%      equal to r = 2M+1.
%    - Valid outer construction methods are: 'random', 'bounce'
%    - Valid inner construction methods are: 
%         'random', 'symmetric', 'randomNoloop', 'symmetricNoloop'
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function Prw = create_rand_Prw( M, sPrwOuterConstructionMethod, ...
	sPrwInnerConstructionMethod )

r = 2*M + 1;
Prw = zeros( r, r );

switch lower( sPrwOuterConstructionMethod )
case 'random'
	% -- Make the 1st and last rows random
	disp( 'outer rows: random' );
	Prw( 1, 1 ) = rand;
	Prw( 1, 2 ) = 1 - Prw( 1, 1 );
	Prw( r, r ) = rand;
	Prw( r, r - 1 ) = 1 - Prw( r, r );
case 'bounce'
	% -- Make the first row [0 1 0 ... 0] and the last row [0 ... 0 1 0]
	disp( 'outer rows: bounce' );
	Prw( 1, 2 ) = 1;
	Prw( r, r - 1 ) = 1;
otherwise
	error( 'invalid sPrwOuterConstructionMethod' );
end

switch lower( sPrwInnerConstructionMethod )
case 'random'
	% -- Make the inner rows arbitrary
	disp( 'inner rows: random' );
	for i = 2:(r-1)
		Prw( i, i ) = rand;
		fracLeft = 1 - Prw( i, i );
		Prw( i, i - 1 ) = fracLeft*rand;
		Prw( i, i + 1 ) = 1 - Prw( i, i ) - Prw( i, i - 1 );
	end
case 'randomnoloop'
	% -- Make the inner row arbitrary, but there are no loops from one
	% -- state back to itself
	disp( 'inner rows: randomNoLoop' );
	for i = 2:(r-1)
		Prw( i, i - 1 ) = rand;
		Prw( i, i + 1 ) = 1 - Prw( i, i - 1 );
	end
case 'symmetric'
	% -- Make the inner rows symmetric about the 0-level
	disp( 'inner rows: symmetric' );
	for i = 2:M
		Prw( i, i ) = rand;
		fracLeft = 1 - Prw( i, i );
		Prw( i, i - 1 ) = fracLeft*rand;
		Prw( i, i + 1 ) = 1 - Prw( i, i ) - Prw( i, i - 1 );
	end
	Prw( M+1, M+1 ) = rand;
	Prw( M+1, M ) = (1 - Prw( M+1, M+1 ))/2;
	Prw( M+1, M+2 ) = Prw( M+1, M );
	for i = (M+2):(r-1)
		j = r + 1 - i;
		Prw( i, i ) = Prw( j, j );
		Prw( i, i - 1 ) = Prw( j, j + 1 );
		Prw( i, i + 1 ) = Prw( j, j - 1 );
	end
case 'symmetricnoloop'
	% -- Make the inner rows symmetric about the 0-level, and impose the
	% -- constraint that there are no self-loops
	disp( 'inner rows: symmetricNoloop' );
	for i = 2:M
		Prw( i, i - 1 ) = rand;
		Prw( i, i + 1 ) = 1 - Prw( i, i - 1 );
	end
	Prw( M+1, M ) = 0.5;
	Prw( M+1, M+2 ) = 0.5;
	%Prw( M+1, M ) = rand;
	%Prw( M+1, M+2 ) = 1 - Prw( M+1, M );
	for i = (M+2):(r-1)
		j = r + 1 - i;
		Prw( i, i - 1 ) = Prw( j, j + 1 );
		Prw( i, i + 1 ) = Prw( j, j - 1 );
	end
otherwise
	error( 'invalid sPrwInnerConstructionMethod' );
end


