% rw_anal: analyze a random walk process
%-------------------------------------------------------------------------------
% [PrwTilde,U,Sigma] = rw_anal( snrdB, bPlot, bMagic )
%    - snrdB = snr in dB
%    - bPlot controls whether to plot or not
%    - bMagic usage: see source
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [PrwTilde,U,Sigma] = rw_anal( snrdB, bPlot, bMagic )

if ( nargin ~= 3 )
	error( 'missing arguments' );
end

close all; % Close all figures

% Step size
s = 1.58114e-5;
% Number of positive levels
%M = 58;
M = 5;
% Number of states
r = 2*M + 1;
% Column vector of all 1s
vOnes = ones( r, 1 );

K1 = 0.52;
K2 = 0.48;
if ( bMagic == 0 )
	disp( 'XXX: creating H1, H2 symmetric' );
	H2 = K1;
	H1 = K2;
else
	disp( 'XXX: creating H1, H2 non-symmetric' );
	H1 = K1;
	H2 = K2;
end

psiRange = s * [-M:M]';
Mpsi = diag( psiRange );

% Create Prw
%Prw = create_fixed_Prw( M, K1, K2, H1, H2 );
Prw = create_rand_Prw( M, 'bounce', 'symmetricNoloop' );
%Prw = create_rand_Prw( M, 'bounce', 'randomNoloop' );
disp( 'Prw:' );
disp( Prw );

% Find the steady-state probability densities
[PV, PD] = eig( Prw' );
[PDVals, PDIndex] = sort( diag(PD) );
disp( 'Eigenvalues of Prw:' );
disp( PDVals' );
biggestEigenval = PDVals(r);
ssDist = PV( :, PDIndex(r) );
% Normalize ssDist so that the elements add to 1
ssDist = ssDist / sum( ssDist );
disp( sprintf( 'biggest eigenvalue of Prw is %g', biggestEigenval ) );
if ( bPlot == 1 )
	figure, subplot(211);
	plot( [-M:M], ssDist, 'bo' ), xlabel( 'state' ), ...
		ylabel( 'steady-state probability' );
	% Adjust y-axis so that it's from 0 to the max value
	figAxis = axis;
	axis( [figAxis(1) figAxis(2) 0 figAxis(4)] );
else
	disp( 'Skipping plot' );
	disp( 'Steady-state distribution:' );
	disp( ssDist' );
end

%-------------------------------------------------------------------------------
% From the steady-state probability values, compute the value of sigma for
% the AWGN noise. Define SNR = lim_{i->\infty} E[\mu_i^2]/\sigma^2, where
% \mu_i is the random walk process.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
num = sum( (psiRange.^2) .* ssDist );
snr = 10^(snrdB/10);
sigma = sqrt( num/snr );
disp( sprintf( 's = %g, snr = %g dB --> sigma = %g', s, snrdB, sigma ) );

%% Construct g, R1, S1
g = exp( -psiRange.^2/2/sigma^2 );
%R1 = zeros( r, 1 );
%R1( M ) = 0.5;
%R1( M + 2 ) = 0.5;
R1 = ones( r, 1 ) * (1/r);
S1 = g .* R1;

% Construct PrwTilde
PrwTilde = zeros( r, r );
for i = 1:r
	PrwTilde( :, i ) = g(i) * Prw( :, i );
end

% Eigendecompose PrwTilde
[U,Sigma] = eig( PrwTilde );
if ( bPlot == 1 )
	subplot(212), plot( sort(diag(Sigma)), 'r*' );
	ylabel( 'eigenvalues of PrwTilde' );
else
	disp( 'Eigenvalues of PrwTilde:' );
	disp( sort( diag(Sigma)' ) );
end

%-------------------------------------------------------------------------------
% Don't do a detailed decomposition
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if 1
A = Mpsi*PrwTilde;
Atilde = inv(U)*A*U;
M11 = zeros( r, r );
M11( 1, 1 ) = 1;
M22 = zeros( r, r );
M22( 2, 2 ) = 1;
xT = ones( 1, r ) * PrwTilde' * U;
y = inv(U)*S1;
xTilde1 = xT*(M11 + M22)*Atilde;
xTilde2 = xT*(M11 - M22)*Atilde;
yTilde1 = Atilde*(M11 + M22)*y;
yTilde2 = Atilde*(M11 - M22)*y;

xy11 = xTilde1' .* yTilde1;
xy12 = xTilde1' .* yTilde2;
xy21 = xTilde2' .* yTilde1;
xy22 = xTilde2' .* yTilde2;
disp( 'Cases' );
disp( sprintf( '11: x~1y~1 = %g, x~2y~2 = %g', xy11(1), xy11(2) ) );
disp( sprintf( '12: x~1y~1 = %g, x~2y~2 = %g', xy12(1), xy12(2) ) );
disp( sprintf( '21: x~1y~1 = %g, x~2y~2 = %g', xy21(1), xy21(2) ) );
disp( sprintf( '22: x~1y~1 = %g, x~2y~2 = %g', xy22(1), xy22(2) ) );

disp( sprintf( 'x1 = %g, x2 = %g', xT(1), xT(2) ) );
disp( sprintf( 'y1 = %g, y2 = %g', y(1), y(2) ) );

%B1 = zeros( 2, 2 );
%B1( 1, 1 ) = Atilde(1,1)^2;
%B1( 1, 2 ) = Atilde(1,1)*Atilde(1,2);
%B1( 2, 1 ) = Atilde(1,1)*Atilde(2,1);
%B1( 2, 2 ) = Atilde(2,1)*Atilde(1,2);
%disp( 'B1:' ); disp( B1 );
%B2 = zeros( 2, 2 );
%B2( 1, 1 ) = Atilde(1,2)*Atilde(2,1);
%B2( 1, 2 ) = Atilde(1,2)*Atilde(2,2);
%B2( 2, 1 ) = Atilde(2,1)*Atilde(2,2);
%B2( 2, 2 ) = Atilde(2,2)^2;
%disp( 'B2:' ); disp( B2 );
%disp( 'B2-B1:' ); disp( B2-B1 );

%disp( 'Atilde( 1:2, 1:2 ):' );
%disp( sprintf( '   %g   %g', Atilde(1,1), Atilde(1,2) ) );
%disp( sprintf( '   %g   %g', Atilde(2,1), Atilde(2,2) ) );

rx = xT(2)/xT(1); ry = y(2)/y(1);
m22 = Atilde(2,2)/Atilde(1,1);
m12 = Atilde(1,2)/Atilde(1,1);
m21 = Atilde(2,1)/Atilde(1,1);
disp( sprintf( 'rx = %g, ry = %g', rx, ry ) );
disp( sprintf( 'm12 = %g, m21 = %g, m22 = %g', m12, m21, m22 ) );
disp( sprintf( '  > m12 = %g, rx*m22 = %g', m12, rx*m22 ) );
disp( sprintf( '  > m21 = %g, ry*m22 = %g', m21, ry*m22 ) );
disp( sprintf( '  > rx*m21 = %g, ry*m12 = %g', rx*m21, ry*m12 ) );

disp( sprintf( 'x~1y~1: %g | %g | %g | %g', ...
	xT(1)*y(1)*Atilde(1,1)^2, ...
	xT(2)*y(1)*Atilde(1,1)*Atilde(2,1), ...
	xT(1)*y(2)*Atilde(1,1)*Atilde(1,2), ...
	xT(2)*y(2)*Atilde(2,1)*Atilde(1,2) ) );
disp( sprintf( 'x~2y~2: %g | %g | %g | %g', ...
	xT(1)*y(1)*Atilde(1,2)*Atilde(2,1), ...
	xT(2)*y(1)*Atilde(2,1)*Atilde(2,2), ...
	xT(1)*y(2)*Atilde(1,2)*Atilde(2,2), ...
	xT(2)*y(2)*Atilde(2,2)^2 ) );
end
