% psf_mesh: create a mesh structure 
%-------------------------------------------------------------------------------
% The mesh structure has the following fields:
%    num{X,Y,Z}Pts - number of pts in the x-, y-, and z-dir. Corresponds to
%                    I,J,K resp.
%			 				[dimensionless]
%    delta{X,Y,Z}       - mesh step size in the x,y,z dir resp.		
%							[nm]
%    vecOriginCoords    - should we choose to, the position where the
%                         mesh is centred							
%							[nm^3]
%    vecLowerLeftCoords - should ... ..., the ... where the lower left corner
%                         of the mesh is positioned					
%							[nm^3]
%    sMeshPosition - how to position the mesh. The options are
%                    (case-insensitive):
%
%                    'origin' -> position the middle of the mesh on [0,0,par.z0]
%                    'lowerLeft' -> position the lower left corner of the mesh
%                                   on (-par.xSpan,-par.ySpan,par.z0)
%
% 
% mesh = psf_mesh( par, spinPos, vecNumPts )
%    par       - parameter structure
%    spinPos   - coords of a molecule, returned from pdb_cl_img
%    vecNumPts - vector of # pts
%
% Output
%    mesh - mesh structure
%
function mesh = psf_mesh( par, spinPos, vecNumPts )

B_VERBOSE = 1;
DELTA_FAC = 3;

if ( nargin < 3 )
    warning( 'Setting default values for vecNumPts' );
    vecNumPts = [ 100 100 1 ];
end

vecDelta = DELTA_FAC * [ par.xSpan/vecNumPts(1), par.ySpan/vecNumPts(2), 0 ];

% Setting vecDelta(3) is subjective and depends on the parameters of
% psf_param3. Basically, do it on an ad-hoc basis.
% vecDelta(3) = min( [vecDelta(1), vecDelta(2)] );
vecDelta(3) = 0.03;
 
vecOriginCoords = [0, 0, par.z0];
vecLowerLeftCoords = [-par.xSpan, -par.ySpan, par.z0];

% If a molecule has been supplied, fit the lower left coords to it. 

if ( ( nargin >= 2 ) & ( ~isempty( spinPos ) ) )
  
    vecLowerLeftCoords = [ min( -par.xSpan, min( spinPos(:,1) ) ), ...
        min( -par.ySpan, min( spinPos(:,2) ) ), ...
        0.7*par.z0 ];

	disp( sprintf( 'Initial vecLowerLeftCoords: (%g,%g,%g)', ...
        vecLowerLeftCoords(1), ...
		vecLowerLeftCoords(2), ...
        vecLowerLeftCoords(3) ) );
    
	spinPos = pdb_lower_left( vecLowerLeftCoords, spinPos, 1 );

	% DEBUG
	disp( 'spinPos:' ), disp( spinPos );

    vecNewDelta(1) = check_mesh_delta( 'x', vecDelta(1), vecNumPts(1), ...
        vecLowerLeftCoords(1), max( spinPos(:,1) ), 0.5 );
    vecNewDelta(2) = check_mesh_delta( 'y', vecDelta(2), vecNumPts(2), ...
        vecLowerLeftCoords(2), max( spinPos(:,2) ), 0.5 );
    vecNewDelta(3) = check_mesh_delta( 'z', vecDelta(3), vecNumPts(3), ...
        min( spinPos(:,3) ), max( spinPos(:,3) ), 0.5 );

	for i = 1:3
		if ( vecNewDelta(i) > vecDelta(i) ), vecDelta(i) = vecNewDelta(i); end
	end

	vecLowerLeftCoords(3) = min( spinPos(:,3) );
end

% Create the mesh structure
mesh = struct( 'numPts', vecNumPts, ...
    'delta', vecDelta, ...
	'originCoord', vecOriginCoords, ...
	'lowerLeftCoord', vecLowerLeftCoords, ...
	'szPosition', 'lowerLeft' );

%...............................................................................
function newDelta = check_mesh_delta( sCoord, delta, numPtsAvail, startPos, ...
	molPos, fracMolPos )
%...............................................................................

if ( delta <= 0 ), error( 'check_mesh_delta: delta <= 0' ), end;
if ( numPtsAvail < 1 ), error( 'check_mesh_delta: numPtsAvail < 1' ), end;
if ( startPos > molPos ), error( 'check_mesh_delta: startPos > molPos' ), end;

numPtsNeeded = 1 + floor( (molPos - startPos) / delta );
N = max( [1, floor( fracMolPos*numPtsAvail )] );
if ( N == 1 ), error( 'N == 1??' ), end;

if ( numPtsNeeded > N )
    % Compute a new delta such that we don't use up more than half the
    % available points
    newDelta = (molPos - startPos) / (N - 1);
    warning( sprintf( 'Need more than %g of available pts', fracMolPos ) );
    disp( sprintf( '%s-dir: delta = %.5e, newDelta = %.5e', ...
        sCoord, delta, newDelta ) );
else
    newDelta = delta;
end

