% psf_param3: derived from psf_param2
%-------------------------------------------------------------------------------
%
% {{{ History }}}
%
% 30-jul-04: Want to derive optimal parameters. Also, renormalize variables
% so that they're reported in nm.
%
% 06-feb-06: Removed mesh parameters from the par structure. This will
% probably break some old code, but oh well.
%
% sPsf = psf_param3( bComputeOpt [, ArgBres] [, Argd] [, bNoMsgs] )
%    bComputeOpt - true if we want to compute optimal parameters for a given
%                  Bres and d
%    ArgBres     - Bres [Gauss]
%    Argd        - d    [nm]
%    bNoMsgs     - boolean flag to suppress printed output
%

function sPsf = psf_param3( bComputeOpt, ArgBres, Argd, bNoMsgs )

% Default parameters, taken from psf_param2
ParBext = 2e4;					% [Gauss]
ParBres = 2.25e4;				% [Gauss]
ParR0 = 6.017e1;				% [nm]
%ParM = 800;					% [???] dunno what units this is
ParM = 1700;
Parm = 4*pi/3*ParM*ParR0^3;		% [emu] Assume spherical tip
Parz0 = 8.35e1;					% [nm]
Pard = -1;						% [nm] Not previously given
ParxPk = 2e-1;					% [nm]
ParxOpt = -1;					% [nm] Not previously computed
ParxSpan = 10;					% [nm]
ParxInc = .02;					% dimensionless
% EXPERIMENTAL
%ParxInc = .02;
ParGMax = -1;					% [Gauss/nm] Not previously computed 

B_USE_SUBOPTIMAL_R0 = 1;
B_USE_UNREALISTIC_B0 = 1;

% Compute optimal parameters for a given Bres and d
if ( bComputeOpt == 1 )
	if ( nargin < 3 ), error( 'insufficient number of arguments' ), end;
	if ( ArgBres <= 0 | Argd <= 0 ), error( 'invalid args' ), end;

	% Set Bres and d to the argument values
	ParBres = ArgBres;
	Pard = Argd;

	% DEBUG
	disp( sprintf( 'psf_param3: using Bres = %.3e, d = %.3e', ParBres, Pard ) );

	optR0Fac = 3; % optimal R0 factor
	ParR0 = optR0Fac*Pard;

	% XXX: because the optimal R0 is too large, go with something smaller.
	% This causes the resonant slice to shrink in radius
	if ( B_USE_SUBOPTIMAL_R0 == 1 )
		expR0Fac = 1;
		ParR0 = expR0Fac*Pard;
		if ( ( nargin < 4 ) | ( ( nargin == 4 ) & ( bNoMsgs == 0 ) ) )
			disp( sprintf( 'Using expR0Fac=%g => R0=%g', expR0Fac, ParR0 ) );
			%beep;
		end
	end

	Parm = 4*pi/3*ParM*ParR0^3;
	Parz0 = ParR0 + Pard;

	optXFac = -( sqrt(35) - sqrt(19) )/4; % optimal x-pos factor for max G
	ParxOpt = optXFac*Parz0;
	ParxSpan = abs( 2*ParxOpt );

	optGFac = 2.7428; % max factor for G in terms of m,z0
	ParGMax = optGFac * Parm / Parz0^4;

	deltaB0 = 50;				% [Gauss]

	% XXX: because 50 is too large, go with something smaller
	if ( B_USE_UNREALISTIC_B0 == 1 )
		deltaB0 = 20;
		if ( ( nargin < 4 ) | ( ( nargin == 4 ) & ( bNoMsgs == 0 ) ) )
			disp( sprintf( 'Using deltaB0=%g', deltaB0 ) );
			%beep;
		end
	end

	ParxPk = deltaB0/ParGMax;

	% Finally, compute Bext
	a1 = -(2 - optXFac^2)/(1 + optXFac^2)^(5/2);
	a2 = -9 * optXFac^2 / (1 + optXFac^2)^5;
	quadSol1 = a1 * (Parm/Parz0^3) + sqrt( ParBres^2 + a2*(Parm/Parz0^3)^2 );
	quadSol2 = a1 * (Parm/Parz0^3) - sqrt( ParBres^2 + a2*(Parm/Parz0^3)^2 );
	if 0 % DEBUG
		disp( sprintf( '%12s %12s %12s %12s', 'a1', 'a2', 'quadSol1', ...
			'quadSol2' ) );
		disp( sprintf( '%12.3e %12.3e %12.3e %12.3e', a1, a2, quadSol1, ...
			quadSol2 ) );
	end
	ParBext = quad_pos_sol( quadSol1, quadSol2 );

	if ( ( nargin < 4 ) | ( ( nargin == 4 ) & ( bNoMsgs == 0 ) ) )
		disp( 'psf_param3: computing optimal parameters' );
	end
else	
	if ( ( nargin < 4 ) | ( ( nargin == 4 ) & ( bNoMsgs == 0 ) ) )
		disp( 'psf_param3: using non-optimized parameters' );
	end
end

sPsf = struct( ... 
	'Bext', ParBext, ... % Simulation parameters
	'Bres', ParBres, ...
	'R0', ParR0, ...
	'M', ParM, ...	
	'm', Parm, ...	
	'xPk', ParxPk, ...	
	'xOpt', ParxOpt, ...
	'z0', Parz0, ...
	'd', Pard, ...
	'GMax', ParGMax, ...
	'xSpan', ParxSpan, ... % Specify the grid for the x-y plane
	'ySpan', ParxSpan, ...			
	'deltaZUnc', 0.1*Parz0 ... % Uncertainty in z for random images
    );

% Add xInc for legacy code
sPsf.xInc = ParxInc;
sPsf.yInc = sPsf.xInc;
