% psf_vert: PSF for a vertical cantilever
%-------------------------------------------------------------------------------
% PSF for the vertical cantilever. See the report "Point Response Function of
% an MRFM Tip" by Mamin et al.
%
% An alternative is mex_psf_vert, which is faster.
%
% [nStatus, Bmag, G, s, P] = psf_vert( x, y, z, ParBext, ParBres, Parm, 
%    ParxPk, ParbVerbose )
%
%    (x, y, z) - position of spin relative to tip centre
%    ParBext   - externally applied field [Gauss]
%    ParBres   - resonant field = gamma*omega_rf [Gauss]
%    Parm      - total moment of the tip = 4*pi/3*M*R_0^3, where M is the
%                magnetization density (800 emu/cm^3) and R_0 is the tip radius
%                (emu)
%    ParxPk    - cantilever peak oscillation amplitude [cm]
%
% Output:
%    nStatus       - 0 if ok; -1 if error(s) occurred
%    Bmag, G, s, P - the first three are intermediate qtys, the last is the
%                    PSF output
%
function [nStatus, Bmag, G, s, P] = psf_vert( x, y, z, ParBext, ParBres, ...
	Parm, ParxPk, ParbVerbose )
		
global Bext Bres m bVerbose

Bext = ParBext;
Bres = ParBres;
m = Parm;
bVerbose = ParbVerbose;

if ( ParxPk <= 0 ), error( 'invalid parameters' ), end;

% Initialize
nStatus = 0; % Ok
Bmag = 0; 
G = 0; 
s = 0; 
P = 0;

[Bmag, bError] = fnBmag( x, y, z );
if ( bError == 1 ), print_err( x, y, z ), nStatus = -1; return, end;

[G, bError] = fnG( x, y, z );
if ( bError == 1 ), print_err( x, y, z ), nStatus = -1; return, end;

[s, bError] = fnS( Bmag, G );
if ( bError == 1 ), print_err( x, y, z ), nStatus = -1; return, end;

% If everything is fine so far, compute PSF(x,y,z). Note that, for now, we
% take G0 = 1.
if ( abs( s ) <= ParxPk )
	P = G^2*( 1 - (s/ParxPk)^2 );
else
	P = 0;
end 

% Check that it is non-negative
if ( P < 0 ), error( 'invalid PVal' ), end;


%-------------------------------------------------------------------------------
function print_err( x, y, z )
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
global bVerbose
if ( bVerbose == 1 )
	disp( sprintf( 'encountered error at x=%g, y=%g, z=%g', x, y, z ) );
end

%-------------------------------------------------------------------------------
function [rc, bError] = fnBmag( x, y, z )
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
global Bext m bVerbose
r = sqrt( x^2 + y^2 + z^2 );
if ( r == 0 )
	if ( bVerbose == 1 ), warning( 'r=0 in fnBmag' ), end;
	rc = -1;
	bError = 1;
	return;
end
r5 = r^5;
a1 = 3*x*z*m/r5;
a2 = 3*y*z*m/r5;
a3 = (2*z^2 - x^2 - y^2)*m/r5 + Bext;
rc = sqrt( a1^2 + a2^2 + a3^2 );
bError = 0;

%-------------------------------------------------------------------------------
function [rc, bError] = fnG( x, y, z )
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
global Bext m bVerbose

r = sqrt( x^2 + y^2 + z^2 );
if ( r == 0 )
	if ( bVerbose == 1 ), warning( 'r=0 in fnG' ), end;
	rc = -1;
	bError = 1;
	return;
end

r12 = r^12;
r10 = r^10;
r5 = r^5;
r7 = r^7;
n1 = -x^2 - y^2 + 2*z^2;

a1 = -90*m^2*x^3*z^2/r12 - 90*m^2*x*y^2*z^2/r12 + 18*m^2*x*z^2/r10;
a2 = -2*m*x/r5 - 5*m*x*n1/r7;
a3 = Bext + m*n1/r5;
a4 = 9*m^2*x^2*z^2/r10 + 9*m^2*y^2*z^2/r10;
a5 = a3^2;

rc = (a1 + 2*a2*a3) / (2 * sqrt( a4 + a5 ));
bError = 0;

%-------------------------------------------------------------------------------
function [rc, bError] = fnS( Bmag, G )
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
global Bres bVerbose
if ( G == 0 )
	if ( bVerbose == 1 ), warning( 'G=0 in s' ), end;
	rc = -1;
	bError = 1;
	return;
end
rc = ( Bres - Bmag ) / G;
bError = 0;
