% read_h_op: read the h operator from disk
%-------------------------------------------------------------------------------
% Reads the specified filename and returns it in hDisk if the disk version of
% par and stMesh match the ones supplied in the function call.
%
% hDisk = read_h_op( par, stMesh, sFilename )
%    par       - parameter structure
%    stMesh    - mesh structure
%    sFilename - filename containing the saved h operator
%
% Output
%    hDisk     - contains the h operator if a match occurred; otherwise,
%                the empty matrix
%
function hDisk = read_h_op( par, stMesh, sFilename )

hDisk = []; % Initialize

% Try to see if the file exists
fid = fopen( sFilename, 'r' );
if ( fid == -1 ), warning( 'cannot open file for reading' ), return, end;
fclose( fid );

% Try loading the file
diskInfo = load( sFilename );

if ( ~isfield( diskInfo.par ) | ~isfield( diskInfo.stMesh ) | ...
	 ~isfield( diskInfo.h ) | ~isfield( diskInfo.sSaveTimestamp ) )
	warning( sprintf( 'read_h_op: lack of data in %s', sFilename ) );
	return;
end

% Check that the relevant fields of par and stMesh are the same. For this, 
% refer to get_h_op().

if ( ~( (par.Bext == diskInfo.par.Bext) & ...
	  	(par.Bres == diskInfo.par.Bres) & ...
		(par.m == diskInfo.par.m) & ...
		(par.xPk == diskInfo.par.xPk) ) )
	warning( 'par values not the same' );
	return;
end

if ( ~( (stMesh.numXPts == diskInfo.stMesh.numXPts) & ...
		(stMesh.numYPts == diskInfo.stMesh.numYPts) & ...
		(stMesh.numZPts == diskInfo.stMesh.numZPts) & ...
		(stMesh.deltaX == diskInfo.stMesh.deltaX) & ...
		(stMesh.deltaY == diskInfo.stMesh.deltaY) & ...
		(stMesh.deltaZ == diskInfo.stMesh.deltaZ) ) )
	warning( 'mesh grid values not the same' );
	return;
end

if ( strcmp( stMesh.sMeshPosition, diskInfo.stMesh.sMeshPosition ) ~= 1 )
	warning( 'mesh position not the same' );
	return;
end

switch ( lower(stMesh.sMeshPosition) )
case 'origin'
	vecInput = stMesh.vecOriginCoords;
	vecDisk = diskInfo.stMesh.vecOriginCoords;
case 'lowerleft'
	vecInput = stMesh.vecLowerLeftCoords;
	vecDisk = diskInfo.stMesh.vecLowerLeftCoords;
otherwise
	error( 'invalid mesh position' );
end

for i = 1:3
	if ( vecInput(i) ~= vecDisk(i) )
		warning( 'mesh position vectors not the same' );
		return;
	end
end

% If all the checks have passed, then we can safely return the disk value
hDisk = diskInfo.h;
