% test_bsolv: tests solvers for binary image reconstruction
%-------------------------------------------------------------------------------
% [vecGain, vecInit] = test_bsolv( n, sigma )
%    n       - dimensionality of the problem
%    sigma   - variance of noise
%
% Output
%    vecGain - vector of improvements. See source code for specifics.
%    vecInit - see source code
%
function [vecGain, vecInit] = test_bsolv( n, sigma )

% Globals
global ATA xML LARGE_NUMBER

% Constants
LARGE_NUMBER = 1e9;
nTrials = 5e3;

% Parameters
nRandomTries = 20 ;
nMaxGradIters = 50;
nRandomBitFlips = min( [round( 0.1*n ), 1] );

vecGain = zeros( nTrials, 1 );

for trialInd = 1:nTrials
	% Construct a suitable A that is not too singular
	for attemptInd = 1:10
		A = toeplitz( 2*rand( 1, n ) - 1 );
		if ( cond( A ) < 1e4 ), break, end;
	end
	if ( cond( A ) > 1e4 ), error( 'could not generate suitable A' ), end;

	ATA = A'*A;
	xKnown = ( rand( n, 1 ) > 0.75 );
	yObs = A*xKnown + sigma*randn( n, 1 );
	xML = A\yObs;

	% Formulate an initial guess
	xInit = bstr_bin_test( xML );

	%x = bstr_steepest_desc( nMaxGradIters, xInit, 0, 'notgreedy', @phi, ...
	%	@phiRowGrad );

	x = bstr_conj_grad( nMaxGradIters, xInit, 0, @phi, @phiRowGrad );

	%[xBest, wBest] = rand_ics( nMaxGradIters, 20, 'greedy' );
	%[xBest, wBest] = rand_cent_ics( nMaxGradIters, xInit, 20, ...
	%	round( .2 * n ), 'greedy' );	

	%vecGain( trialInd ) = ( phi( xInit ) - phi( x ) ) / phi( xInit ) * 100;
	%vecInit = [];

	%vecGain( trialInd ) = sum( ( xML - x ).^2 / sigma^2 );
	%vecInit( trialInd ) = sum( ( xML - xInit ).^2 / sigma^2 );

	vecGain( trialInd ) = sum( abs( xKnown - x ) );
	vecInit( trialInd ) = sum( abs( xKnown - xInit ) );
end

%...............................................................................
% Instead of having one single initial condition to use with the steepest
% descent method, use random initial conditions to find the minimum.
%...............................................................................
function [xBest, wBest] = rand_ics( nMaxGradIters, nRandomTries, sType )

global LARGE_NUMBER

xBest = []; % Initialize
wBest = LARGE_NUMBER;

for i = 1:nRandomTries
	xTry = ( rand( n, 1 ) > 0.5 );
	x = bstr_steepest_desc( nMaxGradIters, xTry, 0, sType, @phi, @phiRowGrad );
	wTry = phi( x );
	if ( wTry < wBest )
		wBest = wTry;
		xBest = x;
	end
end

if ( wBest == LARGE_NUMBER ), error( 'need to increase LARGE_NUMBER' ), end;

%...............................................................................
% This version also uses random initial conditions, but they are centred
% around an initial point xInit.
%...............................................................................
function [xBest, wBest] = rand_cent_ics( nMaxGradIters, xInit, ...
	nRandomTries, nRandomBitFlips, sType )

global LARGE_NUMBER

xBest = []; % Initialize
wBest = LARGE_NUMBER;
thres = 1 - nRandomBitFlips/n;

for i = 1:nRandomTries
	flipMask = ( rand( n, 1 ) > thres );
	xTry = xor( xInit, flipMask );
	x = bstr_steepest_desc( nMaxGradIters, xTry, 0, sType, @phi, @phiRowGrad );
	wTry = phi( x );
	if ( wTry < wBest )
		wBest = wTry;
		xBest = x;
	end
end

if ( wBest == LARGE_NUMBER ), error( 'need to increase LARGE_NUMBER' ), end;

%...............................................................................
% The objective function to minimize, which is:
% phi( x ) = (x - xML)' * (A'*A) (x - xML) = | A(x - xML) |^2
%...............................................................................
function rc = phi( x )
global ATA xML
rc = ( x - xML )' * ATA * ( x - xML );

%...............................................................................
% The row gradient of the objective function above
%...............................................................................
function rc = phiRowGrad( x )
global ATA xML
rc = 2 * ( x - xML )' * ATA; 
