%
% PSF for the vertical cantilever. See the report "Point Response Function of
% an MRFM Tip" by Mamin et al.
%
% p = psf_vert( x, y, z, Bext, Bres, m )
%    (x, y, z) - position of spin relative to tip centre
%    p_Bext    - externally applied field [Gauss]
%    p_Bres    - resonant field = gamma*omega_rf [Gauss]
%    p_m       - total moment of the tip = 4*pi/3*M*R_0^3, where M is the
%                magnetization density (800 emu/cm^3) and R_0 is the tip radius
%                (emu)
%    p_xpk     - cantilever peak oscillation amplitude [cm]
%
function p = psf_vert( x, y, z, p_Bext, p_Bres, p_m, p_xpk )
global Bext Bres m
Bext = p_Bext;
Bres = p_Bres;
m = p_m;

sparam = s( x, y, z );
if ( sparam >= p_xpk )
	p = 0;
else
	Gparam = G( x, y, z );
	% XXX: don't normalize by G_0
	p = Gparam^2 * ( 1 - sparam^2/p_xpk^2 );
end

function r = Bmag( x, y, z )
global Bext m
d5halfs = (x^2 + y^2 + z^2)^(5/2);
a1 = 3*x*z*m/d5halfs;
a2 = 3*y*z*m/d5halfs;
a3 = (2*z^2 - x^2 - y^2)*m/d5halfs + Bext;
r = sqrt( a1^2 + a2^2 + a3^2 );

function r = G( x, y, z )
global Bext m

d5 = (x^2 + y^2 + z^2)^5;
d6 = d5 * (x^2 + y^2 + z^2);
d5halfs = sqrt( d5 );
d7halfs = d5halfs * (x^2 + y^2 + z^2);
n1 = -1*x^2 - y^2 + 2*z^2;

a1 = -90*m^2*x^3*z^2/d6 - 90*m^2*x*y^2*z^2/d6 + 18*m^2*x*z^2/d5;
a2 = -2*m*x/d5halfs - 5*m*x*n1/d7halfs;
a3 = Bext + m*n1/d5halfs;
a4 = 9*m^2*x^2*z^2/d5 + 9*m^2*y^2*z^2/d5;
a5 = (Bext + m*n1/d5halfs)^2;

r = (a1 + 2*a2*a3) / 2 / sqrt( a4 + a5 );

function r = s( x, y, z )
global Bres
r = ( Bres - Bmag( x, y, z ) ) / G( x, y, z );
