%-------------------------------------------------------------------------------
% P = rw_anal2( snrdB, bPlot, bMagic )
%    - snrdB = snr in dB
%    - bPlot controls whether to plot or not
%    - bMagic usage: see source
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function P = rw_anal2( snrdB, bPlot, bMagic )

if ( nargin ~= 3 )
	error( 'missing arguments' );
end

% Step size
s = 1.58114e-5;
% Number of positive levels
M = 58;
%M = 5;
% Number of states
r = 2*M + 1;
% Column vector of all 1s
vOnes = ones( r, 1 );
% Initial probability distribution
vPi = (1/r) * ones( r, 1 );
% Number of values of n to examine when looking at f1(n) and f2(n)
maxCrossDiff = 100;
epsFudgeFactor = 30;

rwK1 = 0.45;
rwK2 = 0.55;
if ( bMagic == 0 )
	disp( '   XXX: creating symmetric prob trans matrix' );
	rwH2 = rwK1;
	rwH1 = rwK2;
else
	disp( '   XXX: creating non-symmetric prob trans matrix' );
	rwH1 = rwK1;
	rwH2 = rwK2;
end

% Vector of possible states
psiRange = s * [-M:M]';

% Create Prw: the transition probability matrix of the Random Walk
matP = create_fixed_Prw( M, rwK1, rwK2, rwH1, rwH2 );
%Prw = create_rand_Prw( M, 'bounce', 'randomNoloop' );
%Prw = create_rand_Prw( M, 'bounce', 'symmetricNoloop' );
P = matP;

% Find the steady-state probability densities
[matPEigenvecs, matPEigenvals] = eig( matP );
[PEigenvalsSorted, PEigenvalsSortedIndex] = sort( real(diag(matPEigenvals)) );
disp( 'Eigenvalues of P:' );
compactArrayPrint( PEigenvalsSorted );
biggestEigenval = PEigenvalsSorted(r);
ssDist = matPEigenvecs( :, PEigenvalsSortedIndex(r) );
% Normalize ssDist so that the elements add to 1
ssDist = ssDist / sum( ssDist );
disp( sprintf( 'Biggest eigenvalue of P is %g', biggestEigenval ) );
% Sanity on the biggest eigenvalue: it should be 1
if ( abs( biggestEigenval - 1 ) > eps*epsFudgeFactor )
	warning( 'sanity check failed' );
	disp( sprintf( 'biggestEigenval=%g', biggestEigenval ) );
	error( '[1] bailing out...' );
end

if ( bPlot == 1 )
	figure( 1 ); %%% Figure 1
	subplot( 211 ), plot( [-M:M], ssDist, 'bo:' ); 
	xlabel( 'state' ), ylabel( 'steady-state probability' );
	% Adjust y-axis so that it's from 0 to the max value
	figAxis = axis;
	axis( [figAxis(1) figAxis(2) 0 figAxis(4)] );
else
	disp( 'Skipping plot' );
	disp( 'Steady-state distribution:' );
	compactArrayPrint( ssDist );
end

%-------------------------------------------------------------------------------
% From the steady-state probability values, compute the value of sigma for
% the AWGN noise. Define SNR = lim_{i->\infty} E[\mu_i^2]/\sigma^2, where
% \mu_i is the random walk process.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
signalPower = sum( (psiRange.^2) .* ssDist );
snr = 10^(snrdB/10);
sigma = sqrt( signalPower/snr );
disp( sprintf( 'signalPower=%g, snr=%g dB --> sigma=%g', ...
	signalPower, snrdB, sigma ) );

matMpsi = diag( psiRange );
matMpsipsi = diag( exp( -psiRange.^2/(2*sigma^2) ) );
matQ = matP * matMpsipsi;
matR = matP * matMpsipsi * matMpsi;
[matQEigenvecs,matQEigenvals] = eig( matQ );
[QEigenvalsSorted, QEigenvalsSortedIndex] = sort( real(diag(matQEigenvals)) );
matLambdaQ = diag( QEigenvalsSorted(end:-1:1) );
matUQ = zeros( r, r );
for i = 1:r
	matUQ( :, i ) = matQEigenvecs( :, QEigenvalsSortedIndex(r+1-i) );
end

% Check that Q = U_Q * Lambda_Q * (U_Q)^{-1}
matUQInverse = inv( matUQ );
matQCheck = matUQ * matLambdaQ * matUQInverse;
normCheck = norm( matQ - matQCheck, 1 );
if ( normCheck > eps*epsFudgeFactor*r )
	warning( 'matQ is not equal to matQCheck' );
	disp( sprintf( 'normCheck=%g', normCheck ) );
	error( '[2] bailing out...' );
else
	disp( '   ***matQCheck passed***' );
end

matRTilde = matUQInverse * matR * matUQ;
d = matUQInverse * vOnes;
vPiTilde = (vPi' * matMpsipsi * matUQ)';
kappa = real( diag( matLambdaQ ) );
kappaPrime = kappa / kappa(1);
disp( 'kappa:' );
compactArrayPrint( kappa );

foo = 2 / log( 1 / kappa(1) );
disp( sprintf( 'XXX: foo=%g', foo ) );

if ( bPlot == 1 ) % Plot eigenvalues of P and Q
	figure( 2 ); %%% Figure 2
	subplot( 211 ), plot( sort(matPEigenvals), zeros( r, 1 ), 'ko' );
	ylabel( 'eigvals of P' );
	axis( [ 0.99, 1.01, -0.1, 0.1 ] );
	subplot( 212 ), plot( sort(matQEigenvals), zeros( r, 1 ), 'kx' );
	ylabel( 'eigvals of Q' );
	axis( [ 0.99, 1.01, -0.1, 0.1 ] );
end

vF1 = zeros( maxCrossDiff, 1 );
vF2 = zeros( maxCrossDiff, 2 );
vF1Plus = zeros( maxCrossDiff, 1 );
vF1Minus = zeros( maxCrossDiff, 1 );

for n = 1:maxCrossDiff
	vF1Plus(n) = f1Plus( n, vPiTilde, matRTilde, d, kappaPrime, r );
	vF1Minus(n) = f1Minus( n, vPiTilde, matRTilde, d, kappaPrime, r );
	vF1(n) = min( [vF1Plus(n), vF1Minus(n)] );
	vF2(n) = f2( n, vPiTilde, matRTilde, d, kappaPrime, r );
end

if ( bPlot == 1 )
	xRange = [1:maxCrossDiff];

	figure( 1 );
	subplot( 212 ), plot( xRange, vF1Plus, 'k+', xRange, vF1Minus, 'mo' );
	xlabel( 'n' ), ylabel( 'f_1^+(n)/black, f_1^-(n)/magenta' );

	figure( 3 ); %%% Figure 3
	subplot( 121 ), plot( xRange, vF1, 'b*', xRange, vF2, 'r+' );
	xlabel( 'n' ), ylabel( 'f_1(n)/blue, f_2(n)/red' );
	subplot( 122 ), semilogy( xRange, vF1, 'b*', xRange, vF2, 'r+' );
	xlabel( 'n' ), ylabel( 'semilog scale' );
end

%------------------------------
function compactArrayPrint( x )
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
sTmp = sprintf( '%.9g, ', x(1:(end-1)) );
fprintf( sTmp );
sTmp = sprintf( '%.9g\n', x(end) );
fprintf( sTmp );

%--------------------------------------------------------------
function y = f1Plus( n, vPiTilde, matRTilde, d, kappaPrime, r )
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
c1 = 0; % Component 1
c2 = 0; % Component 2
for l = 1:r
	c1 = c1 + vPiTilde(1)*matRTilde(1,l)*matRTilde(l,1)*d(1)*...
		(kappaPrime(l))^n;
	c2 = c2 + vPiTilde(r)*matRTilde(r,r+1-l)*matRTilde(r+1-l,r)*d(r)*...
		(kappaPrime(l))^n;
end
y = c1 + c2;

%---------------------------------------------------------------
function y = f1Minus( n, vPiTilde, matRTilde, d, kappaPrime, r )
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
c1 = 0; % Component 1
c2 = 0; % Component 2
for l = 1:r
	c1 = c1 + vPiTilde(1)*matRTilde(1,l)*matRTilde(l,1)*d(1)*...
		(kappaPrime(l))^n;
	c2 = c2 + vPiTilde(r)*matRTilde(r,r+1-l)*matRTilde(r+1-l,r)*d(r)*...
		(kappaPrime(l))^n;
end
y = c1 - c2;

%----------------------------------------------------------
function y = f2( n, vPiTilde, matRTilde, d, kappaPrime, r )
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
c1 = 0; % Component 1
c2 = 0; % Component 2
for l = 1:r
	c1 = c1 + vPiTilde(1)*matRTilde(1,l)*matRTilde(l,r)*d(r)*...
		(kappaPrime(l))^n;
	c2 = c2 + vPiTilde(r)*matRTilde(r,l)*matRTilde(l,1)*d(1)*...
		(kappaPrime(l))^n;
end
s1 = abs( c1 + c2 );
s2 = abs( c1 - c2 );
%y = max( [s1, s2] );	
y = abs(c1) + abs(c2);
