% Function to compute the cumulative coherence function. This is more
% compute intensive than the coherence function.
% Inputs: A, m
% Outputs: A_mu1
function A_mu1 = dict_cumcoherence( A, m )

[ M, N ] = size( A );

% Check that 1 <= m <= N
if ( ( m < 1 ) | ( m > N ) )
    error( 'invalid value for m' );
end

% This part is the same as dict_coherence
warning( 'dict_coherence: *assuming* that columns of A have unit norm' );

global Acorr

% Actually, don't need the columns to have unit l_2 norm because corr
% handles that.
Acorr = corr( A );

if ( ( size( Acorr, 1 ) ~= N ) | ( size( Acorr, 2 ) ~= N ) )
    warning( 'corr returned unexpected result' );
end

% The way that we'll enumerate through the first maximization is to form a
% binary vector of length N that has the first m values set to 1, and the
% rest are 0s. Then, we start moving the right-most 1 all the way to the
% right while leaving the other 1s alone. After there's a 1 in the
% right-most position of the N-vector start moving the 2nd right-most 1.
% And so forth.

binaryColSelVec = zeros( 1, N );
binaryColSelVec( 1:m ) = 1;

A_mu1 = 0; % Initialize

while ( true ) 
    
    % SANITY: check that there are only m 1s in binaryColSelVec
    if ( sum( binaryColSelVec ) ~= m )
        error( 'sanity on binaryColSelVec failed' );
    end
    
    test_value = max_over_psi( binaryColSelVec );
    if ( test_value > A_mu1 ) 
        A_mu1 = test_value;
    end
    
    % Next, shift the 1s to the next position
    [ bDone, nextVec ] = binvec_next( binaryColSelVec, m );
    
    if ( bDone ), break;
    else, binaryColSelVec = nextVec;
    end
    
end

% After the end of the for loop above, A_mu1 should be correct

% Once the set Lambda is fixed, compute the value over the second
% maximization.
%////////////////////////////////////////////////////////////////
function rc = max_over_psi( binaryColSelVec )
global Acorr

rc = 0; % Initialize return value
N = length( binaryColSelVec );
colSelVec = find( binaryColSelVec == 1 );

for colInd = 1:N
    if ( binaryColSelVec( colInd ) == 0  )
        rcTest = sum( abs( Acorr( colInd, colSelVec ) ) );
        if ( rcTest > rc )
            rc = rcTest;
        end
    end
end

return;

% Shift the vector binaryColSelVec to the next position. Returns true if
% this is the last one. We also return the empty vector for nextVec
%///////////////////////////////////////////////////////////////////////
function [ bDone, nextVec ] = binvec_next( binaryColSelVec, m )

N = length( binaryColSelVec );
colSelVec = find( binaryColSelVec == 1 );

if ( length( colSelVec ) ~= m )
    error( 'unexpected length for colSelVec' );
end

bDone = false; % Initialize
nextVec = []; 

% Check to see if all the 1s are stacked to the right of binaryColSelVec.
% If so, then we are done and there is nothing to do.
if ( sum( colSelVec == (N-m+1):N ) == m )
    bDone = true;
    return;
end

bFound = false; % Extra sanity flag

% Locate the right-most 1 not at the end of binaryColSelVec
for testInd = m:-1:1
    if ( colSelVec( testInd ) ~= ( N - m + testInd ) )
        bFound = true;
        break;
    end
end

if ( ~bFound )
    error( 'failed to located right-most 1 to shift' );
end

colSelVec( testInd ) = colSelVec( testInd ) + 1; % Increase by 1

nextVec = zeros( size( binaryColSelVec ) );
nextVec( colSelVec ) = 1;

return;