% // 2008-06-26 // Landweber iteration with decimation. Based on
%                  explandweb.m
% Keys: psf, psfshift, y, x0, addnvar, maxiter, l2normstop, interp_fac,
%       zlices, offset, zslicemin, zslicemax, psfmat, bverbose, bnonneg
% Output: sLandweb
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function sLandweb = explandwebdec( varargin )

% Defaults
%/////////

psfMat          = [];
GPsf            = [];
y               = [];
x0              = [];

psfShiftAmount  = [];

interpFactor    = [];
interpFactorLen = 3;

offset          = [];
offsetLen       = 3;

zSlices         = [];

addNoiseVar     = 0;
maxIter         = 1e4;
l2NormStop      = 0;

yDeltaNorm      = [];
xDeltaNorm      = [];

yDeltaNormSmallest = -1;

c               = 0;
cScaleFactor    = 0.5;
cGiven          = -1;

zSliceMax       = -1;
zSliceMin       = -1;

bVerbose        = true;
bNonNeg         = false;

sLowestObsErr   = struct([]);

% Parse arguments
%////////////////
argNum = floor( length( varargin ) / 2 );

for argInd = 1:argNum
    szKey = varargin{ 2*argInd-1 };
    ip = varargin{ 2*argInd };
    switch szKey
        case 'psf', GPsf = ip;
        case 'psfshift', psfShiftAmount = ip;
        case 'psfmat', psfMat = ip;
        case 'y', y = ip;
        case 'x0', x0 = ip;
        case 'addnvar', addNoiseVar = ip;
        case 'maxiter', maxIter = ip;
        case 'l2normstop', l2NormStop = ip;
        case 'interp_fac', interpFactor = ip; 
        case 'offset', offset = ip;
        case 'zslicemax', zSliceMax = ip;
        case 'zslicemin', zSliceMin = ip;
        case 'bverbose', bVerbose = logical( ip );
        case 'bnonneg', bNonNeg = logical( ip );
        case 'c', cGiven = ip;
        otherwise, error( sprintf( 'Unrecognized key: %s', szKey ) );
    end
end

% Check arguments
%////////////////

if ( l2NormStop == 0 )
    l2NormStop = norm( y(:), 2 ) * 1e-2;
end

if ( length( interpFactor ) ~= interpFactorLen )
    disp( 'interpFactor:' ), disp( interpFactor );
    error( sprintf( 'expected length of interpFactor to be %g', ...
        interpFactorLen ) );
end

if ( length( offset ) ~= offsetLen )
    disp( 'offset:' ), disp( offset );
    error( sprintf( 'expected length of offset is %g', ...
        offsetLen ) );
end    
    
xCurr = x0;   
xCurr = zero_zslices( xCurr, zSliceMin, zSliceMax );
    
% Assume that the l2 norm of the linear operator H is unchanged by the
% downsampling and interpolation. 
psfDft = fftn( psfMat );
psfDftProp = emsep_comp_alpha( psfDft, sqrt( addNoiseVar ), 1 );
cComp = (cScaleFactor * psfDftProp.alphaUpperBound)^2 / addNoiseVar;
fprintf( 1, 'computed c=%g\n', cComp ); 

if ( cGiven < 0 )
    c = cComp;    
else
    c = cGiven;
end

fprintf( 1, 'using c=%g\n', c );

% yEstRough should have the same dimensions as xCurr. In each xy plane, we
% will be downsampling. Compute the max multiple in the x & y directions.

maxMult = floor( size( xCurr ) ./ interpFactor );

bStop = false;

for nIter = 1:maxIter
    
    % Form Hx
    yEstRough = GPsf * xCurr;
    
    if ~isempty( psfShiftAmount ) % If provided, shift
        yEstRough = circshift( yEstRough, psfShiftAmount );
    end
    
    yEstInterp = mydecimate3( yEstRough, interpFactor, 3*interpFactor );
    
    try
        yComp = yEstInterp( offset(1)+[1:size(y,1)], ...
            offset(2)+[1:size(y,2)], ...
            offset(3)+[1:size(y,3)] );
        yDelta = y - yComp;
    catch
        error( sprintf( 'Cannot get yComp, error msg is: %s', lasterr ) );
    end

    yDeltaNorm = [ yDeltaNorm, norm( yDelta(:), 2 ) ]; % Keep track
    
    if ( yDeltaNormSmallest < 0 )
        yDeltaNormSmallest = yDeltaNorm( end ); % Initialize
    end
    
    if ( 0 & ( yDeltaNorm(end) < l2NormStop ) )
        fprintf( 1, '\tyDeltaNorm < l2NormStop\n' );
        bStop = true;
    end
    
    HtyDelta = zeros( size( xCurr ) );
       
    % Form H'(y-Hx). Do not interpolate but rather just upsample.
    
    if ( true )
        HtyDelta( interpFactor(1)*( offset(1)+[1:size(y,1)] ), ...
            interpFactor(2)*( offset(2)+[1:size(y,2)] ), ...
            interpFactor(3)*( offset(3)+[1:size(y,3)] ) ) = yDelta;
    else
        for hInd = 1:3
            hIndRange{ hInd } = interpFactor( hInd ) * ...
                ( offset( hInd ) + [1, size( y, hInd )] );
        end
        HtyDelta( hIndRange{1}(1): hIndRange{1}(2), ...
            hIndRange{2}(1): hIndRange{2}(2), ...
            hIndRange{3}(1): hIndRange{3}(2) ) = myinterp3( yDelta, interpFactor );
    end
    
    HtyDelta = GPsf' * HtyDelta;
    
    % In this step, compute x + cH'(y-Hx)
    xDelta = c * HtyDelta;
    xDeltaNorm = [ xDeltaNorm, norm( xDelta(:), 2 ) ]; % Keep track
    xNext = xCurr + xDelta;

    if ( bVerbose )
        fprintf( 1, '\titer %5g: |yDelta|_2=%g, |xDelta|_2=%g\n', ...
            nIter, yDeltaNorm( end ), xDeltaNorm( end ) );
    end
    
    if ( yDeltaNormSmallest >= yDeltaNorm( end ) )
        sLowestObsErr = struct( 'x', xCurr, ...
            'yRough', yEstRough, ...
            'y', yEstInterp, ...
            'yDeltaNorm', yDeltaNorm( end ) );
        yDeltaNormSmallest = yDeltaNorm( end );
    end
    
    % Replace xCurr with xNext
    xNextZeroed = zero_zslices( xNext, zSliceMin, zSliceMax );
        
    if ( bVerbose )
        fprintf( 1, '|xCurr|_2=%g, |xNext|_2=%g, |xNextZeroed|_2=%g\n', ...
            norm( xCurr(:), 2 ), norm( xNext(:), 2 ), norm( xNextZeroed(:), 2 ) );
    end
        
    xCurr = xNextZeroed;
                    
    % Enforce non-negativity on the image?
    if ( bNonNeg )
        xCurr = xCurr .* ( xCurr > 0 );
    end
    
    if ( bStop ), break, end
    
end

sLast = struct( 'x', xCurr, ...
    'yRough', yEstRough, ...
    'y', yEstInterp, ...
    'yDeltaNorm', yDeltaNorm( end ) );

sLandweb = struct( 'lowestObsErr', sLowestObsErr, 'last', sLast, ...
    'xDeltaNorm', xDeltaNorm, 'yDeltaNorm', yDeltaNorm, 'cComp', cComp, ...
    'c', c );
  




        