% // 2008-05-30 // Landweber iteration with downsampling.
%
% Keys: psf, psfshift, y, x0, addnvar, maxiter, l2normstop, interp_fac,
%       zlices, offset, zslicemin, zslicemax, psfmat, bverbose, bnonneg
% Output: sLandweb
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function sLandweb = explandwebds( varargin )

% Defaults
%/////////

psfMat              = [];
GPsf                = [];
y                   = [];
x0                  = [];

psfShiftAmount      = [];

interpFactor        = [];
interpFactorLen     = 2;
offset              = [];
offsetLen           = 2;

addNoiseVar         = 0;
maxIter             = 1e4;
l2NormStop          = 0;

yDeltaNorm          = [];
xDeltaNorm          = [];
yDeltaNormSmallest  = -1;

c                   = 0;
cScaleFactor        = 0.5;
cGiven              = -1;

zSlices             = [];
zSliceMax           = -1;
zSliceMin           = -1;

bVerbose            = true;
bNonNeg             = false;
bPlot               = true;
nPlotMult           = 30;

sLowestObsErr       = struct([]);

xMinThres           = 1e2*eps;

% Parse arguments
%////////////////
argNum = floor( length( varargin ) / 2 );

for argInd = 1:argNum
    szKey = varargin{ 2*argInd-1 };
    ip = varargin{ 2*argInd };
    switch szKey
        case 'psf', GPsf = ip;
        case 'psfshift', psfShiftAmount = ip;
        case 'psfmat', psfMat = ip;
        case 'y', y = ip;
        case 'x0', x0 = ip;
        case 'addnvar', addNoiseVar = ip;
        case 'maxiter', maxIter = ip;
        case 'l2normstop', l2NormStop = ip;
        case 'interp_fac', interpFactor = ip; 
        case 'offset', offset = ip;
        case 'zslices', zSlices = ip;
        case 'zslicemax', zSliceMax = ip;
        case 'zslicemin', zSliceMin = ip;
        case 'bverbose', bVerbose = logical( ip );
        case 'bnonneg', bNonNeg = logical( ip );
        case 'bplot', bPlot = logical( ip );
        case 'c', cGiven = ip;
        otherwise, error( sprintf( 'Unrecognized key: %s', szKey ) );
    end
end

% Check arguments
%////////////////

if ( l2NormStop == 0 )
    l2NormStop = norm( y(:), 2 ) * 1e-2;
end

if ( length( interpFactor ) ~= interpFactorLen )
    disp( 'interpFactor:' ), disp( interpFactor );
    error( sprintf( 'expected length of interpFactor to be %g', ...
        interpFactorLen ) );
end

if ( length( offset ) ~= offsetLen )
    disp( 'offset:' ), disp( offset );
    error( sprintf( 'expected length of offset is %g', ...
        offsetLen ) );
end    

if ( isempty( zSlices ) )
    error( 'zSlices is empty' );
end

xCurr = x0;   
    
if ( cGiven < 0 )
    
    % Assume that the l2 norm of the linear operator H is unchanged by the
    % downsampling and interpolation.
    psfDft = fftn( psfMat );
    psfDftProp = emsep_comp_alpha( psfDft, sqrt( addNoiseVar ), 0 );
    
    cComp = (cScaleFactor * psfDftProp.alphaUpperBound)^2 / addNoiseVar;
    fprintf( 1, 'computed c=%g\n', cComp );
    c = cComp;
    
else
    c = cGiven;
end

fprintf( 1, 'using c=%g\n', c );

% yEstRough should have the same dimensions as xCurr. In each xy plane, we
% will be downsampling. Compute the max multiple in the x & y directions.

maxMult = [ floor( size( xCurr, 1 )/interpFactor( 1 ) ), ...
            floor( size( xCurr, 2 )/interpFactor( 2 ) ) ];

xSel = interpFactor(1) * ( offset(1) + [1:size(y,1)] )
ySel = interpFactor(2) * ( offset(2) + [1:size(y,2)] )

% The downsampling need not start from the beginning. Ideally, we position
% the downsampled points in the "middle". Ensure that there are enough
% points to support this.
if ( ( xSel(end)/interpFactor(1) > maxMult(1) ) | ...
     ( ySel(end)/interpFactor(2) > maxMult(2) ) )
    error( '{x,y}Sel out of bounds' );
end    

bStop = false;

for nIter = 1:maxIter
    
    % Form Hx
    yEstRough = GPsf * xCurr;
    
    if ( bPlot & ( mod( nIter, nPlotMult ) == 0 ) )
        xysliceplot( 'x', yEstRough, 'nfig', 1, 'b_pause', true, ...
            'sz_title', 'yEstRough' );
    end    
    
    yEstInterp = mydownsamp3( yEstRough, xSel, ySel, zSlices );
    
    if ( bPlot & ( mod( nIter, nPlotMult ) == 0 ) )
        xysliceplot( 'x', yEstInterp, 'nfig', 2, 'b_pause', true, ...
            'sz_title', 'yEstInterp' );
    end

    if ( any( size( yEstInterp ) ~= size( y ) ) )        
        error( 'incompatible sizes for yEstInterp and y' );
    end

    yDelta = y - yEstInterp;
    yDeltaNorm = [ yDeltaNorm, norm( yDelta(:), 2 ) ]; % Keep track
    
    if ( yDeltaNormSmallest < 0 )
        yDeltaNormSmallest = yDeltaNorm( end ); % Initialize
    end
    
    if ( bPlot & ( mod( nIter, nPlotMult ) == 0 ) )
        xysliceplot( 'x', yDelta, 'nfig', 3, 'b_pause', true, ...
            'sz_title', 'yDelta' );
    end
    
    if ( 0 & ( yDeltaNorm(end) < l2NormStop ) )
        fprintf( 1, '\tyDeltaNorm < l2NormStop\n' );
        bStop = true;
    end
    
    HtyDelta = zeros( size( xCurr ) );
    
    % Form H'(y-Hx)
    for zInd = 1:length( zSlices )
        HtyDelta( xSel, ySel, zSlices(zInd) ) = HtyDelta( xSel, ySel, zSlices(zInd) ) + ...
            yDelta( :, :, zInd );
    end
    
    HtyDelta = GPsf' * HtyDelta;
    
    if ( bVerbose )
        fprintf( 1, '\t  |yDelta|_2=%g, |H^T*yDelta|_2=%g, ratio=%g\n', ...
            norm( yDelta(:), 2 ), norm( HtyDelta(:), 2 ), ...
            norm( HtyDelta(:), 2 ) / norm( yDelta(:), 2 ) );
    end
    
    % In this step, compute x + cH'(y-Hx)
    xDelta = c * HtyDelta;
    xDeltaNorm = [ xDeltaNorm, norm( xDelta(:), 2 ) ]; % Keep track
    xNext = xCurr + xDelta;

    if ( bPlot & ( mod( nIter, nPlotMult ) == 0 ) )
        xysliceplot( 'x', circshift( xDelta, -2*psfShiftAmount ), ...
            'nfig', 4, 'b_pause', true, 'sz_title', 'xDelta', 'b_hist', true );
        xysliceplot( 'x', circshift( xNext, -2*psfShiftAmount ), ...
            'nfig', 5, 'b_pause', true, 'sz_title', 'xNext', 'b_hist', true );
    end
    
    if ( bVerbose )
        fprintf( 1, '\titer %5g: |yDelta|_2=%g, |xDelta|_2=%g\n', ...
            nIter, yDeltaNorm( end ), xDeltaNorm( end ) );
    end
    
    if ( yDeltaNormSmallest >= yDeltaNorm( end ) )
        sLowestObsErr = struct( 'x', xCurr, ...
            'yRough', yEstRough, ...
            'y', yEstInterp, ...
            'yDeltaNorm', yDeltaNorm( end ) );
        yDeltaNormSmallest = yDeltaNorm( end );
    end
    
    % Replace xCurr with xNext
    xCurr = xNext;
    if false, xCurr = zero_zslices( xCurr, zSliceMin, zSliceMax ); end
    
    % Enforce non-negativity on the image?
    if ( bNonNeg )
        xCurr = xCurr .* ( xCurr > 0 );
    end
    
    % Do a hard threshold of small values
    xCurr = ( xCurr ) .* ( abs( xCurr ) > xMinThres );
    
    if ( bStop ), break, end
    
end

sLast = struct( 'x', xCurr, ...
    'yRough', yEstRough, ...
    'y', yEstInterp, ...
    'yDeltaNorm', yDeltaNorm( end ) );

sLandweb = struct( 'lowestObsErr', sLowestObsErr, 'last', sLast, ...
    'xDeltaNorm', xDeltaNorm, 'yDeltaNorm', yDeltaNorm, 
    'c', c );
  




        