function labelplot( p , o , str )
%LABELPLOT Creates an automatic legend based on 'Tag'
%
%  For every plotted data run, set the 'Tag' string to the label of the data.
%  E.g.:     p1 = plot(x,y);
%            p2 = plot(a,b,'Tag','A vs. B');
%            p3 = plot(c,d,'Tag','C vs. D');
%  Calling LABELPLOT will create a legend equivalent to:
%            legend([p2 p3],'A vs. B','C vs. D')
%
%  Optional additional arguments 1 & 2 specify legend attributes: 
%  (see `help legend' for more information)
%    labelplot( <Location> , <Orientation> )
%
%  NEW and IMPROVED!
%  Optional argument 3 specifies a legend label:
%    labelplot( ... , ... , <label string> )

%% Set up default legend placement
% These can/should be adjusted to suit.
if nargin < 1
  p = 'northoutside';
end
if nargin < 2
  if isequal(p,'northoutside')
    o = 'horizontal';
  else
    o = 'vertical';
  end
end
if nargin < 3
  str = '';
end

%% Grab the data plots and their 'Tag' strings, make the legend
% For empty 'Tag's, delete the corresponding data plot from the
% legend index.
ch = get(gca,'Children');
strings = get(ch,'Tag');
legend_index = ones(size(ch));
if length(ch) > 1
  for ii = 1:length(ch)
    if length(strings{ii}) == 0
      legend_index(ii) = 0;
    end
  end	
  ch = ch(find(legend_index));
  strings = strings(find(legend_index));
  % Turns out that legends are created first in, last out with
  % respect to the ordering of the respective axes children vector.
  % So we reverse the order to get it back to normal.
  legend_h = legend(ch(end:-1:1),strings(end:-1:1),'location',p,'orientation',o);
elseif length(strings) > 0
  legend_h = legend(ch,strings,'location',p,'orientation',o);
end

% Don't put a box around the legend:
legend boxoff

%% Place the legend label
%
% The font and fontsize is the same as the legend, which is inherited from
% the figure. This can be altered by something like
%    set(findobj('Tag','legendtitle'),'FontSize',14)
% but note that savefig.m will overwrite this kind of thing.

if exist('legend_h','var')==1 & ~isequal(str,'')
  fontname = get(legend_h,'FontName');
  fontsize = get(legend_h,'FontSize');
  legendtitle = text(0.5,1,str,...
    'Parent',legend_h,...
    'HorizontalAlignment','center',...
    'VerticalAlignment','bottom',...
    'Tag','legendlabel',...
    {'FontName','FontSize'},{fontname,fontsize});
end

% Copyright 2006 Will Robertson
%
% This MATLAB function may distributed and/or modified
% under the conditions of the LaTeX Project Public License, 
% version 1.3c or higher (at your discretion). 
%   <http://www.latex-project.org/lppl.txt>
% It consists of the file labelplot.m and is currently
% maintained by Will Robertson.
