% // 2008-06-29 // Experiment to determine the l_2 norm of the psf
% Keys: psf, psfmat, x0, interp_fac, offset, bverbose, maxiter, addnvar, y
function sPsfProp = psf_l2norm( varargin )

% Defaults
%/////////

psfMat          = [];
GPsf            = [];
y               = [];
x0              = [];

bVerbose        = true;

interpFactor    = [];
interpFactorLen = 2;

offset          = [];
offsetLen       = 2;

addNoiseVar     = 0;
maxIter         = 1e3;
xNormHist       = [];
yNormHist       = [];

zSlices         = [];
zSliceMax       = -1;
zSliceMin       = -1;

% Parse arguments
%////////////////
argNum = floor( length( varargin ) / 2 );

for argInd = 1:argNum
    szKey = varargin{ 2*argInd-1 };
    ip = varargin{ 2*argInd };
    switch szKey
        case 'psf', GPsf = ip;
        case 'psfmat', psfMat = ip;
        case 'x0', x0 = ip;
        case 'y', y = ip;
        case 'interp_fac', interpFactor = ip; 
        case 'offset', offset = ip;
        case 'bverbose', bVerbose = logical( ip );
        case 'maxiter', maxIter = ip;
        case 'addnvar', addNoiseVar = ip;
        case 'zslices', zSlices = ip;
        case 'zslicemax', zSliceMax = ip;
        case 'zslicemin', zSliceMin = ip;
        otherwise, error( sprintf( 'Unrecognized key: %s', szKey ) );
    end
end

% Check arguments
%////////////////

if ( length( interpFactor ) ~= interpFactorLen )
    disp( 'interpFactor:' ), disp( interpFactor );
    error( sprintf( 'expected length of interpFactor to be %g', ...
        interpFactorLen ) );
end

if ( length( offset ) ~= offsetLen )
    disp( 'offset:' ), disp( offset );
    error( sprintf( 'expected length of offset is %g', ...
        offsetLen ) );
end 

if ( isempty( zSlices ) )
    error( 'zSlices is empty' );
end

% Assume that the l2 norm of the linear operator H is unchanged by the
% downsampling and interpolation. 
psfDft = fftn( psfMat );
psfDftProp = emsep_comp_alpha( psfDft, sqrt( addNoiseVar ), 0 );

if ( bVerbose )
    fprintf( 1, '\tmax l_2 norm of psfMat is %g\n', psfDftProp.norm2 );
end

xCurr = x0;     % Dummy variable. We don't really use it, but merely 
                % extract its size.

% yEstRough should have the same dimensions as xCurr. In each xy plane, we
% will be downsampling. Compute the max multiple in the x & y directions.

maxMult = [ floor( size( xCurr, 1 )/interpFactor( 1 ) ), ...
            floor( size( xCurr, 2 )/interpFactor( 2 ) ) ];

xSel = interpFactor(1) * ( offset(1) + [1:size(y,1)] )
ySel = interpFactor(2) * ( offset(2) + [1:size(y,2)] )

% The downsampling need not start from the beginning. Ideally, we position
% the downsampled points in the "middle". Ensure that there are enough
% points to support this.
if ( ( xSel(end)/interpFactor(1) > maxMult(1) ) | ...
     ( ySel(end)/interpFactor(2) > maxMult(2) ) )
    error( '{x,y}Sel out of bounds' );
end   

% First, find lower bound for |H|_2
% fprintf( 1, '\tGenerating random points for |H|_2...\n' );

xNormHist = zeros( 1, maxIter );
yNormHist = zeros( size( xNormHist ) );

% for nIter = 1:maxIter
%     
%     % Form Hx
%     xTry = randn( size( xCurr ) );
%     xTry = zero_zslices( xTry, zSliceMin, zSliceMax );
%     
%     yEstRough = GPsf * xTry;
%     
%     % Downsample
%     yEstDownsamp = mydownsamp3( yEstRough, xSel, ySel, zSlices );
%     
%     xNormHist( nIter ) = norm( xTry(:), 2 );
%     yNormHist( nIter ) = norm( yEstDownsamp(:), 2 );
%     
% end
% 
% randNorm2 = max( yNormHist ./ xNormHist );
% 
% if ( bVerbose )
%     fprintf( 1, '\trandNorm2(H) = %g\n', randNorm2 );
% end

randNorm2 = -1;

sPsfProp = struct( 'maxIter', maxIter, 'dftProp', psfDftProp, ...
    'dftNorm2', psfDftProp.norm2, 'HxNormHist', xNormHist, ...
    'HyNormHist', yNormHist, 'HrandNorm2', randNorm2 );

% Next, get lower bound for |H'|_2
fprintf( 1, '\tGenerating random points for |H^T|_2...\n' );

for nIter = 1:maxIter
    
    % Form H'y
    yTry = randn( size( y ) );
    % XXX there's no equivalent of zero_zslices for yTry. HMM: I guess this
    % means that we really shouldn't zero out any xy slices then.
    
    xEstUpsamp = zeros( size( x0 ) );
    
    % Upsample
    for zInd = 1:length( zSlices )
        xEstUpsamp( xSel, ySel, zSlices( zInd ) ) = xEstUpsamp( xSel, ySel, zSlices( zInd ) ) + ...
            yTry( :, :, zInd );
    end

    xEstRough = GPsf' * xEstUpsamp;
    
    yNormHist( nIter ) = norm( yTry(:), 2 );
    xNormHist( nIter ) = norm( xEstRough(:), 2 );
    
end

randNorm2 = max( xNormHist ./ yNormHist );

if ( bVerbose )
    fprintf( 1, '\trandNorm2(H^T) = %g\n', randNorm2 );
end

sPsfProp.HtyNormHist = yNormHist;
sPsfProp.HtxNormHist = xNormHist;
sPsfProp.HtrandNorm2 = randNorm2;

