%
% Compute the HHS-SURE criterion.
%
function sOut = comp_hhs_sure( varargin )

global XOrig GX % X and its Gram matrix

% Initialize
y = [];
thetaHat = [];
noiseSigma = -1;
phi1 = -1;
phi2 = -1;
bComputeApprox = 0;

% Process input arguments

lenVarargin = length( varargin );
numArgsIn = floor( lenVarargin/2 );

if 2*numArgsIn ~= lenVarargin, error( 'even ip args expected' ), end

for argInInd = 1:numArgsIn    
    szKey = lower( varargin{2*argInInd-1} );    
    switch szKey        
        case 'y', y = varargin{2*argInInd};                  
        case 'thetahat', thetaHat = varargin{2*argInInd };        
        case 'noisesigma', noiseSigma = varargin{2*argInInd};
        case 'phi1', phi1 = varargin{2*argInInd};
        case 'phi2', phi2 = varargin{2*argInInd};
        case 'bcomputeapprox', bComputeApprox = varargin{2*argInInd};
        otherwise
            error( sprintf( 'unrecognized option %s', szKey ) );
    end
end

if ( abs( phi1 - phi2 ) < 1e3*eps ) | ( ( phi1 - phi2 ) < eps )
    error( 'phi2 < phi1 not satisfied' );
end

y = y(:); % Make column vectors
thetaHat = thetaHat(:);
N = length( y );
M = length( thetaHat );

e = y - XOrig * thetaHat;
rss = sum( e.^2 );

hhsSureVal0 = N*noiseSigma^2 + rss; % Initialize

% Compute approximation
thetaHatThres = soft_thres( thetaHat, phi1 - phi2 );
nnzThetaHatThres = sum( abs( thetaHatThres ) > 1e3*eps );
hhsSureValApprox = hhsSureVal0 + 2*noiseSigma^2*nnzThetaHatThres;

% If we only want the approximate value, return immediately
if bComputeApprox
    sOut = struct( 'hhsSureVal', -1, 'hhsSureValApprox', hhsSureValApprox );
    return;
end

% Otherwise, spend more cpu cycles to compute the exact approximation. 

% Create U matrix
uVec = rect( thetaHat / ( 2 * (phi1-phi2) ) );
U = spdiag( uVec );

% Find the zero-valued elements of thetaHat
thetaHatNzInd = find( abs(thetaHat) < 1e2*eps );
r = length( thetaHatNzInd );

if ( r == 0 ) % No reduction
    disp( 'ALERT! r = 0' );
    K = GX;
    J = -0.5 * U;
else
    % The first thing to do is build the permutation matrices P and Q. They are of
    % size M x M. Only need to build P, since Q = P'.
    P = speye( M, M );
    
    disp( sprintf( 'r = %g', r ) );
    
    for i = 1:r
        % Same thing => no need to perform swap
        if ( i == thetaHatNzInd(i) ), continue, end;
        % Swap the i-th row with the thetaHatNzInd_i-th row of P.
        P( [i,thetaHatNzInd(i)], : ) = P( [thetaHatNzInd(i),i], : );
    end
    
    Kpre = P * GX * P';
    Jpre = -0.5 * P * U * P';
    
    % The matrix K, J is the lower right (M-r) x (M-r) submatrix of the
    % respective pre variable
    
    K = Kpre( (r+1):end, (r+1):end );
    J = Jpre( (r+1):end, (r+1):end );            
end

% Verify that K and J are of sizes (M-r) x (M-r)
if ( size( K, 1 ) ~= (M-r) ) | ( size( K, 2 ) ~= (M-r) ) | ...
   ( size( J, 1 ) ~= (M-r) ) | ( size( J, 2 ) ~= (M-r) )
    error( 'sizes of K and J are incorrect' );
end

hhsSureVal = hhsSureVal0 + 2*noiseSigma^2*trace( K*inv(K+J) );

sOut = struct( 'hhsSureVal', hhsSureVal, 'hhsSureValApprox', hhsSureValApprox );

% Print the two values
disp( sprintf( 'phi1=%g, phi2=%g', phi1, phi2 ) );
disp( sprintf( '=> hhsSureVal=%g, hhsSureValApprox=%g', ...
    hhsSureVal, hhsSureValApprox ) );
pause( 5 );

