% Finds local extrema in (x,y) plot. Usage is:
%   [xLocExt, yLocExt] = find_local_extrema( varargin )
% with options 'x', 'y', 'span', 'sztype'. Note that definition or local 
% min/max is strict, and would not work with NOISY data points. Options for
% sztype are {slope/thres}{Min/Max}.

function [xLocExt, yLocExt] = find_local_extrema( varargin )

% Initialize

x = [];
y = [];

xLocExt = [];
yLocExt = [];
numExtPoints = 0;

% Process input arguments

lenVarargin = length( varargin );
numArgsIn = floor( lenVarargin/2 );

if 2*numArgsIn ~= lenVarargin, error( 'even ip args expected' ), end

for argInInd = 1:numArgsIn
    
    szKey = lower( varargin{2*argInInd-1} );    
    switch szKey        
		case 'x', x = varargin{2*argInInd};
		case 'y', y = varargin{2*argInInd};
		case 'span', nSpan = varargin{2*argInInd};
		case 'sztype', szType = varargin{2*argInInd};
        otherwise
            error( sprintf( 'unrecognized option %s', szKey ) );
    end
    
end

% For slopeMax, slopeMin
% ======================
%
% Find local min/max in the sequence with span given by nSpan. A local min
% x* has span nSpan if the previous nSpan points together with x* are
% monotone decreasing, and x* together with the next nSpan points are
% monotone increasing. A local max is similarly defined.

% For thresMax, thresMin
% ======================
%
% Something less strict. A local min x* has span nSpan if the previous
% nSpan points and the next nSpan points are strictly greater than x*.

% Check arguments

if isempty( x ), x = 1:length(y); end
if length( x ) ~= length( y ), error( 'length of x != length of y' ), end
if nSpan <= 0, error( 'span should be >= 1' ), end
if length( x ) < (2*nSpan + 1), error( 'insufficient data' ), end

if strcmp( szType(1:5), 'slope' ) == 1

    % Define states
    N_STATE_DEC = 0;
    N_STATE_INC = 1;

    % Initialize nState
    if ( y(2) > y(1) ), nState = N_STATE_INC;
    elseif ( y(2) < y(1) ), nState = N_STATE_DEC;
    else
        % Random rule
        if rand > 0.5, nState = N_STATE_INC;
        else, nState = N_STATE_DEC;
        end
    end

    numDecStates = 0;
    numIncStates = 0;
    
    for i = 3:length( y )
        switch ( nState )
            case N_STATE_DEC
                if ( y(i) <= y(i-1) )
                    numDecStates = numDecStates + 1;
                else
                    nState = N_STATE_INC; % Change state
                    numIncStates = 1;
                end

            case N_STATE_INC
                if ( y(i) >= y(i-1) )
                    numIncStates = numIncStates + 1;
                    if strcmp( szType(6:8), 'Min' ) == 1 % Looking for min
                        if ( ( numDecStates >= nSpan ) & ( numIncStates >= nSpan ) )
                            % Point indexed by (i-numIncStates) qualifies as a local min with
                            % span nSpan, so add to list.
                            yLocExt = [ yLocExt, y(i-numIncStates) ];
                            xLocExt = [ xLocExt, x(i-numIncStates) ];
                            numExtPoints = numExtPoints + 1;
                        end
                    end
                else
                    nState = N_STATE_DEC; % Change state
                    numDecStates = 1;
                end
        end
    end
    
end

if strcmp( szType(1:5), 'thres' ) == 1
    
    indStart = nSpan + 1;
    indEnd = length( y ) - nSpan;
    
    if strcmp( szType(6:8), 'Min' ) == 1 % Looking for min
        
        for i = indStart:indEnd
            bComparePre = y((i-nSpan):(i-1)) > y(i);
            bComparePost = y((i+1):(i+nSpan)) > y(i);
            if ( sum( bComparePre ) == nSpan ) & ( sum( bComparePost ) == nSpan )
                % Point indexed by i qualifies as a local min with span
                % nSpan.
                yLocExt = [ yLocExt, y(i) ];
                xLocExt = [ xLocExt, x(i) ];
                numExtPoints = numExtPoints + 1;
            end
        end
        
    end
    
end

disp( sprintf( 'Found %g local extrema points', numExtPoints ) );
