%
% hs_strat2: fix phi1 and change phi2/phi1 from 0 to 1.
%
% The following elements of sParam are expected:
%    X, XColumnNorm, nPhi2, numEmIters
%    bigCHatHist
%    psf, y, noiseSigma, betaTrue
%    phi0, phi0Cost, phi0Beta

function sOut = hs_strat2( varargin )

global XOrig GX % The original X, and the Gram matrix of X = X'*X

ipImgDim = [];
opImgDim = [];

% Process input arguments

lenVarargin = length( varargin );
numArgsIn = floor( lenVarargin/2 );

if 2*numArgsIn ~= lenVarargin, error( 'even ip args expected' ), end

for argInInd = 1:numArgsIn    
    szKey = lower( varargin{2*argInInd-1} );    
    switch szKey        
        case 'sparam', sParam = varargin{2*argInInd};           
		case 'ip_img_dim', ipImgDim = varargin{2*argInInd};
		case 'op_img_dim', opImgDim = varargin{2*argInInd};
        otherwise
            error( sprintf( 'unrecognized option %s', szKey ) );
    end
end

% Initialize

if ~exist( 'sParam', 'var' ), error( 'sParam uninitialized' ), end;

phi1 = sParam.phi0(1);
noiseSigma = sParam.noiseSigma;
N = length( sParam.y );

% Must make sure that the i/p and o/p image dimensions are identical
if  ( length( ipImgDim ) ~= length( opImgDim ) ) | ...
	( length( ipImgDim ) == 0 ) | ...
	( length( opImgDim ) == 0 )
	error( 'i/p and o/p image dimensions must have equal length>0' );
else
	for i = 1:length( ipImgDim )
		if ipImgDim( i ) ~= opImgDim( i )
			error( sprintf( 'i/p and o/p image do not agree for dim %g', i ) );
		end
	end
end

% SANITY
if ( phi1 < 1e6*eps ), error( 'phi1 = 0' ), end;
phi0Ratio = sParam.phi0(2)/sParam.phi0(1);
% DEBUG
disp( sprintf( 'phi0Ratio=%g', phi0Ratio ) );

% The valid range of phi2 is 0 <= phi2/phi1 <= 1. Note that:
%    phi2/phi1=1 => Soft threshold
%    phi2/phi0=0 => Hard threshold

phi2FracVec = linspace( 0, 1, sParam.nPhi2-1 );
indPhi0 = find( phi2FracVec == phi0Ratio );
if isempty( indPhi0 )
    phi2FracVec = [ phi2FracVec, phi0Ratio ];
else
    if ( indPhi0 == ( sParam.nPhi2-1 ) )
        phi2FracVec = [ phi2FracVec, (phi2FracVec(indPhi0-1)+phi2FracVec(indPhi0))/2 ];
    else
        phi2FracVec = [ phi2FracVec, (phi2FracVec(indPhi0)+phi2FracVec(indPhi0+1))/2 ];
    end
end
phi2FracVec = sort( phi2FracVec );

% Cost vector for hybrid hard-soft thresholding function using search
% strategy 2
vecCost2 = zeros( size( phi2FracVec ) );
vecCost2Approx = zeros( size( vecCost2 ) );

% Initialize the element that coincides with phi0
indPhi0 = find( phi2FracVec == phi0Ratio );
if length( indPhi0 ) ~= 1
    error( 'did not find exactly one phi0Ratio in phi2FracVec' );
else
    vecCost2( indPhi0 ) = sParam.phi0Cost;
    vecCost2Approx( indPhi0 ) = sParam.phi0Cost;
end

% The best beta according to the hybrid thresholding function. The index
% points to the element of phi2FracVec that minimizes the SURE criterion.
betaHatBest2 = sParam.phi0Beta; % Initialize
betaHatBest2Ind = indPhi0;

% Precompute XOrig and the Gram matrix of X
disp( 'Precomputing XOrig and G(X)' );

% Recall the columns of X were normalized to unit norm. Hence must
% premultiply by the norms to restore the original X.
XOrig = sParam.X .* repmat( sParam.XColumnNorm(:)', N, 1 );

% The Gram matrix of the *original* X
GX = XOrig' * XOrig;

% Go through and sweep phi2 through the remaining values.

for i = 1:sParam.nPhi2

    if ( i == indPhi0 ), continue, end; % Skip
    
    disp( sprintf( 'Doing frac=%g', phi2FracVec(i) ) );
    pause( 5 );

    % Compute using Daubechies' iterations
    [sDIter, sDIterOptions] = emgauss_restore( 'forward_op', sParam.psf, ...
        'noisy_output', reshape( sParam.y, opImgDim ), ...
        'noise_sigma', noiseSigma, ...
        'theta_true', reshape( sParam.betaTrue, ipImgDim ), ...
        'm_selector', 'daub_hard_soft', ...
        'num_em_iters', sParam.numEmIters, ...
        'num_max_grad_search_failures', -1, ...
        'theta0_selector', 'Hty2', ...
        'theta0', [], ...
        'a_val', -1, ...
        'w_val', -1, ...
        't_val', -1, ...
        'sparse_ind', 'w', ...
        'stopping_cond', 'fixed', ...
        'stopping_cond_tol', noiseSigma, ...
        'stopping_cond_win', 3e3, ...
        'stopping_cond_last_fac', 2, ...
        'thetastar', 1, ...
        'soft_thres_t', -1, ...
        'hparam_rest_interval', 1e2, ...
        'bseparate_hparam_est', 0, ...
        'lambda', 2*phi1, ...
        'hard_soft_frac', phi2FracVec( i ) ...
        );

    % Compute the cost function
    daubSol = sDIter.last(:);
    
    sOut = comp_hhs_sure( 'y', sParam.y, ...
        'thetaHat', daubSol, ...       
        'noiseSigma', noiseSigma, ...
        'phi1', phi1, ...
        'phi2', phi2FracVec(i)*phi1, ...
        'bComputeApprox', 0 );
    vecCost2( i ) = sOut.hhsSureVal;
    vecCost2Approx( i ) = sOut.hhsSureValApprox;
    
    % Check to see if we have a better solution

    if ( vecCost2( betaHatBest2Ind ) > vecCost2( i ) )
        betaHatBest2Ind = i;
        betaHatBest2 = daubSol;
    end

end % for

% Defines the structure of sOut

phiNew = [ phi1, phi1*phi2FracVec(betaHatBest2Ind) ];

sOutElems = { 'phi', 'phiNew'; ...
    'betaHatBest', 'betaHatBest2'; ...
    'betaHatBestInd', 'betaHatBest2Ind'; ...
    'cost', 'vecCost2'; ...
    'costApprox', 'vecCost2Approx'; ...
    'phi2FracVec', 'phi2FracVec' ...   
    };

for elemInd = 1:size( sOutElems, 1 )
    szCmd = sprintf( 'sOut.%s = %s;', ...
        char( sOutElems(elemInd,1) ), ...       
        char( sOutElems(elemInd,2) ) );
    try
        eval( szCmd );
    catch
        warning( sprintf( 'Could not execute: %s', szCmd ) );
    end
end


