%
% hs_strat3: fix phi2 and vary phi1
%
% The following elements of sParam are expected:
%    X, XColumnNorm, nPhi1, numEmIters, szPhi1Vec
%    bigCHatHist
%    psf, y, noiseSigma, betaTrue
%    phi0, phi0LarsIndex1, phi0Cost, phi0Beta

function sOut = hs_strat3( varargin )

global XOrig GX % The original X, and the Gram matrix of X = X'*X

ipImgDim = [];
opImgDim = [];

% Process input arguments

lenVarargin = length( varargin );
numArgsIn = floor( lenVarargin/2 );

if 2*numArgsIn ~= lenVarargin, error( 'even ip args expected' ), end

for argInInd = 1:numArgsIn    
    szKey = lower( varargin{2*argInInd-1} );    
    switch szKey        
        case 'sparam', sParam = varargin{2*argInInd};           
		case 'ip_img_dim', ipImgDim = varargin{2*argInInd};
		case 'op_img_dim', opImgDim = varargin{2*argInInd};
        otherwise
            error( sprintf( 'unrecognized option %s', szKey ) );
    end
end

% Initialize

if ~exist( 'sParam', 'var' ), error( 'sParam uninitialized' ), end;

phi1Init = sParam.phi0(1);
phi1Vec = [phi1Init];
noiseSigma = sParam.noiseSigma;
N = length( sParam.y );

% Must make sure that the i/p and o/p image dimensions are identical
if  ( length( ipImgDim ) ~= length( opImgDim ) ) | ...
	( length( ipImgDim ) == 0 ) | ...
	( length( opImgDim ) == 0 )
	error( 'i/p and o/p image dimensions must have equal length>0' );
else
	for i = 1:length( ipImgDim )
		if ipImgDim( i ) ~= opImgDim( i )
			error( sprintf( 'i/p and o/p image do not agree for dim %g', i ) );
		end
	end
end

% Precompute XOrig and the Gram matrix of X
disp( 'Precomputing XOrig and G(X)' );

% Recall the columns of X were normalized to unit norm. Hence must
% premultiply by the norms to restore the original X.
XOrig = sParam.X .* repmat( sParam.XColumnNorm(:)', N, 1 );

% The Gram matrix of the *original* X
GX = XOrig' * XOrig;

switch sParam.szPhi1Vec
    case 'max_alpha'
        % Sweep phi1 from the given value of phi0(1) to the max alpha value

        [sDIter, sDIterOptions] = emgauss_restore( 'forward_op', sParam.psf, ...
            'noisy_output', reshape( sParam.y, opImgDim ), ...
            'noise_sigma', noiseSigma, ...
            'theta_true', reshape( sParam.betaTrue, ipImgDim ), ...
            'm_selector', 'daub_hard_soft', ...
            'num_em_iters', 1, ...
            'num_max_grad_search_failures', -1, ...
            'theta0_selector', 'Hty2', ...
            'theta0', [], ...
            'a_val', -1, ...
            'w_val', -1, ...
            't_val', -1, ...
            'sparse_ind', 'w', ...
            'stopping_cond', 'fixed', ...
            'stopping_cond_tol', noiseSigma, ...
            'stopping_cond_win', 3e3, ...
            'stopping_cond_last_fac', 2, ...
            'thetastar', 1, ...
            'soft_thres_t', -1, ...
            'hparam_rest_interval', 1e2, ...
            'bseparate_hparam_est', 0, ...
            'lambda', 2*phi1Init, ...
            'hard_soft_frac', 1 ...
            );

        phi1Max = sqrt(2*log(M))*sDIterOptions.alpha;
        disp( sprintf( 'phi1Max=%g', phi1Max ) );
        if ( phi1Init >= phi1Max )
            error( sprintf( 'phi1Init=%g >= phi1Max=%g', phi1Init, phi1Max ) );
        else
            phi1Vec = linspace( phi1Init, phi1Max, sParam.nPhi1 );
        end

    case 'max_chat'
        % Sweep phi1 from the given value of phi0(1) to the first value of
        % bigCHatHist
        indLast = sParam.phi0LarsIndex1;
        indDelta = floor( (indLast-1)/sParam.nPhi1 );
        if ( indDelta < 1 )
            indDelta = 1;
            indFirst = 1;
            sParam.nPhi1 = indLast;
            warning( sprintf( 'Cannot find a suitable indDelta: setting nPhi1=%g', sParam.nPhi1 ) );
        else
            indFirst = indLast - indDelta * ( sParam.nPhi1 - 1 );
        end
        if ( indFirst < 1 ), error( 'unexpected: indFirst < 1' ), end;
        phi1Vec = sParam.XColumnNorm(1) * sParam.bigCHatHist( indFirst:indDelta:indLast );

    otherwise
        error( 'unrecognized szPhi1Vec' );
end

% SANITY
if ( length( phi1Vec ) ~= sParam.nPhi1 )
    error( 'phi1Vec failed sanity' );
end

% DEBUG
%disp( 'phi1Vec:' ), disp( phi1Vec ), disp( 'Pausing...' ), pause;

phi2 = sParam.phi0(2);

% Cost vector for hybrid hard-soft thresholding function using search
% strategy 3
vecCost3 = zeros( size( phi1Vec ) );
vecCost3Approx = zeros( size( vecCost3 ) );

% Initialize the last element, which corresponds to phi0Cost
vecCost3( end ) = sParam.phi0Cost;
vecCost3Approx( end ) = vecCost3( end );

% The best beta according to the hybrid thresholding function. Note that
% betaHatBest3Ind points to the element of phi1Vec that minimizes the SURE
% criterion.
betaHatBest3 = sParam.phi0Beta;
betaHatBest3Ind = length( vecCost3 ); % Last entry

% SANITY
if ( phi1Vec( betaHatBest3Ind ) ~= sParam.phi0(1) )
    error( 'sanity failed on phi1Vec' );
end

% Go through and sweep phi1

for i = 1:(sParam.nPhi1-1)

    disp( sprintf( 'Doing phi1=%g', phi1Vec(i) ) );
    pause( 5 );

    % Compute using Daubechies' iterations
    [sDIter, sDIterOptions] = emgauss_restore( 'forward_op', sParam.psf, ...
        'noisy_output', reshape( sParam.y, opImgDim ), ...
        'noise_sigma', noiseSigma, ...
        'theta_true', reshape( sParam.betaTrue, ipImgDim ), ...
        'm_selector', 'daub_hard_soft', ...
        'num_em_iters', sParam.numEmIters, ...
        'num_max_grad_search_failures', -1, ...
        'theta0_selector', 'Hty2', ...
        'theta0', [], ...
        'a_val', -1, ...
        'w_val', -1, ...
        't_val', -1, ...
        'sparse_ind', 'w', ...
        'stopping_cond', 'fixed', ...
        'stopping_cond_tol', noiseSigma, ...
        'stopping_cond_win', 3e3, ...
        'stopping_cond_last_fac', 2, ...
        'thetastar', 1, ...
        'soft_thres_t', -1, ...
        'hparam_rest_interval', 1e2, ...
        'bseparate_hparam_est', 0, ...
        'lambda', 2*phi1Vec(i), ...
        'hard_soft_frac', phi2/phi1Vec(i) ...
        );

    % Compute the cost function
    daubSol = sDIter.last(:);
    
    sOut = comp_hhs_sure( 'y', sParam.y, ...
        'thetaHat', daubSol, ...
        'noiseSigma', noiseSigma, ...
        'phi1', phi1Vec(i), ...
        'phi2', phi2, ...
        'bComputeApprox', 0 );
    vecCost3( i ) = sOut.hhsSureVal;
    vecCost3Approx( i ) = sOut.hhsSureValApprox;

    % Check to see if we have a better solution

    if ( vecCost3( betaHatBest3Ind ) > vecCost3( i ) )
        betaHatBest3Ind = i;
        betaHatBest3 = daubSol;
    end

end

% Defines the structure of sOut

phiNew = [ phi1Vec( betaHatBest3Ind ), phi2 ];

sOutElems = { 'phi', 'phiNew'; ...
    'betaHatBest', 'betaHatBest3'; ...
    'betaHatBestInd', 'betaHatBest3Ind'; ...
    'cost', 'vecCost3'; ...
    'costApprox', 'vecCost3Approx'; ...
    'phi1Vec', 'phi1Vec' ...   
    };

for elemInd = 1:size( sOutElems, 1 )
    szCmd = sprintf( 'sOut.%s = %s;', ...
        char( sOutElems(elemInd,1) ), ...       
        char( sOutElems(elemInd,2) ) );
    try
        eval( szCmd );
    catch
        warning( sprintf( 'Could not execute: %s', szCmd ) );
    end
end

    

