%
% Straight-forward LARS. Skeleton copied from sbl_restore. Usage is
% sOut = lars1( varargin )
% 

function sOut = lars1( varargin )

% Default restore options

sRestoreOptions = struct( 'X', [], ...
    'XColumnNorm', [], ...
    'B_DEBUG', 1, ...
    'bNormX', 0 ...
    );
    
% Process input arguments

lenVarargin = length( varargin );
numArgsIn = floor( lenVarargin/2 );

if 2*numArgsIn ~= lenVarargin, error( 'even ip args expected' ), end

for argInInd = 1:numArgsIn
    
    szKey = lower( varargin{2*argInInd-1} );    
    switch szKey
        case 'forward_op', hForwardOp = varargin{2*argInInd};
        case 'noisy_output', y = varargin{2*argInInd};      
        case 'beta_true', betaTrue = varargin{2*argInInd};                          
        case 'num_iters', sRestoreOptions.NUM_ITERS = varargin{2*argInInd};    
		case 'x', sRestoreOptions.X = varargin{2*argInInd};   
        case 'x_column_norm', sRestoreOptions.XColumnNorm = varargin{2*argInInd};
        case 'bnormx', sRestoreOptions.bNormX = 1;
            
        otherwise
            error( sprintf( 'unrecognized option %s', szKey ) );
    end
end

% Check validity of arguments

% Note that we expect to be given the zero-padded psf
N = size( hForwardOp, 1 ); 
if ( N ~= size( hForwardOp, 2 ) ), error( 'h is not a square matrix' ), end;
NN = N^2; % The matrix X will be of size NN x NN

% Check that vector y is of length NN
if length( y ) ~= NN
    error( sprintf( 'length(y)=%g but NN=%g??', length(y), NN ) );
end

if ( sRestoreOptions.NUM_ITERS > NN ) | ( sRestoreOptions.NUM_ITERS < 1 )
    error( 'invalid NUM_ITERS' );
elseif ( sRestoreOptions.NUM_ITERS == NN )
    disp( 'Running LARS to LS solution' );
else
    disp( 'Running LARS, but stopping short of LS solution' );
end

% Allocate space for system matrix. Bad idea: Matlab cannot build matrices
% whose N^2 value exceeds the max representable integer. Use sparse
% matrices instead. 

bCreatedX = 0;

% Build the system matrix. Note that the operations below are valid when we
% assume that the matrix can be represented by a straightforward
% convolution kernel. That is, it can be diagonalized using DFTs.

nLastColumnInd = 0; % Index pointing to the last element of the previous column

% Variables to keep track of non-zero elements
nzNum = 0;
nzI = [];
nzJ = [];
nzS = [];

if isempty( sRestoreOptions.X ), disp( 'Creating matrix X' )
    
    sRestoreOptions.XColumnNorm = zeros( NN, 1 );
    
    for j = 1:NN

        % Iterate over the j-th column to find any non-zero entries

        % Keep track of sum of squared values of each column
        columnSSV = 0;

        for i = 1:NN
            hInd1 = 1 + mod( mod(i,N) - mod(j,N), N );
            hInd2 = 1 + mod( floor((i-1)/N) - floor((j-1)/N), N );
            if ( abs( hForwardOp( hInd1, hInd2 ) ) > 1e3*eps )
                % DEBUG
                %disp( sprintf( 'i=%g, j=%g  --> hInd1=%g, hInd2=%g', i, j, hInd1, hInd2 ) );
                nzNum = nzNum + 1;
                nzI = [nzI; i];
                nzJ = [nzJ; j];
                nzS = [nzS; hForwardOp( hInd1, hInd2 )];
                columnSSV = columnSSV + nzS(end)^2;
            end
        end

        % Normalize the columns so that each is of unit norm
        columnL2Norm = sqrt( columnSSV );        
        if ( abs( columnL2Norm ) < 1e3*eps )
            error( 'encountered zero column!' );
        end
        
        startInd = nLastColumnInd + 1;
        nzS( startInd:end ) = nzS( startInd:end ) / columnL2Norm;
        sRestoreOptions.XColumnNorm( j ) = columnL2Norm;        
        nLastColumnInd = length( nzS );
        
    end       

    disp( sprintf( 'nzNum=%g', nzNum ) );

	% Allocate space for sparse system matrix
	sRestoreOptions.X = sparse( nzI, nzJ, nzS, NN, NN );
	bCreatedX = 1;
    
else, disp( 'Using supplied X' );
    
    if sRestoreOptions.bNormX 
        
        disp( 'Computing XColumnNorm' );
        [ nzI, nzJ, nzS ] = find( sRestoreOptions.X );
        columnSSV = zeros( NN, 1 );
        nnz = length( nzI );
        for i = 1:nnz
            j = nzJ( i );
            columnSSV( j ) = columnSSV( j ) + nzS( i )^2;
        end
        sRestoreOptions.XColumnNorm = sqrt( columnSSV );
        
        disp( 'Normalizing X' );
        for i = 1:nnz
            j = nzJ( i );
            nzS( i ) = nzS( i ) / sRestoreOptions.XColumnNorm( j );
        end
        sRestoreOptions.X = sparse( nzI, nzJ, nzS, NN, NN );
        
    else
        disp( 'Re-using XColumnNorm' );
    end
end

if 0
    yNoiseless = sRestoreOptions.X * (sRestoreOptions.XColumnNorm .* betaTrue);
    imagesc( reshape( yNoiseless, N, N ) ), colorbar;
end

% Pre-compute XTX = X' * X
XTX = (sRestoreOptions.X)' * (sRestoreOptions.X);

activeSet = [];
muHatPrevActiveSet = zeros( NN, 1 );
muHatNextActiveSet = zeros( NN, 1 );
gammaMat = zeros( NN, 2 );

% Time the iterations
tCpuTimeStart = cputime;

for i = 1:sRestoreOptions.NUM_ITERS
    
    if sRestoreOptions.B_DEBUG, disp( sprintf( 'i=%g', i ) ), end;
    
    % Step 1: Compute the correlations
    cHat = sRestoreOptions.X' * ( y - muHatPrevActiveSet );    
    s = sign( cHat );
            
    % Step 2: Get the active set and sort in ascending order of indices
    if ( i == 1 )
        % In the first iteration, use max to locate the active set
        [BigCHat, activeSet] = max( abs( cHat ) );
    else
        % In subsequent iterations, add to the active set manually. It
        % turns out that there are some numerical differences in the cHats
        % that are supposed to be the same.
        BigCHat = max( abs( cHat ) );
        activeSet = [activeSet, jHat];
    end
    
    [activeSetSorted, activeSetSortedInd] = sort( activeSet, 1, 'ascend' );
    activeSet = activeSetSorted;
    nActiveSetSize = length( activeSet );
    BigCHat = BigCHat( 1 ); % Want this to be a real
            
    % Plot the cHat and the max BigCHat
    if sRestoreOptions.B_DEBUG
        figure( 1 ), clf, plot( cHat, 'bx' ), hold on;
        plot( [1, length(cHat)], [BigCHat, BigCHat], 'k--' );
        title( sprintf( 'iter %g', i ) );
        ylabel( 'cHat' );
    end
    
    % Print info to console
    disp( sprintf( 'iter=%g, size of active set=%g', i, nActiveSetSize ) );
    disp( sprintf( 'BigCHat: %g', BigCHat ) );
    disp( 'active set:' ), disp( activeSet );
        
    % Step 3: Construct the variables related to the active set A
    
    clear nzI nzJ nzS;
    nzI = [];
    nzJ = [];
    nzS = [];
    
    for j = 1:nActiveSetSize
        nzI = [ nzI; activeSet(j) ];
        nzJ = [ nzJ; j ];
        nzS = [ nzS; s( activeSet(j) ) ];
    end
    
    % Allocate space for SActiveSet, which is of size NN x nActiveSetSize
	SActiveSet = sparse( nzI, nzJ, nzS, NN, nActiveSetSize );
    
    % Compute the other variables
    GActiveSet = SActiveSet' * XTX * SActiveSet;
    %disp( 'GActiveSet:' ), disp( GActiveSet );
    GInvActiveSet = inv( GActiveSet );
    OneVecActiveSet = ones( nActiveSetSize, 1 );
    AActiveSet = ( OneVecActiveSet' * GInvActiveSet * OneVecActiveSet )^(-1/2);
    wActiveSet = AActiveSet * GInvActiveSet * OneVecActiveSet;
    uActiveSet = sRestoreOptions.X * SActiveSet * wActiveSet;
    a = sRestoreOptions.X' * uActiveSet;
    %disp( 'a:' ), disp( a );
    
    % Step 4: Compute gammaHat and jHat
    
    if ( i < NN )
        
        % We can only do this if i < NN. Here, compute
        % gammaHat = min^+_{j \in A^c} { (CHat - chat_j)/(A_A - a_j), (CHat
        % + chat_j)/(A_A + a_j) }
        
        %gammaMat( :, 1 ) = ( BigCHat - cHat ) ./ ( AActiveSet - a );
        %gammaMat( :, 2 ) = ( BigCHat + cHat ) ./ ( AActiveSet + a );
        
        activeSetC = setdiff( 1:NN, activeSet );
        clear gammaMat2
        gammaMat2 = zeros( length( activeSetC ), 2 );
        tmp = ( BigCHat - cHat(activeSetC) ) ./ ( AActiveSet - a(activeSetC) );
        gammaMat2( :, 1 ) = tmp(:);
        tmp = ( BigCHat + cHat(activeSetC) ) ./ ( AActiveSet + a(activeSetC) );
        gammaMat2( :, 2 ) = tmp(:);
        %gammaMat2 = gammaMat( activeSetC, : );
        %disp( 'gammaMat2:' ), disp( gammaMat2 );        

        % Take the min over each column
        gammaMatPosInd = find( gammaMat2(:,1) > 0 );
        if ~isempty( gammaMatPosInd )
            [gammaMatCol1Min, gammaMatCol1MinInd] = min( gammaMat2( gammaMatPosInd, 1 ) );
            if length( gammaMatCol1MinInd ) > 1
                error( 'cannot handle length( gammaMatCol1MinInd ) > 1' );
            end
        else
            gammaMatCol1Min = inf;
        end
        gammaMatPosInd = find( gammaMat2(:,2) > 0 );
        if ~isempty( gammaMatPosInd )
            [gammaMatCol2Min, gammaMatCol2MinInd] = min( gammaMat2( gammaMatPosInd, 2 ) );
            if length( gammaMatCol2MinInd ) > 1
                error( 'cannot handle length( gammaMatCol2MinInd ) > 1' );
            end
        else
            gammaMatCol2Min = inf;
        end

        % And then combine to find the min over the entire matrix
        [gmm, gmmInd] = min( [gammaMatCol1Min, gammaMatCol2Min] );
        if isfinite( gmm )
            gammaHat = gmm;
            if ( gmmInd == 1 ), jHat = activeSetC( gammaMatCol1MinInd );
            else, jHat = activeSetC( gammaMatCol2MinInd );
            end
        else
            error( 'could not find gammaHat and jHat' );
        end

    else
        
        % Do this if i == NN. This is the last step
        gammaHat = BigCHat/AActiveSet;
        jHat = -1;
        
    end
    
    % Step 5: Form the next estimator of mu
    disp( sprintf( 'gammaHat = %g', gammaHat' ) );
    muHatNextActiveSet = muHatPrevActiveSet + gammaHat * uActiveSet;
    
    % Step 6: Update state variables
    muHatPrevActiveSet = muHatNextActiveSet;
    
    disp( 'End of iter, press key' ), pause;
end

tCpuTimeStop = cputime;

% Defines the structure of sRestore
sRestoreElems = { 'last', 'muHatPrevActiveSet'; ...
    'cputime', '[tCpuTimeStart, tCpuTimeStop]' ...
    };

% Optionally, solve mu = X*beta
if 1
    beta = cgs( sRestoreOptions.X, muHatPrevActiveSet, [], 1e4 );
    sRestoreElems = vertcat( sRestoreElems, { 'beta', 'beta' } );
end

if ( bCreatedX ) | ( sRestoreOptions.bNormX )

	% If we had to create X, return the system matrix in sRestore. Else,
	% the caller already has phi, so there's no need to return it in
	% sRestore.

    sRestoreElems = vertcat( sRestoreElems, ...
        { 'x', 'sRestoreOptions.X'; ...
          'xcolnorm', 'sRestoreOptions.XColumnNorm' } );
    
else
    sRestoreElems = vertcat( sRestoreElems, { 'x', '[]'; 'xcolnorm', '[]' } );
end

for elemInd = 1:size( sRestoreElems, 1 )
    szCmd = sprintf( 'sOut.%s = %s;', ...
        char( sRestoreElems(elemInd,1) ), ...       
        char( sRestoreElems(elemInd,2) ) );
    try
        eval( szCmd );
    catch
        warning( sprintf( 'Could not execute: %s', szCmd ) );
    end
end

    
    

    
    
    


