%
% LARS-LASSO. Adapted from LARS algorithm: see lars1. Usage is
% sOut = lars2( varargin )
% 
% The problem we consider is
% y = X * beta + n
% y ~ N-dim vector, beta ~ M-dim vector, X ~ N x M
%

function sOut = lars2( varargin )

% Default restore options

sRestoreOptions = struct( 'X', [], 'XColumnNorm', [], 'B_DEBUG', 0 );    
        
% Process input arguments

lenVarargin = length( varargin );
numArgsIn = floor( lenVarargin/2 );

if 2*numArgsIn ~= lenVarargin, error( 'even ip args expected' ), end

for argInInd = 1:numArgsIn
    
    szKey = lower( varargin{2*argInInd-1} );    
    switch szKey        
        case 'noisy_output', y = varargin{2*argInInd};      
        case 'beta_true', betaTrue = varargin{2*argInInd};                          
        case 'num_iters', sRestoreOptions.NUM_ITERS = varargin{2*argInInd};    
		case 'x', sRestoreOptions.X = varargin{2*argInInd};   
        case 'x_column_norm', sRestoreOptions.XColumnNorm = varargin{2*argInInd};  
        case 'n_debug', sRestoreOptions.N_DEBUG = varargin{2*argInInd};
        case 'psf', psf = varargin{2*argInInd};
            
        otherwise
            error( sprintf( 'unrecognized option %s', szKey ) );
    end
end

% Check validity of arguments
N = length( y );
M = length( betaTrue );
if ( size( sRestoreOptions.X, 1 ) ~= N ) | ( size( sRestoreOptions.X, 2 ) ~= M )
    error( 'size of X incompatible with y, betaTrue' );
end

if ( sRestoreOptions.NUM_ITERS > M ) | ( sRestoreOptions.NUM_ITERS < 1 )
    error( 'invalid NUM_ITERS' );
elseif ( sRestoreOptions.NUM_ITERS == M )
    disp( 'Running LARS to LS solution' );
else
    disp( 'Running LARS, but stopping short of LS solution' );
end

if 0
    yNoiseless = sRestoreOptions.X * (sRestoreOptions.XColumnNorm .* betaTrue);
    imgSize = sqrt( N );
    if ( imgSize*imgSize ~= N ), error( 'length(y) not a perfect square' ), end;
    figure( 1 ), clf, imagesc( reshape( yNoiseless, imgSize, imgSize ) ), colorbar;
    ylabel( 'Sanity on X, XColumnNorm' );
    title( 'yNoiseless' );
    disp( 'Pausing...' ), pause;
end

% Pre-compute XTX = X' * X
XTX = (sRestoreOptions.X)' * (sRestoreOptions.X);

activeSet = [];
muHatPrevActiveSet = zeros( N, 1 );
muHatNextActiveSet = zeros( N, 1 );
betaHatPrevActiveSet = zeros( M, 1 );
betaHatNextActiveSet = zeros( M, 1 );
gammaMat = zeros( M, 2 );

bLassoKludge = 0; % Flag whenever we have to do a LARS-LASSO kludge
jHat = -1;
jKludge = -1;

% Time the iterations
tCpuTimeStart = cputime;

for i = 1:sRestoreOptions.NUM_ITERS
    
    if sRestoreOptions.N_DEBUG, disp( sprintf( 'i=%g', i ) ), end;      
    
    % Step 1: Compute the correlations
    cHat = sRestoreOptions.X' * ( y - muHatPrevActiveSet );    
    s = sign( cHat );
    
    % LARS sanity: check that sign(beta_j) = sign(c_j) = s_j, for j in the
    % active set
    betaNzInd = find( abs( betaHatPrevActiveSet ) > eps ); % Find non-zero indices
    vecMismatch = ( s(betaNzInd) ~= sign( betaHatPrevActiveSet( betaNzInd ) ) );
    disp( 'betaNzInd:' ), disp( betaNzInd(:)' );
    disp( 'c-values:' ), disp( cHat(betaNzInd) );
    disp( 'beta-values:' ), disp( betaHatPrevActiveSet(betaNzInd) );
    
    if sum( vecMismatch ) ~= 0
        figure( 2 ), clf, plot( s(betaNzInd), 'kx' ), hold on;
        plot( sign(betaHatPrevActiveSet(betaNzInd)), 'ro' );
        legend( 's', 'sign(beta)', 0 );
        error( 'LASSO solution sign contraint violated' );
    end
                   
    % Step 2: Get the active set and sort in ascending order of indices
    disp( 'STEP 2: Adjusting active set' );
    
    if ( i == 1 )
        % In the first iteration, use max to locate the active set
        [BigCHat, activeSet] = max( abs( cHat ) );
    else
        
        % Do this step in subsequent iterations        
        BigCHat = max( abs( cHat ) );

        if ~bLassoKludge
            % In subsequent iterations, add to the active set manually. It
            % turns out that there are some numerical differences in the cHats
            % that are supposed to be the same.
            activeSet = [activeSet, jHat];
        else
            activeSet = setdiff( activeSet, jTilde );
        end
        
    end
                
    bLassoKludge = 0; % Re-initialize
    
    [activeSetSorted, activeSetSortedInd] = sort( activeSet );
    activeSet = activeSetSorted;
    nActiveSetSize = length( activeSet );
    BigCHat = BigCHat( 1 ); % Want this to be a real
        
    % Print correlations
    disp( 'cHat(activeSet):' ), disp( cHat(activeSet) );
    
    % Plot the cHat and the max BigCHat
    if sRestoreOptions.N_DEBUG
        figure( 1 ), clf, plot( cHat, 'bo' ), hold on;
        plot( [1, length(cHat)], [BigCHat, BigCHat], 'k--' );
        plot( [1, length(cHat)], [-BigCHat, -BigCHat], 'k--' );
        title( sprintf( 'iter %g', i ) );
        ylabel( 'cHat' );
    end
    
    % Print info to console
    disp( sprintf( 'iter=%g, size of active set=%g', i, nActiveSetSize ) );
    disp( sprintf( 'BigCHat: %g', BigCHat ) );
    disp( 'active set:' ), disp( activeSet );
        
    % Step 3: Construct the variables related to the active set A
    
    clear nzI nzJ nzS;
    nzI = zeros( nActiveSetSize, 1 );
    nzJ = zeros( size( nzI ) );
    nzS = zeros( size( nzI ) );
    
    for j = 1:nActiveSetSize
        nzI(j) = activeSet(j);
        nzJ(j) = j;
        nzS(j) = s( activeSet(j) );
    end
    
    % Allocate space for SActiveSet, which is of size M x nActiveSetSize
	SActiveSet = sparse( nzI, nzJ, nzS, M, nActiveSetSize );
    
    % Compute the other variables
    GActiveSet = SActiveSet' * XTX * SActiveSet; % size |A| x |A|
    %disp( 'GActiveSet:' ), disp( GActiveSet );
    GInvActiveSet = inv( GActiveSet );
    OneVecActiveSet = ones( nActiveSetSize, 1 ); % |A|-vec
    tmp = full( OneVecActiveSet' * GInvActiveSet * OneVecActiveSet );
    AActiveSet = 1/sqrt(tmp);
    disp( sprintf( 'SANITY: AActiveSet=%g', AActiveSet ) );
    
    wActiveSet = AActiveSet * GInvActiveSet * OneVecActiveSet; % |A|-vec
    uActiveSet = (sRestoreOptions.X * SActiveSet) * wActiveSet; % N-vec
    disp( sprintf( 'SANITY: |u|=%g', norm( uActiveSet ) ) );
    tmp = full( ( SActiveSet' * sRestoreOptions.X' ) * uActiveSet );
    disp( 'SANITY: X_A^T * u_A:' ), disp( tmp(:)' );
    
    a = sRestoreOptions.X' * uActiveSet; % M-vec
    %disp( 'a:' ), disp( a );
    
    % Construct dHat
    dHat = zeros( M, 1 ); % M-vec
    for j = 1:nActiveSetSize
        jj = activeSet( j );
        dHat( jj ) = s( jj ) * wActiveSet( j );
    end
    
    % Step 4a: Compute gammaHat and jHat
    
    jHatColumn = -1; % DEBUG variable
    
    if ( i < M )
        
        % We can only do this if i < NN. Here, compute
        % gammaHat = min^+_{j \in A^c} { (CHat - chat_j)/(A_A - a_j), (CHat
        % + chat_j)/(A_A + a_j) }
        
        activeSetC = setdiff( 1:M, activeSet );
        % DEBUG
        %disp( 'activeSetC:' ), disp( activeSetC(:)' );
        clear gammaMat2;
        gammaMat2 = zeros( length( activeSetC ), 2 );
        
        % Compute the 1st column of gammaMat2
        numer = BigCHat - cHat( activeSetC );
        denom = AActiveSet - a( activeSetC );
        tmp = numer(:) ./ denom(:);
        % SANITY
        if length( tmp(:) ) ~= length( activeSetC ), error( 'Sanity on tmp failed' ), end;
        gammaMat2( :, 1 ) = tmp(:);
        
        % Compute the 2nd column of gammaMat2
        numer = BigCHat + cHat( activeSetC );
        denom = AActiveSet + a( activeSetC );
        tmp = numer(:) ./ denom(:);
        % SANITY
        if length( tmp(:) ) ~= length( activeSetC ), error( 'Sanity on tmp failed' ), end;
        gammaMat2( :, 2 ) = tmp(:);

        % Take the min over each column
        gammaMatPosInd = find( ( gammaMat2(:,1) > eps ) & isfinite( gammaMat2(:,1) ) );
        if ~isempty( gammaMatPosInd )
            gammaMatCol1Min = min( gammaMat2( gammaMatPosInd, 1 ) );
            gammaMatCol1MinInd = find( gammaMat2(:,1) == gammaMatCol1Min );
            if length( gammaMatCol1MinInd ) > 1
                error( 'cannot handle length( gammaMatCol1MinInd ) > 1' );
            end            
            % SANITY
            disp( 'gammaMatCol1Min:' ), disp( gammaMatCol1Min );
            disp( 'gammaMatCol1MinInd:' ), disp( gammaMatCol1MinInd );
            disp( 'jHat of min:' ), disp( activeSetC( gammaMatCol1MinInd ) );
            disp( 'row of gammaMat2:' ), disp( gammaMat2( gammaMatCol1MinInd, : ) );
        else
            gammaMatCol1Min = inf;
        end
        
        gammaMatPosInd = find( ( gammaMat2(:,2) > eps ) & isfinite( gammaMat2(:,2) ) );
        if ~isempty( gammaMatPosInd )
            gammaMatCol2Min = min( gammaMat2( gammaMatPosInd, 2 ) );
            gammaMatCol2MinInd = find( gammaMat2(:,2) == gammaMatCol2Min );
            if length( gammaMatCol2MinInd ) > 1
                error( 'cannot handle length( gammaMatCol2MinInd ) > 1' );
            end
            % SANITY
            disp( 'gammaMatCol2Min:' ), disp( gammaMatCol2Min );
            disp( 'gammaMatCol2MinInd:' ), disp( gammaMatCol2MinInd );
            disp( 'jHat of min:' ), disp( activeSetC( gammaMatCol2MinInd ) );
            disp( 'row of gammaMat2:' ), disp( gammaMat2( gammaMatCol2MinInd, : ) );
        else
            gammaMatCol2Min = inf;
        end

        % And then combine to find the min over the entire matrix
        [gmm, gmmInd] = min( [gammaMatCol1Min, gammaMatCol2Min] );
        if isfinite( gmm )
            gammaHat = gmm;
            if ( gmmInd == 1 )
                % DEBUG
                disp( 'Getting jHat from 1st column of gammaMat2' );
                jHatColumn = 1;
                jHat = activeSetC( gammaMatCol1MinInd );
            else
                % DEBUG
                disp( 'Getting jHat from 2nd column of gammaMat2' );
                jHatColumn = 2;
                jHat = activeSetC( gammaMatCol2MinInd );
            end
        else
            jHatColumn = -1; % Invalid value
            error( 'could not find gammaHat and jHat' );
        end

    else
        
        % Do this if i == M. This is the last step
        gammaHat = BigCHat/AActiveSet;
        jHat = -1;
        
    end
    
    % Step 4b: compute gammaTilde
 
    warning off; % XXX: We are going to get some divide-by-zero errors here!
    gammaTildeVec = -betaHatPrevActiveSet ./ dHat; % This is a M-vector
    warning on;

    % Find the min gamma s.t. it is positive 
    gammaTildePosInd = find( ( gammaTildeVec > eps ) & isfinite( gammaTildeVec ) );
    gtm = min( gammaTildeVec( gammaTildePosInd ) );
    
    switch length( gtm )
        case 0
            % No positive elements in gammaTilde
            gammaTilde = inf;
            jTilde = -1;
        case 1
            % This is the "normal" case
            gtmInd = find( gammaTildeVec == gtm );
            if length( gtmInd ) > 1
                error( 'cannot handle length( gtmInd ) > 1' );
            elseif length( gtmInd ) == 1
                gammaTilde = gtm;
                jTilde = gtmInd;
                % Sanity: check that jTilde is in the active set
                if ~ismember( jTilde, activeSet )
                    error( 'jTilde not in the active set' );
                end
            else
                error( 'Cannot find minimum in gammaTildeVec?' );
            end
        otherwise
            error( 'Multiple minima in gammaTildeVec?' );
    end

    % Step 4c: get gammaConst
    if ( gammaTilde < gammaHat )
        bLassoKludge = 1;
        gammaConst = gammaTilde;
        disp( '*** LARS-LASSO KLUDGE ***' );
    else
        gammaConst = gammaHat;
    end
    
    disp( [ sprintf( 'gammaHat=%g, jHat=%g/', gammaHat, jHat ), ...
        sprintf( 'gammaTilde=%g, jTilde=%g\n', gammaTilde, jTilde ), ...
        sprintf( 'gammaConst=%g', gammaConst ) ] );    
    
    % SANITY
    nextBigCHat = BigCHat - gammaConst * AActiveSet;
    disp( sprintf( 'nextBigCHat = %g', nextBigCHat ) );
    
    % Check that the new correlation equals nextBigCHat
    if ~bLassoKludge
        cjHat = cHat( jHat ) - gammaConst * a( jHat );
        disp( sprintf( 'cjHat = %g', cjHat ) );
        % Find entry in gammaMat2
        jHatGammaMatIndex = find( activeSetC == jHat );
        if length( jHatGammaMatIndex ) ~= 1, error( 'failed to find jHat in activeSetC' ), end;
        disp( 'jHatGammaMatIndex:' ), disp( jHatGammaMatIndex );
        disp( 'jHat entry in gammaMat2:' ), disp( gammaMat2( jHatGammaMatIndex, : ) );
        % Check if cjHat equals nextBigCHat. It should!
        if abs( cjHat - nextBigCHat ) > eps
            warning( 'cjHat not equal to nextBigCHat??' );                        
            % Re-compute gamma
            switch jHatColumn
                case 1
                    gamma2 = ( BigCHat - cHat( jHat ) ) / ( AActiveSet - a( jHat ) );
                case 2
                    gamma2 = ( BigCHat + cHat( jHat ) ) / ( AActiveSet + a( jHat ) );
                otherwise
                    error( 'trying to re-compute gammaHat with invalid jHatColumn' );
            end
            disp( sprintf( 'Re-computed gammaHat = %g', gamma2 ) );            
        end
    end
    
    
    % Step 5: Form the next estimator of mu and beta
    muHatNextActiveSet = muHatPrevActiveSet + gammaConst * uActiveSet;    
    betaHatNextActiveSet = betaHatPrevActiveSet + gammaConst * dHat;
    
    % MASSIVE DEBUG
    if ( sRestoreOptions.N_DEBUG > 1 ) & ( mod( i, 5 ) == 1 ) & ( i > 1 )
        tSoft = BigCHat * sRestoreOptions.XColumnNorm( activeSet(1) );
        disp( sprintf( 'BigCHat = %g, tSoft=%g', BigCHat, tSoft ) );

        % Compute using Daubechies' iterations
        [sDIter, sDIterOptions] = emgauss_restore( 'forward_op', psf, ...
            'noisy_output', reshape( y, 32, 32 ), ...
            'noise_sigma', 1, ...
            'theta_true', reshape( betaTrue, 32, 32 ), ...
            'm_selector', 'soft_fixed', ...
            'num_em_iters', 1e4, ...
            'num_max_grad_search_failures', -1, ...
            'theta0_selector', 'Hty2', ...
            'theta0', [], ...
            'a_val', -1, ...
            'w_val', -1, ...
            't_val', -1, ...
            'sparse_ind', 'w', ...
            'b_use_history', 0, ...
            'b_save_history', 0, ...
            'stopping_cond', 'fixed', ...
            'stopping_cond_tol', 1, ...
            'stopping_cond_win', 3e3, ...
            'stopping_cond_last_fac', 2, ...
            'thetastar', 1, ...
            'soft_thres_t', tSoft, ...
            'hparam_rest_interval', 1e2 ...
            );
        
        figure( 2 ), imagesc( reshape( betaHatNextActiveSet./sRestoreOptions.XColumnNorm, 32, 32 ) ), colorbar;
        title( 'LARS-LASSO' );        
        figure( 3 ), imagesc( sDIter.last ), colorbar;
        title( 'Daubechies' );
        
        vecDIter = reshape( sDIter.last, 1024, 1 );
        vecMismatch = ( sign( reshape( betaHatNextActiveSet, 1024, 1 ) ) ~= sign( vecDIter ) );
        if sum( vecMismatch ) ~= 0
            indMismatch = find( vecMismatch ~= 0 );
            disp( 'Mismatch in LARS-LASSSO vs. Daubechies' );
            disp( 'indMismatch:' ), disp( indMismatch );
            disp( 'Mismatched values/LARS-LASSO:' ); 
            disp( betaHatNextActiveSet( indMismatch ) );
            disp( 'Mismatched values/Daubechies:' );
            disp( vecDIter( indMismatch ) );
        end
        
    end

    % Step 6: Update state variables
    muHatPrevActiveSet = muHatNextActiveSet;
    betaHatPrevActiveSet = betaHatNextActiveSet;
    
    if sRestoreOptions.N_DEBUG, disp( 'End of iter, press key' ), pause, end;
    
end

tCpuTimeStop = cputime;

% Re-scale beta
betaHat = betaHatPrevActiveSet ./ sRestoreOptions.XColumnNorm;
% muCheck should equal muHatPrevActiveSet
muCheck = sRestoreOptions.X * betaHatPrevActiveSet;

% Defines the structure of sRestore
sRestoreElems = { ...
    'last', 'betaHat'; ...
    'muHat', 'muHatPrevActiveSet'; ...
    'muCheck', 'muCheck'; ...
    'bigCHat', 'BigCHat'; ...
    'activeSet', 'activeSet'; ...
    'cputime', '[tCpuTimeStart, tCpuTimeStop]' ...
    };

for elemInd = 1:size( sRestoreElems, 1 )
    szCmd = sprintf( 'sOut.%s = %s;', ...
        char( sRestoreElems(elemInd,1) ), ...       
        char( sRestoreElems(elemInd,2) ) );
    try
        eval( szCmd );
    catch
        warning( sprintf( 'Could not execute: %s', szCmd ) );
    end
end

    
    

    
    
    


