%
% LARS-LASSO-SURE. Adapted from lars3. Usage is
% sOut = lars3ext( varargin )
% 
% The problem we consider is
% y = X * beta + n
% y ~ N-dim vector, beta ~ M-dim vector, X ~ N x M
%

function sOut = lars3ext( varargin )

% Default restore options

sRestoreOptions = struct( 'fnhX', [], ...
    'fnhXColumnNorm', [], ...    
    'nConvMethod', 1, ...
    'ipImgDim', [], ...
    'opImgDim', [], ...
    'constXColumnNorm', -1, ...
	'bCheckDaubechies', 1 );

noiseSigma = -1;

% Process input arguments

lenVarargin = length( varargin );
numArgsIn = floor( lenVarargin/2 );

if 2*numArgsIn ~= lenVarargin, error( 'even ip args expected' ), end

for argInInd = 1:numArgsIn
    
    szKey = lower( varargin{2*argInInd-1} );    
    switch szKey        
        case 'noisy_output', y = varargin{2*argInInd};      
        case 'beta_true', betaTrue = varargin{2*argInInd}; 
        case 'num_iters', sRestoreOptions.NUM_ITERS = varargin{2*argInInd};    
		case 'fnh_x', sRestoreOptions.fnhX = varargin{2*argInInd};   
        case 'fnh_x_column_norm', sRestoreOptions.fnhXColumnNorm = varargin{2*argInInd};  
        case 'n_debug', sRestoreOptions.N_DEBUG = varargin{2*argInInd};
        case 'psf', psf = varargin{2*argInInd};
        case 'noisesigma', noiseSigma = varargin{2*argInInd};
		case 'ip_img_dim', sRestoreOptions.ipImgDim = varargin{2*argInInd}; 
		case 'op_img_dim', sRestoreOptions.opImgDim = varargin{2*argInInd};
        case 'n_conv_method', sRestoreOptions.nConvMethod = varargin{2*argInInd};
            
        otherwise
            error( sprintf( 'unrecognized option %s', szKey ) );
    end
end

N = length( y );
M = length( betaTrue );

NON_ZERO_THRES1 = 1e3*eps; % Threshold for non-zero coefficients
NON_ZERO_THRES2 = 1e-6; % A looser threshold

N_FIG_CHAT = 1;
N_FIG_MUHAT = 2;
N_FIG_SIGN = 3;
N_FIG_DAUB = 4;

%sRestoreOptions.reconSanityInterval = sRestoreOptions.NUM_ITERS - 1;
sRestoreOptions.reconSanityInterval = 5;

sRestoreOptions.constXColumnNorm = sRestoreOptions.fnhXColumnNorm( psf, 1 );
disp( sprintf( 'constXColumnNorm = %g', sRestoreOptions.constXColumnNorm ) );

% Check that the column norm is strictly positive
if sRestoreOptions.constXColumnNorm < NON_ZERO_THRES1
    error( 'const x column norm not positive' );
end

% SANITY: check another random column to see that the column norm is the
% same.
colRandom = 1 + floor( rand * M );
columnNormCheck = sRestoreOptions.fnhXColumnNorm( psf, colRandom );
disp( sprintf( 'sanity: colRandom=%g, columnNormCheck=%g', colRandom, columnNormCheck ) );
if ( abs( columnNormCheck - sRestoreOptions.constXColumnNorm ) > 0.05*columnNormCheck )
    error( 'column sanity check failed' );
end
    
if ( sRestoreOptions.NUM_ITERS > M ) | ( sRestoreOptions.NUM_ITERS < 1 )
    error( 'invalid NUM_ITERS' );
elseif ( sRestoreOptions.NUM_ITERS == M )
    disp( 'Running LARS to LS solution' );
else
    disp( 'Running LARS, but stopping short of LS solution' );
end

if ( noiseSigma <= 0 ), error( 'noiseSigma must be >0' ), end;

if ( prod( sRestoreOptions.opImgDim ) ~= N ) | ...
   ( prod( sRestoreOptions.ipImgDim ) ~= M )
	error( 'op or ip image dimensions are inconsistent' );
end

if sRestoreOptions.N_DEBUG > 1
	if ( M ~= N ), error( 'sorry, for N_DEBUG>1, must have M=N' ), end;
end

% Metrics for convergence of iterations

thetaMaxDiffHist = []; % l_inf norm of difference of 2 consecutive estimates
thetaSumDiffHist = []; % l_1 norm of difference ...

errorNorm1Hist = []; % l_1 norm of difference
errorNorm2Hist = []; % l_2 norm ...

bigCHatHist = [];

% Initialize Gblur object. 
%//////////////////////////////////////////////////////////////////////////

if sRestoreOptions.nConvMethod == 1
	GX = Gblur( ones( sRestoreOptions.opImgDim ), ...
        'psf', psf, ...
        'type', 'conv,fft', ...
        'nDim', length( sRestoreOptions.ipImgDim ) );
elseif nConvMethod == 2
	GX = Gblur( ones( size( imgNoisy ) ), ...
        'psf', psf, ...
        'type', 'conv,same', ...
        'nDim', length( sRestoreOptions.ipImgDim ) );
else
    error( 'illegal nConvMethod' );
end

if 0 % DEBUG
    yNoiseless = GX * ...
        (sRestoreOptions.constXColumnNorm * reshape( betaTrue, sRestoreOptions.ipImgDim ) );
    yNoiseless = yNoiseless / sRestoreOptions.constXColumnNorm;
    if ( length( sRestoreOptions.opImgDim ) == 3 )
        plotSlices( 1, yNoiseless, sRestoreOptions.opImgDim(3), 0, 2 );
        disp( 'Pausing...' ), pause;
    else
        warning( 'Not plotting yNoiseless...' );
    end    
end

activeSet = [];
activeSetPrev = [];
vecAActiveSet = [];

muHatPrevActiveSet = zeros( N, 1 );
muHatNextActiveSet = zeros( size( muHatPrevActiveSet ) );

betaHatPrevActiveSet = zeros( M, 1 );
betaHatNextActiveSet = zeros( size( betaHatPrevActiveSet ) );
betaHatBest = zeros( size( betaHatPrevActiveSet ) );
betaHatBestInd = -1;

gammaMat = zeros( M, 2 );

% Keep track of various statistics during the iterations.

% SURE cost criterion
vecCost = zeros( sRestoreOptions.NUM_ITERS, 1 ); 
% l_0 of estimate
vecL0 = zeros( size( vecCost ) ); 
% soft threshold
vecSoftThres = zeros( size( vecCost ) );

bLassoKludge = 0; % Flag whenever we have to do a LARS-LASSO kludge
jHat = -1;
jKludge = -1;
nextBigCHat = -1;

disp( 'Checks passed...Starting iterations' );

% Time the iterations
tCpuTimeStart = cputime;

for i = 1:sRestoreOptions.NUM_ITERS
              
    %%% Step 1: Compute the correlations
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    disp( sprintf( '*** Iter %g/%g ***', i, sRestoreOptions.NUM_ITERS ) );
    
    residual = reshape( y - muHatPrevActiveSet, sRestoreOptions.opImgDim );
    cHat = full( reshape( GX' * residual, N, 1 ) );
    cHat = cHat / sRestoreOptions.constXColumnNorm;
    cHatSign = sign( cHat );
    
    BigCHat = max( abs( cHat ) );
    BigCHat = BigCHat( 1 ); % Want this to be a real
    bigCHatHist = [ bigCHatHist; BigCHat ]; % Save history
        
    %%% Step 2: Construct {mu,beta}HatNextActiveSet
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
       
    if ( length(activeSet) == 0 )

        % length of active set is 0. This occurs in the first iteration.     

        [dummy, activeSet] = max( abs( cHat ) );
        if ( length( activeSet ) > 1 )
            error( 'in 1st iter, size of activeSet>1?' );
        end
        
    else
        
        % In subsequent iterations, the size of the activeSet MUST be
        % strictly bigger than 1.
        
        if ( length(activeSet) <= 1 )
            error( 'subsequent iter, size of activeSet<=1?' );
        end
        
    end
    
    disp( 'active set:' ), disp( activeSet );
    disp( 'cHat(activeSet):' ), disp( cHat(activeSet)' );    
    
    % SANITY: check that BigCHat matches the predicted one
    if ( i > 1 ), disp( 'Sanity: BigCHat matches nextBigCHat' );
        CHatDiff = abs( BigCHat - nextBigCHat );
        if CHatDiff > NON_ZERO_THRES1
            warning( sprintf( 'BigCHat=%g != predicted value=%g', BigCHat, nextBigCHat ) );         
            disp( 'FAILED! Pausing...' ), pause;
        end
    end
    
    % SANITY: Check that the absolute value of the correlations in the
    % active set match BigCHat
    disp( 'Sanity: |cHat| of active set matches BigCHat' );
    for k = 1:length( activeSet )
		mismatchDelta = abs( abs(cHat(activeSet(k))) - BigCHat );
        if ( mismatchDelta > 2e-2*BigCHat )
            error( sprintf( 'FAILED! |cHat(%g)|=%g != BigCHat=%g, delta=%g', ...
				activeSet(k), cHat( activeSet(k) ), BigCHat, mismatchDelta ) );
        end
    end
    
    % Plot vector of cHats and (the max) BigCHat
    if 0
        figure( N_FIG_CHAT ), clf, plot( cHat, 'bo' ), hold on;
        plot( [1, length(cHat)], [BigCHat, BigCHat], 'k--' );
        plot( [1, length(cHat)], [-BigCHat, -BigCHat], 'k--' );
        title( sprintf( 'iter %g', i ) );
        ylabel( 'cHat' );
    end
    
    % Obtain X_A.
    [X_A, sInfo] = sRestoreOptions.fnhX( psf, cHatSign, activeSet );
    % Check dimensions of X_A are correct
    if ( size( X_A, 1 ) ~= N ) | ( size( X_A, 2 ) ~= length(activeSet) )
        error( sprintf( 'X_A ~ %g x %g => not of correct dimensions', ...
            size( X_A, 1 ), size( X_A, 2 ) ) );
    end
    % Divide X_A by the constant column norm to make the columns unit norm
    X_A = X_A / sRestoreOptions.constXColumnNorm;
    GActiveSet = full( X_A' * X_A );
    disp( sprintf( 'INFO: cond number of G_A: %g', cond( GActiveSet ) ) );
    %disp( 'GActiveSet:' ), disp( GActiveSet );

    OneVecActiveSet = ones( length(activeSet), 1 ); % |A|-vec
    
    %GInvActiveSet = inv( GActiveSet );
    %tmp = full( OneVecActiveSet' * GInvActiveSet * OneVecActiveSet );
    tmp = OneVecActiveSet' * ( GActiveSet\OneVecActiveSet );
    AActiveSet = 1/sqrt(tmp);
    disp( sprintf( 'INFO: AActiveSet=%g', AActiveSet ) );
    % Save history of A_A
    vecAActiveSet = [ vecAActiveSet, AActiveSet ];
    
    % SANITY: check that AActiveSet is non-increasing if we didn't apply
    % the LASSO kludge.
    disp( 'Sanity: check AActiveSet is non-increasing' );
    if ( length( vecAActiveSet ) > 1 )
        if ( vecAActiveSet(end) > vecAActiveSet(end-1) )
            warning( 'FAILED! AActiveSet increased' );
            disp( 'Pausing...' ), pause;
        end
    end

    %wActiveSet = AActiveSet * GInvActiveSet * OneVecActiveSet; % |A|-vec
    wActiveSet = AActiveSet * ( GActiveSet\OneVecActiveSet );

    % Compute the equiangular vector
    uActiveSet = full( X_A * wActiveSet );

    % Run sanity on equiangular vector uActiveSet
    if 1, disp( 'Sanity: check uActiveSet' );

        % SANITY: check that |uActiveSet| = 1
        normUActiveSet = norm( uActiveSet, 2 );
        if abs( normUActiveSet - 1 ) > NON_ZERO_THRES2
            error( 'Magnitude of equiangular vect |u| != 1' );
        end

        % SANITY: next, check that X_A' * u_A = A_A * 1_A
        uActiveSetDotProd = full( X_A' * uActiveSet );
        % Each element of uActiveSetDotProd should equal AActiveSet
        for k = 1:length(activeSet)
            errMargin = abs( uActiveSetDotProd(k) - AActiveSet );
            if errMargin > abs(1e-6*AActiveSet)
                error( sprintf( '(X_A^T * u_A)(%g) != AActiveSet [%g]', k, errMargin ) );
            end
        end
        
    end

    %%% Step 3: compute a
    %%%%%%%%%%%%%%%%%%%%%
    a = reshape( full( GX' * reshape( uActiveSet, sRestoreOptions.opImgDim ) ), M, 1 );
    a = a / sRestoreOptions.constXColumnNorm;
    %disp( 'a:' ), disp( a );
       
    %%% Step 4: Compute gammaHat and jHat
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    jHatColumn = -1; % DEBUG variable

    % Want to compute:
    % gammaHat = min^+_{j \in A^c} { 
    %    (CHat - chat_j)/(A_A - a_j), 
    %    (CHat + chat_j)/(A_A + a_j) }

    if ( length( activeSet ) < M )
                
        activeSetC = setdiff( 1:M, activeSet );
        % DEBUG
        %disp( 'activeSetC:' ), disp( activeSetC(:)' );
        if ( length( activeSetC ) == 0 ), error( 'length of activeSetC==0' ), end;
        clear gammaMat;
        gammaMat = zeros( length( activeSetC ), 2 );
        
        % Compute the 1st column of gammaMat
        numer = BigCHat - cHat( activeSetC );
        denom = AActiveSet - a( activeSetC );
        tmp = numer(:) ./ denom(:);
        % SANITY
        if length( tmp(:) ) ~= length( activeSetC ), error( 'Sanity on length of tmp failed' ), end;
        gammaMat( :, 1 ) = tmp(:);
        
        % Compute the 2nd column of gammaMat
        numer = BigCHat + cHat( activeSetC );
        denom = AActiveSet + a( activeSetC );
        tmp = numer(:) ./ denom(:);
        % SANITY
        if length( tmp(:) ) ~= length( activeSetC ), error( 'Sanity on length of tmp failed' ), end;
        gammaMat( :, 2 ) = tmp(:);

        % Take the min over the 1st column
        gammaMatPosInd = find( ( gammaMat(:,1) > NON_ZERO_THRES2 ) & isfinite( gammaMat(:,1) ) );
        if ~isempty( gammaMatPosInd )
            gammaMatCol1Min = min( gammaMat( gammaMatPosInd, 1 ) );
            gammaMatCol1MinInd = find( gammaMat(:,1) == gammaMatCol1Min );
            if length( gammaMatCol1MinInd ) > 1
                error( 'cannot handle length( gammaMatCol1MinInd ) > 1' );
            end            
            % SANITY
%             disp( 'gammaMatCol1Min:' ), disp( gammaMatCol1Min );
%             disp( 'gammaMatCol1MinInd:' ), disp( gammaMatCol1MinInd );
%             disp( 'jHat of min:' ), disp( activeSetC( gammaMatCol1MinInd ) );
%             disp( 'row of gammaMat:' ), disp( gammaMat( gammaMatCol1MinInd, : ) );
        else
            gammaMatCol1Min = inf;
        end
        
        % Then the min over the 2nd column
        gammaMatPosInd = find( ( gammaMat(:,2) > NON_ZERO_THRES2 ) & isfinite( gammaMat(:,2) ) );
        if ~isempty( gammaMatPosInd )
            gammaMatCol2Min = min( gammaMat( gammaMatPosInd, 2 ) );
            gammaMatCol2MinInd = find( gammaMat(:,2) == gammaMatCol2Min );
            if length( gammaMatCol2MinInd ) > 1
                error( 'cannot handle length( gammaMatCol2MinInd ) > 1' );
            end
            % SANITY
%             disp( 'gammaMatCol2Min:' ), disp( gammaMatCol2Min );
%             disp( 'gammaMatCol2MinInd:' ), disp( gammaMatCol2MinInd );
%             disp( 'jHat of min:' ), disp( activeSetC( gammaMatCol2MinInd ) );
%             disp( 'row of gammaMat:' ), disp( gammaMat( gammaMatCol2MinInd, : ) );
        else
            gammaMatCol2Min = inf;
        end

        % And then combine to find the min over the entire matrix
        [gmm, gmmInd] = min( [gammaMatCol1Min, gammaMatCol2Min] );
        if isfinite( gmm )
            gammaHat = gmm;
            if ( gmmInd == 1 )
                % DEBUG
                %disp( 'Getting jHat from 1st column of gammaMat' );
                jHatColumn = 1;
                jHat = activeSetC( gammaMatCol1MinInd );
                disp( sprintf( 'min gamma entry is (%g,%g) = %g', ...
                    gammaMatCol1MinInd, 1, ...
                    gammaMat( gammaMatCol1MinInd, 1 ) ) );
            else
                % DEBUG
                %disp( 'Getting jHat from 2nd column of gammaMat' );
                jHatColumn = 2;
                jHat = activeSetC( gammaMatCol2MinInd );
                disp( sprintf( 'min gamma entry is (%g,%g) = %g', ...
                    gammaMatCol2MinInd, 2, ...
                    gammaMat( gammaMatCol2MinInd, 2 ) ) );
            end
        else
            jHatColumn = -1; % Invalid value
            error( 'could not find gammaHat and jHat' );
        end

    else % The activeSet is the entire range of valid indices 1, ..., M
        
        gammaHat = BigCHat/AActiveSet;
        jHat = -1;
        
    end
    
    % Form the next estimator of mu 
    muHatNextActiveSet = muHatPrevActiveSet + gammaHat * uActiveSet; 
    
    % Update the active set   
    activeSetPrev = activeSet;
    activeSet = [activeSet, jHat];

    % Always sort the column indices in the active set in ascending order
    [activeSetSorted, activeSetSortedInd] = sort( activeSet );
    activeSet = activeSetSorted;

    %%% Step 5: compute gammaTilde
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    % Construct dHat, which is the next direction that LARS takes in
    % beta-space.
    dHat = zeros( M, 1 ); % M-vec
    for j = 1:length( activeSetPrev )
        jj = activeSetPrev( j );
        dHat( jj ) = cHatSign( jj ) * wActiveSet( j );
        if sRestoreOptions.N_DEBUG >= 2
            disp( sprintf( 'j=%g, jj=%g, cHatSign=%g, wActiveSet=%g', ...
                j, jj, cHatSign(jj), wActiveSet(j) ) );
        end
    end

    % SANITY: print dHat
%     if 1
%         disp( 'dHat:' ), disp( sparse(dHat) );
%     end
  
    % Compute -\hat{\beta}_j/\hat{d}_j for j in the PREV. ACTIVE SET
    gammaTildeVec = zeros( length( activeSetPrev ), 1 ); % |A|-vec
    for j = 1:length( activeSetPrev )
        jj = activeSetPrev( j );
        gammaTildeVec(j) = -1*betaHatPrevActiveSet( jj ) / dHat( jj );
    end
          
    % Find the min gamma s.t. it is positive 
    gammaTildePosInd = find( ( gammaTildeVec > NON_ZERO_THRES2 ) & isfinite( gammaTildeVec ) );
    gtm = min( gammaTildeVec( gammaTildePosInd ) );
    
    switch length( gtm )
        case 0
            % No positive elements in gammaTilde
            gammaTilde = inf;
            jTilde = -1;
        case 1
            % This is the "normal" case
            gtmInd = find( gammaTildeVec == gtm );
            if length( gtmInd ) > 1
                error( 'cannot handle length( gtmInd ) > 1' );
            elseif length( gtmInd ) == 1
                gammaTilde = gtm;
                jTilde = activeSetPrev( gtmInd );
                disp( sprintf( 'min gamma kludge is %g-th entry: %g', gtmInd(1), gtm ) );
            else
                error( 'Cannot find minimum in gammaTildeVec?' );
            end
        otherwise
            error( 'Multiple minima in gammaTildeVec?' );
    end

    bLassoKludge = 0; % Re-initialize
    
    if ( gammaTilde < gammaHat ), disp( '==> LARS-LASSO KLUDGE' );

        % Form the next estimator of mu
        muHatNextActiveSet = muHatPrevActiveSet + gammaTilde * uActiveSet;

        % Restore the previous active set
        activeSet = activeSetPrev;

        % Subtract jTilde from the active set
        activeSet = setdiff( activeSet, jTilde );

        % Always sort the column indices in the active set in ascending order
        [activeSetSorted, activeSetSortedInd] = sort( activeSet );
        activeSet = activeSetSorted;

        % Form the estimate of beta
        betaHatNextActiveSet = betaHatPrevActiveSet + gammaTilde * dHat;
        
        bLassoKludge = 1; % Set the kludge flag
        gammaConst = gammaTilde;
        
    else
        
        % No need to twiddle anything
        betaHatNextActiveSet = betaHatPrevActiveSet + gammaHat * dHat;
        gammaConst = gammaHat;
    end

    disp( [ sprintf( 'gammaHat=%g, jHat=%g/', gammaHat, jHat ), ...
        sprintf( 'gammaTilde=%g, jTilde=%g\n', gammaTilde, jTilde ) ] );
                   
    % Next predicted BigCHat
    nextBigCHat = BigCHat - gammaConst * AActiveSet;
    disp( sprintf( 'nextBigCHat = %g', nextBigCHat ) );
    
    % Check that the new correlation equals nextBigCHat if no LASSO kludge
    % was performed.
    if ~bLassoKludge, disp( 'Sanity: checking new correlation value' );
        cjHat = cHat( jHat ) - gammaConst * a( jHat );
        disp( sprintf( 'SANITY: cjHat = %g', cjHat ) );
        
        % Find entry in gammaMat
        jHatGammaMatIndex = find( activeSetC == jHat );
        if length( jHatGammaMatIndex ) ~= 1, error( 'failed to find jHat in activeSetC' ), end;
        %disp( 'jHatGammaMatIndex:' ), disp( jHatGammaMatIndex );
        %disp( 'jHat entry in gammaMat:' ), disp( gammaMat( jHatGammaMatIndex, : ) );
        
        % Check that |cjHat| equals nextBigCHat. 
        if abs( abs(cjHat) - nextBigCHat ) > NON_ZERO_THRES1                                 
            % Re-compute gamma
            switch jHatColumn
                case 1
                    gamma2 = ( BigCHat - cHat( jHat ) ) / ( AActiveSet - a( jHat ) );
                case 2
                    gamma2 = ( BigCHat + cHat( jHat ) ) / ( AActiveSet + a( jHat ) );
                otherwise
                    error( 'trying to re-compute gammaHat with invalid jHatColumn' );
            end
            disp( sprintf( 'Re-computed gammaHat = %g', gamma2 ) );
            warning( 'cjHat not equal to nextBigCHat??' );
        end
    end
                  
    % SANITY: check that mu = X * beta
    disp( 'Sanity: check that mu = X * beta' );
    muCheck = reshape( GX * reshape( betaHatNextActiveSet, sRestoreOptions.ipImgDim ), ...
        N, 1 );
    muCheck = muCheck / sRestoreOptions.constXColumnNorm;
    
    if ( norm( muCheck - muHatNextActiveSet ) > NON_ZERO_THRES1 ) 
        figure( N_FIG_MUHAT ), clf; 
        plot( muHatNextActiveSet, 'ko' ), hold on;
        plot( muCheck, 'rx' );
        error( 'mu check failed' );
    end        
    
    % Find non-zero indices of beta. This should be equal to the elements
    % of the active set. XXX: or does it?
    
    betaNzInd = find( abs( betaHatNextActiveSet ) > NON_ZERO_THRES1 ); 
    badInd = setxor( betaNzInd(:)', activeSet(:)' );

	disp( 'non-zero indices of beta:' ), disp( betaNzInd(:)' );
	disp( 'active set:' ), disp( activeSet(:)' );
	disp( 'discrepencies:' ), disp( badInd(:)' );
	disp( 'beta(active set):' ), disp( betaHatNextActiveSet(activeSet) ); 

%     if ( length( badInd ) > 0 )
%         warning( 'set of non-zero indices of beta != active set' );
%         disp( 'activeSet:' ), disp( activeSet );
%         disp( 'betaNzInd:' ), disp( betaNzInd );
%         disp( 'badInd:' ), disp( badInd );
%         error( 'aborting' );
%     end
        
   	% Update metrics
    errorNorm1Hist = [ errorNorm1Hist, norm( betaHatNextActiveSet(:) - betaTrue(:), 1 ) ];
    errorNorm2Hist = [ errorNorm2Hist, norm( betaHatNextActiveSet(:) - betaTrue(:), 2 ) ];
	betaHatDiff = gammaConst * dHat(:);
    % The l_infinity norm of the difference
    thetaMaxDiffHist = [thetaMaxDiffHist; max( abs( betaHatDiff ) )];
    % The l_1 norm of the difference
    thetaSumDiffHist = [thetaSumDiffHist; sum( abs( betaHatDiff ) )];
  
    lambda = 2* BigCHat * sRestoreOptions.constXColumnNorm;
    vecSoftThres(i) = lambda/2;
    disp( sprintf( 'current BigCHat = %g, lambda=%g', BigCHat, lambda ) );
    
    if ( ( sRestoreOptions.bCheckDaubechies == 1 ) & ...
         ( mod( i, sRestoreOptions.reconSanityInterval ) == 0 ) & ...
         ( i > 1 ) )
            
        % Compute using Daubechies' iterations. Here, we MUST have the size of the i/p and
		% o/p images be the same, i.e., M=N.
        [sDIter, sDIterOptions] = emgauss_restore( 'forward_op', psf, ...
            'noisy_output', reshape( y, sRestoreOptions.opImgDim ), ...
            'noise_sigma', noiseSigma, ...
            'theta_true', reshape( betaTrue, sRestoreOptions.ipImgDim ), ...
            'm_selector', 'daub2', ...
            'num_em_iters', 2e4, ...
            'num_max_grad_search_failures', -1, ...
            'theta0_selector', 'Hty2', ...
            'theta0', [], ...
            'a_val', -1, ...
            'w_val', -1, ...
            't_val', -1, ...
            'sparse_ind', 'w', ...       
            'stopping_cond', 'fixed', ...
            'stopping_cond_tol', noiseSigma, ...
            'stopping_cond_win', 3e3, ...
            'stopping_cond_last_fac', 2, ...
            'thetastar', 1, ...
            'soft_thres_t', -1, ...
            'hparam_rest_interval', 1e2, ...
            'bseparate_hparam_est', 0, ...
            'gspecial', -1, ...
            'lambda', lambda ...
            );
        
        larsSol = betaHatNextActiveSet / sRestoreOptions.constXColumnNorm;
        daubSol = sDIter.last;
%        figure( N_FIG_DAUB ), clf;
%        plot( larsSol(:), 'bx' ), hold on;
%        plot( daubSol(:), 'ro' );
%        title( 'LARS-LASSO vs. Daubechies' );
%        disp( 'Pausing...' ), pause;
		solDelta = larsSol(:) - daubSol(:);
		disp( sprintf( 'l_inf norm of solDelta: %g', norm( solDelta, inf ) ) );
		disp( sprintf( 'l_1 norm of solDelta: %g', norm( solDelta, 1 ) ) );
    end

    betaOriginalScale = betaHatNextActiveSet(:) / sRestoreOptions.constXColumnNorm;
    vecL0( i ) = sum( abs( betaOriginalScale ) > NON_ZERO_THRES1 ); % l_0(beta)
    vecL1( i ) = sum( abs( betaOriginalScale ) ); % l_1(beta)

    % Compute the cost function  
    predictionErr = y - muHatNextActiveSet;
    rss = sum( predictionErr(:).^2 );
    vecCost( i ) = N*noiseSigma^2 + rss + 2*noiseSigma^2*vecL0(i);
    
    if ( ( betaHatBestInd < 1 ) | ...
         ( (betaHatBestInd >= 1) & (vecCost(betaHatBestInd) > vecCost(i)) ) )
        betaHatBestInd = i;
        betaHatBest = betaHatNextActiveSet;
    end
    
    if 1, disp( 'Sanity: LASSO sign' );
        % LASSO sanity: check that sign(beta_j) = sign(c_j) = s_j, for j in the
        % active set.
        
        % Recompute sign(c_j)
        residualTmp = reshape( y - muHatNextActiveSet, sRestoreOptions.opImgDim );
        cHatTmp = full( reshape( GX' * residualTmp, N, 1 ) );
        cHatTmp = cHatTmp / sRestoreOptions.constXColumnNorm;
        cHatSignTmp = sign( cHatTmp );
      
	  	if ~bLassoKludge, activeSetToConsider = activeSetPrev;
		else, activeSetToConsider = activeSet;
		end 

        % Find the number of mismatches
        vecMismatch = ( cHatSignTmp( activeSetToConsider ) ~= sign( betaHatNextActiveSet( activeSetToConsider ) ) );
        if sum( vecMismatch ) ~= 0           
            disp( 'c-values:' ), disp( cHatTmp(activeSetToConsider) );
            disp( 'beta-values:' ), disp( betaHatNextActiveSet(activeSetToConsider) );
            error( 'LASSO solution sign contraint violated!!!' );
        end
    end     
    
    %%% Step 6: Update state variables
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
         
    muHatPrevActiveSet = muHatNextActiveSet;
    betaHatPrevActiveSet = betaHatNextActiveSet;
    
end

tCpuTimeStop = cputime;

% Re-scale last beta and best beta
betaHat = betaHatPrevActiveSet / sRestoreOptions.constXColumnNorm;
betaHatBest = betaHatBest / sRestoreOptions.constXColumnNorm;

% Defines the structure of sRestore
sRestoreElems = { ...
    'lastIter', 'betaHat'; ...
    'last', 'betaHatBest'; ...
    'lastInd1', 'betaHatBestInd'; ...
    'muHat', 'muHatPrevActiveSet'; ...
    'bigCHat', 'bigCHatHist'; ...
    'activeSet', 'activeSet'; ...
    'vecL1', 'vecL1'; ...
    'vecL0', 'vecL0'; ...
    'vecSoftThres', 'vecSoftThres'; ...
    'vecCost', 'vecCost'; ...
    'cputime', '[tCpuTimeStart, tCpuTimeStop]'; ...
    'thetaMaxDiff', 'thetaMaxDiffHist'; ...
    'thetaSumDiff', 'thetaSumDiffHist'; ...
    'errorNorm1', 'errorNorm1Hist'; ...
    'errorNorm2', 'errorNorm2Hist'; ...
    };

for elemInd = 1:size( sRestoreElems, 1 )
    szCmd = sprintf( 'sOut.%s = %s;', ...
        char( sRestoreElems(elemInd,1) ), ...       
        char( sRestoreElems(elemInd,2) ) );
    try
        eval( szCmd );
    catch
        warning( sprintf( 'Could not execute: %s', szCmd ) );
    end
end

    
    

    
    
    


