%
% LARS-LASSO-SURE-ext. Adapted from lars3. Usage is
% sOut = lars4( varargin )
% 
% The problem we consider is
% y = X * beta + n
% y ~ N-dim vector, beta ~ M-dim vector, X ~ N x M
%

function sOut = lars4( varargin )

% Default restore options

sRestoreOptions = struct( 'X', [], 'XColumnNorm', [], ...
    'B_DEBUG', 0, 'reconSanityInterval', 10 );
noiseSigma = -1;

% Process input arguments

lenVarargin = length( varargin );
numArgsIn = floor( lenVarargin/2 );

if 2*numArgsIn ~= lenVarargin, error( 'even ip args expected' ), end

for argInInd = 1:numArgsIn
    
    szKey = lower( varargin{2*argInInd-1} );    
    switch szKey        
        case 'noisy_output', y = varargin{2*argInInd};      
        case 'beta_true', betaTrue = varargin{2*argInInd}; 

        case 'num_iters', sRestoreOptions.NUM_ITERS = varargin{2*argInInd};    
		case 'x', sRestoreOptions.X = varargin{2*argInInd};   
        case 'x_column_norm', sRestoreOptions.XColumnNorm = varargin{2*argInInd};  
        case 'n_debug', sRestoreOptions.N_DEBUG = varargin{2*argInInd};
        case 'psf', psf = varargin{2*argInInd};
        case 'noisesigma', noiseSigma = varargin{2*argInInd};
        case 'n_phi2', sRestoreOptions.N_PHI2 = varargin{2*argInInd};
        case 'n_phi1', sRestoreOptions.N_PHI1 = varargin{2*argInInd};
        case 'vec_strat', sRestoreOptions.vecStrat = varargin{2*argInInd};
            
        otherwise
            error( sprintf( 'unrecognized option %s', szKey ) );
    end
end

% Check validity of arguments
N = length( y );
M = length( betaTrue );
if ( size( sRestoreOptions.X, 1 ) ~= N ) | ( size( sRestoreOptions.X, 2 ) ~= M )
    error( 'size of X incompatible with y, betaTrue' );
end

if ( sRestoreOptions.NUM_ITERS > M ) | ( sRestoreOptions.NUM_ITERS < 1 )
    error( 'invalid NUM_ITERS' );
elseif ( sRestoreOptions.NUM_ITERS == M )
    disp( 'Running LARS to LS solution' );
else
    disp( 'Running LARS, but stopping short of LS solution' );
end

if ( noiseSigma <= 0 ), error( 'noiseSigma must be >0' ), end;

if 0
    yNoiseless = sRestoreOptions.X * (sRestoreOptions.XColumnNorm .* betaTrue);
    imgSize = sqrt( N );
    if ( imgSize*imgSize ~= N ), error( 'length(y) not a perfect square' ), end;
    figure( 1 ), clf, imagesc( reshape( yNoiseless, imgSize, imgSize ) ), colorbar;
    ylabel( 'Sanity on X, XColumnNorm' );
    title( 'yNoiseless' );
    disp( 'Pausing...' ), pause;
end

% Metrics for convergence of iterations

thetaMaxDiffHist = []; % l_inf norm of difference of 2 consecutive estimates
thetaSumDiffHist = []; % l_1 norm of difference ...

errorNorm1Hist = []; % l_1 norm of difference
errorNorm2Hist = []; % l_2 norm ...

bigCHatHist = [];

% Pre-compute XTX = X' * X
XTX = (sRestoreOptions.X)' * (sRestoreOptions.X);

activeSet = [];
vecAActiveSet = [];

muHatPrevActiveSet = zeros( N, 1 );
muHatNextActiveSet = zeros( size( muHatPrevActiveSet ) );

betaHatPrevActiveSet = zeros( M, 1 );
betaHatNextActiveSet = zeros( size( betaHatPrevActiveSet ) );

% The best l_1 penalty estimator according to the SURE criterion
betaHatBest = zeros( size( betaHatPrevActiveSet ) );

% LARS-LASSO index corresponding to the best betaHat
betaHatBestInd = -1;

gammaMat = zeros( M, 2 );

% Cost vector for l_1 penalty
vecCost = zeros( sRestoreOptions.NUM_ITERS, 1 );

vecT = zeros( size( vecCost ) );
vecL0 = zeros( size( vecCost ) );

bLassoKludge = 0; % Flag whenever we have to do a LARS-LASSO kludge
jHat = -1;
jKludge = -1;
nextBigCHat = -1;

% Time the iterations
tCpuTimeStart = cputime;

for i = 1:sRestoreOptions.NUM_ITERS
    
    if sRestoreOptions.N_DEBUG, disp( sprintf( 'i=%g', i ) ), end;      
    
    % Step 1: Compute the correlations
    cHat = sRestoreOptions.X' * ( y - muHatPrevActiveSet );    
    s = sign( cHat );
    
    betaNzInd = find( abs( betaHatPrevActiveSet ) > eps ); % Find non-zero indices
    vecMismatch = ( s(betaNzInd) ~= sign( betaHatPrevActiveSet( betaNzInd ) ) );
    %disp( 'betaNzInd:' ), disp( betaNzInd(:)' );
    %disp( 'c-values:' ), disp( cHat(betaNzInd) );
    %disp( 'beta-values:' ), disp( betaHatPrevActiveSet(betaNzInd) );
    
    % LARS sanity: check that sign(beta_j) = sign(c_j) = s_j, for j in the
    % active set
    if sum( vecMismatch ) ~= 0
        figure( 2 ), clf, plot( s(betaNzInd), 'kx' ), hold on;
        plot( sign(betaHatPrevActiveSet(betaNzInd)), 'ro' );
        legend( 's', 'sign(beta)', 0 );
        error( 'LASSO solution sign contraint violated' );
    end
                   
    % Step 2: Get the active set and sort in ascending order of indices
    disp( 'STEP 2: Adjusting active set' );
    
    if ( i == 1 )
        % In the first iteration, use max to locate the active set
        [BigCHat, activeSet] = max( abs( cHat ) );
    else
        
        % Do this step in subsequent iterations        
        BigCHat = max( abs( cHat ) );

        if ~bLassoKludge
            % In subsequent iterations, add to the active set manually. It
            % turns out that there are some numerical differences in the cHats
            % that are supposed to be the same.
            activeSet = [activeSet, jHat];
        else
            activeSet = setdiff( activeSet, jTilde );
        end
        
    end                    
    
    [activeSetSorted, activeSetSortedInd] = sort( activeSet );
    activeSet = activeSetSorted;
    nActiveSetSize = length( activeSet );
    
    BigCHat = BigCHat( 1 ); % Want this to be a real        
    % SANITY: check that BigCHat matches the predicted one
    if ( i > 1 )
        if abs( BigCHat - nextBigCHat ) > 1e3*eps
            error( sprintf( 'BigCHat=%g != predicted value=%g', BigCHat, nextBigCHat ) );
        end
    end
    
    % Print correlations
    %disp( 'cHat(activeSet):' ), disp( cHat(activeSet) );
    
    % SANITY: Check that the absolute value of the correlations in the
    % active set match BigCHat
    for n = 1:nActiveSetSize
		mismatchDelta = abs( abs(cHat(activeSet(n))) - BigCHat );
        if ( mismatchDelta > 1e-3*BigCHat )
            error( sprintf( '|cHat(%g)|=%g != BigCHat=%g, delta=%g', ...
				activeSet(n), cHat( activeSet(n) ), BigCHat, mismatchDelta ) );
        end
    end
    
    % Plot the cHat and the max BigCHat
    if sRestoreOptions.N_DEBUG
        figure( 1 ), clf, plot( cHat, 'bo' ), hold on;
        plot( [1, length(cHat)], [BigCHat, BigCHat], 'k--' );
        plot( [1, length(cHat)], [-BigCHat, -BigCHat], 'k--' );
        title( sprintf( 'iter %g', i ) );
        ylabel( 'cHat' );
    end
    
    % Print info to console
    disp( sprintf( 'iter=%g, size of active set=%g', i, nActiveSetSize ) );
    disp( sprintf( 'BigCHat: %g', BigCHat ) );
    disp( 'active set:' ), disp( activeSet );
        
    % Step 3: Construct the variables related to the active set A
    
    clear nzI nzJ nzS;
    nzI = zeros( nActiveSetSize, 1 );
    nzJ = zeros( size( nzI ) );
    nzS = zeros( size( nzI ) );
    
    for j = 1:nActiveSetSize
        nzI(j) = activeSet(j);
        nzJ(j) = j;
        nzS(j) = s( activeSet(j) );
    end
    
    % Allocate space for SActiveSet, which is of size M x nActiveSetSize
	SActiveSet = sparse( nzI, nzJ, nzS, M, nActiveSetSize );
    
    % Compute the other variables
    GActiveSet = SActiveSet' * XTX * SActiveSet; % size |A| x |A|
    %disp( 'GActiveSet:' ), disp( GActiveSet );
    GInvActiveSet = inv( GActiveSet );
    OneVecActiveSet = ones( nActiveSetSize, 1 ); % |A|-vec
    tmp = full( OneVecActiveSet' * GInvActiveSet * OneVecActiveSet );
    AActiveSet = 1/sqrt(tmp);
    
    disp( sprintf( 'SANITY: AActiveSet=%g', AActiveSet ) );
    vecAActiveSet = [ vecAActiveSet, AActiveSet ];
    
    % SANITY: check that AActiveSet is non-increasing if we didn't apply
    % the LASSO kludge.
    if ( ( ~bLassoKludge ) & ( length( vecAActiveSet ) > 1 ) )
        if ( vecAActiveSet(end) > vecAActiveSet(end-1) )
            warning( 'Sanity failed: AActiveSet increased' );
            disp( 'Pausing...' ), pause;
        end
    end
    
    wActiveSet = AActiveSet * GInvActiveSet * OneVecActiveSet; % |A|-vec
    uActiveSet = (sRestoreOptions.X * SActiveSet) * wActiveSet; % N-vec
    normUActiveSet = norm( uActiveSet, 2 );
    disp( sprintf( 'SANITY: |u|=%g', normUActiveSet ) );
    % SANITY: check that |uActiveSet| = 1
    if abs( normUActiveSet - 1 ) > 1e-6
        error( 'Sanity failed: |u| != 1' );
    end
    
    % This should be equal AActiveSet * 1
    tmp = full( ( SActiveSet' * sRestoreOptions.X' ) * uActiveSet );
    %disp( 'SANITY: X_A^T * u_A:' ), disp( tmp(:)' );
    % SANITY: check that each element of tmp equals AActiveSet
    for n = 1:nActiveSetSize
        errMargin = abs( tmp(n) - AActiveSet );
        if errMargin > abs(1e-6*AActiveSet)
            error( sprintf( 'Sanity failed: (X_A^T * u_A)(%g) != AActiveSet [%g]', n, errMargin ) );
        end
    end
    
    a = sRestoreOptions.X' * uActiveSet; % M-vec
    %disp( 'a:' ), disp( a );
    
    % Construct dHat
    dHat = zeros( M, 1 ); % M-vec
    for j = 1:nActiveSetSize
        jj = activeSet( j );
        dHat( jj ) = s( jj ) * wActiveSet( j );
    end
    
    % Step 4a: Compute gammaHat and jHat
    
    jHatColumn = -1; % DEBUG variable
    
    if ( i < M )
        
        % We can only do this if i < NN. Here, compute
        % gammaHat = min^+_{j \in A^c} { (CHat - chat_j)/(A_A - a_j), (CHat
        % + chat_j)/(A_A + a_j) }
        
        activeSetC = setdiff( 1:M, activeSet );
        % DEBUG
        %disp( 'activeSetC:' ), disp( activeSetC(:)' );
        clear gammaMat;
        gammaMat = zeros( length( activeSetC ), 2 );
        
        % Compute the 1st column of gammaMat
        numer = BigCHat - cHat( activeSetC );
        denom = AActiveSet - a( activeSetC );
        tmp = numer(:) ./ denom(:);
        % SANITY
        if length( tmp(:) ) ~= length( activeSetC ), error( 'Sanity on length of tmp failed' ), end;
        gammaMat( :, 1 ) = tmp(:);
        
        % Compute the 2nd column of gammaMat
        numer = BigCHat + cHat( activeSetC );
        denom = AActiveSet + a( activeSetC );
        tmp = numer(:) ./ denom(:);
        % SANITY
        if length( tmp(:) ) ~= length( activeSetC ), error( 'Sanity on length of tmp failed' ), end;
        gammaMat( :, 2 ) = tmp(:);

        % Take the min over each column
        gammaMatPosInd = find( ( gammaMat(:,1) > eps ) & isfinite( gammaMat(:,1) ) );
        if ~isempty( gammaMatPosInd )
            gammaMatCol1Min = min( gammaMat( gammaMatPosInd, 1 ) );
            gammaMatCol1MinInd = find( gammaMat(:,1) == gammaMatCol1Min );
            if length( gammaMatCol1MinInd ) > 1
                error( 'cannot handle length( gammaMatCol1MinInd ) > 1' );
            end            
            % SANITY
            %disp( 'gammaMatCol1Min:' ), disp( gammaMatCol1Min );
            %disp( 'gammaMatCol1MinInd:' ), disp( gammaMatCol1MinInd );
            %disp( 'jHat of min:' ), disp( activeSetC( gammaMatCol1MinInd ) );
            %disp( 'row of gammaMat:' ), disp( gammaMat( gammaMatCol1MinInd, : ) );
        else
            gammaMatCol1Min = inf;
        end
        
        gammaMatPosInd = find( ( gammaMat(:,2) > eps ) & isfinite( gammaMat(:,2) ) );
        if ~isempty( gammaMatPosInd )
            gammaMatCol2Min = min( gammaMat( gammaMatPosInd, 2 ) );
            gammaMatCol2MinInd = find( gammaMat(:,2) == gammaMatCol2Min );
            if length( gammaMatCol2MinInd ) > 1
                error( 'cannot handle length( gammaMatCol2MinInd ) > 1' );
            end
            % SANITY
            %disp( 'gammaMatCol2Min:' ), disp( gammaMatCol2Min );
            %disp( 'gammaMatCol2MinInd:' ), disp( gammaMatCol2MinInd );
            %disp( 'jHat of min:' ), disp( activeSetC( gammaMatCol2MinInd ) );
            %disp( 'row of gammaMat:' ), disp( gammaMat( gammaMatCol2MinInd, : ) );
        else
            gammaMatCol2Min = inf;
        end

        % And then combine to find the min over the entire matrix
        [gmm, gmmInd] = min( [gammaMatCol1Min, gammaMatCol2Min] );
        if isfinite( gmm )
            gammaHat = gmm;
            if ( gmmInd == 1 )
                % DEBUG
                %disp( 'Getting jHat from 1st column of gammaMat' );
                jHatColumn = 1;
                jHat = activeSetC( gammaMatCol1MinInd );
            else
                % DEBUG
                %disp( 'Getting jHat from 2nd column of gammaMat' );
                jHatColumn = 2;
                jHat = activeSetC( gammaMatCol2MinInd );
            end
        else
            jHatColumn = -1; % Invalid value
            error( 'could not find gammaHat and jHat' );
        end

    else
        
        % Do this if i == M. This is the last step
        gammaHat = BigCHat/AActiveSet;
        jHat = -1;
        
    end
    
    % Step 4b: compute gammaTilde
 
    warning off; % XXX: We are going to get some divide-by-zero errors here!
    gammaTildeVec = -betaHatPrevActiveSet ./ dHat; % This is a M-vector
    warning on;

    % Find the min gamma s.t. it is positive 
    gammaTildePosInd = find( ( gammaTildeVec > eps ) & isfinite( gammaTildeVec ) );
    gtm = min( gammaTildeVec( gammaTildePosInd ) );
    
    switch length( gtm )
        case 0
            % No positive elements in gammaTilde
            gammaTilde = inf;
            jTilde = -1;
        case 1
            % This is the "normal" case
            gtmInd = find( gammaTildeVec == gtm );
            if length( gtmInd ) > 1
                error( 'cannot handle length( gtmInd ) > 1' );
            elseif length( gtmInd ) == 1
                gammaTilde = gtm;
                jTilde = gtmInd;
                % Sanity: check that jTilde is in the active set
                if ~ismember( jTilde, activeSet )
                    error( 'jTilde not in the active set' );
                end
            else
                error( 'Cannot find minimum in gammaTildeVec?' );
            end
        otherwise
            error( 'Multiple minima in gammaTildeVec?' );
    end

    % Step 4c: get gammaConst
    
    bLassoKludge = 0; % Re-initialize
    
    if ( gammaTilde < gammaHat )
        bLassoKludge = 1;
        gammaConst = gammaTilde;
        disp( '*** LARS-LASSO KLUDGE ***' );
    else
        gammaConst = gammaHat;
    end
    
    disp( [ sprintf( 'gammaHat=%g, jHat=%g/', gammaHat, jHat ), ...
        sprintf( 'gammaTilde=%g, jTilde=%g\n', gammaTilde, jTilde ), ...
        sprintf( 'gammaConst=%g', gammaConst ) ] );    
    
    % SANITY
    nextBigCHat = BigCHat - gammaConst * AActiveSet;
    disp( sprintf( 'nextBigCHat = %g', nextBigCHat ) );
    
    % Check that the new correlation equals nextBigCHat
    if ~bLassoKludge
        cjHat = cHat( jHat ) - gammaConst * a( jHat );
        disp( sprintf( 'cjHat = %g', cjHat ) );
        
        % Find entry in gammaMat
        jHatGammaMatIndex = find( activeSetC == jHat );
        if length( jHatGammaMatIndex ) ~= 1, error( 'failed to find jHat in activeSetC' ), end;
        %disp( 'jHatGammaMatIndex:' ), disp( jHatGammaMatIndex );
        %disp( 'jHat entry in gammaMat:' ), disp( gammaMat( jHatGammaMatIndex, : ) );
        
        % Check if |cjHat| equals nextBigCHat. It should!
        if abs( abs(cjHat) - nextBigCHat ) > eps                                    
            % Re-compute gamma
            switch jHatColumn
                case 1
                    gamma2 = ( BigCHat - cHat( jHat ) ) / ( AActiveSet - a( jHat ) );
                case 2
                    gamma2 = ( BigCHat + cHat( jHat ) ) / ( AActiveSet + a( jHat ) );
                otherwise
                    error( 'trying to re-compute gammaHat with invalid jHatColumn' );
            end
            disp( sprintf( 'Re-computed gammaHat = %g', gamma2 ) );
            error( 'cjHat not equal to nextBigCHat??' );
        end
    end
    
    
    % Step 5: Form the next estimator of mu and beta
    muHatNextActiveSet = muHatPrevActiveSet + gammaConst * uActiveSet;    
    betaHatNextActiveSet = betaHatPrevActiveSet + gammaConst * dHat;
   
   	% Update metrics
    errorNorm1Hist = [ errorNorm1Hist, norm( betaHatNextActiveSet(:) - betaTrue(:), 1 ) ];
    errorNorm2Hist = [ errorNorm2Hist, norm( betaHatNextActiveSet(:) - betaTrue(:), 2 ) ];
	betaHatDiff = gammaConst * dHat(:);
    % The l_infinity norm of the difference
    thetaMaxDiffHist = [thetaMaxDiffHist; max( abs( betaHatDiff ) )];
    % The l_1 norm of the difference
    thetaSumDiffHist = [thetaSumDiffHist; sum( abs( betaHatDiff ) )];

    bigCHatHist = [ bigCHatHist; BigCHat ];
    
    % MASSIVE DEBUG. XXX: make sure the dimensions of y and betaTrue are
    % accurate.
    
    if ( ( sRestoreOptions.N_DEBUG > 1 ) & ...
         ( mod( i, sRestoreOptions.reconSanityInterval ) == 1 ) & ...
         ( i > 1 ) )
    
        lambda = 2* BigCHat * sRestoreOptions.XColumnNorm( activeSet(1) );
        
        disp( sprintf( 'BigCHat = %g, lambda=%g', BigCHat, lambda ) );

        % Compute using Daubechies' iterations. XXX: This probably won't
        % work as is. Needs to be modified--TODO.
        
        [sDIter, sDIterOptions] = emgauss_restore( 'forward_op', psf, ...
            'noisy_output', reshape( y, 32, 32 ), ...
            'noise_sigma', noiseSigma, ...
            'theta_true', reshape( betaTrue, 32, 32 ), ...
            'm_selector', 'daub2', ...
            'num_em_iters', 2e4, ...
            'num_max_grad_search_failures', -1, ...
            'theta0_selector', 'Hty2', ...
            'theta0', [], ...
            'a_val', -1, ...
            'w_val', -1, ...
            't_val', -1, ...
            'sparse_ind', 'w', ...          
            'stopping_cond', 'fixed', ...
            'stopping_cond_tol', noiseSigma, ...
            'stopping_cond_win', 3e3, ...
            'stopping_cond_last_fac', 2, ...
            'thetastar', 1, ...
            'soft_thres_t', -1, ...
            'hparam_rest_interval', 1e2, ...
            'lambda', lambda ...
            );
        
        larsSol = betaHatNextActiveSet./sRestoreOptions.XColumnNorm;
        daubSol = sDIter.last;
        figure( 2 ), imagesc( larsSol, 32, 32 ), colorbar;
        title( 'LARS-LASSO' );        
        figure( 3 ), imagesc( daubSol ), colorbar;
        title( 'Daubechies' );
        figure( 4 ), plot( larsSol(:), 'bx' ), hold on, plot( daubSol(:), 'ro' );
        title( 'LARS-LASSO vs. Daubechies' );
        disp( 'Pausing...' ), pause;
        
%         vecDIter = reshape( sDIter.last, 1024, 1 );
%         vecMismatch = ( sign( reshape( betaHatNextActiveSet, 1024, 1 ) ) ~= sign( vecDIter ) );
%         if sum( vecMismatch ) ~= 0
%             indMismatch = find( vecMismatch ~= 0 );
%             disp( 'Mismatch in LARS-LASSO vs. Daubechies' );
%             disp( 'indMismatch:' ), disp( indMismatch );
%             disp( 'Mismatched values/LARS-LASSO:' ); 
%             disp( betaHatNextActiveSet( indMismatch ) );
%             disp( 'Mismatched values/Daubechies:' );
%             disp( vecDIter( indMismatch ) );
%         end
        
    end
    
    vecL1( i ) = sum( abs( betaHatNextActiveSet(:) ) ); % l_1(beta)
    vecL0( i ) = sum( abs( betaHatNextActiveSet(:) ) > 1e3*eps ); % l_0(beta)
    
    % Compute the cost function  
    predictionErr = y - muHatNextActiveSet;
    rss = sum( predictionErr(:).^2 );
    vecCost( i ) = N*noiseSigma^2 + rss + 2*noiseSigma^2*vecL0(i);
    
    % Check to see if we have a better solution
    
    if ( ( betaHatBestInd < 1 ) | ...
         ( (betaHatBestInd >= 1) & (vecCost(betaHatBestInd) > vecCost(i)) ) )
        betaHatBestInd = i;
        betaHatBest = betaHatNextActiveSet;
    end
       
    % Step 6: Update state variables
    muHatPrevActiveSet = muHatNextActiveSet;
    betaHatPrevActiveSet = betaHatNextActiveSet;
    
    if sRestoreOptions.N_DEBUG, disp( 'End of iter, press key' ), pause, end;
    
end

% Re-scale last beta and best beta
betaHat = betaHatPrevActiveSet ./ sRestoreOptions.XColumnNorm;
betaHatBest = betaHatBest ./ sRestoreOptions.XColumnNorm;
% muCheck should equal muHatPrevActiveSet
muCheck = sRestoreOptions.X * betaHatPrevActiveSet;

N_SEARCH_STRAT2 = sRestoreOptions.vecStrat(1);
N_SEARCH_STRAT3 = sRestoreOptions.vecStrat(2);
N_SEARCH_STRAT4 = sRestoreOptions.vecStrat(3);

%--------------------------------------------------------------------------
%
% Fix phi1 and change phi2/phi1 from 0 to 1.
%
%--------------------------------------------------------------------------

% Initialize
phi1 = sRestoreOptions.XColumnNorm(1) * bigCHatHist( betaHatBestInd );

% The valid range of phi2 is 0 <= phi2/phi1 <= 1
% phi2/phi1=1 => Soft threshold
% phi2/phi0=0 => Hard threshold

phi2FracVec = linspace( 1, 0, sRestoreOptions.N_PHI2 );

% Cost vector for hybrid hard-soft thresholding function using search
% strategy 2
vecCost2 = zeros( size( phi2FracVec ) );

% Initialize the first element, which coincides with the corresponding
% value in vecCost.
vecCost2( 1 ) = vecCost( betaHatBestInd );

% The best beta according to the hybrid thresholding function
betaHatBest2 = betaHatBest; % Initialize
betaHatBest2Ind = 1;

% Go through and sweep phi2 through the remaining values.

if N_SEARCH_STRAT2 == 1

    for i = 2:sRestoreOptions.N_PHI2

        disp( sprintf( 'Doing frac=%g', phi2FracVec(i) ) );
        pause( 5 );

        % Compute using Daubechies' iterations
        [sDIter, sDIterOptions] = emgauss_restore( 'forward_op', psf, ...
            'noisy_output', reshape( y, 32, 32 ), ...
            'noise_sigma', noiseSigma, ...
            'theta_true', reshape( betaTrue, 32, 32 ), ...
            'm_selector', 'daub_hard_soft', ...
            'num_em_iters', 1e4, ...
            'num_max_grad_search_failures', -1, ...
            'theta0_selector', 'Hty2', ...
            'theta0', [], ...
            'a_val', -1, ...
            'w_val', -1, ...
            't_val', -1, ...
            'sparse_ind', 'w', ...
            'stopping_cond', 'fixed', ...
            'stopping_cond_tol', noiseSigma, ...
            'stopping_cond_win', 3e3, ...
            'stopping_cond_last_fac', 2, ...
            'thetastar', 1, ...
            'soft_thres_t', -1, ...
            'hparam_rest_interval', 1e2, ...
            'bseparate_hparam_est', 0, ...
            'lambda', 2*phi1, ...
            'hard_soft_frac', phi2FracVec( i ) ...
            );

        % Compute the cost function
        daubSol = sDIter.last(:);
        % Recall that the columns of X have been normalized to unit norm.
        % Hence, we must premultiply by the original column norms
        predictionErr = y - sRestoreOptions.X*(sRestoreOptions.XColumnNorm.*daubSol);
        % Compute the residual sum of square errors
        rss = sum( predictionErr(:).^2 );

        daubSolThres = soft_thres( daubSol, (1-phi2FracVec(i))*phi1 );
        nnzSoftThres = sum( abs( daubSolThres(:) ) > 1e3*eps );
        vecCost2( i ) = N*noiseSigma^2 + rss + 2*noiseSigma^2*nnzSoftThres;

        % Check to see if we have a better solution

        if ( vecCost2( betaHatBest2Ind ) > vecCost2( i ) )
            betaHatBest2Ind = i;
            betaHatBest2 = daubSol;
        end

    end % for

end

%--------------------------------------------------------------------------
%
% Fix phi2 and vary phi1
%
%--------------------------------------------------------------------------

% Initialize
phi1Init = sRestoreOptions.XColumnNorm(1) * bigCHatHist( betaHatBestInd );
phi1Vec = [phi1Init];

if N_SEARCH_STRAT3 == 1

    [sDIter, sDIterOptions] = emgauss_restore( 'forward_op', psf, ...
        'noisy_output', reshape( y, 32, 32 ), ...
        'noise_sigma', noiseSigma, ...
        'theta_true', reshape( betaTrue, 32, 32 ), ...
        'm_selector', 'daub_hard_soft', ...
        'num_em_iters', 1e2, ...
        'num_max_grad_search_failures', -1, ...
        'theta0_selector', 'Hty2', ...
        'theta0', [], ...
        'a_val', -1, ...
        'w_val', -1, ...
        't_val', -1, ...
        'sparse_ind', 'w', ...
        'stopping_cond', 'fixed', ...
        'stopping_cond_tol', noiseSigma, ...
        'stopping_cond_win', 3e3, ...
        'stopping_cond_last_fac', 2, ...
        'thetastar', 1, ...
        'soft_thres_t', -1, ...
        'hparam_rest_interval', 1e2, ...
        'bseparate_hparam_est', 0, ...
        'lambda', 2*phi1Init, ...
        'hard_soft_frac', 1 ...
        );

    phi1Max = sqrt(2*log(M))*sDIterOptions.alpha;
    if ( phi1Init >= phi1Max )
        error( sprintf( 'phi1Init=%g >= phi1Max=%g', phi1Init, phi1Max ) );
    end
    disp( sprintf( 'phi1Max=%g', phi1Max ) );

    % Sweep phi1 from the optimal l_1 value to the max
    %phi1Vec = linspace( phi1Init, phi1Max, sRestoreOptions.N_PHI1 );
    
    % Sweep phi1 from the optimal l_1 value to larger values
    indLast = betaHatBestInd;
    indDelta = floor( (indLast-1)/(sRestoreOptions.N_PHI1-1) );
    if ( indDelta < 1 ), error( 'Cannot find a suitable indDelta' ), end;
    indFirst = indLast - indDelta * ( sRestoreOptions.N_PHI1 - 1 );
    if ( indFirst < 1 ), error( 'unexpected: indFirst < 1' ), end;
    phi1Vec = sRestoreOptions.XColumnNorm(1) * bigCHatHist( indFirst:indDelta:indLast );
    
    % SANITY
    if ( length( phi1Vec ) ~= sRestoreOptions.N_PHI1 )
        error( 'phi1Vec failed sanity' );
    end
    
    % DEBUG
    %disp( 'phi1Vec:' ), disp( phi1Vec ), disp( 'Pausing...' ), pause;
    
end

phi2 = phi1Init;

% Cost vector for hybrid hard-soft thresholding function using search
% strategy 3
vecCost3 = zeros( size( phi1Vec ) );

% Initialize the last element, which coincides with the corresponding
% value in vecCost.
vecCost3( end ) = vecCost( betaHatBestInd );

% The best beta according to the hybrid thresholding function
betaHatBest3 = betaHatBest; % Initialize
betaHatBest3Ind = length( vecCost3 ); % Last entry

% Go through and sweep phi1 

if N_SEARCH_STRAT3 == 1

    for i = 1:(sRestoreOptions.N_PHI1-1)

        disp( sprintf( 'Doing phi1=%g', phi1Vec(i) ) );
        pause( 5 );

        % Compute using Daubechies' iterations
        [sDIter, sDIterOptions] = emgauss_restore( 'forward_op', psf, ...
            'noisy_output', reshape( y, 32, 32 ), ...
            'noise_sigma', noiseSigma, ...
            'theta_true', reshape( betaTrue, 32, 32 ), ...
            'm_selector', 'daub_hard_soft', ...
            'num_em_iters', 1e4, ...
            'num_max_grad_search_failures', -1, ...
            'theta0_selector', 'Hty2', ...
            'theta0', [], ...
            'a_val', -1, ...
            'w_val', -1, ...
            't_val', -1, ...
            'sparse_ind', 'w', ...
            'stopping_cond', 'fixed', ...
            'stopping_cond_tol', noiseSigma, ...
            'stopping_cond_win', 3e3, ...
            'stopping_cond_last_fac', 2, ...
            'thetastar', 1, ...
            'soft_thres_t', -1, ...
            'hparam_rest_interval', 1e2, ...
            'bseparate_hparam_est', 0, ...
            'lambda', 2*phi1Vec(i), ...
            'hard_soft_frac', phi2/phi1Vec(i) ...
            );

        % Compute the cost function
        daubSol = sDIter.last(:);
        % Recall that the columns of X have been normalized to unit norm.
        % Hence, we must premultiply by the original column norms
        predictionErr = y - sRestoreOptions.X*(sRestoreOptions.XColumnNorm.*daubSol);
        % Compute the residual sum of square errors
        rss = sum( predictionErr(:).^2 );

        daubSolThres = soft_thres( daubSol, phi1Vec(i) - phi2 );
        nnzSoftThres = sum( abs( daubSolThres(:) ) > 1e3*eps );
        vecCost3( i ) = N*noiseSigma^2 + rss + 2*noiseSigma^2*nnzSoftThres;

        % Check to see if we have a better solution

        if ( vecCost3( betaHatBest3Ind ) > vecCost3( i ) )
            betaHatBest3Ind = i;
            betaHatBest3 = daubSol;
        end

    end

end

%--------------------------------------------------------------------------
%
% Vary phi1 and phi2
%
%--------------------------------------------------------------------------

if N_SEARCH_STRAT4 == 1
    
    [sDIter, sDIterOptions] = emgauss_restore( 'forward_op', psf, ...
        'noisy_output', reshape( y, 32, 32 ), ...
        'noise_sigma', noiseSigma, ...
        'theta_true', reshape( betaTrue, 32, 32 ), ...
        'm_selector', 'daub_hard_soft', ...
        'num_em_iters', 1e2, ...
        'num_max_grad_search_failures', -1, ...
        'theta0_selector', 'Hty2', ...
        'theta0', [], ...
        'a_val', -1, ...
        'w_val', -1, ...
        't_val', -1, ...
        'sparse_ind', 'w', ...
        'stopping_cond', 'fixed', ...
        'stopping_cond_tol', noiseSigma, ...
        'stopping_cond_win', 3e3, ...
        'stopping_cond_last_fac', 2, ...
        'thetastar', 1, ...
        'soft_thres_t', -1, ...
        'hparam_rest_interval', 1e2, ...
        'bseparate_hparam_est', 0, ...
        'lambda', 2*phi1Init, ...
        'hard_soft_frac', 1 ...
        );
    
    phi1Max = sqrt(2*log(M))*sDIterOptions.alpha;
    disp( sprintf( 'phi1Max=%g', phi1Max ) );
    if ( phi1Init >= phi1Max )
        error( sprintf( 'phi1Init=%g >= phi1Max=%g', phi1Init, phi1Max ) );
    end
        
end

if mod( sRestoreOptions.N_PHI1, 2 ) == 0 % Even
    numPaddingBefore = sRestoreOptions.N_PHI1/2;
    numPaddingAfter = sRestoreOptions.N_PHI1/2 - 1;

else
    numPaddingBefore = (sRestoreOptions.N_PHI1-1)/2;
    numPaddingAfter = numPaddingBefore;
end

% Make sure that we reach the first index of bigCHatHist
indDelta4 = floor( (betaHatBestInd-1)/numPaddingBefore );
if ( indDelta4 < 1 ), error( 'Cannot find a suitable indDelta4' ), end;

% DEBUG
disp( sprintf( 'padding before/after=%g/%g, delta4=%g, betaHatBestInd=%g', ...
    numPaddingBefore, numPaddingAfter, indDelta4, betaHatBestInd ) );

indFirst = betaHatBestInd - numPaddingBefore*indDelta4;
indLast = betaHatBestInd + numPaddingAfter*indDelta4;
    
phi1VecStrat4 = []; % Initialize
matCost4 = [];
betaHatBest4 = [];
betaHatBest4Ind = [];

if N_SEARCH_STRAT4 == 1

    if ( indFirst < 1 ), error( 'indFirst too small' ), end;
    if ( indLast > length( bigCHatHist ) ), error( 'indLast too large' ), end;
    phi1VecStrat4 = sRestoreOptions.XColumnNorm(1) * bigCHatHist( indFirst:indDelta4:indLast );

    % DEBUG
    if 1, disp( 'phi1VecStrat4:' ), disp( phi1VecStrat4 ), disp( 'Pausing' ), pause, end;
    
    % Matrix of costs
    matCost4 = zeros( length( phi1VecStrat4 ), length( phi2FracVec ) );
    matCost4( :, 1 ) = reshape( vecCost( indFirst:indDelta4:indLast ), sRestoreOptions.N_PHI1, 1 );

    betaHatBest4 = betaHatBest;
    betaHatBest4Ind = [ numPaddingBefore + 1, 1 ];

    % SANITY
    if matCost4( betaHatBest4Ind(1), betaHatBest4Ind(2) ) ~= vecCost( betaHatBestInd )
        error( 'sanity on matCost4 failed' );
    end

    % Go through and sweep phi1 and phi2

    for i = 1:sRestoreOptions.N_PHI1
        for j = 2:sRestoreOptions.N_PHI2
          
            disp( sprintf( '(%g,%g) Doing phi1=%g, phi2 frac=%g', i, j, phi1VecStrat4(i), phi2FracVec(j) ) );
            pause( 5 );

            % Compute using Daubechies' iterations
            [sDIter, sDIterOptions] = emgauss_restore( 'forward_op', psf, ...
                'noisy_output', reshape( y, 32, 32 ), ...
                'noise_sigma', noiseSigma, ...
                'theta_true', reshape( betaTrue, 32, 32 ), ...
                'm_selector', 'daub_hard_soft', ...
                'num_em_iters', 1e4, ...
                'num_max_grad_search_failures', -1, ...
                'theta0_selector', 'Hty2', ...
                'theta0', [], ...
                'a_val', -1, ...
                'w_val', -1, ...
                't_val', -1, ...
                'sparse_ind', 'w', ...
                'stopping_cond', 'fixed', ...
                'stopping_cond_tol', noiseSigma, ...
                'stopping_cond_win', 3e3, ...
                'stopping_cond_last_fac', 2, ...
                'thetastar', 1, ...
                'soft_thres_t', -1, ...
                'hparam_rest_interval', 1e2, ...
                'bseparate_hparam_est', 0, ...
                'lambda', 2*phi1VecStrat4(i), ...
                'hard_soft_frac', phi2FracVec(j) ...
                );

            % Compute the cost function
            daubSol = sDIter.last(:);
            % Recall that the columns of X have been normalized to unit norm.
            % Hence, we must premultiply by the original column norms
            predictionErr = y - sRestoreOptions.X*(sRestoreOptions.XColumnNorm.*daubSol);
            % Compute the residual sum of square errors
            rss = sum( predictionErr(:).^2 );

            daubSolThres = soft_thres( daubSol, (1-phi2FracVec(j))*phi1VecStrat4(i) );
            nnzSoftThres = sum( abs( daubSolThres(:) ) > 1e3*eps );
            matCost4( i, j ) = N*noiseSigma^2 + rss + 2*noiseSigma^2*nnzSoftThres;

            % Check to see if we have a better solution

            if ( matCost4( betaHatBest4Ind(1), betaHatBest4Ind(2) ) > matCost4( i, j ) )
                betaHatBest4Ind = [i, j];
                betaHatBest4 = daubSol;
            end

        end % loop phi2
    end % loop phi1
    
end

tCpuTimeStop = cputime;

% Defines the structure of sRestore

sRestoreElems = { ...
    'bestBeta4', 'betaHatBest4'; ...
    'lastInd4', 'betaHatBest4Ind'; ...
    'cost4', 'matCost4'; ...
    'phi1VecStrat4', 'phi1VecStrat4'; ...
    'bestBeta3', 'betaHatBest3'; ...
    'lastInd3', 'betaHatBest3Ind'; ...
    'cost3', 'vecCost3'; ...
    'phi1Vec', 'phi1Vec'; ...
    'bestBeta2', 'betaHatBest2'; ...
    'lastInd2', 'betaHatBest2Ind'; ...
    'cost2', 'vecCost2'; ...
    'phi2FracVec', 'phi2FracVec'; ...
    'bestBeta', 'betaHatBest'; ...
    'lastInd', 'betaHatBestInd'; ...
    'cost', 'vecCost'; ...       
    'bigCHat', 'bigCHatHist'; ...
    'l1', 'vecL1'; ...
    'l0', 'vecL0'; ...
    'cputime', '[tCpuTimeStart, tCpuTimeStop]'; ...
    'thetaMaxDiff', 'thetaMaxDiffHist'; ...
    'thetaSumDiff', 'thetaSumDiffHist'; ...
    'errorNorm1', 'errorNorm1Hist'; ...
    'errorNorm2', 'errorNorm2Hist'; ...
    };

for elemInd = 1:size( sRestoreElems, 1 )
    szCmd = sprintf( 'sOut.%s = %s;', ...
        char( sRestoreElems(elemInd,1) ), ...       
        char( sRestoreElems(elemInd,2) ) );
    try
        eval( szCmd );
    catch
        warning( sprintf( 'Could not execute: %s', szCmd ) );
    end
end

    
    

    
    
    


