%
% LARS-LASSO-SURE-ext. Adapted from lars4. Usage is
% sOut = lars5( varargin )
% 
% The problem we consider is
% y = X * beta + n
% y ~ N-dim vector, beta ~ M-dim vector, X ~ N x M
%

function sOut = lars5( varargin )

% Default restore options

sRestoreOptions = struct( 'X', [], 'XColumnNorm', [], ...
    'B_DEBUG', 0, 'reconSanityInterval', 10 );
noiseSigma = -1;

% Process input arguments

lenVarargin = length( varargin );
numArgsIn = floor( lenVarargin/2 );

if 2*numArgsIn ~= lenVarargin, error( 'even ip args expected' ), end

for argInInd = 1:numArgsIn
    
    szKey = lower( varargin{2*argInInd-1} );    
    switch szKey        
        case 'noisy_output', y = varargin{2*argInInd};
        case 'beta_true', betaTrue = varargin{2*argInInd};
        case 'num_iters', sRestoreOptions.NUM_ITERS = varargin{2*argInInd};
        case 'x', sRestoreOptions.X = varargin{2*argInInd};
        case 'x_column_norm', sRestoreOptions.XColumnNorm = varargin{2*argInInd};
        case 'n_debug', sRestoreOptions.N_DEBUG = varargin{2*argInInd};
        case 'psf', psf = varargin{2*argInInd};
        case 'noisesigma', noiseSigma = varargin{2*argInInd};
		case 'ip_img_dim', ipImgDim = varargin{2*argInInd}; 
		case 'op_img_dim', opImgDim = varargin{2*argInInd};
        case 'n_phi1', sRestoreOptions.n_phi1 = varargin{2*argInInd};
        case 'n_phi2', sRestoreOptions.n_phi2 = varargin{2*argInInd};
        case 'n_strat', sRestoreOptions.n_strat = varargin{2*argInInd};
            
        otherwise
            error( sprintf( 'unrecognized option %s', szKey ) );
    end
    
end

% Check validity of arguments
N = length( y );
M = length( betaTrue );
if ( size( sRestoreOptions.X, 1 ) ~= N ) | ( size( sRestoreOptions.X, 2 ) ~= M )
    error( 'size of X incompatible with y, betaTrue' );
end

if ( sRestoreOptions.NUM_ITERS > M ) | ( sRestoreOptions.NUM_ITERS < 1 )
    error( 'invalid NUM_ITERS' );
elseif ( sRestoreOptions.NUM_ITERS == M )
    disp( 'Running LARS to LS solution' );
else
    disp( 'Running LARS, but stopping short of LS solution' );
end

if ( noiseSigma <= 0 ), error( 'noiseSigma must be >0' ), end;

if 0
    yNoiseless = sRestoreOptions.X * (sRestoreOptions.XColumnNorm .* betaTrue);
    imgSize = sqrt( N );
    if ( imgSize*imgSize ~= N ), error( 'length(y) not a perfect square' ), end;
    figure( 1 ), clf, imagesc( reshape( yNoiseless, imgSize, imgSize ) ), colorbar;
    ylabel( 'Sanity on X, XColumnNorm' );
    title( 'yNoiseless' );
    disp( 'Pausing...' ), pause;
end

if ( prod( opImgDim ) ~= N ) | ( prod( ipImgDim ) ~= M )
	error( 'op or ip image dimensions are inconsistent' );
end

if sRestoreOptions.N_DEBUG > 1
	if ( M ~= N ), error( 'sorry, for N_DEBUG>1, must have M=N' ), end;
end

% Metrics for convergence of iterations

thetaMaxDiffHist = []; % l_inf norm of difference of 2 consecutive estimates
thetaSumDiffHist = []; % l_1 norm of difference ...

errorNorm1Hist = []; % l_1 norm of difference
errorNorm2Hist = []; % l_2 norm ...

bigCHatHist = [];

% Pre-compute XTX = X' * X
XTX = (sRestoreOptions.X)' * (sRestoreOptions.X);

activeSet = [];
vecAActiveSet = [];

muHatPrevActiveSet = zeros( N, 1 );
muHatNextActiveSet = zeros( size( muHatPrevActiveSet ) );

betaHatPrevActiveSet = zeros( M, 1 );
betaHatNextActiveSet = zeros( size( betaHatPrevActiveSet ) );

% The best l_1 penalty estimator according to the SURE criterion
betaHatBest = zeros( size( betaHatPrevActiveSet ) );

% LARS-LASSO index corresponding to the best betaHat
betaHatBestInd = -1;

gammaMat = zeros( M, 2 );

% Cost vector for l_1 penalty
vecCost = zeros( sRestoreOptions.NUM_ITERS, 1 );

vecT = zeros( size( vecCost ) );
vecL0 = zeros( size( vecCost ) );

bLassoKludge = 0; % Flag whenever we have to do a LARS-LASSO kludge
jHat = -1;
jKludge = -1;
nextBigCHat = -1;

% Time the iterations
tCpuTimeStart = cputime;

for i = 1:sRestoreOptions.NUM_ITERS
    
    if sRestoreOptions.N_DEBUG, disp( sprintf( 'i=%g', i ) ), end;      
    
    % Step 1: Compute the correlations
    cHat = sRestoreOptions.X' * ( y - muHatPrevActiveSet );    
    s = sign( cHat );
    
    betaNzInd = find( abs( betaHatPrevActiveSet ) > eps ); % Find non-zero indices
    vecMismatch = ( s(betaNzInd) ~= sign( betaHatPrevActiveSet( betaNzInd ) ) );
    %disp( 'betaNzInd:' ), disp( betaNzInd(:)' );
    %disp( 'c-values:' ), disp( cHat(betaNzInd) );
    %disp( 'beta-values:' ), disp( betaHatPrevActiveSet(betaNzInd) );
    
    % LARS sanity: check that sign(beta_j) = sign(c_j) = s_j, for j in the
    % active set
    if sum( vecMismatch ) ~= 0
        figure( 2 ), clf, plot( s(betaNzInd), 'kx' ), hold on;
        plot( sign(betaHatPrevActiveSet(betaNzInd)), 'ro' );
        legend( 's', 'sign(beta)', 0 );
        error( 'LASSO solution sign contraint violated' );
    end
                   
    % Step 2: Get the active set and sort in ascending order of indices
    disp( 'STEP 2: Adjusting active set' );
    
    if ( i == 1 )
        % In the first iteration, use max to locate the active set
        [BigCHat, activeSet] = max( abs( cHat ) );
    else
        
        % Do this step in subsequent iterations        
        BigCHat = max( abs( cHat ) );

        if ~bLassoKludge
            % In subsequent iterations, add to the active set manually. It
            % turns out that there are some numerical differences in the cHats
            % that are supposed to be the same.
            activeSet = [activeSet, jHat];
        else
            activeSet = setdiff( activeSet, jTilde );
        end
        
    end                    
    
    [activeSetSorted, activeSetSortedInd] = sort( activeSet );
    activeSet = activeSetSorted;
    nActiveSetSize = length( activeSet );
    
    BigCHat = BigCHat( 1 ); % Want this to be a real        
    % SANITY: check that BigCHat matches the predicted one
    if ( i > 1 )
        if abs( BigCHat - nextBigCHat ) > 1e3*eps
            error( sprintf( 'BigCHat=%g != predicted value=%g', BigCHat, nextBigCHat ) );
        end
    end
    
    % Print correlations
    %disp( 'cHat(activeSet):' ), disp( cHat(activeSet) );
    
    % SANITY: Check that the absolute value of the correlations in the
    % active set match BigCHat
    for n = 1:nActiveSetSize
		mismatchDelta = abs( abs(cHat(activeSet(n))) - BigCHat );
        if ( mismatchDelta > 1e-3*BigCHat )
            error( sprintf( '|cHat(%g)|=%g != BigCHat=%g, delta=%g', ...
				activeSet(n), cHat( activeSet(n) ), BigCHat, mismatchDelta ) );
        end
    end
    
    % Plot the cHat and the max BigCHat
    if sRestoreOptions.N_DEBUG
        figure( 1 ), clf, plot( cHat, 'bo' ), hold on;
        plot( [1, length(cHat)], [BigCHat, BigCHat], 'k--' );
        plot( [1, length(cHat)], [-BigCHat, -BigCHat], 'k--' );
        title( sprintf( 'iter %g', i ) );
        ylabel( 'cHat' );
    end
    
    % Print info to console
    disp( sprintf( 'iter=%g, size of active set=%g', i, nActiveSetSize ) );
    disp( sprintf( 'BigCHat: %g', BigCHat ) );
    disp( 'active set:' ), disp( activeSet );
        
    % Step 3: Construct the variables related to the active set A
    
    clear nzI nzJ nzS;
    nzI = zeros( nActiveSetSize, 1 );
    nzJ = zeros( size( nzI ) );
    nzS = zeros( size( nzI ) );
    
    for j = 1:nActiveSetSize
        nzI(j) = activeSet(j);
        nzJ(j) = j;
        nzS(j) = s( activeSet(j) );
    end
    
    % Allocate space for SActiveSet, which is of size M x nActiveSetSize
	SActiveSet = sparse( nzI, nzJ, nzS, M, nActiveSetSize );
    
    % Compute the other variables
    GActiveSet = SActiveSet' * XTX * SActiveSet; % size |A| x |A|
    %disp( 'GActiveSet:' ), disp( GActiveSet );
    GInvActiveSet = inv( GActiveSet );
    OneVecActiveSet = ones( nActiveSetSize, 1 ); % |A|-vec
    tmp = full( OneVecActiveSet' * GInvActiveSet * OneVecActiveSet );
    AActiveSet = 1/sqrt(tmp);
    
    disp( sprintf( 'SANITY: AActiveSet=%g', AActiveSet ) );
    vecAActiveSet = [ vecAActiveSet, AActiveSet ];
    
    % SANITY: check that AActiveSet is non-increasing if we didn't apply
    % the LASSO kludge.
    if ( ( ~bLassoKludge ) & ( length( vecAActiveSet ) > 1 ) )
        if ( vecAActiveSet(end) > vecAActiveSet(end-1) )
            warning( 'Sanity failed: AActiveSet increased' );
            disp( 'Pausing...' ), pause;
        end
    end
    
    wActiveSet = AActiveSet * GInvActiveSet * OneVecActiveSet; % |A|-vec
    uActiveSet = (sRestoreOptions.X * SActiveSet) * wActiveSet; % N-vec
    normUActiveSet = norm( uActiveSet, 2 );
    disp( sprintf( 'SANITY: |u|=%g', normUActiveSet ) );
    % SANITY: check that |uActiveSet| = 1
    if abs( normUActiveSet - 1 ) > 1e-6
        error( 'Sanity failed: |u| != 1' );
    end
    
    % This should be equal AActiveSet * 1
    tmp = full( ( SActiveSet' * sRestoreOptions.X' ) * uActiveSet );
    %disp( 'SANITY: X_A^T * u_A:' ), disp( tmp(:)' );
    % SANITY: check that each element of tmp equals AActiveSet
    for n = 1:nActiveSetSize
        errMargin = abs( tmp(n) - AActiveSet );
        if errMargin > abs(1e-6*AActiveSet)
            error( sprintf( 'Sanity failed: (X_A^T * u_A)(%g) != AActiveSet [%g]', n, errMargin ) );
        end
    end
    
    a = sRestoreOptions.X' * uActiveSet; % M-vec
    %disp( 'a:' ), disp( a );
    
    % Construct dHat
    dHat = zeros( M, 1 ); % M-vec
    for j = 1:nActiveSetSize
        jj = activeSet( j );
        dHat( jj ) = s( jj ) * wActiveSet( j );
    end
    
    % Step 4a: Compute gammaHat and jHat
    
    jHatColumn = -1; % DEBUG variable
    
    if ( i < M )
        
        % We can only do this if i < NN. Here, compute
        % gammaHat = min^+_{j \in A^c} { (CHat - chat_j)/(A_A - a_j), (CHat
        % + chat_j)/(A_A + a_j) }
        
        activeSetC = setdiff( 1:M, activeSet );
        % DEBUG
        %disp( 'activeSetC:' ), disp( activeSetC(:)' );
        clear gammaMat;
        gammaMat = zeros( length( activeSetC ), 2 );
        
        % Compute the 1st column of gammaMat
        numer = BigCHat - cHat( activeSetC );
        denom = AActiveSet - a( activeSetC );
        tmp = numer(:) ./ denom(:);
        % SANITY
        if length( tmp(:) ) ~= length( activeSetC ), error( 'Sanity on length of tmp failed' ), end;
        gammaMat( :, 1 ) = tmp(:);
        
        % Compute the 2nd column of gammaMat
        numer = BigCHat + cHat( activeSetC );
        denom = AActiveSet + a( activeSetC );
        tmp = numer(:) ./ denom(:);
        % SANITY
        if length( tmp(:) ) ~= length( activeSetC ), error( 'Sanity on length of tmp failed' ), end;
        gammaMat( :, 2 ) = tmp(:);

        % Take the min over each column
        gammaMatPosInd = find( ( gammaMat(:,1) > eps ) & isfinite( gammaMat(:,1) ) );
        if ~isempty( gammaMatPosInd )
            gammaMatCol1Min = min( gammaMat( gammaMatPosInd, 1 ) );
            gammaMatCol1MinInd = find( gammaMat(:,1) == gammaMatCol1Min );
            if length( gammaMatCol1MinInd ) > 1
                error( 'cannot handle length( gammaMatCol1MinInd ) > 1' );
            end            
            % SANITY
            %disp( 'gammaMatCol1Min:' ), disp( gammaMatCol1Min );
            %disp( 'gammaMatCol1MinInd:' ), disp( gammaMatCol1MinInd );
            %disp( 'jHat of min:' ), disp( activeSetC( gammaMatCol1MinInd ) );
            %disp( 'row of gammaMat:' ), disp( gammaMat( gammaMatCol1MinInd, : ) );
        else
            gammaMatCol1Min = inf;
        end
        
        gammaMatPosInd = find( ( gammaMat(:,2) > eps ) & isfinite( gammaMat(:,2) ) );
        if ~isempty( gammaMatPosInd )
            gammaMatCol2Min = min( gammaMat( gammaMatPosInd, 2 ) );
            gammaMatCol2MinInd = find( gammaMat(:,2) == gammaMatCol2Min );
            if length( gammaMatCol2MinInd ) > 1
                error( 'cannot handle length( gammaMatCol2MinInd ) > 1' );
            end
            % SANITY
            %disp( 'gammaMatCol2Min:' ), disp( gammaMatCol2Min );
            %disp( 'gammaMatCol2MinInd:' ), disp( gammaMatCol2MinInd );
            %disp( 'jHat of min:' ), disp( activeSetC( gammaMatCol2MinInd ) );
            %disp( 'row of gammaMat:' ), disp( gammaMat( gammaMatCol2MinInd, : ) );
        else
            gammaMatCol2Min = inf;
        end

        % And then combine to find the min over the entire matrix
        [gmm, gmmInd] = min( [gammaMatCol1Min, gammaMatCol2Min] );
        if isfinite( gmm )
            gammaHat = gmm;
            if ( gmmInd == 1 )
                % DEBUG
                %disp( 'Getting jHat from 1st column of gammaMat' );
                jHatColumn = 1;
                jHat = activeSetC( gammaMatCol1MinInd );
            else
                % DEBUG
                %disp( 'Getting jHat from 2nd column of gammaMat' );
                jHatColumn = 2;
                jHat = activeSetC( gammaMatCol2MinInd );
            end
        else
            jHatColumn = -1; % Invalid value
            error( 'could not find gammaHat and jHat' );
        end

    else
        
        % Do this if i == M. This is the last step
        gammaHat = BigCHat/AActiveSet;
        jHat = -1;
        
    end
    
    % Step 4b: compute gammaTilde
 
    warning off; % XXX: We are going to get some divide-by-zero errors here!
    gammaTildeVec = -betaHatPrevActiveSet ./ dHat; % This is a M-vector
    warning on;

    % Find the min gamma s.t. it is positive 
    gammaTildePosInd = find( ( gammaTildeVec > eps ) & isfinite( gammaTildeVec ) );
    gtm = min( gammaTildeVec( gammaTildePosInd ) );
    
    switch length( gtm )
        case 0
            % No positive elements in gammaTilde
            gammaTilde = inf;
            jTilde = -1;
        case 1
            % This is the "normal" case
            gtmInd = find( gammaTildeVec == gtm );
            if length( gtmInd ) > 1
                error( 'cannot handle length( gtmInd ) > 1' );
            elseif length( gtmInd ) == 1
                gammaTilde = gtm;
                jTilde = gtmInd;
                % Sanity: check that jTilde is in the active set
                if ~ismember( jTilde, activeSet )
                    error( 'jTilde not in the active set' );
                end
            else
                error( 'Cannot find minimum in gammaTildeVec?' );
            end
        otherwise
            error( 'Multiple minima in gammaTildeVec?' );
    end

    % Step 4c: get gammaConst
    
    bLassoKludge = 0; % Re-initialize
    
    if ( gammaTilde < gammaHat )
        bLassoKludge = 1;
        gammaConst = gammaTilde;
        disp( '*** LARS-LASSO KLUDGE ***' );
    else
        gammaConst = gammaHat;
    end
    
    disp( [ sprintf( 'gammaHat=%g, jHat=%g/', gammaHat, jHat ), ...
        sprintf( 'gammaTilde=%g, jTilde=%g\n', gammaTilde, jTilde ), ...
        sprintf( 'gammaConst=%g', gammaConst ) ] );    
    
    % SANITY
    nextBigCHat = BigCHat - gammaConst * AActiveSet;
    disp( sprintf( 'nextBigCHat = %g', nextBigCHat ) );
    
    % Check that the new correlation equals nextBigCHat
    if ~bLassoKludge
        cjHat = cHat( jHat ) - gammaConst * a( jHat );
        disp( sprintf( 'cjHat = %g', cjHat ) );
        
        % Find entry in gammaMat
        jHatGammaMatIndex = find( activeSetC == jHat );
        if length( jHatGammaMatIndex ) ~= 1, error( 'failed to find jHat in activeSetC' ), end;
        %disp( 'jHatGammaMatIndex:' ), disp( jHatGammaMatIndex );
        %disp( 'jHat entry in gammaMat:' ), disp( gammaMat( jHatGammaMatIndex, : ) );
        
        % Check if |cjHat| equals nextBigCHat. It should!
        if abs( abs(cjHat) - nextBigCHat ) > eps                                    
            % Re-compute gamma
            switch jHatColumn
                case 1
                    gamma2 = ( BigCHat - cHat( jHat ) ) / ( AActiveSet - a( jHat ) );
                case 2
                    gamma2 = ( BigCHat + cHat( jHat ) ) / ( AActiveSet + a( jHat ) );
                otherwise
                    error( 'trying to re-compute gammaHat with invalid jHatColumn' );
            end
            disp( sprintf( 'Re-computed gammaHat = %g', gamma2 ) );
            error( 'cjHat not equal to nextBigCHat??' );
        end
    end
    
    
    % Step 5: Form the next estimator of mu and beta
    muHatNextActiveSet = muHatPrevActiveSet + gammaConst * uActiveSet;    
    betaHatNextActiveSet = betaHatPrevActiveSet + gammaConst * dHat;
   
   	% Update metrics
    errorNorm1Hist = [ errorNorm1Hist, norm( betaHatNextActiveSet(:) - betaTrue(:), 1 ) ];
    errorNorm2Hist = [ errorNorm2Hist, norm( betaHatNextActiveSet(:) - betaTrue(:), 2 ) ];
	betaHatDiff = gammaConst * dHat(:);
    % The l_infinity norm of the difference
    thetaMaxDiffHist = [thetaMaxDiffHist; max( abs( betaHatDiff ) )];
    % The l_1 norm of the difference
    thetaSumDiffHist = [thetaSumDiffHist; sum( abs( betaHatDiff ) )];

    bigCHatHist = [ bigCHatHist; BigCHat ];
    
    % MASSIVE DEBUG. XXX: make sure the dimensions of y and betaTrue are
    % accurate.
    
    if ( ( sRestoreOptions.N_DEBUG > 1 ) & ...
         ( mod( i, sRestoreOptions.reconSanityInterval ) == 1 ) & ...
         ( i > 1 ) )
    
        lambda = 2* BigCHat * sRestoreOptions.XColumnNorm( activeSet(1) );
        
        disp( sprintf( 'BigCHat = %g, lambda=%g', BigCHat, lambda ) );

        % Compute using Daubechies' iterations. Here, we MUST have the size of the i/p and
		% o/p images be the same, i.e., M=N.
        
        [sDIter, sDIterOptions] = emgauss_restore( 'forward_op', psf, ...
            'noisy_output', reshape( y, opImgDim ), ...
            'noise_sigma', noiseSigma, ...
            'theta_true', reshape( betaTrue, ipImgDim ), ...
            'm_selector', 'daub2', ...
            'num_em_iters', 2e4, ...
            'num_max_grad_search_failures', -1, ...
            'theta0_selector', 'Hty2', ...
            'theta0', [], ...
            'a_val', -1, ...
            'w_val', -1, ...
            't_val', -1, ...
            'sparse_ind', 'w', ...          
            'stopping_cond', 'fixed', ...
            'stopping_cond_tol', noiseSigma, ...
            'stopping_cond_win', 3e3, ...
            'stopping_cond_last_fac', 2, ...
            'thetastar', 1, ...
            'soft_thres_t', -1, ...
            'hparam_rest_interval', 1e2, ...
            'lambda', lambda ...
            );
        
        larsSol = betaHatNextActiveSet./sRestoreOptions.XColumnNorm;
        daubSol = sDIter.last;
        figure( 2 ), imagesc( larsSol, 32, 32 ), colorbar;
        title( 'LARS-LASSO' );        
        figure( 3 ), imagesc( daubSol ), colorbar;
        title( 'Daubechies' );
        figure( 4 ), plot( larsSol(:), 'bx' ), hold on, plot( daubSol(:), 'ro' );
        title( 'LARS-LASSO vs. Daubechies' );
        disp( 'Pausing...' ), pause;
        
%         vecDIter = reshape( sDIter.last, 1024, 1 );
%         vecMismatch = ( sign( reshape( betaHatNextActiveSet, 1024, 1 ) ) ~= sign( vecDIter ) );
%         if sum( vecMismatch ) ~= 0
%             indMismatch = find( vecMismatch ~= 0 );
%             disp( 'Mismatch in LARS-LASSO vs. Daubechies' );
%             disp( 'indMismatch:' ), disp( indMismatch );
%             disp( 'Mismatched values/LARS-LASSO:' ); 
%             disp( betaHatNextActiveSet( indMismatch ) );
%             disp( 'Mismatched values/Daubechies:' );
%             disp( vecDIter( indMismatch ) );
%         end
        
    end
    
    vecL1( i ) = sum( abs( betaHatNextActiveSet(:) ) ); % l_1(beta)
    vecL0( i ) = sum( abs( betaHatNextActiveSet(:) ) > 1e3*eps ); % l_0(beta)
    
    % Compute the cost function. For L1-SURE, this is painless.
    predictionErr = y - muHatNextActiveSet;
    rss = sum( predictionErr(:).^2 );
    vecCost( i ) = N*noiseSigma^2 + rss + 2*noiseSigma^2*vecL0(i);
    
    % Check to see if we have a better solution
    
    if ( ( betaHatBestInd < 1 ) | ...
         ( (betaHatBestInd >= 1) & (vecCost(betaHatBestInd) > vecCost(i)) ) )
        betaHatBestInd = i;
        betaHatBest = betaHatNextActiveSet;
    end
       
    % Step 6: Update state variables
    muHatPrevActiveSet = muHatNextActiveSet;
    betaHatPrevActiveSet = betaHatNextActiveSet;
    
    if sRestoreOptions.N_DEBUG, disp( 'End of iter, press key' ), pause, end;
    
end

% Re-scale last beta and best beta
betaHat = betaHatPrevActiveSet ./ sRestoreOptions.XColumnNorm;
betaHatBest = betaHatBest ./ sRestoreOptions.XColumnNorm;
% muCheck should equal muHatPrevActiveSet
muCheck = sRestoreOptions.X * betaHatPrevActiveSet;

lambda = 2*sRestoreOptions.XColumnNorm(1)*bigCHatHist( betaHatBestInd );

switch sRestoreOptions.n_strat
    case 2
        sParam2 = struct( 'X', sRestoreOptions.X, ...
            'XColumnNorm', sRestoreOptions.XColumnNorm, ...
            'nPhi2', sRestoreOptions.n_phi2, ...
            'numEmIters', 2e4, ...
            'bigCHatHist', bigCHatHist, ...
            'psf', psf, ...
            'y', y, ...
            'noiseSigma', noiseSigma, ...
            'betaTrue', betaTrue, ...
            'phi0', [lambda/2, lambda/2], ...
            'phi0Cost', vecCost( betaHatBestInd ), ...
            'phi0Beta', betaHatBest );
        sStratOut = hs_strat2( 'sparam', sParam2 );
        sOut.sStratOut = sStratOut;

    case 3
        sParam3 = struct( 'X', sRestoreOptions.X, ...
            'XColumnNorm', sRestoreOptions.XColumnNorm, ...
            'nPhi1', sRestoreOptions.n_phi1, ...
            'numEmIters', 2e4, ...
            'szPhi1Vec', 'max_chat', ...
            'bigCHatHist', bigCHatHist, ...
            'psf', psf, ...
            'y', y, ...
            'noiseSigma', noiseSigma, ...
            'betaTrue', betaTrue, ...
            'phi0', [lambda/2, lambda/2], ...
            'phi0LarsIndex1', betaHatBestInd, ...
            'phi0Cost', vecCost( betaHatBestInd ), ...
            'phi0Beta', betaHatBest );
        sStratOut = hs_strat3( 'sparam', sParam3 );
        sOut.sStratOut = sStratOut;

    case 4
        sParam4 = struct( 'X', sRestoreOptions.X, ...
            'XColumnNorm', sRestoreOptions.XColumnNorm, ...
            'nPhi1', sRestoreOptions.n_phi1, ...
            'nPhi2', sRestoreOptions.n_phi2, ...
            'numEmIters', 2e4, ...
            'szPhi1Vec', 'max_chat', ...
            'bigCHatHist', bigCHatHist, ...
            'psf', psf, ...
            'y', y, ...
            'noiseSigma', noiseSigma, ...
            'betaTrue', betaTrue, ...
            'phi0', [lambda/2, lambda/2], ...
            'phi0LarsIndex1', betaHatBestInd, ...
            'phi0Cost', vecCost( betaHatBestInd ), ...
            'phi0Beta', betaHatBest );
        sStratOut = hs_strat4( 'sparam', sParam4 );
        sOut.sStratOut = sStratOut;
        
    case 32
        sParam3 = struct( 'X', sRestoreOptions.X, ...
            'XColumnNorm', sRestoreOptions.XColumnNorm, ...
            'nPhi1', sRestoreOptions.n_phi1, ...
            'numEmIters', 2e4, ...
            'szPhi1Vec', 'max_chat', ...
            'bigCHatHist', bigCHatHist, ...
            'psf', psf, ...
            'y', y, ...
            'noiseSigma', noiseSigma, ...
            'betaTrue', betaTrue, ...
            'phi0', [lambda/2, lambda/2], ...
            'phi0LarsIndex1', betaHatBestInd, ...
            'phi0Cost', vecCost( betaHatBestInd ), ...
            'phi0Beta', betaHatBest );
        sStratOutStep1 = hs_strat3( 'sparam', sParam3 );
        sOut.sStratOut1 = sStratOutStep1;

        sParam2 = struct( 'X', sRestoreOptions.X, ...
            'XColumnNorm', sRestoreOptions.XColumnNorm, ...
            'nPhi2', sRestoreOptions.n_phi2, ...
            'numEmIters', 2e4, ...
            'bigCHatHist', bigCHatHist, ...
            'psf', psf, ...
            'y', y, ...
            'noiseSigma', noiseSigma, ...
            'betaTrue', betaTrue, ...
            'phi0', sStratOutStep1.phi, ...
            'phi0Cost', sStratOutStep1.cost( sStratOutStep1.betaHatBestInd ), ...
            'phi0Beta', sStratOutStep1.betaHatBest );
        sStratOut = hs_strat2( 'sparam', sParam2 ); % Last optimization should be called sStratOut
        sOut.sStratOut2 = sStratOut;

    case 323
        sParam3 = struct( 'X', sRestoreOptions.X, ...
            'XColumnNorm', sRestoreOptions.XColumnNorm, ...
            'nPhi1', sRestoreOptions.n_phi1, ...
            'numEmIters', 2e4, ...
            'szPhi1Vec', 'max_chat', ...
            'bigCHatHist', bigCHatHist, ...
            'psf', psf, ...
            'y', y, ...
            'noiseSigma', noiseSigma, ...
            'betaTrue', betaTrue, ...
            'phi0', [lambda/2, lambda/2], ...
            'phi0LarsIndex1', betaHatBestInd, ...
            'phi0Cost', vecCost( betaHatBestInd ), ...
            'phi0Beta', betaHatBest );
        sStratOutStep1 = hs_strat3( 'sparam', sParam3 );
        sOut.sStratOut1 = sStratOutStep1;

        sParam2 = struct( 'X', sRestoreOptions.X, ...
            'XColumnNorm', sRestoreOptions.XColumnNorm, ...
            'nPhi2', sRestoreOptions.n_phi2, ...
            'numEmIters', 2e4, ...
            'bigCHatHist', bigCHatHist, ...
            'psf', psf, ...
            'y', y, ...
            'noiseSigma', noiseSigma, ...
            'betaTrue', betaTrue, ...
            'phi0', sStratOutStep1.phi, ...
            'phi0Cost', sStratOutStep1.cost( sStratOutStep1.betaHatBestInd ), ...
            'phi0Beta', sStratOutStep1.betaHatBest );
        sStratOutStep2 = hs_strat2( 'sparam', sParam2 ); 
        sOut.sStratOut2 = sStratOutStep2;
        
        sParam3.phi0 = sStratOutStep2.phi;
        sParam3.phi0Cost = sStratOutStep2.cost( sStratOutStep2.betaHatBestInd );
        sParam3.phi0Beta = sStratOutStep2.betaHatBest;
                 
        [minVal, minValInd] = min( abs( bigCHatHist - sStratOutStep2.phi(1)/sRestoreOptions.XColumnNorm(1) ) );
        if length( minValInd ) > 1
            error( 'More than 1 value appears to match phi(1)' );
        end
        sParam3.phi0LarsIndex1 = minValInd;
        if abs( sRestoreOptions.XColumnNorm(1)*bigCHatHist( sParam3.phi0LarsIndex1 ) - sParam3.phi0(1) ) > 1e3*eps
            error( 'could not find phi0LarsIndex1' );
        end
        
        sStratOut = hs_strat3( 'sparam', sParam3 ); % Last optimization should be called sStratOut
        sOut.sStratOut3 = sStratOut;

        % Last optimization should be called sStratOut
    otherwise
        error( 'invalid n_strat' );
end

tCpuTimeStop = cputime;

% Defines the structure of sOut

sOutElems = { 'phi', 'sStratOut.phi'; ...
    'bestBeta', 'sStratOut.betaHatBest'; ...
    'bestBetaInd', 'sStratOut.betaHatBestInd'; ...
    'cost', 'sStratOut.cost'; ...
    'bestBeta1', 'betaHatBest'; ...
    'lastInd1', 'betaHatBestInd'; ...
    'cost1', 'vecCost'; ...       
    'lambda', 'lambda'; ...
    'bigCHat', 'bigCHatHist'; ...
    'l1', 'vecL1'; ...
    'l0', 'vecL0'; ...
    'cputime', '[tCpuTimeStart, tCpuTimeStop]'; ...
    'thetaMaxDiff', 'thetaMaxDiffHist'; ...
    'thetaSumDiff', 'thetaSumDiffHist'; ...
    'errorNorm1', 'errorNorm1Hist'; ...
    'errorNorm2', 'errorNorm2Hist'; ...
    };

for elemInd = 1:size( sOutElems, 1 )
    szCmd = sprintf( 'sOut.%s = %s;', ...
        char( sOutElems(elemInd,1) ), ...       
        char( sOutElems(elemInd,2) ) );
    try
        eval( szCmd );
    catch
        warning( sprintf( 'Could not execute: %s', szCmd ) );
    end
end

