
% Function that normalizes the X matrix for LARS-LASSO implementation so
% that each column of X is of unit norm.

function [X, XColumnNorm] = lars_normx( varargin )

% Defaults
bConstructX = 0;
ZERO_THRES = 1e3*eps;

% Process input arguments

lenVarargin = length( varargin );
numArgsIn = floor( lenVarargin/2 );

if 2*numArgsIn ~= lenVarargin, error( 'even ip args expected' ), end

for argInInd = 1:numArgsIn
    
    szKey = lower( varargin{2*argInInd-1} );    
    switch szKey
        case 'forward_op', hForwardOp = varargin{2*argInInd}; bConstructX = 1;
        case 'x', X = varargin{2*argInInd};           
        otherwise
            error( sprintf( 'unrecognized option %s', szKey ) );
    end
end

if ~bConstructX

    % No need to construct X, just normalize columns to have unit norm.
    % NOTE that we expect X to be a sparse matrix
        
    disp( 'Not constructing X, simply re-normalizing columns' );
    numRows = size( X, 1 );
    numCols = size( X, 2 );
    columnSSV = zeros( numCols, 1 );
    
    [ nzI, nzJ, nzS ] = find( X );    
    nnz = length( nzS );
    
    for i = 1:nnz
        j = nzJ( i );
        columnSSV( j ) = columnSSV( j ) + nzS( i )^2;
    end
    XColumnNorm = sqrt( columnSSV );

    % Check to see if X has constant column norm
    bConstCNorm = 1;
    c = XColumnNorm( 1 );
    for i = 2:numCols
        if ( XColumnNorm( i ) - c ) > 1e2*eps
            disp( 'X does not have constant column norm' );            
            bConstCNorm = 0;
            break;
        end
    end
    
    if bConstCNorm
        disp( 'X has constant column norm' );
    else
        figure( 1 ), clf, plot( XColumnNorm, 'bx' );
        disp( 'Pausing...' ), pause;
    end
    
    % Normalize columns of X to unit norm
    for i = 1:nnz
        j = nzJ( i );
        nzS( i ) = nzS( i ) / XColumnNorm( j );
    end
    X = sparse( nzI, nzJ, nzS, numRows, numCols );
        
else
    
    % Need to construct X from zero-padded psf. ASSUME that the matrix
    % represents a convolution kernel. That is, it can be diagonalized
    % using DFTs. 
    
    % We expect the psf to be a square zero-padded matrix.
    HM = size( hForwardOp, 1 );
    HN = size( hForwardOp, 2 );
    if ( HM ~= HN ), error( 'h is not a square matrix' ), end;    
    XSize = HM*HN; % X will be of size XSize x XSize
    N = HM;

    % Variables to keep track of non-zero elements
    nzNum = 0;
    nzI = [];
    nzJ = [];
    nzS = [];
    XColumnNorm = zeros( XSize, 1 );

    % Index pointing to the last element of the previous column
    nLastColumnInd = 0;
        
    for j = 1:XSize

        % Iterate over the j-th column to find any non-zero entries

        % Keep track of sum of squared values of each column
        columnSSV = 0;

        for i = 1:XSize
            hInd1 = 1 + mod( mod(i,N) - mod(j,N), N );
            hInd2 = 1 + mod( floor((i-1)/N) - floor((j-1)/N), N );
            hVal = hForwardOp( hInd1, hInd2 );
            if ( abs( hVal ) > ZERO_THRES )
                % DEBUG
                %disp( sprintf( 'i=%g, j=%g  --> hInd1=%g, hInd2=%g', i, j, hInd1, hInd2 ) );
                nzNum = nzNum + 1;
                nzI = [nzI, i];
                nzJ = [nzJ, j];
                nzS = [nzS, hVal];
                columnSSV = columnSSV + nzS(end)^2;
            end
        end

        % Normalize the columns so that each is of unit norm
        columnL2Norm = sqrt( columnSSV );        
        if ( abs( columnL2Norm ) < ZERO_THRES )
            error( 'encountered zero column in X!' );
        end
        
        startInd = nLastColumnInd + 1;
        nzS( startInd:end ) = nzS( startInd:end ) / columnL2Norm;
        sRestoreOptions.XColumnNorm( j ) = columnL2Norm;        
        nLastColumnInd = length( nzS ); % Update nLastColumnInd
        
    end       

    disp( sprintf( 'nzNum=%g', nzNum ) );

	% Allocate space for sparse system matrix
	sRestoreOptions.X = sparse( nzI, nzJ, nzS, XSize, XSize );
	
end