%
% LARS-LASSO-SURE. Adapted from ll_small. Usage is
% sOut = ll_big( varargin )
% 

% The problem we consider is
% y = X * beta + n
% y ~ N-dim vector, beta ~ M-dim vector, X ~ N x M

% Spin-off from ll_small.m The strategy we have for getting this to work
% with large X is to make the minimal amount of changes necessary. No point
% in getting bitten by the evil coding bug. To regression test, apply
% ll_big to problems that ll_small can solve, e.g., a 2d psf.

% 20-jan-2007. Argh, we have to recover the the crash of the Maxtor drive.
% Fortunately, I think that minimal changes were made to ll_big relative to
% ll_small.

function sOut = ll_big( varargin )

% Constants

ll_defs; % Load from file

noiseSigma = -1;

% Default restore options. Put all the tolerances in sRestoreOptions.
sRestoreOptions = struct( 'X', [], ...
    'XColumnNorm', [], ...
    'reconSanityInterval', 10, ...
    'debugWord', 0, ...
    'nConvMethod', 1, ...
    'colNormPercentTol', 1e-6, 'betaTol', 1e-9, 'CHatTol', 1e-6, ...
    'gammaTol', 1e-9, 'daubPercentTol', 1e-4, 'l0Tol', 1e4*eps, ...
    'UActiveSetTol', 1e-6, 'AActiveSetPercentTol', 1e-6, 'corrPercentTol', 1e-6, ...
    'NUM_DAUB_ITERS', 2e4, ...
    'PAUSE_NUM_SECS', 5, ...
    'B_USE_PSF2MAT_XA', 0 );

% Defines the structure of sRestore
sRestoreElems = { ...
    'lastIter',     'betaHat'; ...
    'last',         'betaHatBest'; ...
    'lastInd1',     'betaHatBestInd'; ...
    'muHat',        'muHatPrevActiveSet'; ...
    'bigCHat',      'bigCHatHist'; ...
    'activeSet',    'activeSet'; ...
    'vecL1',        'vecL1'; ...
    'vecL0',        'vecL0'; ...
    'vecSoftThres', 'vecSoftThres'; ...
    'vecCost',      'vecCost'; ...
    'cputime',      '[tCpuTimeStart, tCpuTimeStop]'; ...
    'elapsedtime',  '[tElapsedTimeStart, tElapsedTimeStop]'; ...
    'thetaMaxDiff', 'thetaMaxDiffHist'; ...
    'thetaSumDiff', 'thetaSumDiffHist'; ...
    'errorNorm1',   'errorNorm1Hist'; ...
    'errorNorm2',   'errorNorm2Hist'; ...
    };


% Process input arguments

lenVarargin = length( varargin );
numArgsIn = floor( lenVarargin/2 );

if 2*numArgsIn ~= lenVarargin, error( 'even ip args expected' ), end

for argInInd = 1:numArgsIn
    
    switch varargin{2*argInInd-1}      
        case 'noisy_output', y = varargin{2*argInInd};      
        case 'beta_true', betaTrue = varargin{2*argInInd}; 
        case 'num_iters', sRestoreOptions.NUM_ITERS = varargin{2*argInInd};    
        case 'n_debug', sRestoreOptions.N_DEBUG = varargin{2*argInInd};
        case 'psf', psf = varargin{2*argInInd};
        case 'noise_sigma', noiseSigma = varargin{2*argInInd};
		case 'ip_img_dim', ipImgDim = varargin{2*argInInd}; 
		case 'op_img_dim', opImgDim = varargin{2*argInInd};
        case 'recon_sanity_interval', sRestoreOptions.reconSanityInterval = varargin{2*argInInd};
        case 'debug_word', sRestoreOptions.debugWord = varargin{2*argInInd};
        case 'b_use_psf2mat_xa', sRestoreOptions.B_USE_PSF2MAT_XA = varargin{2*argInInd};
            
        otherwise
            error( sprintf( 'unrecognized option %s', szKey ) );
    end
end

% Check validity of arguments
%%%

N = length( y );
M = length( betaTrue );

if ( sRestoreOptions.NUM_ITERS > M ) | ( sRestoreOptions.NUM_ITERS < 1 )
    error( 'invalid NUM_ITERS' );
elseif ( sRestoreOptions.NUM_ITERS == M )
    disp( 'Running LARS to LS solution' );
else
    disp( 'Running LARS, but stopping short of LS solution' );
end

if ( noiseSigma <= 0 ), error( 'noiseSigma must be >0' ), end;

if ( prod( opImgDim ) ~= N ) | ( prod( ipImgDim ) ~= M )
	error( 'op or ip image dimensions are inconsistent' );
end

disp( sprintf( '#elems in op: %g, ip: %g', N, M ) )

if ( M ~= N ), error( 'must have M=N' ), end;

% Metrics for convergence of iterations
%%%

thetaMaxDiffHist = [];  % l_inf norm of difference of 2 consecutive estimates
thetaSumDiffHist = [];  % l_1 norm of difference ...

errorNorm1Hist = [];    % l_1 norm of difference
errorNorm2Hist = [];    % l_2 norm ...

bigCHatHist = [];

activeSet = [];
vecAActiveSet = [];
XActiveSet = [];

muHatPrevActiveSet = zeros( N, 1 );
muHatNextActiveSet = zeros( size( muHatPrevActiveSet ) );

betaHatPrevActiveSet = zeros( M, 1 );
betaHatNextActiveSet = zeros( size( betaHatPrevActiveSet ) );
betaHatBest = zeros( size( betaHatPrevActiveSet ) );
betaHatBestInd = -1;

gammaMat = zeros( M, 2 );

% Keep track of various statistics during the iterations.
%%%

% SURE cost criterion
vecCost = zeros( sRestoreOptions.NUM_ITERS, 1 ); 
% l_0 of estimate
vecL0 = zeros( size( vecCost ) ); 
% soft threshold
vecSoftThres = zeros( size( vecCost ) );

bLassoKludge = 0; % Flag whenever we have to do a LARS-LASSO kludge
jHat = -1;
jKludge = -1;
nextBigCHat = -1;

% As we do not have access to the matrix X, we shall compute X*u and X'*u
% using the Gblur object. But since X is assumed to have unit l_2 column
% norm, we have to figure out what the column norm of the specified psf is.
% Let W be the original matrix specified by the psf. ASSUMING that W has
% constant column norm, then 
%    X = ( w_1/||w_1||, w_2/||w_2||, ... )
%      = ( w_1/w, w_2/w, ... )
%      = (1/w) W
% where w := ||w_1|| = ||w_2|| = ...

% Compute the l_2 norm of several random columns and check that they are
% all equal

nRandColumns = min( [ floor( 0.01 * M ), 20 ] );
disp( sprintf( 'Computing the l_2 norm of %g columns of X', nRandColumns ) )
colNorm = psf2mat_cnorm( psf, 1 + floor( rand( 1, nRandColumns ) * M ) );

colNormDelta = sum( abs( colNorm - colNorm(1) ) );
disp( sprintf( 'colNormDelta: %g', colNormDelta ) )

if ( colNormDelta > nRandColumns * sRestoreOptions.colNormPercentTol * colNorm(1) )
    warning( 'X matrix does not have equal l_2 column norms' )
    disp( 'colNorm:' ), disp( colNorm )
    error( 'FATAL ERROR' )
end

% Also note that below, where we used to treat XColumnNorm as a VECTOR in
% ll_small, now we treat it as a CONSTANT. In ll_small, however, the
% vector should have had all the same entries.
sRestoreOptions.XColumnNorm = colNorm( 1 );
disp( sprintf( 'Column norm is %g', sRestoreOptions.XColumnNorm ) )

% Initialize Gblur object. Copied from test_xa.m.

if sRestoreOptions.nConvMethod == 1
	GX = Gblur( ones( opImgDim ), ...
        'psf', psf / sRestoreOptions.XColumnNorm, ...
        'type', 'conv,fft', ...
        'nDim', length( ipImgDim ) );
elseif sRestoreOptions.nConvMethod == 2
	GX = Gblur( ones( opImgDim ), ...
        'psf', psf / sRestoreOptions.XColumnNorm, ...
        'type', 'conv,same', ...
        'nDim', length( ipImgDim ) );
else
    error( 'illegal nConvMethod' );
end

% EXTRA SANITY: Check that GX represents a matrix with unit l_2 column norm
betaTest = zeros( M, 1 );
betaTest( 1 + floor( rand * M ) ) = 1;
muTest = GX * reshape( betaTest, ipImgDim );
muTest = muTest(:);
disp( sprintf( '|muTest|_2 = %g', norm( muTest, 2 ) ) )

% Time the iterations
tCpuTimeStart = cputime;
tElapsedTimeStart = clock;

for i = 1:sRestoreOptions.NUM_ITERS
    
    if bitand( sRestoreOptions.debugWord, LL_DEBUG_LOOP1 )
        disp( sprintf( '\nITERATION %g\n', i ) )
    end
    
    % Step 1: Compute the correlations
    disp( 'STEP 1: Computing correlations' );   
    
    % NB. 
    %    (i) every time we compute X*u, we must reshape u to have
    %        dimensions ipImgDim. 
    %    (ii) every time we compute X'*v, we must reshape v to have
    %         dimensions opImgDim
    
    cHat = GX' * reshape( full( y - muHatPrevActiveSet ), opImgDim );
    cHat = cHat(:); % Convert back to column vector
    
    s = sign( cHat );
    
    betaNzInd = find( abs( betaHatPrevActiveSet ) > sRestoreOptions.betaTol ); % Find non-zero indices
    vecMismatch = ( s(betaNzInd) ~= sign( betaHatPrevActiveSet( betaNzInd ) ) );
    if bitand( sRestoreOptions.debugWord, LL_DEBUG_LOOP2 )
        disp( 'betaNzInd:' ), disp( betaNzInd(:)' );
        disp( 'c-values:' ), tmp = cHat( betaNzInd ); disp( tmp(:)' );
%         disp( 'beta-values:' ), tmp = betaHatPrevActiveSet(betaNzInd); disp( tmp(:)' );
    end
    
    % SANITY: check that sign(beta_j) = sign(c_j) = s_j, for j in the
    % active set
    if sum( vecMismatch ) ~= 0
        figure( 2 ), clf, plot( s(betaNzInd), 'kx' ), hold on;
        plot( sign(betaHatPrevActiveSet(betaNzInd)), 'ro' );
        legend( 's', 'sign(beta)', 0 );
        error( 'LASSO solution sign contraint violated' );
    end
                   
    % Step 2: Get the active set. 
    
    disp( 'STEP 2: Adjusting active set' );
    
    % This is another part where ll_big is different from ll_small. We no
    % longer have access to X, so must compute XActiveSet through another
    % means. In this case, psf2mat_xa.

    if ( i == 1 )
        
        % In the first iteration, use max to locate the active set
        [BigCHat, activeSet] = max( abs( cHat ) );
        
        if sRestoreOptions.B_USE_PSF2MAT_XA
            XActiveSet1 = psf2mat_xa( psf/sRestoreOptions.XColumnNorm, activeSet, s( activeSet ) );
            XActiveSet = XActiveSet1(:);
        else
            tmp = zeros( M, 1 );
            tmp( activeSet ) = 1;
            XActiveSet2 = GX * reshape( tmp, ipImgDim );
            XActiveSet = s( activeSet ) * XActiveSet2(:);
        end

        % DEBUG
        %%%
        
%         if ( norm( XActiveSet1(:) - XActiveSet2(:), 2 ) > 1e3*eps )
%             figure( 2 ), clf
%             plot( full( XActiveSet1 ), 'bs' ), hold on
%             plot( XActiveSet2, 'ro' )
%             legend( 'XActiveSet1', 'XActiveSet2', 0 )
%             error( 'XActiveSet differs' )
%         end

    else
        
        % In subsequent iterations, do not rely on the active set being, by
        % definition, the values for which |cHat| equals the max. It
        % turns out that there are some numerical differences in the cHats,
        % so that now all of the "maximum" |cHat|s are the same.

        [BigCHat, activeSetTheory] = max( abs( cHat ) );

        if ~bLassoKludge
            
            % SANITY: see if jHat is already in the active set. It
            % shouldn't be.
            if ~isempty( find( activeSet == jHat ) )
                error( 'jHat is already in activeSet' )
            end
            
            % Add to the active set
            activeSet = [activeSet, jHat];
            
            % Get new column for XActiveSet
            if sRestoreOptions.B_USE_PSF2MAT_XA
                newCol1 = psf2mat_xa( psf/sRestoreOptions.XColumnNorm, jHat, s( jHat ) );
                newCol = newCol1(:);
            else
                tmp = zeros( M, 1 );
                tmp( jHat ) = 1;
                newCol2 = GX * reshape( tmp, ipImgDim );
                newCol = s( jHat ) * newCol2(:);
            end
            
            % Add to XActiveSet
            XActiveSet = [ XActiveSet, newCol ];

            % DEBUG
            %%%
                     
%             if ( norm( newCol1(:) - newCol2, 2 ) > 1e3*eps )
%                 figure( 1 ), clf
%                 plot( full( newCol1 ), 'bs' ), hold on
%                 plot( newCol2, 'ro' )
%                 legend( 'newCol1', 'newCol2', 0 )
%                 error( 'newCol differs' )
%             end
            
        else
            
            % Delete from the active set
            colToDelete = find( activeSet == jTilde );
            if length( colToDelete ) > 1, error( 'repeated values in activeSet' ), end
            if isempty( colToDelete ), error( 'cannot find jTilde in activeSet' ), end
            
            activeSet( colToDelete ) = [];
            XActiveSet( :, colToDelete ) = [];
            
        end
        
    end                    

    % No need to sort the active set    
    nActiveSetSize = length( activeSet );        
        
    BigCHat = BigCHat( 1 ); % Want this to be a real        
    
    % Print info to console
    disp( sprintf( 'Size of active set=%g', nActiveSetSize ) );
    
    if bitand( sRestoreOptions.debugWord, LL_DEBUG_DISP_ACTIVESET )
        disp( 'active set:' ), disp( activeSet );
    end
    
    if bitand( sRestoreOptions.debugWord, LL_DEBUG_DISP_CHAT )
        disp( 'cHat(activeSet):' ), disp( cHat(activeSet)' );
    end
    
    % SANITY: check that BigCHat matches the predicted one
    if ( i > 1 )
        CHatDiff = abs( BigCHat - nextBigCHat )
        if CHatDiff > sRestoreOptions.CHatTol
            warning( sprintf( 'BigCHat=%g != predicted value=%g', BigCHat, nextBigCHat ) );
            disp( 'activeSetTheory:' ), disp( activeSetTheory );
            error( 'ABORTING' )
        end
    end    

    % SANITY: Check that the absolute value of the correlations in the
    % active set match BigCHat
    nError = 0;
    
    for n = 1:nActiveSetSize
		mismatchDelta = abs( abs(cHat(activeSet(n))) - BigCHat );
        if ( mismatchDelta > sRestoreOptions.corrPercentTol*BigCHat )
            warning( sprintf( '|cHat(%g)|=%g != BigCHat=%g, delta=%g', ...
				activeSet(n), cHat( activeSet(n) ), BigCHat, mismatchDelta ) );
            nError = nError + 1;
        end
    end
    
    if ( nError > 0 ), error( 'ABORTING' ), end
    
    % Plot the cHat and the max BigCHat
    if bitand( sRestoreOptions.debugWord, LL_DEBUG_PLOT )
        figure( 1 ), clf, plot( cHat, 'mx' ), hold on;
        plot( [1, length(cHat)], [BigCHat, BigCHat], 'k--' );
        plot( [1, length(cHat)], [-BigCHat, -BigCHat], 'k--' );
        title( sprintf( 'iter %g', i ) );
        ylabel( 'cHat' );
    end
               
    % Step 3: Construct the variables related to the active set A
    
    if ( size( XActiveSet, 1 ) ~= N | size( XActiveSet, 2 ) ~= nActiveSetSize )
        error( sprintf( 'XActiveSet has incorrect size: %g x %g, expected %g x %g', ...
            size( XActiveSet, 1 ), size( XActiveSet, 2 ), N, nActiveSetSize ) )
    end
    
    GActiveSet = XActiveSet' * XActiveSet; % size |A| x |A|
    
    if bitand( sRestoreOptions.debugWord, LL_DEBUG_DISP_GA )
        disp( 'GActiveSet:' ), disp( GActiveSet )
        % Also display X_A
        figure( 2 ), clf
        for colInd = 1:nActiveSetSize
            plot( XActiveSet( :, colInd ), 'r<' )
            title( sprintf( 'X_A, column %g', colInd ) )
            disp( sprintf( '||X_A(:,%g)||_2 = %g', colInd, norm( XActiveSet( :, colInd ), 2 ) ) )
            disp( 'Press key' ), pause
        end
    end
    
    % Solve the least-squares problem: do not take an explicit inverse
    OneVecActiveSet = ones( nActiveSetSize, 1 ); % |A|-vec
    GInvTimesOneVec = GActiveSet \ OneVecActiveSet;
    AActiveSet = 1/sqrt( OneVecActiveSet' * GInvTimesOneVec );
    
    disp( sprintf( 'SANITY: AActiveSet=%g', AActiveSet ) );
    vecAActiveSet = [ vecAActiveSet, AActiveSet ];
    
    % SANITY: check that AActiveSet is non-increasing if we didn't apply
    % the LASSO kludge.
    if ( ( ~bLassoKludge ) & ( length( vecAActiveSet ) > 1 ) )
        if ( vecAActiveSet(end) > vecAActiveSet(end-1) )
            warning( 'Sanity failed: AActiveSet increased' );
            disp( 'Pausing...' ), pause;
        end
    end
    
    % As before, avoid using GInv
    wActiveSet = AActiveSet * GInvTimesOneVec;
    uActiveSet = XActiveSet * wActiveSet; % N-vec
    normUActiveSet = norm( uActiveSet, 2 );
    disp( sprintf( 'SANITY: |u|=%g', normUActiveSet ) );
    % SANITY: check that |uActiveSet| = 1
    if abs( normUActiveSet - 1 ) > sRestoreOptions.UActiveSetTol
        error( 'Sanity failed: |u| != 1' );
    end
    
    % This should be equal AActiveSet * 1
    equiAngleCheckVec = full( XActiveSet' * uActiveSet );
    if ( length( equiAngleCheckVec ) ~= nActiveSetSize )
        error( sprintf( 'equiAngleCheckVec has length %g', length( equiAngleCheckVec ) ) )
    end
    
    % SANITY: check that each element of tmp equals AActiveSet
    for n = 1:nActiveSetSize
        AActiveSetDelta = abs( equiAngleCheckVec(n) - AActiveSet );
        if AActiveSetDelta > sRestoreOptions.AActiveSetPercentTol * AActiveSet
            warning( sprintf( 'Sanity failed: (X_A^T * u_A)(%g) != AActiveSet [%g]', n, AActiveSetDelta ) );
            disp( 'equiAngleCheckVec:' ), disp( equiAngleCheckVec )
            error( 'Fatal error' )
        end
    end
    
    a = GX' * reshape( full( uActiveSet ), opImgDim );
    a = a(:); % M-vec
    
    % Construct dHat
    dHat = zeros( M, 1 ); % M-vec
    for j = 1:nActiveSetSize
        jj = activeSet( j );
        dHat( jj ) = s( jj ) * wActiveSet( j );
    end
    
    % Step 4a: Compute gammaHat and jHat
    
    jHatColumn = -1; % DEBUG variable
    
    if ( i < M )
        
        % We can only do this if i < NN. Here, compute
        % gammaHat = min^+_{j \in A^c} { (CHat - chat_j)/(A_A - a_j), (CHat
        % + chat_j)/(A_A + a_j) }
        
        activeSetC = setdiff( 1:M, activeSet );
        % DEBUG
        %disp( 'activeSetC:' ), disp( activeSetC(:)' );
        
        % Free and reallocate gammaMat
        clear gammaMat;
        gammaMat = zeros( length( activeSetC ), 2 );
        
        % Compute the 1st column of gammaMat
        numer = BigCHat - cHat( activeSetC );
        denom = AActiveSet - a( activeSetC );
        tmp = numer(:) ./ denom(:);
        % SANITY
        if length( tmp(:) ) ~= length( activeSetC ), error( 'Sanity on length of tmp failed' ), end;
        gammaMat( :, 1 ) = tmp(:);
        
        % Compute the 2nd column of gammaMat
        numer = BigCHat + cHat( activeSetC );
        denom = AActiveSet + a( activeSetC );
        tmp = numer(:) ./ denom(:);
        % SANITY
        if length( tmp(:) ) ~= length( activeSetC ), error( 'Sanity on length of tmp failed' ), end;
        gammaMat( :, 2 ) = tmp(:);

        % Take the min over each column
        gammaMatPosInd = find( ( gammaMat(:,1) > sRestoreOptions.gammaTol ) & isfinite( gammaMat(:,1) ) );
        if ~isempty( gammaMatPosInd )
            gammaMatCol1Min = min( gammaMat( gammaMatPosInd, 1 ) );
            gammaMatCol1MinInd = find( gammaMat(:,1) == gammaMatCol1Min );
            if length( gammaMatCol1MinInd ) > 1
                error( 'cannot handle length( gammaMatCol1MinInd ) > 1' );
            end            

            if bitand( sRestoreOptions.debugWord, LL_DEBUG_GAMMA )
                disp( 'gammaMatCol1Min:' ), disp( gammaMatCol1Min );
                disp( 'gammaMatCol1MinInd:' ), disp( gammaMatCol1MinInd );
                disp( 'jHat of min:' ), disp( activeSetC( gammaMatCol1MinInd ) );
                disp( 'row of gammaMat:' ), disp( gammaMat( gammaMatCol1MinInd, : ) );
            end
        else
            gammaMatCol1Min = inf;
        end
        
        gammaMatPosInd = find( ( gammaMat(:,2) > sRestoreOptions.gammaTol ) & isfinite( gammaMat(:,2) ) );
        if ~isempty( gammaMatPosInd )
            gammaMatCol2Min = min( gammaMat( gammaMatPosInd, 2 ) );
            gammaMatCol2MinInd = find( gammaMat(:,2) == gammaMatCol2Min );
            if length( gammaMatCol2MinInd ) > 1
                error( 'cannot handle length( gammaMatCol2MinInd ) > 1' );
            end
            
            if bitand( sRestoreOptions.debugWord, LL_DEBUG_GAMMA )
                disp( 'gammaMatCol2Min:' ), disp( gammaMatCol2Min );
                disp( 'gammaMatCol2MinInd:' ), disp( gammaMatCol2MinInd );
                disp( 'jHat of min:' ), disp( activeSetC( gammaMatCol2MinInd ) );
                disp( 'row of gammaMat:' ), disp( gammaMat( gammaMatCol2MinInd, : ) );
            end
        else
            gammaMatCol2Min = inf;
        end

        % And then combine to find the min over the entire matrix
        [gmm, gmmInd] = min( [gammaMatCol1Min, gammaMatCol2Min] );
        if isfinite( gmm )
            gammaHat = gmm;
            if ( gmmInd == 1 )
                if bitand( sRestoreOptions.debugWord, LL_DEBUG_GAMMA )
                    disp( 'Getting jHat from 1st column of gammaMat' );
                end
                jHatColumn = 1;
                jHat = activeSetC( gammaMatCol1MinInd );
            else
                if bitand( sRestoreOptions.debugWord, LL_DEBUG_GAMMA )
                    disp( 'Getting jHat from 2nd column of gammaMat' );
                end
                jHatColumn = 2;
                jHat = activeSetC( gammaMatCol2MinInd );
            end
        else
            jHatColumn = -1; % Invalid value
            error( 'could not find gammaHat and jHat' );
        end

    else
        
        % Do this if i == M. This is the last step
        gammaHat = BigCHat/AActiveSet;
        jHat = -1;
        
    end
    
    % Step 4b: compute gammaTilde
 
    warning off; % XXX: We are going to get some divide-by-zero errors here!
    gammaTildeVec = -betaHatPrevActiveSet ./ dHat; % This is a M-vector
    warning on;

    % Find the min gamma s.t. it is positive 
    gammaTildePosInd = find( ( gammaTildeVec > sRestoreOptions.gammaTol ) & isfinite( gammaTildeVec ) );
    gtm = min( gammaTildeVec( gammaTildePosInd ) );
    
    switch length( gtm )
        case 0
            % No positive elements in gammaTilde
            gammaTilde = inf;
            jTilde = -1;
        case 1
            % This is the "normal" case
            gtmInd = find( gammaTildeVec == gtm );
            if length( gtmInd ) > 1
                error( 'cannot handle length( gtmInd ) > 1' );
            elseif length( gtmInd ) == 1
                gammaTilde = gtm;
                jTilde = gtmInd;
                % Sanity: check that jTilde is in the active set
                if ~ismember( jTilde, activeSet )
                    error( 'jTilde not in the active set' );
                end
            else
                error( 'Cannot find minimum in gammaTildeVec?' );
            end
        otherwise
            error( 'Multiple minima in gammaTildeVec?' );
    end

    % Step 4c: get gammaConst
    
    bLassoKludge = 0; % Re-initialize
    
    if ( gammaTilde < gammaHat )
        bLassoKludge = 1;
        gammaConst = gammaTilde;
        disp( '*** LARS-LASSO KLUDGE ***' );
    else
        gammaConst = gammaHat;
    end
    
    disp( [ sprintf( 'gammaHat=%g, jHat=%g/', gammaHat, jHat ), ...
        sprintf( 'gammaTilde=%g, jTilde=%g\n', gammaTilde, jTilde ), ...
        sprintf( '=> gammaConst=%g', gammaConst ) ] );    
    
    % SANITY
    nextBigCHat = BigCHat - gammaConst * AActiveSet;
    disp( sprintf( 'nextBigCHat = %g', nextBigCHat ) );
    
    % Check that the new correlation equals nextBigCHat
    if ~bLassoKludge
        cjHat = cHat( jHat ) - gammaConst * a( jHat );
        disp( sprintf( 'cjHat = %g', cjHat ) );
        
        % Find entry in gammaMat
        jHatGammaMatIndex = find( activeSetC == jHat );
        if length( jHatGammaMatIndex ) ~= 1, error( 'failed to find jHat in activeSetC' ), end;
        if bitand( sRestoreOptions.debugWord, LL_DEBUG_GAMMA )
            disp( 'jHatGammaMatIndex:' ), disp( jHatGammaMatIndex );
            disp( 'jHat entry in gammaMat:' ), disp( gammaMat( jHatGammaMatIndex, : ) );
        end
        
        % Check if |cjHat| equals nextBigCHat. It should!
        if abs( abs(cjHat) - nextBigCHat ) > sRestoreOptions.CHatTol
            % Re-compute gamma
            switch jHatColumn
                case 1
                    gamma2 = ( BigCHat - cHat( jHat ) ) / ( AActiveSet - a( jHat ) );
                case 2
                    gamma2 = ( BigCHat + cHat( jHat ) ) / ( AActiveSet + a( jHat ) );
                otherwise
                    error( 'trying to re-compute gammaHat with invalid jHatColumn' );
            end
            disp( sprintf( 'Re-computed gammaHat = %g', gamma2 ) );
            error( 'cjHat not equal to nextBigCHat??' );
        end
    end
    
    
    % Step 5: Form the next estimator of mu and beta
    muHatNextActiveSet = muHatPrevActiveSet + gammaConst * uActiveSet;    
    betaHatNextActiveSet = betaHatPrevActiveSet + gammaConst * dHat;
   
   	% Update metrics
    errorNorm1Hist = [ errorNorm1Hist, norm( betaHatNextActiveSet(:) - betaTrue(:), 1 ) ];
    errorNorm2Hist = [ errorNorm2Hist, norm( betaHatNextActiveSet(:) - betaTrue(:), 2 ) ];
	betaHatDiff = gammaConst * dHat(:);
    % The l_infinity norm of the difference
    thetaMaxDiffHist = [thetaMaxDiffHist; max( abs( betaHatDiff ) )];
    % The l_1 norm of the difference
    thetaSumDiffHist = [thetaSumDiffHist; sum( abs( betaHatDiff ) )];

    bigCHatHist = [ bigCHatHist; BigCHat ];
    
    lambda = 2* BigCHat * sRestoreOptions.XColumnNorm;
    vecSoftThres(i) = lambda/2;
    disp( sprintf( 'BigCHat = %g, lambda=%g', BigCHat, lambda ) );


    % Compute using Daubechies' iterations. Here, we MUST have the size of
    % the i/p and o/p images be the same, i.e., M=N.

    if ( bitand( sRestoreOptions.debugWord, LL_DEBUG_DAUB ) & ( i > 1 ) )

        if ( M ~= N ), error( 'Can only run Daubechies iterations under M=N' ), end
        
        if ( sRestoreOptions.reconSanityInterval > 1 ) & ...
                ( mod( i, sRestoreOptions.reconSanityInterval ) == 0 )

            disp( sprintf( 'Sanity: running Daubechies at i=%g', i ) );
            
            [sDIter, sDIterOptions] = emgauss_restore( 'forward_op', psf, ...
                'noisy_output', reshape( y, opImgDim ), ...
                'noise_sigma', noiseSigma, ...
                'theta_true', reshape( betaTrue, ipImgDim ), ...
                'm_selector', 'daub2', ...
                'num_em_iters', sRestoreOptions.NUM_DAUB_ITERS, ...
                'num_max_grad_search_failures', -1, ...
                'theta0_selector', 'Hty2', ...
                'theta0', [], ...
                'a_val', -1, ...
                'w_val', -1, ...
                't_val', -1, ...
                'sparse_ind', 'w', ...
                'b_save_history', 0, ...
                'stopping_cond', 'fixed', ...
                'stopping_cond_tol', noiseSigma, ...
                'stopping_cond_win', 0, ...
                'stopping_cond_last_fac', 0, ...
                'thetastar', 0, ...
                'soft_thres_t', -1, ...
                'hparam_rest_interval', 0, ...
                'bseparate_hparam_est', 0, ...
                'lambda', lambda ...
                );

            larsSol = betaHatPrevActiveSet/sRestoreOptions.XColumnNorm;
            daubSol = sDIter.last;
            [deltaLinf, deltaLinfInd] = max( abs( larsSol(:) - daubSol(:) ) );
            disp( sprintf( '|larsSol - daubSol|_inf = %g', deltaLinf ) );
            absDaubVal = abs( daubSol( deltaLinfInd ) );
            
            if ( deltaLinfInd > absDaubVal * sRestoreOptions.daubPercentTol )
                
                warning( sprintf( '|larsSol - daubSol|_inf = %g', ...
                    sRestoreOptions.daubDeltaTolFac*absDaubVal ) );
                
                if length( ipImgDim ) == 2
                    figure( 2 ), clf, imagesc( reshape( larsSol, ipImgDim ) ), colorbar;
                    title( 'LARS-LASSO' );
                    figure( 3 ), clf, imagesc( reshape( daubSol, ipImgDim ) ), colorbar;
                    title( 'Daubechies' );
                else
                    warning( 'Since the ip img is not 2-d, skipping plots #2 and #3' )
                end
                
                figure( 4 ), clf, plot( larsSol(:), 'bx' ), hold on, plot( daubSol(:), 'ro' );
                title( 'LARS-LASSO vs. Daubechies' );
                
                disp( 'Pausing...' ), pause;
                
            else
                disp( 'Pausing briefly' ), pause( sRestoreOptions.PAUSE_NUM_SECS );
            end

        end
        
    end

    betaOriginalScale = betaHatNextActiveSet(:) / sRestoreOptions.XColumnNorm;
    vecL0( i ) = sum( abs( betaOriginalScale ) > sRestoreOptions.l0Tol ); % l_0(beta)
    vecL1( i ) = sum( abs( betaOriginalScale ) ); % l_1(beta)

    % Compute the cost function  
    predictionErr = y - muHatNextActiveSet;
    rss = sum( predictionErr(:).^2 );
    vecCost( i ) = N*noiseSigma^2 + rss + 2*noiseSigma^2*vecL0(i);
    
    if ( ( betaHatBestInd < 1 ) | ...
         ( (betaHatBestInd >= 1) & (vecCost(betaHatBestInd) > vecCost(i)) ) )
        betaHatBestInd = i;
        betaHatBest = betaHatNextActiveSet;
    end
       
    % Step 6: Update state variables
    muHatPrevActiveSet = muHatNextActiveSet;
    betaHatPrevActiveSet = betaHatNextActiveSet;

    if bitand( sRestoreOptions.debugWord, LL_DEBUG_ITER_PAUSE )
        disp( 'Pausing...press key' ), pause
    end

end

tCpuTimeStop = cputime;
tElapsedTimeStop = clock;

disp( 'CPU time:' ), tCpuTimeStop - tCpuTimeStart
disp( 'Elapsed time:' ), etime( tElapsedTimeStop, tElapsedTimeStart )

% Re-scale last beta and best beta
betaHat = betaHatPrevActiveSet / sRestoreOptions.XColumnNorm;
betaHatBest = betaHatBest / sRestoreOptions.XColumnNorm;
% muCheck should equal muHatPrevActiveSet
muCheck = GX * reshape( full( betaHatPrevActiveSet ), ipImgDim );
muCheck = muCheck(:);


% Generate the output structure
for elemInd = 1:size( sRestoreElems, 1 )
    szCmd = sprintf( 'sOut.%s = %s;', ...
        char( sRestoreElems(elemInd,1) ), ...       
        char( sRestoreElems(elemInd,2) ) );
    try
        eval( szCmd );
    catch
        warning( sprintf( 'Could not execute: %s', szCmd ) );
    end
end

    
    

    
    
    


