%
% LARS-LASSO-SURE. Adapted from lars3. Usage is
% sOut = ll_small( varargin )
% 

% The problem we consider is
% y = X * beta + n
% y ~ N-dim vector, beta ~ M-dim vector, X ~ N x M

% NB. This function ASSUMES that the columns of X have unit norm. As well,
% this does NOT work for problems where X is large.

% This is basically lars3, except we are spinning off a new line of code
% changes. The name is an indication that this function will only work if
% we can obtain the entire matrix X.

% Basically, this is a naive implementation of LARS-LASSO-SURE. We have
% both psf and the full matrix X; however, the latter is unnecessary given
% the former. That is rectified in ll_big.

function sOut = ll_small( varargin )

% Constants

ll_defs; % Load from file

noiseSigma = -1;

% Default restore options. Put all the tolerances in sRestoreOptions.
sRestoreOptions = struct( 'X', [], ...
    'XColumnNorm', [], ...
    'reconSanityInterval', 10, ...
    'debugWord', 0, ...
    'betaTol', 1e-9, 'CHatTol', 1e-6, ...
    'gammaTol', 1e-9, 'daubPercentTol', 1e-4, 'l0Tol', 1e4*eps, ...
    'UActiveSetTol', 1e-6, 'AActiveSetPercentTol', 1e-6, 'corrPercentTol', 1e-3, ...
    'NUM_DAUB_ITERS', 2e4, ...
    'PAUSE_NUM_SECS', 5 );

% Defines the structure of sRestore
sRestoreElems = { ...
    'lastIter',     'betaHat'; ...
    'last',         'betaHatBest'; ...
    'lastInd1',     'betaHatBestInd'; ...
    'muHat',        'muHatPrevActiveSet'; ...
    'bigCHat',      'bigCHatHist'; ...
    'activeSet',    'activeSet'; ...
    'vecL1',        'vecL1'; ...
    'vecL0',        'vecL0'; ...
    'vecSoftThres', 'vecSoftThres'; ...
    'vecCost',      'vecCost'; ...
    'cputime',      '[tCpuTimeStart, tCpuTimeStop]'; ...
    'thetaMaxDiff', 'thetaMaxDiffHist'; ...
    'thetaSumDiff', 'thetaSumDiffHist'; ...
    'errorNorm1',   'errorNorm1Hist'; ...
    'errorNorm2',   'errorNorm2Hist'; ...
    };

% Process input arguments

lenVarargin = length( varargin );
numArgsIn = floor( lenVarargin/2 );

if 2*numArgsIn ~= lenVarargin, error( 'even ip args expected' ), end

for argInInd = 1:numArgsIn
    
    switch varargin{2*argInInd-1}      
        case 'noisy_output', y = varargin{2*argInInd};      
        case 'beta_true', betaTrue = varargin{2*argInInd}; 
        case 'num_iters', sRestoreOptions.NUM_ITERS = varargin{2*argInInd};    
        case 'n_debug', sRestoreOptions.N_DEBUG = varargin{2*argInInd};
        case 'psf', psf = varargin{2*argInInd};
        case 'noise_sigma', noiseSigma = varargin{2*argInInd};
		case 'ip_img_dim', ipImgDim = varargin{2*argInInd}; 
		case 'op_img_dim', opImgDim = varargin{2*argInInd};
        case 'recon_sanity_interval', sRestoreOptions.reconSanityInterval = varargin{2*argInInd};
        case 'debug_word', sRestoreOptions.debugWord = varargin{2*argInInd};
            
        % Now these ONLY appear in ll_small and NOT in ll_big. X is the
        % matrix where the columns have already been normalized to have
        % unit norm. The original norm of the columns is in XColumnNorm.
        
        case 'x', sRestoreOptions.X = varargin{2*argInInd};
        case 'x_column_norm', sRestoreOptions.XColumnNorm = varargin{2*argInInd};

        otherwise
            error( sprintf( 'unrecognized option %s', szKey ) );
    end
end

% Check validity of arguments
%%%

N = length( y );
M = length( betaTrue );

if ( size( sRestoreOptions.X, 1 ) ~= N ) | ( size( sRestoreOptions.X, 2 ) ~= M )
    error( 'size of X incompatible with y, betaTrue' );
end

if ( sRestoreOptions.NUM_ITERS > M ) | ( sRestoreOptions.NUM_ITERS < 1 )
    error( 'invalid NUM_ITERS' );
elseif ( sRestoreOptions.NUM_ITERS == M )
    disp( 'Running LARS to LS solution' );
else
    disp( 'Running LARS, but stopping short of LS solution' );
end

if ( noiseSigma <= 0 ), error( 'noiseSigma must be >0' ), end;

if ( prod( opImgDim ) ~= N ) | ( prod( ipImgDim ) ~= M )
	error( 'op or ip image dimensions are inconsistent' );
end

if ( M ~= N ), error( 'must have M=N' ), end;

% Metrics for convergence of iterations
%%%

thetaMaxDiffHist = [];  % l_inf norm of difference of 2 consecutive estimates
thetaSumDiffHist = [];  % l_1 norm of difference ...

errorNorm1Hist = [];    % l_1 norm of difference
errorNorm2Hist = [];    % l_2 norm ...

bigCHatHist = [];

% Pre-compute XTX = X' * X. This is only possible for X SMALL.
XTX = (sRestoreOptions.X)' * (sRestoreOptions.X);

activeSet = [];
vecAActiveSet = [];

muHatPrevActiveSet = zeros( N, 1 );
muHatNextActiveSet = zeros( size( muHatPrevActiveSet ) );

betaHatPrevActiveSet = zeros( M, 1 );
betaHatNextActiveSet = zeros( size( betaHatPrevActiveSet ) );
betaHatBest = zeros( size( betaHatPrevActiveSet ) );
betaHatBestInd = -1;

gammaMat = zeros( M, 2 );

% Keep track of various statistics during the iterations.
%%%

% SURE cost criterion
vecCost = zeros( sRestoreOptions.NUM_ITERS, 1 ); 
% l_0 of estimate
vecL0 = zeros( size( vecCost ) ); 
% soft threshold
vecSoftThres = zeros( size( vecCost ) );

bLassoKludge = 0; % Flag whenever we have to do a LARS-LASSO kludge
jHat = -1;
jKludge = -1;
nextBigCHat = -1;

% Time the iterations
tCpuTimeStart = cputime;

for i = 1:sRestoreOptions.NUM_ITERS
    
    if bitand( sRestoreOptions.debugWord, LL_DEBUG_LOOP1 )
        disp( sprintf( '\nITERATION %g\n', i ) )
    end
    
    % Step 1: Compute the correlations
    disp( 'STEP 1: Computing correlations' );    
    cHat = sRestoreOptions.X' * ( y - muHatPrevActiveSet );    
    s = sign( cHat );
    
    betaNzInd = find( abs( betaHatPrevActiveSet ) > sRestoreOptions.betaTol ); % Find non-zero indices
    vecMismatch = ( s(betaNzInd) ~= sign( betaHatPrevActiveSet( betaNzInd ) ) );
    if bitand( sRestoreOptions.debugWord, LL_DEBUG_LOOP2 )
        disp( 'betaNzInd:' ), disp( betaNzInd(:)' );
        disp( 'c-values:' ), tmp = cHat( betaNzInd ); disp( tmp(:)' );
        disp( 'beta-values:' ), tmp = betaHatPrevActiveSet(betaNzInd); disp( tmp(:)' );
    end
    
    % SANITY: check that sign(beta_j) = sign(c_j) = s_j, for j in the
    % active set
    if sum( vecMismatch ) ~= 0
        figure( 2 ), clf, plot( s(betaNzInd), 'kx' ), hold on;
        plot( sign(betaHatPrevActiveSet(betaNzInd)), 'ro' );
        legend( 's', 'sign(beta)', 0 );
        error( 'LASSO solution sign contraint violated' );
    end
                   
    % Step 2: Get the active set and sort in ascending order of indices
    disp( 'STEP 2: Adjusting active set' );
    
    if ( i == 1 )
        % In the first iteration, use max to locate the active set
        [BigCHat, activeSet] = max( abs( cHat ) );
    else
        
        % Do this step in subsequent iterations        
        [BigCHat, activeSetTheory] = max( abs( cHat ) );

        if ~bLassoKludge
            % In subsequent iterations, add to the active set manually. It
            % turns out that there are some numerical differences in the cHats
            % that are supposed to be the same.
            activeSet = [activeSet, jHat];
        else
            activeSet = setdiff( activeSet, jTilde );
        end
        
    end                    
    
    [activeSetSorted, activeSetSortedInd] = sort( activeSet );
    activeSet = activeSetSorted;
    nActiveSetSize = length( activeSet );
    
    BigCHat = BigCHat( 1 ); % Want this to be a real        
    
    % SANITY: check that BigCHat matches the predicted one
    if ( i > 1 )
        CHatDiff = abs( BigCHat - nextBigCHat )
        if CHatDiff > sRestoreOptions.CHatTol
            warning( sprintf( 'BigCHat=%g != predicted value=%g', BigCHat, nextBigCHat ) );
            disp( 'activeSetTheory:' ), disp( activeSetTheory );
            disp( 'Pausing...' ), pause;
        end
    end
    
    % Print info to console
    disp( sprintf( 'Size of active set=%g', nActiveSetSize ) );
    disp( 'active set:' ), disp( activeSet );
    disp( 'cHat(activeSet):' ), disp( cHat(activeSet)' );

    % SANITY: Check that the absolute value of the correlations in the
    % active set match BigCHat
    for n = 1:nActiveSetSize
		mismatchDelta = abs( abs(cHat(activeSet(n))) - BigCHat );
        if ( mismatchDelta > sRestoreOptions.corrPercentTol*BigCHat )
            error( sprintf( '|cHat(%g)|=%g != BigCHat=%g, delta=%g', ...
				activeSet(n), cHat( activeSet(n) ), BigCHat, mismatchDelta ) );
        end
    end
    
    % Plot the cHat and the max BigCHat
    if bitand( sRestoreOptions.debugWord, LL_DEBUG_PLOT )
        figure( 1 ), clf, plot( cHat, 'c*' ), hold on;
        plot( [1, length(cHat)], [BigCHat, BigCHat], 'k--' );
        plot( [1, length(cHat)], [-BigCHat, -BigCHat], 'k--' );
        title( sprintf( 'iter %g', i ) );
        ylabel( 'cHat' );
    end
               
    % Step 3: Construct the variables related to the active set A
    
    % Free and reallocate memory for nz{I,J,S}
    clear nzI nzJ nzS;
    nzI = zeros( nActiveSetSize, 1 );
    nzJ = zeros( size( nzI ) );
    nzS = zeros( size( nzI ) );
    
    for j = 1:nActiveSetSize
        nzI(j) = activeSet(j);
        nzJ(j) = j;
        nzS(j) = s( activeSet(j) );
    end
    
    % Allocate space for SActiveSet, which is of size M x nActiveSetSize
	SActiveSet = sparse( nzI, nzJ, nzS, M, nActiveSetSize );
    
    % Compute the other variables
    GActiveSet = SActiveSet' * XTX * SActiveSet; % size |A| x |A|
    
    if bitand( sRestoreOptions.debugWord, LL_DEBUG_DISP_GA )
        disp( 'GActiveSet:' ), disp( GActiveSet )
        % Also display X_A
        X_ASign = s( activeSet );
        X_A = sRestoreOptions.X( :, activeSet ) .* repmat( X_ASign(:)', N, 1 );
        disp( 'X_A^T * X_A:' ), disp( X_A' * X_A )
        figure( 2 ), clf
        for colInd = 1:nActiveSetSize
            plot( X_A( :, colInd ), 'bs' )
            title( sprintf( 'X_A, column %g', colInd ) )
            disp( sprintf( '||X_A(:,%g)||_2 = %g', colInd, norm( X_A( :, colInd ), 2 ) ) )
            disp( 'Press key' ), pause
        end
    end
    
    % Solve the least-squares problem: do not take an explicit inverse
    OneVecActiveSet = ones( nActiveSetSize, 1 ); % |A|-vec
    GInvTimesOneVec = GActiveSet \ OneVecActiveSet;
    tmp = OneVecActiveSet' * GInvTimesOneVec;
    AActiveSet = 1/sqrt(tmp);
    
    disp( sprintf( 'SANITY: AActiveSet=%g', AActiveSet ) );
    vecAActiveSet = [ vecAActiveSet, AActiveSet ];
    
    % SANITY: check that AActiveSet is non-increasing if we didn't apply
    % the LASSO kludge.
    if ( ( ~bLassoKludge ) & ( length( vecAActiveSet ) > 1 ) )
        if ( vecAActiveSet(end) > vecAActiveSet(end-1) )
            warning( 'Sanity failed: AActiveSet increased' );
            disp( 'Pausing...' ), pause;
        end
    end
    
    % As before, avoid using GInv
    wActiveSet = AActiveSet * GInvTimesOneVec;
    uActiveSet = (sRestoreOptions.X * SActiveSet) * wActiveSet; % N-vec
    normUActiveSet = norm( uActiveSet, 2 );
    disp( sprintf( 'SANITY: |u|=%g', normUActiveSet ) );
    % SANITY: check that |uActiveSet| = 1
    if abs( normUActiveSet - 1 ) > sRestoreOptions.UActiveSetTol
        error( 'Sanity failed: |u| != 1' );
    end
    
    % This should be equal AActiveSet * 1
    tmp = full( ( SActiveSet' * sRestoreOptions.X' ) * uActiveSet );
    % SANITY: check that each element of tmp equals AActiveSet
    for n = 1:nActiveSetSize
        AActiveSetDelta = abs( tmp(n) - AActiveSet );
        if AActiveSetDelta > sRestoreOptions.AActiveSetPercentTol * AActiveSet
            error( sprintf( 'Sanity failed: (X_A^T * u_A)(%g) != AActiveSet [%g]', n, AActiveSetDelta ) );
        end
    end
    
    a = sRestoreOptions.X' * uActiveSet; % M-vec
    %disp( 'a:' ), disp( a );
    
    % Construct dHat
    dHat = zeros( M, 1 ); % M-vec
    for j = 1:nActiveSetSize
        jj = activeSet( j );
        dHat( jj ) = s( jj ) * wActiveSet( j );
    end
    
    % Step 4a: Compute gammaHat and jHat
    
    jHatColumn = -1; % DEBUG variable
    
    if ( i < M )
        
        % We can only do this if i < NN. Here, compute
        % gammaHat = min^+_{j \in A^c} { (CHat - chat_j)/(A_A - a_j), (CHat
        % + chat_j)/(A_A + a_j) }
        
        activeSetC = setdiff( 1:M, activeSet );
        % DEBUG
        %disp( 'activeSetC:' ), disp( activeSetC(:)' );
        
        % Free and reallocate gammaMat
        clear gammaMat;
        gammaMat = zeros( length( activeSetC ), 2 );
        
        % Compute the 1st column of gammaMat
        numer = BigCHat - cHat( activeSetC );
        denom = AActiveSet - a( activeSetC );
        tmp = numer(:) ./ denom(:);
        % SANITY
        if length( tmp(:) ) ~= length( activeSetC ), error( 'Sanity on length of tmp failed' ), end;
        gammaMat( :, 1 ) = tmp(:);
        
        % Compute the 2nd column of gammaMat
        numer = BigCHat + cHat( activeSetC );
        denom = AActiveSet + a( activeSetC );
        tmp = numer(:) ./ denom(:);
        % SANITY
        if length( tmp(:) ) ~= length( activeSetC ), error( 'Sanity on length of tmp failed' ), end;
        gammaMat( :, 2 ) = tmp(:);

        % Take the min over each column
        gammaMatPosInd = find( ( gammaMat(:,1) > sRestoreOptions.gammaTol ) & isfinite( gammaMat(:,1) ) );
        if ~isempty( gammaMatPosInd )
            gammaMatCol1Min = min( gammaMat( gammaMatPosInd, 1 ) );
            gammaMatCol1MinInd = find( gammaMat(:,1) == gammaMatCol1Min );
            if length( gammaMatCol1MinInd ) > 1
                error( 'cannot handle length( gammaMatCol1MinInd ) > 1' );
            end            

            if bitand( sRestoreOptions.debugWord, LL_DEBUG_GAMMA )
                disp( 'gammaMatCol1Min:' ), disp( gammaMatCol1Min );
                disp( 'gammaMatCol1MinInd:' ), disp( gammaMatCol1MinInd );
                disp( 'jHat of min:' ), disp( activeSetC( gammaMatCol1MinInd ) );
                disp( 'row of gammaMat:' ), disp( gammaMat( gammaMatCol1MinInd, : ) );
            end
        else
            gammaMatCol1Min = inf;
        end
        
        gammaMatPosInd = find( ( gammaMat(:,2) > sRestoreOptions.gammaTol ) & isfinite( gammaMat(:,2) ) );
        if ~isempty( gammaMatPosInd )
            gammaMatCol2Min = min( gammaMat( gammaMatPosInd, 2 ) );
            gammaMatCol2MinInd = find( gammaMat(:,2) == gammaMatCol2Min );
            if length( gammaMatCol2MinInd ) > 1
                error( 'cannot handle length( gammaMatCol2MinInd ) > 1' );
            end
            
            if bitand( sRestoreOptions.debugWord, LL_DEBUG_GAMMA )
                disp( 'gammaMatCol2Min:' ), disp( gammaMatCol2Min );
                disp( 'gammaMatCol2MinInd:' ), disp( gammaMatCol2MinInd );
                disp( 'jHat of min:' ), disp( activeSetC( gammaMatCol2MinInd ) );
                disp( 'row of gammaMat:' ), disp( gammaMat( gammaMatCol2MinInd, : ) );
            end
        else
            gammaMatCol2Min = inf;
        end

        % And then combine to find the min over the entire matrix
        [gmm, gmmInd] = min( [gammaMatCol1Min, gammaMatCol2Min] );
        if isfinite( gmm )
            gammaHat = gmm;
            if ( gmmInd == 1 )
                if bitand( sRestoreOptions.debugWord, LL_DEBUG_GAMMA )
                    disp( 'Getting jHat from 1st column of gammaMat' );
                end
                jHatColumn = 1;
                jHat = activeSetC( gammaMatCol1MinInd );
            else
                if bitand( sRestoreOptions.debugWord, LL_DEBUG_GAMMA )
                    disp( 'Getting jHat from 2nd column of gammaMat' );
                end
                jHatColumn = 2;
                jHat = activeSetC( gammaMatCol2MinInd );
            end
        else
            jHatColumn = -1; % Invalid value
            error( 'could not find gammaHat and jHat' );
        end

    else
        
        % Do this if i == M. This is the last step
        gammaHat = BigCHat/AActiveSet;
        jHat = -1;
        
    end
    
    % Step 4b: compute gammaTilde
 
    warning off; % XXX: We are going to get some divide-by-zero errors here!
    gammaTildeVec = -betaHatPrevActiveSet ./ dHat; % This is a M-vector
    warning on;

    % Find the min gamma s.t. it is positive 
    gammaTildePosInd = find( ( gammaTildeVec > sRestoreOptions.gammaTol ) & isfinite( gammaTildeVec ) );
    gtm = min( gammaTildeVec( gammaTildePosInd ) );
    
    switch length( gtm )
        case 0
            % No positive elements in gammaTilde
            gammaTilde = inf;
            jTilde = -1;
        case 1
            % This is the "normal" case
            gtmInd = find( gammaTildeVec == gtm );
            if length( gtmInd ) > 1
                error( 'cannot handle length( gtmInd ) > 1' );
            elseif length( gtmInd ) == 1
                gammaTilde = gtm;
                jTilde = gtmInd;
                % Sanity: check that jTilde is in the active set
                if ~ismember( jTilde, activeSet )
                    error( 'jTilde not in the active set' );
                end
            else
                error( 'Cannot find minimum in gammaTildeVec?' );
            end
        otherwise
            error( 'Multiple minima in gammaTildeVec?' );
    end

    % Step 4c: get gammaConst
    
    bLassoKludge = 0; % Re-initialize
    
    if ( gammaTilde < gammaHat )
        bLassoKludge = 1;
        gammaConst = gammaTilde;
        disp( '*** LARS-LASSO KLUDGE ***' );
    else
        gammaConst = gammaHat;
    end
    
    disp( [ sprintf( 'gammaHat=%g, jHat=%g/', gammaHat, jHat ), ...
        sprintf( 'gammaTilde=%g, jTilde=%g\n', gammaTilde, jTilde ), ...
        sprintf( '=> gammaConst=%g', gammaConst ) ] );    
    
    % SANITY
    nextBigCHat = BigCHat - gammaConst * AActiveSet;
    disp( sprintf( 'nextBigCHat = %g', nextBigCHat ) );
    
    % Check that the new correlation equals nextBigCHat
    if ~bLassoKludge
        cjHat = cHat( jHat ) - gammaConst * a( jHat );
        disp( sprintf( 'cjHat = %g', cjHat ) );
        
        % Find entry in gammaMat
        jHatGammaMatIndex = find( activeSetC == jHat );
        if length( jHatGammaMatIndex ) ~= 1, error( 'failed to find jHat in activeSetC' ), end;
        if bitand( sRestoreOptions.debugWord, LL_DEBUG_GAMMA )
            disp( 'jHatGammaMatIndex:' ), disp( jHatGammaMatIndex );
            disp( 'jHat entry in gammaMat:' ), disp( gammaMat( jHatGammaMatIndex, : ) );
        end
        
        % Check if |cjHat| equals nextBigCHat. It should!
        if abs( abs(cjHat) - nextBigCHat ) > sRestoreOptions.CHatTol                                    
            % Re-compute gamma
            switch jHatColumn
                case 1
                    gamma2 = ( BigCHat - cHat( jHat ) ) / ( AActiveSet - a( jHat ) );
                case 2
                    gamma2 = ( BigCHat + cHat( jHat ) ) / ( AActiveSet + a( jHat ) );
                otherwise
                    error( 'trying to re-compute gammaHat with invalid jHatColumn' );
            end
            disp( sprintf( 'Re-computed gammaHat = %g', gamma2 ) );
            error( 'cjHat not equal to nextBigCHat??' );
        end
    end
    
    
    % Step 5: Form the next estimator of mu and beta
    muHatNextActiveSet = muHatPrevActiveSet + gammaConst * uActiveSet;    
    betaHatNextActiveSet = betaHatPrevActiveSet + gammaConst * dHat;
   
   	% Update metrics
    errorNorm1Hist = [ errorNorm1Hist, norm( betaHatNextActiveSet(:) - betaTrue(:), 1 ) ];
    errorNorm2Hist = [ errorNorm2Hist, norm( betaHatNextActiveSet(:) - betaTrue(:), 2 ) ];
	betaHatDiff = gammaConst * dHat(:);
    % The l_infinity norm of the difference
    thetaMaxDiffHist = [thetaMaxDiffHist; max( abs( betaHatDiff ) )];
    % The l_1 norm of the difference
    thetaSumDiffHist = [thetaSumDiffHist; sum( abs( betaHatDiff ) )];

    bigCHatHist = [ bigCHatHist; BigCHat ];
    
    lambda = 2* BigCHat * sRestoreOptions.XColumnNorm( activeSet(1) );
    vecSoftThres(i) = lambda/2;
    disp( sprintf( 'BigCHat = %g, lambda=%g', BigCHat, lambda ) );


    % Compute using Daubechies' iterations. Here, we MUST have the size of
    % the i/p and o/p images be the same, i.e., M=N.

    if ( bitand( sRestoreOptions.debugWord, LL_DEBUG_DAUB ) & ( i > 1 ) )

        if ( M ~= N ), error( 'Can only run Daubechies iterations under M=N' ), end
        
        if ( sRestoreOptions.reconSanityInterval > 1 ) & ...
                ( mod( i, sRestoreOptions.reconSanityInterval ) == 0 )

            disp( sprintf( 'Sanity: running Daubechies at i=%g', i ) );
            
            [sDIter, sDIterOptions] = emgauss_restore( 'forward_op', psf, ...
                'noisy_output', reshape( y, opImgDim ), ...
                'noise_sigma', noiseSigma, ...
                'theta_true', reshape( betaTrue, ipImgDim ), ...
                'm_selector', 'daub2', ...
                'num_em_iters', sRestoreOptions.NUM_DAUB_ITERS, ...
                'num_max_grad_search_failures', -1, ...
                'theta0_selector', 'Hty2', ...
                'theta0', [], ...
                'a_val', -1, ...
                'w_val', -1, ...
                't_val', -1, ...
                'sparse_ind', 'w', ...
                'b_save_history', 0, ...
                'stopping_cond', 'fixed', ...
                'stopping_cond_tol', noiseSigma, ...
                'stopping_cond_win', 0, ...
                'stopping_cond_last_fac', 0, ...
                'thetastar', 0, ...
                'soft_thres_t', -1, ...
                'hparam_rest_interval', 0, ...
                'bseparate_hparam_est', 0, ...
                'lambda', lambda ...
                );

            larsSol = betaHatPrevActiveSet./sRestoreOptions.XColumnNorm;
            daubSol = sDIter.last;
            [deltaLinf, deltaLinfInd] = max( abs( larsSol(:) - daubSol(:) ) );
            disp( sprintf( '|larsSol - daubSol|_inf = %g', deltaLinf ) );
            absDaubVal = abs( daubSol( deltaLinfInd ) );
            
            if ( deltaLinfInd > absDaubVal * sRestoreOptions.daubPercentTol )
                
                warning( sprintf( '|larsSol - daubSol|_inf = %g', ...
                    sRestoreOptions.daubDeltaTolFac*absDaubVal ) );
                
                if length( ipImgDim ) == 2
                    figure( 2 ), clf, imagesc( reshape( larsSol, ipImgSize ) ), colorbar;
                    title( 'LARS-LASSO' );
                    figure( 3 ), clf, imagesc( reshape( daubSol, ipImgSize ) ), colorbar;
                    title( 'Daubechies' );
                else
                    warning( 'Since the ip img is not 2-d, skipping plots #2 and #3' )
                end
                
                figure( 4 ), clf, plot( larsSol(:), 'bx' ), hold on, plot( daubSol(:), 'ro' );
                title( 'LARS-LASSO vs. Daubechies' );
                
                disp( 'Pausing...' ), pause;
                
            else
                disp( 'Pausing briefly' ), pause( sRestoreOptions.PAUSE_NUM_SECS );
            end

        end
        
    end

    betaOriginalScale = betaHatNextActiveSet(:) ./ sRestoreOptions.XColumnNorm(:);
    vecL0( i ) = sum( abs( betaOriginalScale ) > sRestoreOptions.l0Tol ); % l_0(beta)
    vecL1( i ) = sum( abs( betaOriginalScale ) ); % l_1(beta)

    % Compute the cost function  
    predictionErr = y - muHatNextActiveSet;
    rss = sum( predictionErr(:).^2 );
    vecCost( i ) = N*noiseSigma^2 + rss + 2*noiseSigma^2*vecL0(i);
    
    if ( ( betaHatBestInd < 1 ) | ...
         ( (betaHatBestInd >= 1) & (vecCost(betaHatBestInd) > vecCost(i)) ) )
        betaHatBestInd = i;
        betaHatBest = betaHatNextActiveSet;
    end
       
    % Step 6: Update state variables
    muHatPrevActiveSet = muHatNextActiveSet;
    betaHatPrevActiveSet = betaHatNextActiveSet;

    if bitand( sRestoreOptions.debugWord, LL_DEBUG_ITER_PAUSE )
        disp( 'Pausing...press key' ), pause
    end
    
end

tCpuTimeStop = cputime;

% Re-scale last beta and best beta
betaHat = betaHatPrevActiveSet ./ sRestoreOptions.XColumnNorm;
betaHatBest = betaHatBest ./ sRestoreOptions.XColumnNorm;
% muCheck should equal muHatPrevActiveSet
muCheck = sRestoreOptions.X * betaHatPrevActiveSet;

% Generate the output structure
for elemInd = 1:size( sRestoreElems, 1 )
    szCmd = sprintf( 'sOut.%s = %s;', ...
        char( sRestoreElems(elemInd,1) ), ...       
        char( sRestoreElems(elemInd,2) ) );
    try
        eval( szCmd );
    catch
        warning( sprintf( 'Could not execute: %s', szCmd ) );
    end
end

    
    

    
    
    


