
% sure_l1: Use SURE risk with l1 penalty
% 
% Based on emgauss_restore

function [sRestore, sRestoreOptions] = sure_l1( varargin )

nConvMethod = 1; % 1 => use DFTs, 2 => use convolution

% Default restore values
sRestoreOptions = struct( ...
    'NUM_EM_ITERS', 0, ...   
    'THETA0_SELECTOR', 'obs', ...
    'THETA0', [], ...
	'THRES_INTERVAL', [0, 1], ...
	'THRES_NUM_POINTS', 20, ...
	'THRES_VECT', [], ...
	'nConvMethod', nConvMethod ...
    );

% Process input arguments

lenVarargin = length( varargin );
numArgsIn = floor( lenVarargin/2 );

if 2*numArgsIn ~= lenVarargin, error( 'even ip args expected' ), end

for argInInd = 1:numArgsIn
    
    szKey = lower( varargin{2*argInInd-1} );    
    switch szKey
        case 'forward_op', hForwardOp = varargin{2*argInInd};
        case 'noisy_output', imgNoisy = varargin{2*argInInd};
        case 'noise_sigma', noiseSigma = varargin{2*argInInd};
        case 'theta_true', thetaTrue = varargin{2*argInInd};
                                 
        case 'num_em_iters', sRestoreOptions.NUM_EM_ITERS = varargin{2*argInInd};
        case 'theta0_selector', sRestoreOptions.THETA0_SELECTOR = varargin{2*argInInd};
        case 'theta0', sRestoreOptions.THETA0 = varargin{2*argInInd};
		case 'thres_interval', sRestoreOptions.THRES_INTERVAL = varargin{2*argInInd};
		case 'thres_num_points', sRestoreOptions.THRES_NUM_POINTS = varargin{2*argInInd};
		case 'thres_vect', sRestoreOptions.THRES_VECT = varargin{2*argInInd};
            
        otherwise
            error( sprintf( 'unrecognized option %s', szKey ) );
    end
end

% WTF? Convert thetaTrue to double. It seems that sometimes we get a logical
% array!
thetaTrue = double( thetaTrue );

thetaHist = {};

% Use the Aspire-based Gblur method
if nConvMethod == 1
	G = Gblur( ones( size( imgNoisy ) ), 'psf', hForwardOp, 'type', 'conv,fft' );
elseif nConvMethod == 2
	G = Gblur( ones( size( imgNoisy ) ), 'psf', hForwardOp, 'type', 'conv,same' );
else
    error( 'illegal nConvMethod' );
end


% Use DFTs
%imgNoisyDft = fft2( imgNoisy );
imgNoisyDft = fftn( imgNoisy );
% Must zero-pad the psf
%hForwardOpDft = fft2( hForwardOp, size( imgNoisy,1 ), size( imgNoisy,2 ) );
hForwardOpDft = fftn( hForwardOp, size( imgNoisy ) );
sRes = emsep_comp_alpha( hForwardOpDft, imgNoisyDft, noiseSigma );
% Extract upper bound for alpha
alphaUpperBound = sRes.alphaUpperBound;

switch sRestoreOptions.THETA0_SELECTOR 
    case 'deconv'
        % Zero-pad the psf
        psfZeroPadded = zeros( size( imgNoisy ) );
        psfZeroPadded( 1:size(hForwardOp,1), 1:size(hForwardOp,2) ) = hForwardOp;
        % Apply the pseudo-inverse to the noise observation
        deconvSoln = pseudoinv_eqdim( psfZeroPadded, imgNoisy );
        thetaHat0 = deconvSoln;
    case 'zero'
        thetaHat0 = zeros( size( imgNoisy ) );
    case 'checker'
        thetaHat0 = create_checker( size( imgNoisy ), 0.5, 1 );
    case 'custom'
        thetaHat0 = sRestoreOptions.THETA0;
    case 'true'
        thetaHat0 = thetaTrue;
    case 'obs'
        thetaHat0 = imgNoisy;
    case 'Hty'
        thetaHat0 = G' * imgNoisy;
	case 'Hty2'
		thetaHat0 = G' * imgNoisy;
		thetaHat0Sorted = sort( thetaHat0(:), 'descend' );
		k = floor( 0.8 * length( thetaHat0Sorted ) );
		thresh0 = thetaHat0Sorted( k );
		thetaHat0 = ( thetaHat0 - sign(thetaHat0)*thresh0 ) .* ( abs(thetaHat0) > thresh0 );
    otherwise
        error( 'invalid selection for theta0' );
end

sRestoreOptions.alpha = alphaUpperBound;
sRestoreOptions.tau = 1;
thetaHatPrev = thetaHat0;

% Metrics for convergence of iterations

thetaMaxDiffHist = []; % l_inf norm of difference of 2 consecutive estimates
thetaSumDiffHist = []; % l_1 norm of difference ...

% Metrics for accuracy of reconstruction. Note that depending on the method
% used, some of these might not be computed.

errorNorm1Hist = []; % l_1 norm of difference
errorNorm2Hist = []; % l_2 norm ...

% Construct the vector of thresholds, if it wasn't supplied
if isempty( sRestoreOptions.THRES_VECT )
	sRestoreOptions.THRES_VECT = linspace( sRestoreOptions.THRES_INTERVAL(1), ...
		sRestoreOptions.THRES_INTERVAL(2), ...
		sRestoreOptions.THRES_NUM_POINTS );
else
	sRestoreOptions.THRES_NUM_POINTS = length( sRestoreOptions.THRES_VECT );
end

% Vector of cost values
costVect = zeros( size( sRestoreOptions.THRES_VECT ) );

% Matrix with all of the thetaHats
thetaHatMat = zeros( [sRestoreOptions.THRES_NUM_POINTS, numel( thetaHat0 )] );

n = numel( imgNoisy );

tCpuTimeStart = cputime;

for thresInd = 1:sRestoreOptions.THRES_NUM_POINTS

	% Work with this threshold
	thres = sRestoreOptions.THRES_VECT( thresInd );
	disp( sprintf( 'Using thresold %d: %g', thresInd, thres ) );

	% Run the EM algorithm for a FIXED number of iterations
	for emIter = 1:sRestoreOptions.NUM_EM_ITERS
    	%--- E step ---
		zHatNext = thetaHatPrev + sRestoreOptions.tau * ( G' * ( imgNoisy - G*thetaHatPrev ) );
		%--- M step ---
		thetaHatNext = ( zHatNext >= thres ) .* ( zHatNext - thres ) + ...
			( zHatNext <= -thres ) .* ( zHatNext + thres );
		% Update next/previous variables
		thetaHatPrev = thetaHatNext;
	end

	thetaHat = thetaHatPrev; % Output
	thetaHatMat( thresInd, : ) = thetaHat(:)'; % Save output

	% Compute the cost function
	predictionErr = imgNoisy - G * thetaHat;
	rss = sum( predictionErr(:).^2 );
	costVect( thresInd ) = n*noiseSigma^2 + rss + 2*noiseSigma^2*sum( abs(thetaHat(:)) > 1e3*eps );

end

tCpuTimeStop = cputime;

% Defines the structure of sRestore
sRestoreElems = { 'thetaHatMat', 'thetaHatMat'; ...
    'init', 'thetaHat0'; ...
    'thetaMaxDiff', 'thetaMaxDiffHist'; ...
    'thetaSumDiff', 'thetaSumDiffHist'; ...
    'errorNorm1', 'errorNorm1Hist'; ...
    'errorNorm2', 'errorNorm2Hist'; ...
    'cputime', '[tCpuTimeStart, tCpuTimeStop]'; ...
	'costVect', 'costVect' ...
    };

for elemInd = 1:size( sRestoreElems, 1 )
    szCmd = sprintf( 'sRestore.%s = %s;', ...
        char( sRestoreElems(elemInd,1) ), ...       
        char( sRestoreElems(elemInd,2) ) );
    try
        eval( szCmd );
    catch
        warning( sprintf( 'Could not execute: %s', szCmd ) );
    end
end


