% bnb_min: branch-and-bound minimizer
%-------------------------------------------------------------------------------
% This is a proof-of-concept script. Here, we try to minimize the function
% phi(x) = (x-x0)^T A (x-x0), subject to xj in {0,1} for all j. Note that we
% will be using matrix inversion here, so the dimensions of A should not be
% too big.
%
% vecOptim = bnb_min( matA, vecColX0, maxIter [, ArgbVerbose] 
%    [, ArgVecRowSolInitial] )
%    matA        - square matrix compatible with x0
%    vecColX0    - column vector
%    maxIter     - max number of iterations (sanity check)
%    ArgbVerbose - verbose flag
%    ArgVecRowSolInitial - initial solution
%
% Output
%    vecOptim    - the optimal solution
%
function vecOptim = bnb_min( matA, vecColX0, maxIter, ArgbVerbose, ...
	ArgVecRowSolInitial )

% Globals
global stStorage caCoeff vecBCoeffComputed

% Parameters
bVerbose = 1; % by default, be verbose
DISPLAY_COUNT = 5000;
DOUBLE_TOL = max( [eps*1e8, 1e-9] );

disp( 'bnb_min called...' );

% Check arguments
n = length( vecColX0 );

if ( ( ~is_col_vec( vecColX0, n ) ) | ( n < 2 ) )
	error( 'vecColX0 has to be a column vector of length > 1' );
end

matASize = size( matA );
if ( ( matASize(1) ~= n ) | ( matASize(2) ~= n ) )
	error( 'matA is incompatible with x0' );
end

if ( (nargin >= 4) & (ArgbVerbose == 0) )
	bVerbose = 0;
end
 
% Create storage structure
stStorage = struct( 'numElems', 0, 'vecLen', n, 'data', [], 'dataSize', 0 );

caCoeff = cell( 1, n - 1 );

% Array of boolean flags indicated whether we have computed the coeff
% for various values of indFirst
vecBCoeffComputed = zeros( 1, n - 1 );

% If A is positive definite, the optimal solution is x0. Assume that this
% is the case.
vecRowSol0 = vecColX0'; % convert to row vector

% If an initial solution hasn't been provided, construct an incumbent solution 
% for the "bound" part of the algorithm. Basically, we take x0 and round each 
% element to either 0 or 1.
if ( nargin == 5 )
	vecRowSolIncumb = ArgVecRowSolInitial;
else
	vecRowSolIncumb = round( vecRowSol0 );
	for i = 1:n
		if ( vecRowSolIncumb(i) < 0.5 )
			vecRowSolIncumb(i) = 0;
		else
			vecRowSolIncumb(i) = 1;
		end
	end
end

phiIncumb = fnPhi( vecRowSolIncumb', matA, vecColX0 );
disp( sprintf( 'phiIncumb = %.5e', phiIncumb ) );

numIter = 0;
bPrematureExit = 0;

my_stack_push( vecRowSol0 );

%%% Big while loop

while ( 1 )
	numIter = numIter + 1; % Count the number of while loops we took

	% Sanity check: put a limit on the number of iterations
	if ( numIter >= maxIter )
		my_stack_dump;
		warning( 'maxIter reached' )
		bPrematureExit = 1;
	end

	% Sanity check: put a limit on the stack size
	if ( mod( numIter, DISPLAY_COUNT ) == (DISPLAY_COUNT-1) )
		stackSize = my_stack_size;
		disp( sprintf( 'stack size: %3d', stackSize ) );
		if ( stackSize > n )
			my_stack_dump;
			warning( 'stackSize > n???' );
			bPrematureExit = 1;
		end
	end

	if ( bPrematureExit )
		warning( '   !!! Breaking prematurely !!!' );
		disp( sprintf( 'BNB_MIN: used %5d iterations', numIter ) );
		disp( sprintf( 'BNB_MIN: optimal phi value %.5e', phiIncumb ) );
		vecOptim = vecRowSolIncumb';
		return;
	end

	% Are there any more solution spaces in the fifo to look at?
	vecRowSolParent = my_stack_pop;
	if ( length( vecRowSolParent ) == 0 ), break, end;
	if ( bVerbose )
		disp( sprintf( '   POPPED solParent: stack size=%3d', ...
			my_stack_size ) );
		disp( vecRowSolParent );
	end

	% Find the non-binary digits
	ind = [];
	for i = 1:n
		tmp = vecRowSolParent(i);
		if ( abs( tmp ) > DOUBLE_TOL )
			if ( abs( tmp - 1 ) > DOUBLE_TOL )
				ind = [ind, i];
				if ( bVerbose )
					disp( sprintf( 'adding i=%4d, tmp=%.9e', i, tmp ) );
				end
			end
		end
	end

	% All digits are binary. Check if this solution is better than the
	% incumbent.
	if ( length( ind ) == 0 )
		phiParent = fnPhi( vecRowSolParent', matA, vecColX0 );
		if ( phiParent < phiIncumb ) % Replace incumbent solution
			if ( bVerbose ), disp( '   --> replacing incumbent soln' ), end;
			vecRowSolIncumb = vecRowSolParent;
			phiIncumb = phiParent;
		end
		continue;
	end

	% The parent had non-binary digits. Create subproblems to further
	% explore the solution space.
	indFirst = ind( 1 ); % index of first non-binary digit
	if ( bVerbose ), disp( sprintf( 'indFirst = %4d', indFirst ) ), end;

	vecRowSolChild0 = vecRowSolParent;
	vecRowSolChild0( indFirst ) = 0;
	vecRowZ0 = vecRowSolChild0 - vecColX0';

	vecRowSolChild1 = vecRowSolParent;
	vecRowSolChild1( indFirst ) = 1;
	vecRowZ1 = vecRowSolChild1 - vecColX0';

	% Evaluate the optimal value of x for the two subproblems
	if ( indFirst < n )
		coeff = get_coeff( n, indFirst, matA );

		vecColZ0a = vecRowZ0( 1:indFirst )';
		vecColZ0b = coeff*vecColZ0a;
		vecRowXOptim0 = vecColX0' + [vecColZ0a', vecColZ0b'];

		vecColZ1a = vecRowZ1( 1:indFirst )';
		vecColZ1b = coeff*vecColZ1a;
		vecRowXOptim1 = vecColX0' + [vecColZ1a', vecColZ1b'];
	elseif ( indFirst == n )
		vecRowXOptim0 = vecRowSolChild0;
		vecRowXOptim1 = vecRowSolChild1;
	end

	phiOptim0 = fnPhi( vecRowXOptim0', matA, vecColX0 );
	phiOptim1 = fnPhi( vecRowXOptim1', matA, vecColX0 );

	if ( bVerbose )
		disp( 'xOptim0:' );
		disp( vecRowXOptim0 );
		disp( sprintf( 'subprob1: phiOptim1 = %.5e, phiIncumb = %.5e', ...
			phiOptim1, phiIncumb ) );
		disp( 'xOptim1:' );
		disp( vecRowXOptim1 );
		disp( sprintf( 'subprob0: phiOptim0 = %.5e, phiIncumb = %.5e', ...
			phiOptim0, phiIncumb ) );
	end

	% Do the same for subproblem 1
	if ( phiOptim1 < phiIncumb ) % Cannot discard
		my_stack_push( vecRowXOptim1 );
		if ( bVerbose )
			disp( sprintf( '   PUSHED xOptim1...stack size=%3d', ...
				my_stack_size ) );
		end
	else
		if ( bVerbose ), disp( '   Fathoming subprob1' ), end;
	end

	% Check to see if we can fathom subproblem 0
	if ( phiOptim0 < phiIncumb ) % Cannot discard
		my_stack_push( vecRowXOptim0 );
		if ( bVerbose )
			disp( sprintf( '   PUSHED xOptim0...stack size=%3d', ...
				my_stack_size ) );
		end
	else
		if ( bVerbose ), disp( '   Fathoming subprob0' ), end;
	end
end

disp( sprintf( 'BNB_MIN: used %5d iterations', numIter ) );
disp( sprintf( 'BNB_MIN: optimal phi value %.5e', phiIncumb ) );
vecOptim = vecRowSolIncumb';

%...............................................................................
function coeff = get_coeff( n, indFirst, matA )
%...............................................................................
global caCoeff vecBCoeffComputed

if ( (indFirst < 1) | (indFirst >= n) )
	error( 'get_coeff: invalid indFirst' );
end

if ( vecBCoeffComputed(indFirst) == 0 )
	A22 = matA( (indFirst+1):end, (indFirst+1):end );
	A12 = matA( 1:indFirst, (indFirst+1):end );
	A21 = matA( (indFirst+1):end, 1:indFirst );
	tmp = -inv( A22 + A22' )*( A12' + A21 );
	caCoeff(indFirst) = {tmp};
	vecBCoeffComputed(indFirst) = 1;
	if ( indFirst == (n-1) )
		disp( '   >>> last coeff computed <<<' );
		%beep;
	end
end

coeff = caCoeff{indFirst};

%...............................................................................
function my_stack_dump
%...............................................................................
global stStorage
disp( 'dumping stack contents:' );
for i = 1:stStorage.numElems
	disp( stStorage.data( i, : ) );
end

%...............................................................................
function nElems = my_stack_size
%...............................................................................
global stStorage
nElems = stStorage.numElems;

%...............................................................................
function my_stack_push( vecRow )
%...............................................................................
global stStorage

if ( length( vecRow ) ~= stStorage.vecLen )
	error( 'stack_push: incompatible vector' );
end
if ( stStorage.numElems < 0 )
	error( 'stack_push: numElems < 0' );
end

stStorage.numElems = stStorage.numElems + 1;

if ( stStorage.numElems <= stStorage.dataSize )
	% If there is space, add the data element to an existing row
	stStorage.data( stStorage.numElems, : ) = vecRow;
elseif ( stStorage.numElems == (stStorage.dataSize + 1) )
	% Add new elements to the last row of the matrix 'data'
	stStorage.data = [stStorage.data; vecRow];
	stStorage.dataSize = stStorage.dataSize + 1;
else
	error( 'stack_push: unexpected state' );
end

%...............................................................................
function vecRow = my_stack_pop
%...............................................................................
global stStorage

if ( stStorage.numElems == 0 )
	vecRow = [];
elseif ( stStorage.numElems > 0 )
	% Remove elements from the last row of the matrix 'data'
	vecRow = stStorage.data( stStorage.numElems, : );
	stStorage.numElems = stStorage.numElems - 1;
else
	error( 'stack_pop: numElems < 0' );
end

%...............................................................................
function bRc = is_col_vec( vec, n )
%...............................................................................
vecSize = size( vec );
bRc = 0; % Initialize to false
if ( ( vecSize(1) == n ) & ( vecSize(2) == 1 ) )
	bRc = 1;
end

%...............................................................................
function y = fnPhi( x, A, x0 )
%...............................................................................
y = (x - x0)' * A * (x - x0);
